Package: mutscan
Title: Preprocessing and Analysis of Deep Mutational Scanning Data
Version: 1.2.0
Authors@R: 
    c(person(given = "Charlotte",
             family = "Soneson",
             role = c("aut", "cre"),
             email = "charlottesoneson@gmail.com",
             comment = c(ORCID = "0000-0003-3833-2169")),
      person(given = "Michael",
    	       family = "Stadler",
    	       role = c("aut"),
    	       email = "michael.stadler@fmi.ch",
    	       comment = c(ORCID = "0000-0002-2269-4934")),
      person(given = "Friedrich Miescher Institute for Biomedical Research",
             role = "cph"))
Description: Provides functionality for processing and statistical
        analysis of multiplexed assays of variant effect (MAVE) and
        similar data. The package contains functions covering the full
        workflow from raw FASTQ files to publication-ready
        visualizations. A broad range of library designs can be
        processed with a single, unified interface.
Depends: R (>= 4.5.0)
Imports: BiocGenerics, S4Vectors, methods, SummarizedExperiment, Rcpp,
        edgeR (>= 3.42.0), dplyr, Matrix, limma, tidyr, stats, GGally,
        ggplot2, tidyselect (>= 1.2.0), tibble, rlang, grDevices, csaw,
        rmarkdown, xfun, DT, ggrepel, IRanges, utils, DelayedArray,
        tools
Suggests: testthat (>= 3.0.0), BiocStyle, knitr, Biostrings, pwalign,
        plotly, scattermore, BiocManager
SystemRequirements: GNU make
biocViews: GeneticVariability, GenomicVariation, Preprocessing
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
LinkingTo: Rcpp
Config/testthat/edition: 3
URL: https://github.com/fmicompbio/mutscan
BugReports: https://github.com/fmicompbio/mutscan/issues
Config/pak/sysreqs: cmake make libbz2-dev libicu-dev liblzma-dev
        libuv1-dev libssl-dev xz-utils zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 13:05:22 UTC
RemoteUrl: https://github.com/bioc/mutscan
RemoteRef: RELEASE_3_23
RemoteSha: 6aa85729d6ab31b4227c979a742e96f6d33dae62
NeedsCompilation: yes
Packaged: 2026-04-29 23:32:07 UTC; root
Author: Charlotte Soneson [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3833-2169>),
  Michael Stadler [aut] (ORCID: <https://orcid.org/0000-0002-2269-4934>),
  Friedrich Miescher Institute for Biomedical Research [cph]
Maintainer: Charlotte Soneson <charlottesoneson@gmail.com>
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-29 23:36:57 UTC; unix

# Generated by roxygen2: do not edit by hand

export(calcNearestStringDist)
export(calculateFitnessScore)
export(calculateRelativeFC)
export(collapseMutants)
export(collapseMutantsByAA)
export(collapseMutantsBySimilarity)
export(digestFastqs)
export(generateQCReport)
export(groupSimilarSequences)
export(linkMultipleVariants)
export(plotDistributions)
export(plotFiltering)
export(plotMeanDiff)
export(plotPairs)
export(plotTotals)
export(plotVolcano)
export(relabelMutPositions)
export(summarizeExperiment)
importFrom(BiocGenerics,"rownames<-")
importFrom(BiocGenerics,rownames)
importFrom(DT,datatable)
importFrom(DelayedArray,rowsum)
importFrom(GGally,eval_data_col)
importFrom(GGally,ggpairs)
importFrom(IRanges,IntegerList)
importFrom(Matrix,colSums)
importFrom(Matrix,rowMeans)
importFrom(Matrix,rowSums)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,unstrsplit)
importFrom(SummarizedExperiment,"rowData<-")
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(csaw,normOffsets)
importFrom(dplyr,across)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,matches)
importFrom(dplyr,mutate)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,ungroup)
importFrom(edgeR,DGEList)
importFrom(edgeR,estimateDisp)
importFrom(edgeR,getNormLibSizes)
importFrom(edgeR,glmQLFTest)
importFrom(edgeR,glmQLFit)
importFrom(edgeR,normLibSizes)
importFrom(edgeR,predFC)
importFrom(edgeR,scaleOffset)
importFrom(edgeR,topTags)
importFrom(ggplot2,aes)
importFrom(ggplot2,after_stat)
importFrom(ggplot2,annotate)
importFrom(ggplot2,coord_cartesian)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_density)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_fill_continuous)
importFrom(ggplot2,scale_fill_discrete)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_x_log10)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,scale_y_log10)
importFrom(ggplot2,stat_density2d)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,theme_void)
importFrom(ggplot2,ylim)
importFrom(ggrepel,geom_text_repel)
importFrom(grDevices,colorRamp)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,rgb)
importFrom(limma,contrasts.fit)
importFrom(limma,eBayes)
importFrom(limma,lmFit)
importFrom(limma,topTable)
importFrom(limma,voom)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(rlang,.data)
importFrom(rmarkdown,render)
importFrom(stats,cor)
importFrom(stats,setNames)
importFrom(tibble,rownames_to_column)
importFrom(tidyr,gather)
importFrom(tidyr,separate)
importFrom(tidyr,separate_rows)
importFrom(tidyselect,matches)
importFrom(tools,file_ext)
importFrom(utils,globalVariables)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)
importFrom(utils,relist)
importFrom(xfun,Rscript_call)
useDynLib(mutscan, .registration = TRUE)

<img src="man/figures/mutscan.png" align="right" alt="mutscan" width="150"/>

<br>

# mutscan

<br>

<!-- badges: start -->
  [![R-CMD-check](https://github.com/fmicompbio/mutscan/workflows/R-CMD-check/badge.svg)](https://github.com/fmicompbio/mutscan/actions)
  <!-- badges: end -->

`mutscan` is an R package to process, statistically analyse and visualise 
multiplexed assays of variant effect data.
The package is under active development, and the functionality and 
user interface may change. 
For more detailed information about how to run `mutscan`, please consult the 
vignettes. 

## Installation

`mutscan` is available from Bioconductor (from release 3.22) and can be 
installed with:

```r
# install.packages("BiocManager")
BiocManager::install("mutscan")
```

You can also install the development version of `mutscan` from GitHub with:

``` r
# install.packages("devtools")
devtools::install_github("fmicompbio/mutscan")
```

suppressPackageStartupMessages({
  library(ShortRead)
  library(BiocParallel)
})

## Trans, Input
r1 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952429_1.fastq.gz", withIds = TRUE)
r2 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952429_2.fastq.gz", withIds = TRUE)
ShortRead::writeFastq(r1[1:1000], file = "transInput_1.fastq.gz", mode = "w", full = FALSE, compress = TRUE)
ShortRead::writeFastq(r2[1:1000], file = "transInput_2.fastq.gz", mode = "w", full = FALSE, compress = TRUE)

## Trans, Output
r1 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952432_1.fastq.gz", withIds = TRUE)
r2 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952432_2.fastq.gz", withIds = TRUE)
ShortRead::writeFastq(r1[1:1000], file = "transOutput_1.fastq.gz", mode = "w", full = FALSE, compress = TRUE)
ShortRead::writeFastq(r2[1:1000], file = "transOutput_2.fastq.gz", mode = "w", full = FALSE, compress = TRUE)

## Cis, Input
r1 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952435_1.fastq.gz", withIds = TRUE)
r2 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952435_2.fastq.gz", withIds = TRUE)
ShortRead::writeFastq(r1[1:1000], file = "cisInput_1.fastq.gz", mode = "w", full = FALSE, compress = TRUE)
ShortRead::writeFastq(r2[1:1000], file = "cisInput_2.fastq.gz", mode = "w", full = FALSE, compress = TRUE)

## Cis, Output
r1 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952438_1.fastq.gz", withIds = TRUE)
r2 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/SRR5952438_2.fastq.gz", withIds = TRUE)
ShortRead::writeFastq(r1[1:1000], file = "cisOutput_1.fastq.gz", mode = "w", full = FALSE, compress = TRUE)
ShortRead::writeFastq(r2[1:1000], file = "cisOutput_2.fastq.gz", mode = "w", full = FALSE, compress = TRUE)

## Count matrix, cis
meta <- read.delim("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/meta/GSE102901_cis_coldata.tsv", header = TRUE, as.is = TRUE)
sampleIds <- meta$Name
names(sampleIds) <- sampleIds
allSamples <- lapply(sampleIds, function(id) {
  print(id)
  mutscan::digestFastqs(fastqForward = paste0("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/", id, "_1.fastq.gz"),
                        fastqReverse = paste0("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/GSE102901/FASTQ/", id, "_2.fastq.gz"),
                        mergeForwardReverse = TRUE, minOverlap = 0, maxOverlap = 0,
                        maxFracMismatchOverlap = 1, greedyOverlap = TRUE, 
                        revComplForward = FALSE, revComplReverse = TRUE,
                        elementsForward = "SUCV", elementsReverse = "SUCVS",
                        elementLengthsForward = c(1, 10, 18, 96),
                        elementLengthsReverse = c(1, 7, 17, 96, -1),
                        constantForward = "AACCGGAGGAGGGAGCTG",
                        constantReverse = "GAGTTCATCCTGGCAGC", adapterForward = "GGAAGAGCACACGTC", 
                        adapterReverse = "GGAAGAGCGTCGTGT", primerForward = "",
                        primerReverse = "", 
                        avePhredMinForward = 20, avePhredMinReverse = 20,
                        variableNMaxForward = 0, variableNMaxReverse = 0, umiNMax = 0,
                        wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA", 
                        wildTypeReverse = "", 
                        nbrMutatedCodonsMaxForward = 1, nbrMutatedCodonsMaxReverse = 1,
                        forbiddenMutatedCodonsForward = "NNW", forbiddenMutatedCodonsReverse = "NNW", 
                        mutNameDelimiter = ".", maxNReads = -1, verbose = TRUE, nThreads = 6)
})
se <- mutscan::summarizeExperiment(x = allSamples, coldata = meta, countType = "umis")
seaa <- mutscan::collapseMutantsByAA(se)
saveRDS(se, file = "GSE102901_cis_se.rds")

## Leu+Jun
r1 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/LEU_JUN_6rep_3tp/FASTQ/t0/GD09_26674_AACGGT_read1.fastq.gz", withIds = TRUE)
r2 <- ShortRead::readFastq("/tungstenfs/groups/gbioinfo/sonechar/FMIgroups/GroupDiss/Diss_deep_mutational_scanning/data/LEU_JUN_6rep_3tp/FASTQ/t0/GD09_26674_AACGGT_read2.fastq.gz", withIds = TRUE)
ShortRead::writeFastq(r1[22501:23500], file = "leujunt0_1.fastq.gz", mode = "w", full = FALSE, compress = TRUE)
ShortRead::writeFastq(r2[22501:23500], file = "leujunt0_2.fastq.gz", mode = "w", full = FALSE, compress = TRUE)

suppressPackageStartupMessages({
  library(Biostrings)
  library(ShortRead)
})

processReadsLeuJun <- function(Ldef) {
  ## Read data
  fq1 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqForward)
  fq2 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqReverse)
  
  ## Number of reads
  message("Number of reads: ", length(fq1))  ## 1000
  
  ## Primer sequences
  prim <- Biostrings::vcountPattern(Ldef$primerForward, fq1) == 0 |
    Biostrings::vcountPattern(Ldef$primerReverse, fq2) == 0
  message("Number of reads without primers: ", sum(prim))  ## 126
  fq1 <- fq1[!prim]
  fq2 <- fq2[!prim]
  
  ## Average quality in variable sequences too low
  endprimFwd <- unlist(end(Biostrings::vmatchPattern(Ldef$primerForward, fq1)))
  varForward <- Biostrings::subseq(fq1, start = endprimFwd + 1, width = width(fq1) - endprimFwd)
  endprimRev <- unlist(end(Biostrings::vmatchPattern(Ldef$primerReverse, fq2)))
  varReverse <- Biostrings::subseq(fq2, start = endprimRev + 1, width = Ldef$variableLengthReverse)
  avgQualForward <- ShortRead::alphabetScore(Biostrings::quality(varForward))/
    BiocGenerics::width(varForward)
  avgQualReverse <- ShortRead::alphabetScore(Biostrings::quality(varReverse))/
    BiocGenerics::width(varReverse)
  lowqual <- avgQualForward < Ldef$avePhredMinForward | avgQualReverse < Ldef$avePhredMinReverse
  message("Number of reads with low average quality: ", sum(lowqual))  # 0
  fq1 <- fq1[!lowqual]
  fq2 <- fq2[!lowqual]
  varForward <- varForward[!lowqual]
  varReverse <- varReverse[!lowqual]
  
  ## Too many Ns in variable sequences
  nbrNForward <- Biostrings::vcountPattern("N", varForward, fixed = TRUE)
  nbrNReverse <- Biostrings::vcountPattern("N", varReverse, fixed = TRUE)
  toomanynvar <- nbrNForward > Ldef$variableNMaxForward | nbrNReverse > Ldef$variableNMaxReverse
  message("Number of reads with too many N in variable sequence: ", sum(toomanynvar))  # 76
  fq1 <- fq1[!toomanynvar]
  fq2 <- fq2[!toomanynvar]
  varForward <- varForward[!toomanynvar]
  varReverse <- varReverse[!toomanynvar]
  
  ## Compare to wildtype, forward
  bestmatch <- DNAStringSet()
  for (i in seq_along(varForward)) {
    s <- DNAStringSet(Ldef$wildTypeForward)
    s <- Biostrings::subseq(s, start = 1, width = length(varForward[[i]]))
    p <- DNAStringSet(rep(as.character(varForward[[i]]), length(s)))
    pattern_width <- width(p)
    subject_width <- width(s)
    unlisted_ans <- which(as.raw(unlist(p)) == as.raw(unlist(s)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    matchPositions <- nucleotideMatches - offsets
    bestmatch <- c(bestmatch, s[which.max(sapply(matchPositions, length))])
  }
  p <- bestmatch
  pattern_width <- width(p)
  subject_width <- width(varForward)
  unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(varForward)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  mismatchPositions <- nucleotideMismatches - offsets
  mismatchQualities <- Biostrings::quality(varForward)[mismatchPositions]
  lowqualmm <- min(as(mismatchQualities, "IntegerList")) < Ldef$mutatedPhredMinForward
  message("Number of reads with low-quality mismatch bases, forward: ", sum(lowqualmm))  ## 0
  fq1 <- fq1[!lowqualmm]
  fq2 <- fq2[!lowqualmm]
  varForward <- varForward[!lowqualmm]
  varReverse <- varReverse[!lowqualmm]
  mismatchPositions <- mismatchPositions[!lowqualmm]
  p <- p[!lowqualmm]
  ## Number of mutated codons
  codonMismatches <- (mismatchPositions - 1) %/% 3 + 1
  nbrMutCodons <- S4Vectors::elementNROWS(unique(codonMismatches))
  toomanycodons <- nbrMutCodons > Ldef$nbrMutatedCodonsMaxForward
  message("Number of reads with too many mutated codons, forward: ", sum(toomanycodons))  ## 58
  fq1 <- fq1[!toomanycodons]
  fq2 <- fq2[!toomanycodons]
  codonMismatches <- codonMismatches[!toomanycodons]
  varForward <- varForward[!toomanycodons]
  varReverse <- varReverse[!toomanycodons]
  p <- p[!toomanycodons]
  ## Forbidden codons
  # uniqueMutCodons <- unique(codonMismatches)
  # mutCodons <- S4Vectors::endoapply(uniqueMutCodons, 
  #                                   function(v) rep(3 * v, each = 3) + c(-2, -1, 0))
  # mutCodons <- as(varForward, "DNAStringSet")[mutCodons]
  # matchForbidden <- Biostrings::vmatchPattern(
  #   pattern = Ldef$forbiddenMutatedCodonsForward, 
  #   as(mutCodons, "DNAStringSet"), 
  #   fixed = FALSE)
  # forbiddencodon <- sapply(BiocGenerics::relist(start(unlist(matchForbidden)) %% 3==1, 
  #                                               matchForbidden), function(i) any(i))
  # message("Number of reads with forbidden codons, forward: ", sum(forbiddencodon))  ## 6
  # fq1 <- fq1[!forbiddencodon]
  # fq2 <- fq2[!forbiddencodon]
  # mutCodons <- mutCodons[!forbiddencodon]
  # uniqueMutCodons <- uniqueMutCodons[!forbiddencodon]
  # splitMutCodons <- as(strsplit(gsub("([^.]{3})", "\\1\\.", 
  #                                    as.character(mutCodons)), ".", fixed = TRUE), 
  #                      "CharacterList")
  # encodedMutCodonsForward <- S4Vectors::unstrsplit(BiocGenerics::relist(
  #   paste0("f", unlist(uniqueMutCodons), unlist(splitMutCodons)), 
  #   uniqueMutCodons),
  #   sep = "_")
  encodedMutCodonsForward <- names(p)
  
  
  ## Compare to wildtype, reverse
  p <- DNAStringSet(rep(Ldef$wildTypeReverse, length(varReverse)))
  pattern_width <- width(p)
  subject_width <- width(varReverse)
  unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(varReverse)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  mismatchPositions <- nucleotideMismatches - offsets
  mismatchQualities <- Biostrings::quality(varReverse)[mismatchPositions]
  lowqualmm <- min(as(mismatchQualities, "IntegerList")) < Ldef$mutatedPhredMinReverse
  message("Number of reads with low-quality mismatch bases, reverse: ", sum(lowqualmm))  ## 0
  fq1 <- fq1[!lowqualmm]
  fq2 <- fq2[!lowqualmm]
  mismatchPositions <- mismatchPositions[!lowqualmm]
  varForward <- varForward[!lowqualmm]
  varReverse <- varReverse[!lowqualmm]
  encodedMutCodonsForward <- encodedMutCodonsForward[!lowqualmm]
  ## Number of mutated codons
  codonMismatches <- (mismatchPositions - 1) %/% 3 + 1
  nbrMutCodons <- S4Vectors::elementNROWS(unique(codonMismatches))
  toomanycodons <- nbrMutCodons > Ldef$nbrMutatedCodonsMaxReverse
  message("Number of reads with too many mutated codons, reverse: ", sum(toomanycodons))  ## 137
  fq1 <- fq1[!toomanycodons]
  fq2 <- fq2[!toomanycodons]
  codonMismatches <- codonMismatches[!toomanycodons]
  varForward <- varForward[!toomanycodons]
  varReverse <- varReverse[!toomanycodons]
  encodedMutCodonsForward <- encodedMutCodonsForward[!toomanycodons]
  ## Forbidden codons
  uniqueMutCodons <- unique(codonMismatches)
  mutCodons <- S4Vectors::endoapply(uniqueMutCodons, 
                                    function(v) rep(3 * v, each = 3) + c(-2, -1, 0))
  mutCodons <- as(varReverse, "DNAStringSet")[mutCodons]
  matchForbidden <- Biostrings::vmatchPattern(
    pattern = Ldef$forbiddenMutatedCodonsReverse, 
    as(mutCodons, "DNAStringSet"), 
    fixed = FALSE)
  forbiddencodon <- sapply(BiocGenerics::relist(start(unlist(matchForbidden)) %% 3==1, 
                                                matchForbidden), function(i) any(i))
  message("Number of reads with forbidden codons, reverse: ", sum(forbiddencodon))  ## 3
  fq1 <- fq1[!forbiddencodon]
  fq2 <- fq2[!forbiddencodon]
  encodedMutCodonsForward <- encodedMutCodonsForward[!forbiddencodon]
  mutCodons <- mutCodons[!forbiddencodon]
  uniqueMutCodons <- uniqueMutCodons[!forbiddencodon]
  splitMutCodons <- as(strsplit(gsub("([^.]{3})", "\\1\\.", 
                                     as.character(mutCodons)), ".", fixed = TRUE), 
                       "CharacterList")
  encodedMutCodonsReverse <- S4Vectors::unstrsplit(BiocGenerics::relist(
    paste0("r", unlist(uniqueMutCodons), unlist(splitMutCodons)), 
    uniqueMutCodons),
    sep = "_")
  mutNames <- paste(encodedMutCodonsForward, encodedMutCodonsReverse, sep = "_")
  mutNames <- gsub("^_", "", gsub("_$", "", mutNames))
  mutNames[mutNames <- ""] <- "WT"
  message("Number of variants seen twice: ", sum(table(mutNames) == 2))  ## 5
  message("Variants seen twice: ")
  print(which(table(mutNames) == 2))  ## ATF6B_r13CCC, ATF7_r26CTC, CEBPA, FOSB, FOSL1_r3GGC
  
  ## Retained reads
  message("Number of retained reads: ", length(fq1))  ## 600
}

## ----------------------------------------------------------------------------
leu <- c(ATF2 = "GATCCTGATGAAAAAAGGAGAAAGTTTTTAGAGCGAAATAGAGCAGCAGCTTCAAGATGCCGACAAAAAAGGAAAGTCTGGGTTCAGTCTTTAGAGAAGAAAGCTGAAGACTTGAGTTCATTAAATGGTCAGCTGCAGAGTGAAGTCACCCTGCTGAGAAATGAAGTGGCACAGCTGAAACAGCTTCTTCTGGCT",
         ATF7 = "GATCCAGATGAGCGACGGCAGCGCTTTCTGGAGCGCAACCGGGCTGCAGCCTCCCGCTGCCGCCAAAAGCGAAAGCTGTGGGTGTCCTCCCTAGAGAAGAAGGCCGAAGAACTCACTTCTCAGAACATTCAGCTGAGTAATGAAGTCACATTACTACGCAATGAGGTGGCCCAGTTGAAACAGCTACTGTTAGCT",
         CREB5 = "GATCCGGACGAGAGGCGGCGGAAATTTCTGGAACGGAACCGGGCAGCTGCCACCCGCTGCAGACAGAAGAGGAAGGTCTGGGTGATGTCATTGGAAAAGAAAGCAGAAGAACTCACCCAGACAAACATGCAGCTTCAGAATGAAGTGTCTATGTTGAAAAATGAGGTGGCCCAGCTGAAACAGTTGTTGTTAACA",
         ATF3 = "GAAGAAGATGAAAGGAAAAAGAGGCGACGAGAAAGAAATAAGATTGCAGCTGCAAAGTGCCGAAACAAGAAGAAGGAGAAGACGGAGTGCCTGCAGAAAGAGTCGGAGAAGCTGGAAAGTGTGAATGCTGAACTGAAGGCTCAGATTGAGGAGCTCAAGAACGAGAAGCAGCATTTGATATACATGCTCAACCTT",
         JDP2 = "GAGGAAGAGGAGCGAAGGAAAAGGCGCCGGGAGAAGAACAAAGTCGCAGCAGCCCGATGCCGGAACAAGAAGAAGGAGCGCACGGAGTTTCTGCAGCGGGAATCCGAGCGGCTGGAACTCATGAACGCAGAGCTGAAGACCCAGATTGAGGAGCTGAAGCAGGAGCGGCAGCAGCTCATCCTGATGCTGAACCGA",
         ATF4 = "GAGAAACTGGATAAGAAGCTGAAAAAAATGGAGCAAAACAAGACAGCAGCCACTAGGTACCGCCAGAAGAAGAGGGCGGAGCAGGAGGCTCTTACTGGTGAGTGCAAAGAGCTGGAAAAGAAGAACGAGGCTCTAAAAGAGAGGGCGGATTCCCTGGCCAAGGAGATCCAGTACCTGAAAGATTTGATAGAAGAG",
         ATF5 = "ACCCGAGGGGACCGCAAGCAAAAGAAGAGAGACCAGAACAAGTCGGCGGCTCTGAGGTACCGCCAGCGGAAGCGGGCAGAGGGTGAGGCCCTGGAGGGCGAGTGCCAGGGGCTGGAGGCACGGAATCGCGAGCTGAAGGAACGGGCAGAGTCCGTGGAGCGCGAGATCCAGTACGTCAAGGACCTGCTCATCGAG",
         CREBZF = "AGTCCCCGGAAGGCGGCGGCGGCCGCTGCCCGCCTTAATCGACTGAAGAAGAAGGAGTACGTGATGGGGCTGGAGAGTCGAGTCCGGGGTCTGGCAGCCGAGAACCAGGAGCTGCGGGCCGAGAATCGGGAGCTGGGCAAACGCGTACAGGCACTGCAGGAGGAGAGTCGCTACCTACGGGCAGTCTTAGCCAAC",
         BATF2 = "CCCAAGGAGCAACAAAGGCAGCTGAAGAAGCAGAAGAACCGGGCAGCCGCCCAGCGAAGCCGGCAGAAGCACACAGACAAGGCAGACGCCCTGCACCAGCAGCACGAGTCTCTGGAAAAAGACAACCTCGCCCTGCGGAAGGAGATCCAGTCCCTGCAGGCCGAGCTGGCGTGGTGGAGCCGGACCCTGCACGTG",
         BATF3 = "GAGGATGATGACAGGAAGGTCCGAAGGAGAGAAAAAAACCGAGTTGCTGCTCAGAGAAGTCGGAAGAAGCAGACCCAGAAGGCTGACAAGCTCCATGAGGAATATGAGAGCCTGGAGCAAGAAAACACCATGCTGCGGAGAGAGATCGGGAAGCTGACAGAGGAGCTGAAGCACCTGACAGAGGCACTGAAGGAG",
         CEBPE = "AAAGATAGCCTTGAGTACCGGCTGAGGCGGGAGCGCAACAACATCGCCGTGCGCAAGAGCCGAGACAAGGCCAAGAGGCGCATTCTGGAGACGCAGCAGAAGGTGCTGGAGTACATGGCAGAGAACGAGCGCCTCCGCAGCCGCGTGGAGCAGCTCACCCAGGAGCTAGACACCCTCCGCAACCTCTTCCGCCAG",
         BACH1 = "CTGGATTGTATCCATGATATTCGAAGAAGAAGTAAAAACAGAATTGCTGCACAGCGCTGTCGCAAGAGAAAACTTGACTGTATACAGAATCTTGAATCAGAAATTGAGAAGCTGCAAAGTGAAAAGGAGAGCTTGTTGAAGGAAAGAGATCACATTTTGTCAACTCTGGGTGAGACAAAGCAGAACCTAACTGGA",
         BACH2 = "TTAGAGTTTATTCATGATGTCCGACGGCGCAGCAAGAACCGCATCGCGGCCCAGCGCTGCCGCAAAAGGAAACTGGACTGTATTCAGAATTTAGAATGTGAAATCCGCAAATTGGTGTGTGAGAAAGAGAAACTGTTGTCAGAGAGGAATCAACTGAAAGCATGCATGGGGGAACTGTTGGACAACTTCTCCTGC",
         NFE2L1 = "CTGAGCCTCATCCGAGACATCCGGCGCCGGGGCAAGAACAAGATGGCGGCGCAGAACTGCCGCAAGCGCAAGCTGGACACCATCCTGAATCTGGAGCGTGATGTGGAGGACCTGCAGCGTGACAAAGCCCGGCTGCTGCGGGAGAAAGTGGAGTTCCTGCGCTCCCTGCGACAGATGAAGCAGAAGGTCCAGAGC",
         NFE2 = "CTAGCGCTAGTCCGGGACATCCGACGACGGGGCAAAAACAAGGTGGCAGCCCAGAACTGCCGCAAGAGGAAGCTGGAAACCATTGTGCAGCTGGAGCGGGAGCTGGAGCGGCTGACCAATGAACGGGAGCGGCTTCTCAGGGCCCGCGGGGAGGCAGACCGGACCCTGGAGGTCATGCGCCAACAGCTGACAGAG",
         NFIL3 = "AAGAAAGATGCTATGTATTGGGAAAAAAGGCGGAAAAATAATGAAGCTGCCAAAAGATCTCGTGAGAAGCGTCGACTGAATGACCTGGTTTTAGAGAACAAACTAATTGCACTGGGAGAAGAAAACGCCACTTTAAAAGCTGAGCTGCTTTCACTAAAATTAAAGTTTGGTTTAATTAGCTCCACAGCATATGCT",
         FOS = "GAAGAAGAAGAGAAAAGGAGAATCCGAAGGGAAAGGAATAAGATGGCTGCAGCCAAATGCCGCAACCGGAGGAGGGAGCTGACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTAGAGTTCATCCTGGCAGCT",
         FOSB = "GAGGAAGAGGAGAAGCGAAGGGTGCGCCGGGAACGAAATAAACTAGCAGCAGCTAAATGCAGGAACCGGCGGAGGGAGCTGACCGACCGACTCCAGGCGGAGACAGATCAGTTGGAGGAAGAAAAAGCAGAGCTGGAGTCGGAGATCGCCGAGCTCCAAAAGGAGAAGGAACGTCTGGAGTTTGTGCTGGTGGCC",
         FOSL1 = "GAGGAAGAGGAGCGCCGCCGAGTAAGGCGCGAGCGGAACAAGCTGGCTGCGGCCAAGTGCAGGAACCGGAGGAAGGAACTGACCGACTTCCTGCAGGCGGAGACTGACAAACTGGAAGATGAGAAATCTGGGCTGCAGCGAGAGATTGAGGAGCTGCAGAAGCAGAAGGAGCGCCTAGAGCTGGTGCTGGAAGCC",
         FOSL2 = "GAAGAGGAGGAGAAGCGTCGCATCCGGCGGGAGAGGAACAAGCTGGCTGCAGCCAAGTGCCGGAACCGACGCCGGGAGCTGACAGAGAAGCTGCAGGCGGAGACAGAGGAGCTGGAGGAGGAGAAGTCAGGCCTGCAGAAGGAGATTGCTGAGCTGCAGAAGGAGAAGGAGAAGCTGGAGTTCATGTTGGTGGCT",
         MAFB = "GTGATCCGCCTGAAGCAGAAGCGGCGGACCCTGAAGAACCGGGGCTACGCCCAGTCTTGCAGGTATAAACGCGTCCAGCAGAAGCACCACCTGGAGAATGAGAAGACGCAGCTCATTCAGCAGGTGGAGCAGCTTAAGCAGGAGGTGTCCCGGCTGGCCCGCGAGAGAGACGCCTACAAGGTCAAGTGCGAGAAA",
         JUN = "CAGGAGCGGATCAAGGCGGAGAGGAAGCGCATGAGGAACCGCATCGCTGCCTCCAAGTGCCGAAAAAGGAAGCTGGAGAGAATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTTAAACAGAAAGTCATGAAC",
         JUNB = "CAAGAGCGCATCAAAGTGGAGCGCAAGCGGCTGCGGAACCGGCTGGCGGCCACCAAGTGCCGGAAGCGGAAGCTGGAGCGCATCGCGCGCCTGGAGGACAAGGTGAAGACGCTCAAGGCCGAGAACGCGGGGCTGTCGAGTACCGCCGGCCTCCTCCGGGAGCAGGTGGCCCAGCTCAAACAGAAGGTCATGACC",
         JUND = "CAGGAGCGCATCAAGGCGGAGCGCAAGCGGCTGCGCAACCGCATCGCCGCCTCCAAGTGCCGCAAGCGCAAGCTGGAGCGCATCTCGCGCCTGGAAGAGAAAGTGAAGACCCTCAAGAGTCAGAACACGGAGCTGGCGTCCACGGCGAGCCTGCTGCGCGAGCAGGTGGCGCAGCTCAAGCAGAAAGTCCTCAGC",
         CREB3 = "GAACAAATTCTGAAACGTGTGCGGAGGAAGATTCGAAATAAAAGATCTGCTCAAGAGAGCCGCAGGAAAAAGAAGGTGTATGTTGGGGGTTTAGAGAGCAGGGTCTTGAAATACACAGCCCAGAATATGGAGCTTCAGAACAAAGTACAGCTTCTGGAGGAACAGAATTTGTCCCTTCTAGATCAACTGAGGAAA",
         HLF = "CTGAAGGATGACAAGTACTGGGCAAGGCGCAGAAAGAACAACATGGCAGCCAAGCGCTCCCGCGACGCCCGGAGGCTGAAAGAGAACCAGATCGCCATCCGGGCCTCGTTCCTGGAGAAGGAGAACTCGGCCCTCCGCCAGGAGGTGGCTGACTTGAGGAAGGAGCTGGGCAAATGCAAGAACATACTTGCCAAG",
         MAFG = "ATCGTCCAGCTGAAGCAGCGCCGGCGCACGCTCAAGAACCGCGGCTACGCTGCCAGCTGCCGCGTGAAGCGGGTGACGCAGAAGGAGGAGCTGGAGAAGCAGAAGGCGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCCTCAGAGAACGCCAGCATGAAGCTGGAGCTCGACGCGCTGCGCTCCAAGTACGAGGCG",
         MAFK = "GTGACCCGCCTGAAGCAGCGTCGGCGCACACTCAAGAACCGCGGCTACGCGGCCAGCTGCCGCATCAAGCGGGTGACGCAGAAGGAGGAGCTGGAGCGGCAGCGCGTGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCGCGTGAGAACAGCAGCATGCGGCTGGAGCTGGACGCCCTGCGCTCCAAGTACGAGGCG",
         XBP1 = "AGCCCCGAGGAGAAGGCGCTGAGGAGGAAACTGAAAAACAGAGTAGCAGCTCAGACTGCCAGAGATCGAAAGAAGGCTCGAATGAGTGAGCTGGAACAGCAAGTGGTAGATTTAGAAGAAGAGAACCAAAAACTTTTGCTAGAAAATCAGCTTTTACGAGAGAAAACTCATGGCCTTGTAGTTGAGAACCAGGAG",
         ATF6 = "ATTGCTGTGCTAAGGAGACAGCAACGTATGATAAAAAATCGAGAATCCGCTTGTCAGTCTCGCAAGAAGAAGAAAGAATATATGCTAGGGTTAGAGGCGAGATTAAAGGCTGCCCTCTCAGAAAACGAGCAACTGAAGAAAGAAAATGGAACACTGAAGCGGCAGCTGGATGAAGTTGTGTCAGAGAACCAGAGG",
         ATF6B = "GCAAAGCTGCTGAAGCGGCAGCAGCGAATGATCAAGAACCGGGAGTCAGCCTGCCAGTCCCGGAGAAAGAAGAAAGAGTATCTGCAGGGACTGGAGGCTCGGCTGCAAGCAGTACTGGCTGACAACCAGCAGCTCCGCCGAGAGAATGCTGCCCTCCGGCGGCGGCTGGAGGCCCTGCTGGCTGAAAACAGCGAG",
         CEBPA = "AAGAACAGCAACGAGTACCGGGTGCGGCGCGAGCGCAACAACATCGCGGTGCGCAAGAGCCGCGACAAGGCCAAGCAGCGCAACGTGGAGACGCAGCAGAAGGTGCTGGAGCTGACCAGTGACAATGACCGCCTGCGCAAGCGGGTGGAACAGCTGAGCCGCGAACTGGACACGCTGCGGGGCATCTTCCGCCAG",
         CEBPB = "AAGCACAGCGACGAGTACAAGATCCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGATGCGCAACCTGGAGACGCAGCACAAGGTCCTGGAGCTCACGGCCGAGAACGAGCGGCTGCAGAAGAAGGTGGAGCAGCTGTCGCGCGAGCTCAGCACCCTGCGGAACTTGTTCAAGCAG",
         CEBPD = "CGCGGCAGCCCCGAGTACCGGCAGCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGCGGCGCAACCAGGAGATGCAGCAGAAGTTGGTGGAGCTGTCGGCTGAGAACGAGAAGCTGCACCAGCGCGTGGAGCAGCTCACGCGGGACCTGGCCGGCCTCCGGCAGTTCTTCAAGCAG",
         CEBPG = "CGAAACAGTGACGAGTATCGGCAACGCCGAGAGAGGAACAACATGGCTGTGAAAAAGAGCCGGTTGAAAAGCAAGCAGAAAGCACAAGACACACTGCAGAGAGTCAATCAGCTCAAAGAAGAGAATGAACGGTTGGAAGCAAAAATCAAATTGCTGACCAAGGAATTAAGTGTACTCAAAGATTTGTTTCTTGAG",
         CREB1 = "GAAGCAGCACGAAAGAGAGAGGTCCGTCTAATGAAGAACAGGGAAGCAGCTCGAGAGTGTCGTAGAAAGAAGAAAGAATATGTGAAATGTTTAGAAAACAGAGTGGCAGTGCTTGAAAATCAAAACAAGACATTGATTGAGGAGCTAAAAGCACTTAAGGACCTTTACTGCCACAAATCAGAT",
         CREB3L1 = "GAGAAGGCCTTGAAGAGAGTCCGGAGGAAAATCAAGAACAAGATCTCAGCCCAGGAGAGCCGTCGTAAGAAGAAGGAGTATGTGGAGTGTCTAGAAAAGAAGGTGGAGACATTTACATCTGAGAACAATGAACTGTGGAAGAAGGTGGAGACCCTGGAGAATGCCAACAGGACCCTGCTCCAGCAGCTGCAGAAA",
         CREB3L2 = "GAGAAGGCCCTGAAGAAAATTCGGAGGAAGATCAAGAATAAGATTTCTGCTCAGGAAAGTAGGAGAAAGAAGAAAGAATACATGGACAGCCTGGAGAAAAAAGTGGAGTCTTGTTCAACTGAGAACTTGGAGCTTCGGAAGAAGGTAGAGGTTCTAGAGAACACTAATAGGACTCTCCTTCAGCAACTCCAGAAG",
         CREB3L3 = "GAGCGAGTGCTGAAAAAAATCCGCCGGAAAATCCGGAACAAGCAGTCGGCGCAAGAAAGCAGGAAGAAGAAGAAGGAATATATCGATGGCCTGGAGACTCGGATGTCAGCTTGCACTGCTCAGAATCAGGAGTTACAGAGGAAAGTCTTGCATCTCGAGAAGCAAAACCTGTCCCTCTTGGAGCAACTGAAGAAA",
         CREB3L4 = "GAGAGGGTCCTCAAGAAGGTCAGGAGGAAAATCCGTAACAAGCAGTCAGCTCAGGACAGTCGGCGGCGGAAGAAGGAGTACATTGATGGGCTGGAGAGCAGGGTGGCAGCCTGTTCTGCACAGAACCAAGAATTACAGAAAAAAGTCCAGGAGCTGGAGAGGCACAACATCTCCTTGGTAGCTCAGCTCCGCCAG",
         CREBL2 = "CCAGCCAAAATTGACTTGAAAGCAAAACTTGAGAGGAGCCGGCAGAGTGCAAGAGAATGCCGAGCCCGAAAAAAGCTGAGATATCAGTATTTGGAAGAGTTGGTATCCAGTCGAGAAAGAGCTATATGTGCCCTCAGAGAGGAACTGGAAATGTACAAGCAGTGGTGCATGGCAATGGACCAAGGAAAAATCCCT",
         CREBRF = "CCCTTAACAGCCCGACCAAGGTCAAGGAAGGAAAAAAATAAGCTGGCTTCCAGAGCTTGTCGGTTAAAGAAGAAAGCCCAGTATGAAGCTAATAAAGTGAAATTATGGGGCCTCAACACAGAATATGATAATTTATTGTTTGTAATCAACTCCATCAAGCAAGAGATTGTAAACCGGGTACAGAATCCAAGAGAT",
         DBP = "CAGAAGGATGAGAAATACTGGAGCCGGCGGTACAAGAACAACGAGGCAGCCAAGCGGTCCCGTGACGCCCGGCGGCTCAAGGAGAACCAGATATCGGTGCGGGCGGCCTTCCTGGAGAAGGAGAACGCCCTGCTGCGGCAGGAAGTTGTGGCCGTGCGCCAGGAGCTGTCCCACTACCGCGCCGTGCTGTCCCGA",
         NFE2L2 = "CTTGCATTAATTCGGGATATACGTAGGAGGGGTAAGAATAAAGTGGCTGCTCAGAATTGCAGAAAAAGAAAACTGGAAAATATAGTAGAACTAGAGCAAGATTTAGATCATTTGAAAGATGAAAAAGAAAAATTGCTCAAAGAAAAAGGAGAAAATGACAAAAGCCTTCACCTACTGAAAAAACAACTCAGCACC",
         NFE2L3 = "GTCTCACTTATCCGTGACATCAGACGAAGAGGGAAAAATAAAGTTGCTGCGCAGAACTGTCGTAAACGCAAATTGGACATAATTTTGAATTTAGAAGATGATGTATGTAACTTGCAAGCAAAGAAGGAAACTCTTAAGAGAGAGCAAGCACAATGTAACAAAGCTATTAACATAATGAAACAGAAACTGCATGAC",
         TEF = "CAGAAGGATGAAAAGTACTGGACAAGACGCAAGAAGAACAACGTGGCAGCTAAACGGTCACGGGATGCCCGGCGCCTGAAAGAGAATCAGATCACCATCCGGGCAGCCTTCCTGGAGAAGGAGAACACAGCCCTGCGGACGGAGGTGGCCGAGCTACGCAAGGAGGTGGGCAAGTGCAAGACCATCGTGTCCAAG"
)
## Generate values for comparison in unit tests - trans experiment
fqt1 <- system.file("extdata/leujunt0_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/leujunt0_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = -1, skipReverse = -1, 
  umiLengthForward = -1, umiLengthReverse = -1, 
  constantLengthForward = -1, constantLengthReverse = -1, 
  variableLengthForward = -1, variableLengthReverse = 96,
  adapterForward = "", 
  adapterReverse = "",
  primerForward = "GTCAGGTGGAGGCGGATCC",
  primerReverse = "GAAAAAGGAAGCTGGAGAGA",
  wildTypeForward = leu,
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = "", 
  constantReverse = "", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 0,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".", verbose = FALSE
)
processReadsLeuJun(Ldef)

suppressPackageStartupMessages({
  library(Biostrings)
  library(ShortRead)
})

processReadsCis <- function(Ldef) {
  ## Read data
  fq1 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqForward)
  fq2 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqReverse)
  
  ## Number of reads
  message("Number of reads: ", length(fq1))  ## 1000
  
  ## Adapter sequences
  adapt <- Biostrings::vcountPattern(Ldef$adapterForward, fq1) > 0 |
    Biostrings::vcountPattern(Ldef$adapterReverse, fq2) > 0
  message("Number of reads with adapters: ", sum(adapt))  ## 126
  fq1 <- fq1[!adapt]
  fq2 <- fq2[!adapt]
  
  ## Merge sequences
  varForward <-  Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 
                                      Ldef$constantLengthForward + 1, width = Ldef$variableLengthForward)
  varReverse <-  Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 
                                      Ldef$constantLengthReverse + 1, width = Ldef$variableLengthReverse)
  varReverse <- Biostrings::reverseComplement(varReverse)
  replaceAt <- as(ShortRead::FastqQuality(Biostrings::quality(varReverse)), "matrix") > 
    as(ShortRead::FastqQuality(Biostrings::quality(varForward)), "matrix")
  replaceAtList <- as(lapply(seq_len(nrow(replaceAt)),
                             function(x) which(replaceAt[x, ])), "IntegerList")
  consensusSeq <- Biostrings::replaceLetterAt(x = varForward, at = replaceAt, 
                                              letter = varReverse[replaceAtList])
  qualsForward <- as(ShortRead::FastqQuality(Biostrings::quality(varForward)), "matrix")
  qualsReverse <- as(ShortRead::FastqQuality(Biostrings::quality(varReverse)), "matrix")
  qualsForward[replaceAt] <- qualsReverse[replaceAt]
  phredQuals <- apply(qualsForward, 1, function(v) seqTools::ascii2char(v + 33))
  consensusQuals <- Biostrings::PhredQuality(phredQuals)
  varForward <- Biostrings::QualityScaledDNAStringSet(consensusSeq, consensusQuals)
  
  ## Average quality in variable sequences too low
  avgQualForward <- ShortRead::alphabetScore(Biostrings::quality(varForward))/
    BiocGenerics::width(varForward)
  lowqual <- avgQualForward < Ldef$avePhredMinForward
  message("Number of reads with low average quality: ", sum(lowqual))  # 0
  fq1 <- fq1[!lowqual]
  fq2 <- fq2[!lowqual]
  varForward <- varForward[!lowqual]
  
  ## Too many Ns in variable sequences
  nbrNForward <- Biostrings::vcountPattern("N", varForward, fixed = TRUE)
  toomanynvar <- nbrNForward > Ldef$variableNMaxForward
  message("Number of reads with too many N in variable sequence: ", sum(toomanynvar))  # 44
  fq1 <- fq1[!toomanynvar]
  fq2 <- fq2[!toomanynvar]
  varForward <- varForward[!toomanynvar]
  
  ## Too many Ns in UMIs
  umiForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + 1, width = Ldef$umiLengthForward)
  umiReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + 1, width = Ldef$umiLengthReverse)
  umis <- Biostrings::xscat(umiForward, umiReverse)
  nbrNumi <- Biostrings::vcountPattern("N", umis, fixed = TRUE)
  toomanynumi <- nbrNumi > Ldef$umiNMax
  message("Number of reads with too many N in UMI: ", sum(toomanynumi))  # 0
  fq1 <- fq1[!toomanynumi]
  fq2 <- fq2[!toomanynumi]
  varForward <- varForward[!toomanynumi]
  
  ## Compare to wildtype, forward
  p <- DNAStringSet(rep(Ldef$wildTypeForward, length(varForward)))
  pattern_width <- width(p)
  subject_width <- width(varForward)
  unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(varForward)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  mismatchPositions <- nucleotideMismatches - offsets
  mismatchQualities <- Biostrings::quality(varForward)[mismatchPositions]
  lowqualmm <- min(as(mismatchQualities, "IntegerList")) < Ldef$mutatedPhredMinForward
  message("Number of reads with low-quality mismatch bases: ", sum(lowqualmm))  ## 0
  fq1 <- fq1[!lowqualmm]
  fq2 <- fq2[!lowqualmm]
  mismatchPositions <- mismatchPositions[!lowqualmm]
  varForward <- varForward[!lowqualmm]
  ## Number of mutated codons/bases
  codonMismatches <- (mismatchPositions - 1) %/% 3 + 1
  nbrMutCodons <- S4Vectors::elementNROWS(unique(codonMismatches))
  baseMismatches <- mismatchPositions
  nbrMutBases <- S4Vectors::elementNROWS(unique(baseMismatches))
  if (Ldef$nbrMutatedCodonsMaxForward != (-1)) {
    toomanycodons <- nbrMutCodons > Ldef$nbrMutatedCodonsMaxForward
  } else {
    toomanycodons <- nbrMutBases > Ldef$nbrMutatedBasesMaxForward
  }
  message("Number of reads with too many mutated codons: ", sum(toomanycodons))  ## 581
  fq1 <- fq1[!toomanycodons]
  fq2 <- fq2[!toomanycodons]
  codonMismatches <- codonMismatches[!toomanycodons]
  nbrMutCodons <- nbrMutCodons[!toomanycodons]
  baseMismatches <- baseMismatches[!toomanycodons]
  nbrMutBases <- nbrMutBases[!toomanycodons]
  varForward <- varForward[!toomanycodons]
  ## Forbidden codons
  # if (Ldef$nbrMutatedCodonsMaxForward != (-1)) {
  uniqueMutCodons <- unique(codonMismatches)
  uniqueMutBases <- unique(baseMismatches)
  mutBases <- S4Vectors::endoapply(uniqueMutBases,
                                    function(v) v)
  mutBases <- as(varForward, "DNAStringSet")[mutBases]
  mutCodons <- S4Vectors::endoapply(uniqueMutCodons, 
                                    function(v) rep(3 * v, each = 3) + c(-2, -1, 0))
  mutCodons <- as(varForward, "DNAStringSet")[mutCodons]
  matchForbidden <- Biostrings::vmatchPattern(
    pattern = Ldef$forbiddenMutatedCodonsForward, 
    as(mutCodons, "DNAStringSet"), 
    fixed = FALSE)
  forbiddencodon <- sapply(BiocGenerics::relist(start(unlist(matchForbidden)) %% 3==1, 
                                                matchForbidden), function(i) any(i))
  message("Number of reads with forbidden codons: ", sum(forbiddencodon))  ## 82
  fq1 <- fq1[!forbiddencodon]
  fq2 <- fq2[!forbiddencodon]
  varForward <- varForward[!forbiddencodon]
  mutCodons <- mutCodons[!forbiddencodon]
  mutBases <- mutBases[!forbiddencodon]
  uniqueMutCodons <- uniqueMutCodons[!forbiddencodon]
  uniqueMutBases <- uniqueMutBases[!forbiddencodon]
  nbrMutCodons <- nbrMutCodons[!forbiddencodon]
  nbrMutBases <- nbrMutBases[!forbiddencodon]
  if (Ldef$nbrMutatedCodonsMaxForward != (-1)) {
    splitMutCodons <- as(strsplit(gsub("([^.]{3})", "\\1\\.", 
                                       as.character(mutCodons)), ".", fixed = TRUE), 
                         "CharacterList")
    encodedMutCodonsForward <- S4Vectors::unstrsplit(BiocGenerics::relist(
      paste0("f", unlist(uniqueMutCodons), unlist(splitMutCodons)), 
      uniqueMutCodons),
      sep = "_")
  } else {
    splitMutBases <- as(strsplit(gsub("([^.]{1})", "\\1\\.",
                                      as.character(mutBases)), ".", fixed = TRUE),
                        "CharacterList")
    encodedMutCodonsForward <- S4Vectors::unstrsplit(BiocGenerics::relist(
      paste0("f", unlist(uniqueMutBases), unlist(splitMutBases)), 
      uniqueMutBases),
      sep = "_")
  }
  
  mutNames <- encodedMutCodonsForward
  mutNames <- gsub("^_", "", gsub("_$", "", mutNames))
  mutNames[mutNames <- ""] <- "WT"
  message("Number of variants seen thrice: ", sum(table(mutNames) == 3))  ## 5
  message("Variants seen thrice: ")
  print(which(table(mutNames) == 3))  
  
  ## Retained reads
  message("Number of retained reads: ", length(fq1))  ## 167
  
  ## Number of mutated bases and codons
  message("Number of mutated codons per retained read: ")
  print(table(nbrMutCodons))
  message("Number of different reads with each number of mismatched codons: ")
  print(table(nbrMutCodons[!duplicated(varForward)]))
  message("Number of mutated bases per retained read: ")
  print(table(nbrMutBases))
  message("Number of different reads with each number of mismatched bases: ")
  print(table(nbrMutBases[!duplicated(varForward)]))
  
  ## Number of mutated amino acids
  p_aa <- Biostrings::translate(DNAStringSet(rep(Ldef$wildTypeForward, length(varForward))))
  pattern_width_aa <- width(p_aa)
  varForward_aa <- Biostrings::translate(varForward)
  subject_width_aa <- width(varForward_aa)
  unlisted_ans_aa <- which(as.raw(unlist(p_aa)) != as.raw(unlist(varForward_aa)))
  breakpoints_aa <- cumsum(pattern_width_aa)
  ans_eltlens_aa <- tabulate(findInterval(unlisted_ans_aa - 1L,
                                       breakpoints_aa) + 1L,
                          nbins = length(p_aa))
  skeleton_aa <- IRanges::PartitioningByEnd(cumsum(ans_eltlens_aa))
  aaMismatches <- BiocGenerics::relist(unlisted_ans_aa, skeleton_aa)
  message("Number of mutated aas per retained read: ")
  print(table(lengths(aaMismatches)))  ## 97, 70
  message("Number of different reads with each number of mismatched aas: ")
  print(table(lengths(aaMismatches)[!duplicated(varForward)]))  ## 14, 53
  
  ## Error statistics
  ## Forward
  constForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 1, 
                                     width = Ldef$constantLengthForward)
  p <- DNAStringSet(rep(Ldef$constantForward, length(constForward)))
  pattern_width <- width(p)
  subject_width <- width(constForward)
  
  unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(constForward)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  mismatchPositions <- nucleotideMismatches - offsets
  mismatchQualities <- unlist(as(quality(constForward)[mismatchPositions], "IntegerList"))
  
  unlisted_ans <- which(as.raw(unlist(p)) == as.raw(unlist(constForward)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  matchPositions <- nucleotideMatches - offsets
  matchQualities <- unlist(as(quality(constForward)[matchPositions], "IntegerList"))
  
  message("Match qualities, forward:")
  print(table(matchQualities))
  message("Mismatch qualities, forward:")
  print(table(mismatchQualities))
  
  ## Reverse
  constReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 1,
                                     width = Ldef$constantLengthReverse)
  constReverse <- Biostrings::reverseComplement(constReverse)
  p <- DNAStringSet(rep(Ldef$constantReverse, length(constReverse)))
  pattern_width <- width(p)
  subject_width <- width(constReverse)
  
  unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(constReverse)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  mismatchPositions <- nucleotideMismatches - offsets
  mismatchQualities <- unlist(as(quality(constReverse)[mismatchPositions], "IntegerList"))
  
  unlisted_ans <- which(as.raw(unlist(p)) == as.raw(unlist(constReverse)))
  breakpoints <- cumsum(pattern_width)
  ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                       breakpoints) + 1L,
                          nbins = length(p))
  skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
  nucleotideMatches <- BiocGenerics::relist(unlisted_ans, skeleton)
  offsets <- c(0L, breakpoints[-length(breakpoints)])
  matchPositions <- nucleotideMatches - offsets
  matchQualities <- unlist(as(quality(constReverse)[matchPositions], "IntegerList"))
  
  message("Match qualities, reverse:")
  print(table(matchQualities))
  message("Mismatch qualities, reverse:")
  print(table(mismatchQualities))
}  

## ----------------------------------------------------------------------------
## Generate values for comparison in unit tests - cis experiment
fqc1 <- system.file("extdata/cisInput_1.fastq.gz", package = "mutscan")
fqc2 <- system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqc1, fastqReverse = fqc2, 
  mergeForwardReverse = TRUE, revComplForward = FALSE, revComplReverse = TRUE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 7, 
  constantLengthForward = 18, constantLengthReverse = 17, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAGTTCATCCTGGCAGC", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  verbose = FALSE
)
processReadsCis(Ldef)

## Limit number of mismatching bases rather than codons
Ldef <- list(
  fastqForward = fqc1, fastqReverse = fqc2, 
  mergeForwardReverse = TRUE, revComplForward = FALSE, revComplReverse = TRUE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 7, 
  constantLengthForward = 18, constantLengthReverse = 17, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAGTTCATCCTGGCAGC", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = -1,
  nbrMutatedCodonsMaxReverse = -1,
  nbrMutatedBasesMaxForward = 2,
  nbrMutatedBasesMaxReverse = 2,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  verbose = FALSE
)
processReadsCis(Ldef)

suppressPackageStartupMessages({
  library(Biostrings)
  library(ShortRead)
  library(dplyr)
})

collapse_seqs <- function(seqs, umis, reads, maxdist, umimaxdist, minreads, minratio) {
  tokeep <- list()
  while (length(seqs) > 0) {
    if (reads[1] >= minreads) {
      d <- stringdist::stringdist(seqs[1], seqs, method = "hamming")
      w <- union(1, which(d <= maxdist & reads[1] >= minratio * reads))
    } else {
      w <- 1
    }
    tokeep[[seqs[1]]] <- paste(umis[w], collapse = ",")
    seqs <- seqs[-w]
    umis <- umis[-w]
    reads <- reads[-w]
  }
  tokeep <- lapply(tokeep, function(tk) {
    tk <- sort(unique(strsplit(tk, ",")[[1]]))
    if (length(tk) == 1) {
      tk
    } else {
      k <- c()
      while (length(tk) > 0) {
        d <- stringdist::stringdist(tk[1], tk, method = "hamming")
        w <- which(d <= umimaxdist)
        k <- c(k, tk[1])
        tk <- tk[-w]
      }
      k
    }
  })
  tokeep
}


processReadsTrans <- function(Ldef) {
  ## Read data
  fq1 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqForward)
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- Biostrings::readQualityScaledDNAStringSet(Ldef$fastqReverse)
  }
  
  ## Number of reads
  message("Number of reads: ", length(fq1))  ## 1000
  
  ## Adapter sequences
  adapt <- Biostrings::vcountPattern(Ldef$adapterForward, fq1) > 0
  if (!is.null(Ldef$fastqReverse)) {
    adapt2 <- Biostrings::vcountPattern(Ldef$adapterReverse, fq2) > 0
    adapt <- adapt | adapt2
  }
  message("Number of reads with adapters: ", sum(adapt))  ## 314
  fq1 <- fq1[!adapt]
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- fq2[!adapt]
  }
  
  ## Average quality in variable sequences too low
  varForward <-  Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 
                                      Ldef$constantLengthForward + 1, width = Ldef$variableLengthForward)
  avgQualForward <- ShortRead::alphabetScore(Biostrings::quality(varForward))/
    BiocGenerics::width(varForward)
  lowqual <- avgQualForward < Ldef$avePhredMinForward
  if (!is.null(Ldef$fastqReverse)) {
    varReverse <-  Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 
                                        Ldef$constantLengthReverse + 1, width = Ldef$variableLengthReverse)
    avgQualReverse <- ShortRead::alphabetScore(Biostrings::quality(varReverse))/
      BiocGenerics::width(varReverse)
    lowqual <- lowqual | avgQualReverse < Ldef$avePhredMinReverse
  }
  message("Number of reads with low average quality: ", sum(lowqual))  # 7
  fq1 <- fq1[!lowqual]
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- fq2[!lowqual]
  }
  
  ## Too many Ns in variable sequences
  varForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 
                                     Ldef$constantLengthForward + 1, width = Ldef$variableLengthForward)
  nbrNForward <- Biostrings::vcountPattern("N", varForward, fixed = TRUE)
  toomanynvar <- nbrNForward > Ldef$variableNMaxForward
  if (!is.null(Ldef$fastqReverse)) {
    varReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 
                                       Ldef$constantLengthReverse + 1, width = Ldef$variableLengthReverse)
    nbrNReverse <- Biostrings::vcountPattern("N", varReverse, fixed = TRUE)
    toomanynvar <- toomanynvar | nbrNReverse > Ldef$variableNMaxReverse
  }
  message("Number of reads with too many N in variable sequence: ", sum(toomanynvar))  # 0
  fq1 <- fq1[!toomanynvar]
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- fq2[!toomanynvar]
  }
  
  ## Too many Ns in UMIs
  umiForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + 1, width = Ldef$umiLengthForward)
  if (!is.null(Ldef$fastqReverse)) {
    umiReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + 1, width = Ldef$umiLengthReverse)
    umis <- Biostrings::xscat(umiForward, umiReverse)
  } else {
    umis <- umiForward
  }
  nbrNumi <- Biostrings::vcountPattern("N", umis, fixed = TRUE)
  toomanynumi <- nbrNumi > Ldef$umiNMax
  message("Number of reads with too many N in UMI: ", sum(toomanynumi))  # 0
  fq1 <- fq1[!toomanynumi]
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- fq2[!toomanynumi]
  }
  
  ## Compare to wildtype, forward
  mutCodons <- NULL
  uniqueMutCodons <- NULL
  mutBases <- NULL
  uniqueMutBases <- NULL
  nbrMutBasesTot <- NULL
  if (Ldef$wildTypeForward != "") {
    varForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 
                                       Ldef$constantLengthForward + 1, width = Ldef$variableLengthForward)
    p <- DNAStringSet(rep(Ldef$wildTypeForward, length(varForward)))
    pattern_width <- width(p)
    subject_width <- width(varForward)
    unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(varForward)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    mismatchPositions <- nucleotideMismatches - offsets
    ## First check, excluding reads with more than 3x(max nbr mut codons) mismatches
    baseMismatches <- mismatchPositions
    nbrMutBases <- S4Vectors::elementNROWS(unique(baseMismatches))
    toomanybases <- nbrMutBases > 3 * Ldef$nbrMutatedCodonsMaxForward
    message("Number of reads with too many mutated codons, forward (1): ", sum(toomanybases))
    fq1 <- fq1[!toomanybases]
    if (!is.null(Ldef$fastqReverse)) {
      fq2 <- fq2[!toomanybases]
    }
    mismatchPositions <- mismatchPositions[!toomanybases]
    varForward <- varForward[!toomanybases]
    
    mismatchQualities <- Biostrings::quality(varForward)[mismatchPositions]
    lowqualmm <- min(as(mismatchQualities, "IntegerList")) < Ldef$mutatedPhredMinForward
    message("Number of reads with low-quality mismatch bases, forward: ", sum(lowqualmm))  ## 0
    fq1 <- fq1[!lowqualmm]
    if (!is.null(Ldef$fastqReverse)) {
      fq2 <- fq2[!lowqualmm]
    }
    mismatchPositions <- mismatchPositions[!lowqualmm]
    varForward <- varForward[!lowqualmm]
    ## Number of mutated codons
    codonMismatches <- (mismatchPositions - 1) %/% 3 + 1
    nbrMutCodons <- S4Vectors::elementNROWS(unique(codonMismatches))
    toomanycodons <- nbrMutCodons > Ldef$nbrMutatedCodonsMaxForward
    message("Number of reads with too many mutated codons, forward: ", sum(toomanycodons))  ## 287
    fq1 <- fq1[!toomanycodons]
    if (!is.null(Ldef$fastqReverse)) {
      fq2 <- fq2[!toomanycodons]
    }
    codonMismatches <- codonMismatches[!toomanycodons]
    varForward <- varForward[!toomanycodons]
    nbrMutCodons <- nbrMutCodons[!toomanycodons]
    nbrMutBases <- S4Vectors::elementNROWS(unique(mismatchPositions[!toomanycodons]))
    ## Forbidden codons
    uniqueMutCodons <- unique(codonMismatches)
    mutCodons <- S4Vectors::endoapply(uniqueMutCodons, 
                                      function(v) rep(3 * v, each = 3) + c(-2, -1, 0))
    mutCodons <- as(varForward, "DNAStringSet")[mutCodons]
    matchForbidden <- Biostrings::vmatchPattern(
      pattern = Ldef$forbiddenMutatedCodonsForward, 
      as(mutCodons, "DNAStringSet"), 
      fixed = FALSE)
    forbiddencodon <- sapply(BiocGenerics::relist(start(unlist(matchForbidden)) %% 3==1, 
                                                  matchForbidden), function(i) any(i))
    message("Number of reads with forbidden codons, forward: ", sum(forbiddencodon))  ## 6
    fq1 <- fq1[!forbiddencodon]
    if (!is.null(Ldef$fastqReverse)) {
      fq2 <- fq2[!forbiddencodon]
    }
    mutCodons <- mutCodons[!forbiddencodon]
    nbrMutCodonsTot <- nbrMutCodons[!forbiddencodon]
    nbrMutBasesTot <- nbrMutBases[!forbiddencodon]
    uniqueMutCodons <- uniqueMutCodons[!forbiddencodon]
    splitMutCodons <- as(strsplit(gsub("([^.]{3})", "\\1\\.", 
                                       as.character(mutCodons)), ".", fixed = TRUE), 
                         "CharacterList")
    encodedMutCodonsForward <- S4Vectors::unstrsplit(BiocGenerics::relist(
      paste0("f", unlist(uniqueMutCodons), unlist(splitMutCodons)), 
      uniqueMutCodons),
      sep = "_")
  } else {
    encodedMutCodonsForward <- ""
  }
  
  ## Compare to wildtype, reverse
  if (!is.null(Ldef$fastqReverse) && Ldef$wildTypeReverse != "") {
    varReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 
                                       Ldef$constantLengthReverse + 1, width = Ldef$variableLengthReverse)
    p <- DNAStringSet(rep(Ldef$wildTypeReverse, length(varReverse)))
    pattern_width <- width(p)
    subject_width <- width(varReverse)
    unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(varReverse)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    mismatchPositions <- nucleotideMismatches - offsets
    
    ## First check, excluding reads with more than 3x(max nbr mut codons) mismatches
    baseMismatches <- mismatchPositions
    nbrMutBases <- S4Vectors::elementNROWS(unique(baseMismatches))
    toomanybases <- nbrMutBases > 3 * Ldef$nbrMutatedCodonsMaxReverse
    message("Number of reads with too many mutated codons, reverse (1): ", sum(toomanybases))
    fq1 <- fq1[!toomanybases]
    if (!is.null(Ldef$fastqReverse)) {
      fq2 <- fq2[!toomanybases]
    }
    mismatchPositions <- mismatchPositions[!toomanybases]
    varReverse <- varReverse[!toomanybases]
    encodedMutCodonsForward <- encodedMutCodonsForward[!toomanybases]
    nbrMutBasesTot <- nbrMutBasesTot[!toomanybases]
    nbrMutCodonsTot <- nbrMutCodonsTot[!toomanybases]
    
    mismatchQualities <- Biostrings::quality(varReverse)[mismatchPositions]
    lowqualmm <- min(as(mismatchQualities, "IntegerList")) < Ldef$mutatedPhredMinReverse
    message("Number of reads with low-quality mismatch bases, reverse: ", sum(lowqualmm))  ## 0
    fq1 <- fq1[!lowqualmm]
    fq2 <- fq2[!lowqualmm]
    mismatchPositions <- mismatchPositions[!lowqualmm]
    varReverse <- varReverse[!lowqualmm]
    encodedMutCodonsForward <- encodedMutCodonsForward[!lowqualmm]
    nbrMutBasesTot <- nbrMutBasesTot[!lowqualmm]
    nbrMutCodonsTot <- nbrMutCodonsTot[!lowqualmm]
    ## Number of mutated codons
    codonMismatches <- (mismatchPositions - 1) %/% 3 + 1
    nbrMutCodons <- S4Vectors::elementNROWS(unique(codonMismatches))
    toomanycodons <- nbrMutCodons > Ldef$nbrMutatedCodonsMaxReverse
    message("Number of reads with too many mutated codons, reverse: ", sum(toomanycodons))  ## 105
    fq1 <- fq1[!toomanycodons]
    fq2 <- fq2[!toomanycodons]
    codonMismatches <- codonMismatches[!toomanycodons]
    varReverse <- varReverse[!toomanycodons]
    encodedMutCodonsForward <- encodedMutCodonsForward[!toomanycodons]
    nbrMutBasesTot <- nbrMutBasesTot[!toomanycodons]
    nbrMutCodonsTot <- nbrMutCodonsTot[!toomanycodons]
    nbrMutCodons <- nbrMutCodons[!toomanycodons]
    nbrMutBases <- S4Vectors::elementNROWS(unique(mismatchPositions[!toomanycodons]))
    ## Forbidden codons
    uniqueMutCodons <- unique(codonMismatches)
    mutCodons <- S4Vectors::endoapply(uniqueMutCodons, 
                                      function(v) rep(3 * v, each = 3) + c(-2, -1, 0))
    mutCodons <- as(varReverse, "DNAStringSet")[mutCodons]
    matchForbidden <- Biostrings::vmatchPattern(
      pattern = Ldef$forbiddenMutatedCodonsReverse, 
      as(mutCodons, "DNAStringSet"), 
      fixed = FALSE)
    forbiddencodon <- sapply(BiocGenerics::relist(start(unlist(matchForbidden)) %% 3==1, 
                                                  matchForbidden), function(i) any(i))
    message("Number of reads with forbidden codons, reverse: ", sum(forbiddencodon))  ## 2
    fq1 <- fq1[!forbiddencodon]
    fq2 <- fq2[!forbiddencodon]
    encodedMutCodonsForward <- encodedMutCodonsForward[!forbiddencodon]
    mutCodons <- mutCodons[!forbiddencodon]
    nbrMutCodonsTot <- nbrMutCodonsTot[!forbiddencodon]
    nbrMutBasesTot <- nbrMutBasesTot[!forbiddencodon]
    nbrMutCodons <- nbrMutCodons[!forbiddencodon]
    nbrMutBases <- nbrMutBases[!forbiddencodon]
    uniqueMutCodons <- uniqueMutCodons[!forbiddencodon]
    splitMutCodons <- as(strsplit(gsub("([^.]{3})", "\\1\\.", 
                                       as.character(mutCodons)), ".", fixed = TRUE), 
                         "CharacterList")
    encodedMutCodonsReverse <- S4Vectors::unstrsplit(BiocGenerics::relist(
      paste0("r", unlist(uniqueMutCodons), unlist(splitMutCodons)), 
      uniqueMutCodons),
      sep = "_")
    mutNames <- paste(encodedMutCodonsForward, encodedMutCodonsReverse, sep = "_")
    
    nbrMutCodonsTot <- nbrMutCodonsTot + nbrMutCodons
    nbrMutBasesTot <- nbrMutBasesTot + nbrMutBases
    
  } else {
    mutNames <- encodedMutCodonsForward
  }
  
  ## Constant sequence filtering
  keep <- rep(TRUE, length(fq1))
  if (!all(Ldef$constantForward == "") && Ldef$constantMaxDistForward > (-1)) {
    constForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 1, 
                                       width = Ldef$constantLengthForward)
    dists <- as.matrix(pwalign::stringDist(
      c(DNAStringSet(structure(Ldef$constantForward, 
                               names = paste0("C", length(Ldef$constantForward)))),
        DNAStringSet(constForward)
      )
    ))
    dists <- dists[-(1:length(Ldef$constantForward)), 1:length(Ldef$constantForward), drop = FALSE]
    keep <- keep & (apply(dists, 1, min) <= Ldef$constantMaxDistForward)
  }
  if (!all(Ldef$constantReverse == "") && Ldef$constantMaxDistReverse > (-1)) {
    constReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 1, 
                                       width = Ldef$constantLengthReverse)
    dists <- as.matrix(pwalign::stringDist(
      c(DNAStringSet(structure(Ldef$constantReverse, 
                               names = paste0("C", length(Ldef$constantReverse)))),
        DNAStringSet(constReverse)
      )
    ))
    dists <- dists[-(1:length(Ldef$constantReverse)), 1:length(Ldef$constantReverse), drop = FALSE]
    keep <- keep & (apply(dists, 1, min) <= Ldef$constantMaxDistReverse)
  }
  message("Number of reads with too many mismatches in constant seq: ", sum(!keep))
  fq1 <- fq1[keep]
  if (!is.null(Ldef$fastqReverse)) {
    fq2 <- fq2[keep]
  }
  mutCodons <- mutCodons[keep]
  uniqueMutCodons <- uniqueMutCodons[keep]
  mutNames <- mutNames[keep]

  if (!is.null(nbrMutBasesTot)) {
    nbrMutBasesTot <- nbrMutBasesTot[keep]
    nbrMutCodonsTot <- nbrMutCodonsTot[keep]
    ## Number of mutated bases and codons
    varForwardFinal <- Biostrings::subseq(
      fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 
        Ldef$constantLengthForward + 1, width = Ldef$variableLengthForward)
    if (!is.null(Ldef$fastqReverse)) {
      varReverseFinal <- Biostrings::subseq(
        fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 
          Ldef$constantLengthReverse + 1, width = Ldef$variableLengthReverse)
    } else {
      varReverseFinal <- ""
    }
    message("Number of mutated codons per retained read: ")
    print(table(nbrMutCodonsTot))
    message("Number of different reads with each number of mismatched codons: ")
    print(table(nbrMutCodonsTot[!duplicated(paste0(varForwardFinal, "_", varReverseFinal))]))
    message("Number of mutated bases per retained read: ")
    print(table(nbrMutBasesTot))
    message("Number of different reads with each number of mismatched bases: ")
    print(table(nbrMutBasesTot[!duplicated(paste0(varForwardFinal, "_", varReverseFinal))]))
    
    ## Number of mutated amino acids
    ## Forward
    p_aa <- Biostrings::translate(DNAStringSet(rep(Ldef$wildTypeForward, length(varForwardFinal))))
    pattern_width_aa <- width(p_aa)
    varForwardFinal_aa <- Biostrings::translate(varForwardFinal)
    subject_width_aa <- width(varForwardFinal_aa)
    unlisted_ans_aa <- which(as.raw(unlist(p_aa)) != as.raw(unlist(varForwardFinal_aa)))
    breakpoints_aa <- cumsum(pattern_width_aa)
    ans_eltlens_aa <- tabulate(findInterval(unlisted_ans_aa - 1L,
                                            breakpoints_aa) + 1L,
                               nbins = length(p_aa))
    skeleton_aa <- IRanges::PartitioningByEnd(cumsum(ans_eltlens_aa))
    aaMismatchesForward <- BiocGenerics::relist(unlisted_ans_aa, skeleton_aa)
    ## Reverse
    if (!is.null(Ldef$fastqReverse)) {
      p_aa <- Biostrings::translate(DNAStringSet(rep(Ldef$wildTypeReverse, length(varReverseFinal))))
      pattern_width_aa <- width(p_aa)
      varReverseFinal_aa <- Biostrings::translate(varReverseFinal)
      subject_width_aa <- width(varReverseFinal_aa)
      unlisted_ans_aa <- which(as.raw(unlist(p_aa)) != as.raw(unlist(varReverseFinal_aa)))
      breakpoints_aa <- cumsum(pattern_width_aa)
      ans_eltlens_aa <- tabulate(findInterval(unlisted_ans_aa - 1L,
                                              breakpoints_aa) + 1L,
                                 nbins = length(p_aa))
      skeleton_aa <- IRanges::PartitioningByEnd(cumsum(ans_eltlens_aa))
      aaMismatchesReverse <- BiocGenerics::relist(unlisted_ans_aa, skeleton_aa)
     
      message("Number of mutated aas per retained read: ")
      print(table(lengths(aaMismatchesForward) + lengths(aaMismatchesReverse)))
      message("Number of different reads with each number of mismatched aas: ")
      print(table((lengths(aaMismatchesForward) + lengths(aaMismatchesReverse))[!duplicated(paste0(varForwardFinal, "_", varReverseFinal))]))
    } else {
      message("Number of mutated aas per retained read: ")
      print(table(lengths(aaMismatchesForward)))
      message("Number of different reads with each number of mismatched aas: ")
      print(table(lengths(aaMismatchesForward)[!duplicated(varForwardFinal)])) 
      
    }
  }

  mutNames <- gsub("^_", "", gsub("_$", "", mutNames))
  mutNames[mutNames == ""] <- "WT"
  message("Number of variants seen twice: ", sum(table(mutNames) == 2))  ## 2
  message("Variants seen twice: ")
  print(which(table(mutNames) == 2))
  
  ## Retained reads
  message("Number of retained reads: ", length(fq1))  ## 279
  
  ## Error statistics
  ## Forward
  if (length(Ldef$constantForward) == 1) {
    constForward <- Biostrings::subseq(fq1, start = Ldef$skipForward + Ldef$umiLengthForward + 1, 
                                       width = Ldef$constantLengthForward)
    p <- DNAStringSet(rep(Ldef$constantForward, length(constForward)))
    pattern_width <- width(p)
    subject_width <- width(constForward)
    
    unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(constForward)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    mismatchPositions <- nucleotideMismatches - offsets
    mismatchQualities <- unlist(as(quality(constForward)[mismatchPositions], "IntegerList"))
    
    unlisted_ans <- which(as.raw(unlist(p)) == as.raw(unlist(constForward)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    matchPositions <- nucleotideMatches - offsets
    matchQualities <- unlist(as(quality(constForward)[matchPositions], "IntegerList"))
    
    message("Match qualities, forward:")
    print(table(matchQualities))
    message("Mismatch qualities, forward:")
    print(table(mismatchQualities))
  }
  
  ## Reverse
  if (!is.null(Ldef$fastqReverse) && length(Ldef$constantReverse) == 1) {
    constReverse <- Biostrings::subseq(fq2, start = Ldef$skipReverse + Ldef$umiLengthReverse + 1,
                                       width = Ldef$constantLengthReverse)
    #constReverse <- Biostrings::reverseComplement(constReverse)
    p <- DNAStringSet(rep(Ldef$constantReverse, length(constReverse)))
    pattern_width <- width(p)
    subject_width <- width(constReverse)
    
    unlisted_ans <- which(as.raw(unlist(p)) != as.raw(unlist(constReverse)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMismatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    mismatchPositions <- nucleotideMismatches - offsets
    mismatchQualities <- unlist(as(quality(constReverse)[mismatchPositions], "IntegerList"))
    
    unlisted_ans <- which(as.raw(unlist(p)) == as.raw(unlist(constReverse)))
    breakpoints <- cumsum(pattern_width)
    ans_eltlens <- tabulate(findInterval(unlisted_ans - 1L,
                                         breakpoints) + 1L,
                            nbins = length(p))
    skeleton <- IRanges::PartitioningByEnd(cumsum(ans_eltlens))
    nucleotideMatches <- BiocGenerics::relist(unlisted_ans, skeleton)
    offsets <- c(0L, breakpoints[-length(breakpoints)])
    matchPositions <- nucleotideMatches - offsets
    matchQualities <- unlist(as(quality(constReverse)[matchPositions], "IntegerList"))
    
    message("Match qualities, reverse:")
    print(table(matchQualities))
    message("Mismatch qualities, reverse:")
    print(table(mismatchQualities))
  }
  
  ## Collapsing
  if (Ldef$wildTypeForward == "" && Ldef$wildTypeReverse == "") {
    umicoll <- list()
    if (!is.null(Ldef$fastqReverse)) {
      fqseq <- paste0(as.character(varForward), "_", as.character(varReverse))
    } else {
      fqseq <- as.character(varForward)
    }
    tbl <- as.data.frame(table(fqseq)) |> dplyr::arrange(desc(Freq), fqseq) |>
      dplyr::mutate(fqseq = as.character(fqseq))
    tbl$umis <- ""
    for (i in seq_len(nrow(tbl))) {
      tbl$umis[i] <- paste(umis[fqseq == tbl$fqseq[i]], collapse = ",")
    }
    collseqs <- collapse_seqs(tbl$fqseq, tbl$umis, tbl$Freq, 
                              maxdist = Ldef$variableCollapseMaxDist,
                              umimaxdist = Ldef$umiCollapseMaxDist,
                              minreads = Ldef$variableCollapseMinReads, 
                              minratio = Ldef$variableCollapseMinRatio)
    message("Number of unique sequences: ", nrow(tbl))
    message("Number of collapsed sequences: ", length(collseqs))
    message("Total UMI count: ", length(unlist(collseqs)))
    return(tbl)
  }
  return(NULL)
}

## ----------------------------------------------------------------------------
## Generate values for comparison in unit tests - trans experiment
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
processReadsTrans(Ldef)

## ----------------------------------------------------------------------------
## Change some parameter values
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 30.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNA",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 25.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
processReadsTrans(Ldef)

## ----------------------------------------------------------------------------
## Only one input file
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
Ldef <- list(
  fastqForward = fqt1, fastqReverse = NULL, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 30.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNA",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 25.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
processReadsTrans(Ldef)

## Test collapsing
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, 
  minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
  revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  primerForward = "",
  primerReverse = "",
  wildTypeForward = "",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".",
  maxNReads = -1, variableCollapseMaxDist = 6, umiCollapseMaxDist = 4, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
processReadsTrans(Ldef)

fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = NULL, 
  mergeForwardReverse = FALSE, 
  minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
  revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  primerForward = "",
  primerReverse = "",
  wildTypeForward = "",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".",
  variableCollapseMaxDist = 10, umiCollapseMaxDist = 5, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  maxNReads = -1, verbose = FALSE
)
processReadsTrans(Ldef)

## don't add anything between the above and below runs
## same as above but with new collapsing approach
Ldef$variableCollapseMaxDist <- 0
Ldef$umiCollapseMaxDist <- 5 
Ldef$variableCollapseMinReads <- 0
Ldef$variableCollapseMinRatio <- 0
v <- processReadsTrans(Ldef)
cseq <- collapse_seqs(v$fqseq, v$umis, v$Freq, maxdist = 10, 
                      umimaxdist = 0, minreads = 0, minratio = 0)
length(cseq)

## Collapse only highly abundant features
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, 
  minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
  revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  primerForward = "",
  primerReverse = "",
  wildTypeForward = "",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".",
  maxNReads = -1, variableCollapseMaxDist = 2, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 2, variableCollapseMinRatio = 0,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
v <- processReadsTrans(Ldef)

## don't add anything between the above and below runs
## same as above but with new collapsing approach
Ldef$variableCollapseMaxDist <- 0
Ldef$umiCollapseMaxDist <- 0 
Ldef$variableCollapseMinReads <- 0
Ldef$variableCollapseMinRatio <- 0
v <- processReadsTrans(Ldef)
cseq <- collapse_seqs(v$fqseq, v$umis, v$Freq, maxdist = 2, 
                      umimaxdist = 0, minreads = 1.5, minratio = 0)
length(cseq)

## Collapse only features with high enough count ratio
fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
## default arguments
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, 
  minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
  revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  primerForward = "",
  primerReverse = "",
  wildTypeForward = "",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".",
  maxNReads = -1, variableCollapseMaxDist = 3, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 1, variableCollapseMinRatio = 1.5,
  constantMaxDistForward = -1, constantMaxDistReverse = -1,
  verbose = FALSE
)
v <- processReadsTrans(Ldef)

## don't add anything between the above and below runs
## same as above but with new collapsing approach
Ldef$variableCollapseMaxDist <- 0
Ldef$umiCollapseMaxDist <- 0 
Ldef$variableCollapseMinReads <- 0
Ldef$variableCollapseMinRatio <- 0
v <- processReadsTrans(Ldef)
cseq <- collapse_seqs(v$fqseq, v$umis, v$Freq, maxdist = 3, 
                      umimaxdist = 0, minreads = 1, minratio = 1.5)
length(cseq)

## filter based on constant sequences
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = 0, constantMaxDistReverse = 0,
  verbose = FALSE
)
processReadsTrans(Ldef)

## 1 mismatch
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = 1, constantMaxDistReverse = 1,
  verbose = FALSE
)
processReadsTrans(Ldef)

## Multiple constant sequences
Ldef <- list(
  fastqForward = fqt1, fastqReverse = fqt2, 
  mergeForwardReverse = FALSE, revComplForward = FALSE, revComplReverse = FALSE,
  skipForward = 1, skipReverse = 1, 
  umiLengthForward = 10, umiLengthReverse = 8, 
  constantLengthForward = 18, constantLengthReverse = 20, 
  variableLengthForward = 96, variableLengthReverse = 96,
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
  constantForward = c("AACCGGAGGAGGGAGCTG", "AACCGGCGGAGGGAGCTG"), 
  constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  variableCollapseMaxDist = 0, umiCollapseMaxDist = 0, 
  variableCollapseMinReads = 0, variableCollapseMinRatio = 0,
  constantMaxDistForward = 0, constantMaxDistReverse = 0,
  verbose = FALSE
)
processReadsTrans(Ldef)

## ----chunk-opts, include = FALSE----------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----setup, message = FALSE---------------------------------------------------
library(SummarizedExperiment)
library(mutscan)

## ----overview-figure, fig.align = 'center', out.width = "100%", fig.cap = "Overview of the `mutscan` functionality. The `digestFastqs()` function processes each sequencing library, possibly consisting of multiple (pairs of) FASTQ files, separately, and generates an output object containing, among other things, the count table and a summary of the filtering steps. The `summarizeExperiment()` function takes one or more of these objects and combines them into a `SummarizedExperiment` object, that can then be used for downstream analysis such as plotting and statistical testing.", echo = FALSE----
knitr::include_graphics("mutscan-overview.001.png")

## ----list-files---------------------------------------------------------------
datadir <- system.file("extdata", package = "mutscan")
dir(datadir)

## ----trans-digest-------------------------------------------------------------
transInput <- digestFastqs(
    fastqForward = file.path(datadir, "transInput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "transInput_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV",
    elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1,
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = FALSE
)

## ----trans-pars---------------------------------------------------------------
transInput$parameters

## ----trans-filter-------------------------------------------------------------
transInput$filterSummary

## ----trans-summary------------------------------------------------------------
head(transInput$summaryTable)

## ----trans-error--------------------------------------------------------------
transInput$errorStatistics[rowSums(transInput$errorStatistics[, -1]) != 0, ]

## ----trans-error-stats--------------------------------------------------------
(propErrorsConstantF <- sum(transInput$errorStatistics$nbrMismatchForward) /
   (nchar(transInput$parameters$constantForward) * 
        transInput$filterSummary$nbrRetained))
(propErrorsConstantR <- sum(transInput$errorStatistics$nbrMismatchReverse) /
   (nchar(transInput$parameters$constantReverse) * 
        transInput$filterSummary$nbrRetained))

## ----cis-digest---------------------------------------------------------------
cisInput <- digestFastqs(
    fastqForward = file.path(datadir, "cisInput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "cisInput_2.fastq.gz"),
    mergeForwardReverse = TRUE, 
    minOverlap = 0, 
    maxOverlap = 0, 
    maxFracMismatchOverlap = 1, 
    greedyOverlap = TRUE, 
    revComplForward = FALSE, 
    revComplReverse = TRUE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV", 
    elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAGTTCATCCTGGCAGC", 
    primerForward = "", 
    primerReverse = "",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = "", 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = TRUE
)

## ----cis-out------------------------------------------------------------------
cisInput$parameters
cisInput$filterSummary
cisInput$errorStatistics[rowSums(cisInput$errorStatistics[, -1]) != 0, ]

## ----cis-summary--------------------------------------------------------------
head(cisInput$summaryTable)

## ----def-primers--------------------------------------------------------------
leu <- c(
    ATF2 = "GATCCTGATGAAAAAAGGAGAAAGTTTTTAGAGCGAAATAGAGCAGCAGCTTCAAGATGCCGACAAAAAAGGAAAGTCTGGGTTCAGTCTTTAGAGAAGAAAGCTGAAGACTTGAGTTCATTAAATGGTCAGCTGCAGAGTGAAGTCACCCTGCTGAGAAATGAAGTGGCACAGCTGAAACAGCTTCTTCTGGCT",
    ATF7 = "GATCCAGATGAGCGACGGCAGCGCTTTCTGGAGCGCAACCGGGCTGCAGCCTCCCGCTGCCGCCAAAAGCGAAAGCTGTGGGTGTCCTCCCTAGAGAAGAAGGCCGAAGAACTCACTTCTCAGAACATTCAGCTGAGTAATGAAGTCACATTACTACGCAATGAGGTGGCCCAGTTGAAACAGCTACTGTTAGCT",
    CREB5 = "GATCCGGACGAGAGGCGGCGGAAATTTCTGGAACGGAACCGGGCAGCTGCCACCCGCTGCAGACAGAAGAGGAAGGTCTGGGTGATGTCATTGGAAAAGAAAGCAGAAGAACTCACCCAGACAAACATGCAGCTTCAGAATGAAGTGTCTATGTTGAAAAATGAGGTGGCCCAGCTGAAACAGTTGTTGTTAACA",
    ATF3 = "GAAGAAGATGAAAGGAAAAAGAGGCGACGAGAAAGAAATAAGATTGCAGCTGCAAAGTGCCGAAACAAGAAGAAGGAGAAGACGGAGTGCCTGCAGAAAGAGTCGGAGAAGCTGGAAAGTGTGAATGCTGAACTGAAGGCTCAGATTGAGGAGCTCAAGAACGAGAAGCAGCATTTGATATACATGCTCAACCTT",
    JDP2 = "GAGGAAGAGGAGCGAAGGAAAAGGCGCCGGGAGAAGAACAAAGTCGCAGCAGCCCGATGCCGGAACAAGAAGAAGGAGCGCACGGAGTTTCTGCAGCGGGAATCCGAGCGGCTGGAACTCATGAACGCAGAGCTGAAGACCCAGATTGAGGAGCTGAAGCAGGAGCGGCAGCAGCTCATCCTGATGCTGAACCGA",
    ATF4 = "GAGAAACTGGATAAGAAGCTGAAAAAAATGGAGCAAAACAAGACAGCAGCCACTAGGTACCGCCAGAAGAAGAGGGCGGAGCAGGAGGCTCTTACTGGTGAGTGCAAAGAGCTGGAAAAGAAGAACGAGGCTCTAAAAGAGAGGGCGGATTCCCTGGCCAAGGAGATCCAGTACCTGAAAGATTTGATAGAAGAG",
    ATF5 = "ACCCGAGGGGACCGCAAGCAAAAGAAGAGAGACCAGAACAAGTCGGCGGCTCTGAGGTACCGCCAGCGGAAGCGGGCAGAGGGTGAGGCCCTGGAGGGCGAGTGCCAGGGGCTGGAGGCACGGAATCGCGAGCTGAAGGAACGGGCAGAGTCCGTGGAGCGCGAGATCCAGTACGTCAAGGACCTGCTCATCGAG",
    CREBZF = "AGTCCCCGGAAGGCGGCGGCGGCCGCTGCCCGCCTTAATCGACTGAAGAAGAAGGAGTACGTGATGGGGCTGGAGAGTCGAGTCCGGGGTCTGGCAGCCGAGAACCAGGAGCTGCGGGCCGAGAATCGGGAGCTGGGCAAACGCGTACAGGCACTGCAGGAGGAGAGTCGCTACCTACGGGCAGTCTTAGCCAAC",
    BATF2 = "CCCAAGGAGCAACAAAGGCAGCTGAAGAAGCAGAAGAACCGGGCAGCCGCCCAGCGAAGCCGGCAGAAGCACACAGACAAGGCAGACGCCCTGCACCAGCAGCACGAGTCTCTGGAAAAAGACAACCTCGCCCTGCGGAAGGAGATCCAGTCCCTGCAGGCCGAGCTGGCGTGGTGGAGCCGGACCCTGCACGTG",
    BATF3 = "GAGGATGATGACAGGAAGGTCCGAAGGAGAGAAAAAAACCGAGTTGCTGCTCAGAGAAGTCGGAAGAAGCAGACCCAGAAGGCTGACAAGCTCCATGAGGAATATGAGAGCCTGGAGCAAGAAAACACCATGCTGCGGAGAGAGATCGGGAAGCTGACAGAGGAGCTGAAGCACCTGACAGAGGCACTGAAGGAG",
    CEBPE = "AAAGATAGCCTTGAGTACCGGCTGAGGCGGGAGCGCAACAACATCGCCGTGCGCAAGAGCCGAGACAAGGCCAAGAGGCGCATTCTGGAGACGCAGCAGAAGGTGCTGGAGTACATGGCAGAGAACGAGCGCCTCCGCAGCCGCGTGGAGCAGCTCACCCAGGAGCTAGACACCCTCCGCAACCTCTTCCGCCAG",
    BACH1 = "CTGGATTGTATCCATGATATTCGAAGAAGAAGTAAAAACAGAATTGCTGCACAGCGCTGTCGCAAGAGAAAACTTGACTGTATACAGAATCTTGAATCAGAAATTGAGAAGCTGCAAAGTGAAAAGGAGAGCTTGTTGAAGGAAAGAGATCACATTTTGTCAACTCTGGGTGAGACAAAGCAGAACCTAACTGGA",
    BACH2 = "TTAGAGTTTATTCATGATGTCCGACGGCGCAGCAAGAACCGCATCGCGGCCCAGCGCTGCCGCAAAAGGAAACTGGACTGTATTCAGAATTTAGAATGTGAAATCCGCAAATTGGTGTGTGAGAAAGAGAAACTGTTGTCAGAGAGGAATCAACTGAAAGCATGCATGGGGGAACTGTTGGACAACTTCTCCTGC",
    NFE2L1 = "CTGAGCCTCATCCGAGACATCCGGCGCCGGGGCAAGAACAAGATGGCGGCGCAGAACTGCCGCAAGCGCAAGCTGGACACCATCCTGAATCTGGAGCGTGATGTGGAGGACCTGCAGCGTGACAAAGCCCGGCTGCTGCGGGAGAAAGTGGAGTTCCTGCGCTCCCTGCGACAGATGAAGCAGAAGGTCCAGAGC",
    NFE2 = "CTAGCGCTAGTCCGGGACATCCGACGACGGGGCAAAAACAAGGTGGCAGCCCAGAACTGCCGCAAGAGGAAGCTGGAAACCATTGTGCAGCTGGAGCGGGAGCTGGAGCGGCTGACCAATGAACGGGAGCGGCTTCTCAGGGCCCGCGGGGAGGCAGACCGGACCCTGGAGGTCATGCGCCAACAGCTGACAGAG",
    NFIL3 = "AAGAAAGATGCTATGTATTGGGAAAAAAGGCGGAAAAATAATGAAGCTGCCAAAAGATCTCGTGAGAAGCGTCGACTGAATGACCTGGTTTTAGAGAACAAACTAATTGCACTGGGAGAAGAAAACGCCACTTTAAAAGCTGAGCTGCTTTCACTAAAATTAAAGTTTGGTTTAATTAGCTCCACAGCATATGCT",
    FOS = "GAAGAAGAAGAGAAAAGGAGAATCCGAAGGGAAAGGAATAAGATGGCTGCAGCCAAATGCCGCAACCGGAGGAGGGAGCTGACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTAGAGTTCATCCTGGCAGCT",
    FOSB = "GAGGAAGAGGAGAAGCGAAGGGTGCGCCGGGAACGAAATAAACTAGCAGCAGCTAAATGCAGGAACCGGCGGAGGGAGCTGACCGACCGACTCCAGGCGGAGACAGATCAGTTGGAGGAAGAAAAAGCAGAGCTGGAGTCGGAGATCGCCGAGCTCCAAAAGGAGAAGGAACGTCTGGAGTTTGTGCTGGTGGCC",
    FOSL1 = "GAGGAAGAGGAGCGCCGCCGAGTAAGGCGCGAGCGGAACAAGCTGGCTGCGGCCAAGTGCAGGAACCGGAGGAAGGAACTGACCGACTTCCTGCAGGCGGAGACTGACAAACTGGAAGATGAGAAATCTGGGCTGCAGCGAGAGATTGAGGAGCTGCAGAAGCAGAAGGAGCGCCTAGAGCTGGTGCTGGAAGCC",
    FOSL2 = "GAAGAGGAGGAGAAGCGTCGCATCCGGCGGGAGAGGAACAAGCTGGCTGCAGCCAAGTGCCGGAACCGACGCCGGGAGCTGACAGAGAAGCTGCAGGCGGAGACAGAGGAGCTGGAGGAGGAGAAGTCAGGCCTGCAGAAGGAGATTGCTGAGCTGCAGAAGGAGAAGGAGAAGCTGGAGTTCATGTTGGTGGCT",
    MAFB = "GTGATCCGCCTGAAGCAGAAGCGGCGGACCCTGAAGAACCGGGGCTACGCCCAGTCTTGCAGGTATAAACGCGTCCAGCAGAAGCACCACCTGGAGAATGAGAAGACGCAGCTCATTCAGCAGGTGGAGCAGCTTAAGCAGGAGGTGTCCCGGCTGGCCCGCGAGAGAGACGCCTACAAGGTCAAGTGCGAGAAA",
    JUN = "CAGGAGCGGATCAAGGCGGAGAGGAAGCGCATGAGGAACCGCATCGCTGCCTCCAAGTGCCGAAAAAGGAAGCTGGAGAGAATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTTAAACAGAAAGTCATGAAC",
    JUNB = "CAAGAGCGCATCAAAGTGGAGCGCAAGCGGCTGCGGAACCGGCTGGCGGCCACCAAGTGCCGGAAGCGGAAGCTGGAGCGCATCGCGCGCCTGGAGGACAAGGTGAAGACGCTCAAGGCCGAGAACGCGGGGCTGTCGAGTACCGCCGGCCTCCTCCGGGAGCAGGTGGCCCAGCTCAAACAGAAGGTCATGACC",
    JUND = "CAGGAGCGCATCAAGGCGGAGCGCAAGCGGCTGCGCAACCGCATCGCCGCCTCCAAGTGCCGCAAGCGCAAGCTGGAGCGCATCTCGCGCCTGGAAGAGAAAGTGAAGACCCTCAAGAGTCAGAACACGGAGCTGGCGTCCACGGCGAGCCTGCTGCGCGAGCAGGTGGCGCAGCTCAAGCAGAAAGTCCTCAGC",
    CREB3 = "GAACAAATTCTGAAACGTGTGCGGAGGAAGATTCGAAATAAAAGATCTGCTCAAGAGAGCCGCAGGAAAAAGAAGGTGTATGTTGGGGGTTTAGAGAGCAGGGTCTTGAAATACACAGCCCAGAATATGGAGCTTCAGAACAAAGTACAGCTTCTGGAGGAACAGAATTTGTCCCTTCTAGATCAACTGAGGAAA",
    HLF = "CTGAAGGATGACAAGTACTGGGCAAGGCGCAGAAAGAACAACATGGCAGCCAAGCGCTCCCGCGACGCCCGGAGGCTGAAAGAGAACCAGATCGCCATCCGGGCCTCGTTCCTGGAGAAGGAGAACTCGGCCCTCCGCCAGGAGGTGGCTGACTTGAGGAAGGAGCTGGGCAAATGCAAGAACATACTTGCCAAG",
    MAFG = "ATCGTCCAGCTGAAGCAGCGCCGGCGCACGCTCAAGAACCGCGGCTACGCTGCCAGCTGCCGCGTGAAGCGGGTGACGCAGAAGGAGGAGCTGGAGAAGCAGAAGGCGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCCTCAGAGAACGCCAGCATGAAGCTGGAGCTCGACGCGCTGCGCTCCAAGTACGAGGCG",
    MAFK = "GTGACCCGCCTGAAGCAGCGTCGGCGCACACTCAAGAACCGCGGCTACGCGGCCAGCTGCCGCATCAAGCGGGTGACGCAGAAGGAGGAGCTGGAGCGGCAGCGCGTGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCGCGTGAGAACAGCAGCATGCGGCTGGAGCTGGACGCCCTGCGCTCCAAGTACGAGGCG",
    XBP1 = "AGCCCCGAGGAGAAGGCGCTGAGGAGGAAACTGAAAAACAGAGTAGCAGCTCAGACTGCCAGAGATCGAAAGAAGGCTCGAATGAGTGAGCTGGAACAGCAAGTGGTAGATTTAGAAGAAGAGAACCAAAAACTTTTGCTAGAAAATCAGCTTTTACGAGAGAAAACTCATGGCCTTGTAGTTGAGAACCAGGAG",
    ATF6 = "ATTGCTGTGCTAAGGAGACAGCAACGTATGATAAAAAATCGAGAATCCGCTTGTCAGTCTCGCAAGAAGAAGAAAGAATATATGCTAGGGTTAGAGGCGAGATTAAAGGCTGCCCTCTCAGAAAACGAGCAACTGAAGAAAGAAAATGGAACACTGAAGCGGCAGCTGGATGAAGTTGTGTCAGAGAACCAGAGG",
    ATF6B = "GCAAAGCTGCTGAAGCGGCAGCAGCGAATGATCAAGAACCGGGAGTCAGCCTGCCAGTCCCGGAGAAAGAAGAAAGAGTATCTGCAGGGACTGGAGGCTCGGCTGCAAGCAGTACTGGCTGACAACCAGCAGCTCCGCCGAGAGAATGCTGCCCTCCGGCGGCGGCTGGAGGCCCTGCTGGCTGAAAACAGCGAG",
    CEBPA = "AAGAACAGCAACGAGTACCGGGTGCGGCGCGAGCGCAACAACATCGCGGTGCGCAAGAGCCGCGACAAGGCCAAGCAGCGCAACGTGGAGACGCAGCAGAAGGTGCTGGAGCTGACCAGTGACAATGACCGCCTGCGCAAGCGGGTGGAACAGCTGAGCCGCGAACTGGACACGCTGCGGGGCATCTTCCGCCAG",
    CEBPB = "AAGCACAGCGACGAGTACAAGATCCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGATGCGCAACCTGGAGACGCAGCACAAGGTCCTGGAGCTCACGGCCGAGAACGAGCGGCTGCAGAAGAAGGTGGAGCAGCTGTCGCGCGAGCTCAGCACCCTGCGGAACTTGTTCAAGCAG",
    CEBPD = "CGCGGCAGCCCCGAGTACCGGCAGCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGCGGCGCAACCAGGAGATGCAGCAGAAGTTGGTGGAGCTGTCGGCTGAGAACGAGAAGCTGCACCAGCGCGTGGAGCAGCTCACGCGGGACCTGGCCGGCCTCCGGCAGTTCTTCAAGCAG",
    CEBPG = "CGAAACAGTGACGAGTATCGGCAACGCCGAGAGAGGAACAACATGGCTGTGAAAAAGAGCCGGTTGAAAAGCAAGCAGAAAGCACAAGACACACTGCAGAGAGTCAATCAGCTCAAAGAAGAGAATGAACGGTTGGAAGCAAAAATCAAATTGCTGACCAAGGAATTAAGTGTACTCAAAGATTTGTTTCTTGAG",
    CREB1 = "GAAGCAGCACGAAAGAGAGAGGTCCGTCTAATGAAGAACAGGGAAGCAGCTCGAGAGTGTCGTAGAAAGAAGAAAGAATATGTGAAATGTTTAGAAAACAGAGTGGCAGTGCTTGAAAATCAAAACAAGACATTGATTGAGGAGCTAAAAGCACTTAAGGACCTTTACTGCCACAAATCAGAT",
    CREB3L1 = "GAGAAGGCCTTGAAGAGAGTCCGGAGGAAAATCAAGAACAAGATCTCAGCCCAGGAGAGCCGTCGTAAGAAGAAGGAGTATGTGGAGTGTCTAGAAAAGAAGGTGGAGACATTTACATCTGAGAACAATGAACTGTGGAAGAAGGTGGAGACCCTGGAGAATGCCAACAGGACCCTGCTCCAGCAGCTGCAGAAA",
    CREB3L2 = "GAGAAGGCCCTGAAGAAAATTCGGAGGAAGATCAAGAATAAGATTTCTGCTCAGGAAAGTAGGAGAAAGAAGAAAGAATACATGGACAGCCTGGAGAAAAAAGTGGAGTCTTGTTCAACTGAGAACTTGGAGCTTCGGAAGAAGGTAGAGGTTCTAGAGAACACTAATAGGACTCTCCTTCAGCAACTCCAGAAG",
    CREB3L3 = "GAGCGAGTGCTGAAAAAAATCCGCCGGAAAATCCGGAACAAGCAGTCGGCGCAAGAAAGCAGGAAGAAGAAGAAGGAATATATCGATGGCCTGGAGACTCGGATGTCAGCTTGCACTGCTCAGAATCAGGAGTTACAGAGGAAAGTCTTGCATCTCGAGAAGCAAAACCTGTCCCTCTTGGAGCAACTGAAGAAA",
    CREB3L4 = "GAGAGGGTCCTCAAGAAGGTCAGGAGGAAAATCCGTAACAAGCAGTCAGCTCAGGACAGTCGGCGGCGGAAGAAGGAGTACATTGATGGGCTGGAGAGCAGGGTGGCAGCCTGTTCTGCACAGAACCAAGAATTACAGAAAAAAGTCCAGGAGCTGGAGAGGCACAACATCTCCTTGGTAGCTCAGCTCCGCCAG",
    CREBL2 = "CCAGCCAAAATTGACTTGAAAGCAAAACTTGAGAGGAGCCGGCAGAGTGCAAGAGAATGCCGAGCCCGAAAAAAGCTGAGATATCAGTATTTGGAAGAGTTGGTATCCAGTCGAGAAAGAGCTATATGTGCCCTCAGAGAGGAACTGGAAATGTACAAGCAGTGGTGCATGGCAATGGACCAAGGAAAAATCCCT",
    CREBRF = "CCCTTAACAGCCCGACCAAGGTCAAGGAAGGAAAAAAATAAGCTGGCTTCCAGAGCTTGTCGGTTAAAGAAGAAAGCCCAGTATGAAGCTAATAAAGTGAAATTATGGGGCCTCAACACAGAATATGATAATTTATTGTTTGTAATCAACTCCATCAAGCAAGAGATTGTAAACCGGGTACAGAATCCAAGAGAT",
    DBP = "CAGAAGGATGAGAAATACTGGAGCCGGCGGTACAAGAACAACGAGGCAGCCAAGCGGTCCCGTGACGCCCGGCGGCTCAAGGAGAACCAGATATCGGTGCGGGCGGCCTTCCTGGAGAAGGAGAACGCCCTGCTGCGGCAGGAAGTTGTGGCCGTGCGCCAGGAGCTGTCCCACTACCGCGCCGTGCTGTCCCGA",
    NFE2L2 = "CTTGCATTAATTCGGGATATACGTAGGAGGGGTAAGAATAAAGTGGCTGCTCAGAATTGCAGAAAAAGAAAACTGGAAAATATAGTAGAACTAGAGCAAGATTTAGATCATTTGAAAGATGAAAAAGAAAAATTGCTCAAAGAAAAAGGAGAAAATGACAAAAGCCTTCACCTACTGAAAAAACAACTCAGCACC",
    NFE2L3 = "GTCTCACTTATCCGTGACATCAGACGAAGAGGGAAAAATAAAGTTGCTGCGCAGAACTGTCGTAAACGCAAATTGGACATAATTTTGAATTTAGAAGATGATGTATGTAACTTGCAAGCAAAGAAGGAAACTCTTAAGAGAGAGCAAGCACAATGTAACAAAGCTATTAACATAATGAAACAGAAACTGCATGAC",
    TEF = "CAGAAGGATGAAAAGTACTGGACAAGACGCAAGAAGAACAACGTGGCAGCTAAACGGTCACGGGATGCCCGGCGCCTGAAAGAGAATCAGATCACCATCCGGGCAGCCTTCCTGGAGAAGGAGAACACAGCCCTGCGGACGGAGGTGGCCGAGCTACGCAAGGAGGTGGGCAAGTGCAAGACCATCGTGTCCAAG"
)

## ----primers-digest-----------------------------------------------------------
leujunt0 <- digestFastqs(
    fastqForward = file.path(datadir, "leujunt0_1.fastq.gz"),
    fastqReverse = file.path(datadir, "leujunt0_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    elementsForward = "SPV", 
    elementsReverse = "SPVS",
    elementLengthsForward = c(-1, 19, -1),
    elementLengthsReverse = c(-1, 20, 96, -1),
    constantForward = "", 
    constantReverse = "", 
    adapterForward = "",
    adapterReverse = "", 
    primerForward = "GTCAGGTGGAGGCGGATCC", 
    primerReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20, 
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = leu, 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 0, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "",
    forbiddenMutatedCodonsReverse = "NNW",
    mutatedPhredMinForward = 0.0, 
    mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".", 
    verbose = TRUE
)

## ----primers-out--------------------------------------------------------------
leujunt0$parameters
leujunt0$filterSummary

## ----primers-summary----------------------------------------------------------
head(leujunt0$summaryTable)

## ----combine-digest-----------------------------------------------------------
transOutput <- digestFastqs(
    fastqForward = file.path(datadir, "transOutput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "transOutput_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV", 
    elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    primerForward = "", 
    primerReverse = "",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = FALSE
)

## Generate SummarizedExperiment object
se <- summarizeExperiment(
    x = list(sample1 = transInput,
             sample2 = transOutput),
    coldata = data.frame(Name = c("sample1", "sample2"),
                         Condition = c("input", "output"),
                         Replicate = c("R1", "R1"),
                         OD = c(0.05, 1.5))
)

## The SummarizedExperiment contains a count matrix, and annotations 
## for samples and variants in the `colData` and `rowData`, 
## respectively.
head(assay(se, "counts"))
Matrix::colSums(assay(se, "counts"))
head(rowData(se))
colData(se)

## Count type (reads or UMIs)
metadata(se)$countType

## ----collapse-aa--------------------------------------------------------------
se_collapsed <- collapseMutantsByAA(se)
head(assay(se_collapsed, "counts"))
Matrix::colSums(assay(se_collapsed, "counts"))
colData(se_collapsed)

## ----read-se------------------------------------------------------------------
se <- readRDS(file.path(datadir, "GSE102901_cis_se.rds"))

## ----plot-filter-reads, fig.height=10, fig.width=8----------------------------
plotFiltering(se, valueType = "reads", onlyActiveFilters = TRUE, 
              plotType = "remaining", facetBy = "sample", numberSize = 3)

## ----plot-filter-frac, fig.height=10, fig.width=8-----------------------------
plotFiltering(se, valueType = "fractions", onlyActiveFilters = TRUE,
              plotType = "filtered", facetBy = "step", numberSize = 3)

## ----plot-pairs, fig.height=7-------------------------------------------------
plotPairs(se, selAssay = "counts")

## ----plot-more----------------------------------------------------------------
plotTotals(se, selAssay = "counts")
plotDistributions(se, selAssay = "counts", plotType = "density", 
                  pseudocount = 1)

## ----make-qc-report, eval=FALSE-----------------------------------------------
# generateQCReport(se, outFile = tempfile(fileext = ".html"))

## ----calc-ppi-----------------------------------------------------------------
se_collapsed <- collapseMutantsByAA(se)
ppis <- calculateFitnessScore(
    se = se_collapsed, pairingCol = "Replicate", 
    ODCols = c("OD1", "OD2"),
    comparison = c("Condition", "cis_output", "cis_input"),
    WTrows = "f.0.WT"
)
head(ppis[order(abs(rowMeans(ppis)), decreasing = TRUE), , drop = FALSE])

## The PPI score for the WT sequence is 1, by construction
ppis["f.0.WT", , drop = FALSE]

## ----def-mm-------------------------------------------------------------------
model.matrix(~ Replicate + Condition,
             data = colData(se_collapsed))

## ----calc-rel-fc--------------------------------------------------------------
## edgeR
edger_scores <- calculateRelativeFC(
    se = se_collapsed,
    design = model.matrix(~ Replicate + Condition,
                          data = colData(se_collapsed)),
    coef = "Conditioncis_output", pseudocount = 1, WTrows = "f.0.WT",
    method = "edgeR")
head(edger_scores[order(edger_scores$PValue), , drop = FALSE])
## As before, the WT sequence has a logFC close to 0, by construction
edger_scores["f.0.WT", , drop = FALSE]

## limma
limma_scores <- calculateRelativeFC(
    se = se_collapsed,
    design = model.matrix(~ Replicate + Condition,
                          data = colData(se_collapsed)),
    coef = "Conditioncis_output", pseudocount = 1, WTrows = "f.0.WT",
    method = "limma")
head(limma_scores[order(limma_scores$P.Value), , drop = FALSE])
## As before, the WT sequence has a logFC close to 0, by construction
limma_scores["f.0.WT", , drop = FALSE]

## ----plot-rel-fc--------------------------------------------------------------
plotMeanDiff(edger_scores, pointSize = "large")
plotVolcano(edger_scores, pointSize = "large")

## ----session-info-------------------------------------------------------------
sessionInfo()

## Generate test data for linkMultipleVariants
## Construct: C(6) - V(24) - C(10) - V(30) - C(10) - V(40) - C(6)
## Forward read length 90
## Reverse read length 70

nReads <- 100
status <- rep("OK", nReads)

## First constant region
c1 <- rep("ACGTGA", nReads)
set.seed(1)
i1 <- sample(seq_len(nReads), 4)  ## reads with mutation
c1[i1] <- "ACTTGA"
status[i1] <- gsub("OK,", "", paste0(status[i1], ",C1mut"))

## Second constant region
c2 <- rep("TTGATCCCGG", nReads)
set.seed(2)
i2 <- sample(seq_len(nReads), 5)  ## reads with mutation
c2[i2] <- "TTGATCGCGG"
status[i2] <- gsub("OK,", "", paste0(status[i2], ",C2mut"))

## Third constant region
c3 <- rep("TAAATACCGG", nReads)
set.seed(3)
i3 <- sample(seq_len(nReads), 5)  ## reads with mutation
c3[i3] <- "TAGATACCGG"
status[i3] <- gsub("OK,", "", paste0(status[i3], ",C3mut"))

## Fourth constant region
c4 <- rep("ACCTGA", nReads)
set.seed(4)
i4 <- sample(seq_len(nReads), 4)  ## reads with mutation
c4[i4] <- "ACCCGA"
status[i4] <- gsub("OK,", "", paste0(status[i4], ",C4mut"))

## First variable region (barcodes)
tmpstatus <- rep("", nReads)
set.seed(1)
v1true <- rep(c(paste(sample(c("A", "C", "G", "T"), 24, replace = TRUE), collapse = ""),
                paste(sample(c("A", "C", "G", "T"), 24, replace = TRUE), collapse = ""),
                paste(sample(c("A", "C", "G", "T"), 24, replace = TRUE), collapse = "")),
              c(floor(nReads/3), floor(nReads/3), nReads - 2 * floor(nReads/3)))
v1 <- v1true
i1 <- sample(seq_len(nReads), 10)  ## reads with mutation
for (i in i1) {
    tmp <- strsplit(v1[i], "")[[1]]
    i0 <- sample(seq_along(tmp), 1)
    tmp[i0] <- ifelse(tmp[i0] == "A", "T", "A")
    v1[i] <- paste(tmp, collapse = "")
    tmpstatus[i] <- ",V1mut"
}
i2 <- sample(seq_len(nReads), 4)  ## reads with N
for (i in i2) {
    tmp <- strsplit(v1[i], "")[[1]]
    tmp[sample(seq_along(tmp), 1)] <- "N"
    v1[i] <- paste(tmp, collapse = "")
    tmpstatus[i] <- paste0(tmpstatus[i], ",V1N")
}
ordr <- sample(seq_along(v1true), length(v1true))
v1true <- v1true[ordr]
v1 <- v1[ordr]
tmpstatus <- tmpstatus[ordr]
status <- gsub("OK,", "", paste0(status, tmpstatus))

## Second variable region (multiple WT sequences)
tmpstatus <- rep("", nReads)
set.seed(2)
V2v1 <- paste(sample(c("A", "C", "G", "T"), 30, replace = TRUE), collapse = "")
V2v2 <- paste(sample(c("A", "C", "G", "T"), 30, replace = TRUE), collapse = "")
V2v3 <- paste(sample(c("A", "C", "G", "T"), 30, replace = TRUE), collapse = "")
v2true <- rep(c(V2v1, V2v2, V2v3),
              c(floor(nReads/3), floor(nReads/3), nReads - 2 * floor(nReads/3)))
n2true <- rep(c("V2v1", "V2v2", "V2v3"),
              c(floor(nReads/3), floor(nReads/3), nReads - 2 * floor(nReads/3)))
v2 <- v2true
i1 <- sample(seq_len(nReads), 20)   ## reads with mutation
for (i in i1) {
    tmp <- strsplit(v2[i], "")[[1]]
    i0 <- sample(seq_along(tmp), 1)
    repl <- ifelse(tmp[i0] == "A", "T", "A")
    tmp[i0] <- repl
    v2[i] <- paste(tmp, collapse = "")
    tmpstatus[i] <- paste0(",", n2true[i], ".", i0, ".", repl)
}
i3 <- sample(seq_len(nReads), 3)   ## reads with deletion
## don't modify the reads here, since that won't generate an invalid overlap 
## (the modification will be in both the fwd and rev reads). Just record
## which reads to modify
for (i in i3) {
    tmpstatus[i] <- paste0(tmpstatus[i], ",V2del")
}
ordr <- sample(seq_along(v2true), length(v2true))
v2true <- v2true[ordr]
n2true <- n2true[ordr]
v2 <- v2[ordr]
tmpstatus <- tmpstatus[ordr]
status <- gsub("OK,", "", paste0(status, tmpstatus))

## Third variable region (one WT sequence)
tmpstatus <- rep("", nReads)
set.seed(3)
V3 <- paste(sample(c("A", "C", "G", "T"), 40, replace = TRUE), collapse = "")
v3true <- rep(V3, nReads)
v3 <- v3true
i1 <- sample(seq_len(nReads), 30)   ## reads with mutation
for (i in i1) {
    tmp <- strsplit(v3[i], "")[[1]]
    i0 <- sample(seq_along(tmp), 1)
    repl <- ifelse(tmp[i0] == "A", "T", "A")
    tmp[i0] <- repl
    v3[i] <- paste(tmp, collapse = "")
    tmpstatus[i] <- paste0(",V3.", i0, ".", repl)
}
ordr <- sample(seq_along(v3true), length(v3true))
v3true <- v3true[ordr]
v3 <- v3[ordr]
tmpstatus <- tmpstatus[ordr]
status <- gsub("OK,", "", paste0(status, tmpstatus))

## Final reads
reads <- paste0(c1, v1, c2, v2, c3, v3, c4)
qualities <- rep(paste(rep("I", 126), collapse = ""), nReads)

readsR1 <- substr(reads, start = 1, stop = 90)
readsR2 <- substr(reads, start = 57, stop = 126)
qualitiesR1 <- substr(qualities, start = 1, stop = 90)
qualitiesR2 <- substr(qualities, start = 57, stop = 126)

## Add the deletion to the forward reads
idx <- grep("V2del", status)
for (i in idx) {
    tmp <- strsplit(readsR1[i], "")[[1]]
    tmp[41] <- ""
    tmp <- c(tmp, substr(reads[i], start = 91, stop = 91))
    tmp <- paste(tmp, collapse = "")
    readsR1[i] <- tmp
}


fastqR1 <- Biostrings::QualityScaledDNAStringSet(
    x = Biostrings::DNAStringSet(
        x = structure(readsR1, names = paste0("R", seq_len(nReads)))
    ), 
    quality = Biostrings::PhredQuality(
        Biostrings::BStringSet(
            x = structure(qualitiesR1, names = paste0("R", seq_len(nReads)))
        )
    )
)
fastqR2 <- Biostrings::QualityScaledDNAStringSet(
    x = Biostrings::reverseComplement(
        Biostrings::DNAStringSet(
            x = structure(readsR2, names = paste0("R", seq_len(nReads)))
        )
    ), 
    quality = Biostrings::PhredQuality(
        Biostrings::BStringSet(
            x = structure(qualitiesR2, names = paste0("R", seq_len(nReads)))
        )
    )
)

Biostrings::writeQualityScaledXStringSet(fastqR1, filepath = "multipleVariableRegions_R1.fastq.gz", compress = TRUE)
Biostrings::writeQualityScaledXStringSet(fastqR2, filepath = "multipleVariableRegions_R2.fastq.gz", compress = TRUE)

## True sequences
barcode <- v1true
V2var <- vapply(strsplit(status, ","), function(x) {
    i <- grep("V2.*\\.", x)
    if (length(i) > 0) {
        x[i]
    } else {
        NA_character_
    }}, NA_character_)
V2var[is.na(V2var)] <- paste0(n2true[is.na(V2var)], ".0.WT")
V3var <- vapply(strsplit(status, ","), function(x) {
    i <- grep("V3.*\\.", x)
    if (length(i) > 0) {
        x[i]
    } else {
        NA_character_
    }}, NA_character_)
V3var[is.na(V3var)] <- "V3.0.WT"
WTV2 <- c(V2v1 = V2v1, V2v2 = V2v2, V2v3 = V2v3)
WTV3 <- c(V3 = V3)
constFwd <- "ACGTGATTGATCCCGGTAAATACCGG"
constRev <- "TAAATACCGGACCTGA"

truth <- data.frame(read = paste0("R", seq_len(nReads)),
                    trueBarcode = barcode,
                    trueV2 = V2var,
                    trueV3 = V3var,
                    obsBarcode = v1,
                    status = status)

saveRDS(list(truth = truth, WTV2 = WTV2, WTV3 = WTV3,
             constFwd = constFwd, constRev = constRev),
        "multipleVariableRegions_truth.rds")

#' Plot the column totals of a selected assay
#' 
#' @param se A \code{SummarizedExperiment} object, typically generated by 
#'     \code{summarizeExperiment()}.
#' @param selAssay Character scalar specifying the assay in \code{se} to 
#'     use for the plotting.
#' @param groupBy Character scalar indicating a column in 
#'     \code{rowData(se)} to group the features by before calculating the 
#'     column sums.
#' 
#' @export
#' @author Charlotte Soneson
#' 
#' @return A ggplot object.
#' 
#' @importFrom ggplot2 ggplot theme_minimal theme element_text labs 
#'     geom_bar scale_fill_discrete aes
#' @importFrom SummarizedExperiment assay rowData assayNames
#' @importFrom rlang .data
#' 
#' @examples 
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
#'                           package = "mutscan"))[1:200, ]
#' plotTotals(se)
#' 
plotTotals <- function(se, selAssay = "counts", groupBy = NULL) {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = selAssay, type = "character",
                  validValues = assayNames(se))
    if (!is.null(groupBy)) {
        .assertScalar(
            x = groupBy, type = "character",
            validValues = colnames(rowData(se)))
    }
    
    ## Extract the assay matrix and reformat relevant parts of it for plotting
    selAssayMat <- assay(se, selAssay)
    if (!is.null(groupBy)) {
        df <- do.call(
            rbind, 
            lapply(unique(rowData(se)[[groupBy]]), 
                   function(rn) {
                       data.frame(
                           names = colnames(se),
                           category = rn,
                           total = colSums(
                               selAssayMat[rowData(se)[[groupBy]] == rn, ])
                       )
                   })
        )
    } else {
        df <- data.frame(names = colnames(se),
                         category = "all",
                         total = colSums(selAssayMat))
    }
    
    gg <- ggplot(df, aes(x = .data$names, y = .data$total)) + 
        theme_minimal() + 
        theme(
            axis.text.x = element_text(angle = 90, hjust = 1, 
                                       vjust = 0.5, size = 12),
            axis.text.y = element_text(size = 12),
            axis.title = element_text(size = 14)) + 
        labs(x = "", y = paste0("Total (", selAssay, ")"))
    if (!is.null(groupBy)) {
        gg <- gg + 
            geom_bar(stat = "identity", position = "stack", 
                     aes(fill = .data$category)) + 
            scale_fill_discrete(name = groupBy)
    } else {
        gg <- gg + 
            geom_bar(stat = "identity", position = "stack")
    }
    
    gg
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:1000, 1:3]
SummarizedExperiment::rowData(se)$categ <- sample(LETTERS[1:3], nrow(se), replace = TRUE)

test_that("plotTotals fails with incorrect arguments", {
    expect_error(plotTotals(se = 1))
    expect_error(plotTotals(se = "x"))
    expect_error(plotTotals(se = matrix(1:6), 2, 3))
    
    expect_error(plotTotals(se = se, selAssay = 1))
    expect_error(plotTotals(se = se, selAssay = c("counts", "counts")))
    expect_error(plotTotals(se = se, selAssay = TRUE))
    expect_error(plotTotals(se = se, selAssay = "wrong"))
    
    expect_error(plotTotals(se = se, groupBy = 1))
    expect_error(plotTotals(se = se, groupBy = "wrong"))
    expect_error(plotTotals(se = se, groupBy = c("Name", "Condition")))
})

test_that("plotTotals works as expected", {
    ## Defaults
    expect_true(ggplot2::is_ggplot(plotTotals(se = se, selAssay = "counts",
                                              groupBy = NULL)))

    ## Group by column
    expect_true(ggplot2::is_ggplot(plotTotals(se = se, selAssay = "counts",
                                              groupBy = "categ")))
    
})

#' Relabel the positions of mutations in the designated ID
#' 
#' @param se SummarizedExperiment object, with row names of the form
#'     XX\{.\}AA\{.\}NNN, where XX is the name of the reference sequence, 
#'     AA is the position of the mutated codon, and NNN is the mutated codon 
#'     or amino acid. \{.\} is the delimiter, to be specified in the 
#'     \code{mutNameDelimiter} argument. For rows corresponding to sequences 
#'     with multiple mutated codons, the row names contain multiple names of 
#'     the form above in a single string, separated by "_".
#' @param conversionTable \code{data.frame} with at least three columns:
#'     \itemize{
#'     \item seqname The reference sequence name (should match XX in the 
#'     mutation name)
#'     \item position The codon position (should match AA in the mutation name)
#'     \item name The new name for the codon (will replace AA in the mutation 
#'     name, if the reference sequence matches seqname)
#'     } 
#' @param mutNameDelimiter The delimiter used in the mutation name 
#'     (\{.\} above).
#' 
#' @author Charlotte Soneson
#' @export
#' 
#' @return A SummarizedExperiment object with modified row names.
#' 
#' @importFrom S4Vectors unstrsplit
#' @importFrom utils relist
#' @importFrom BiocGenerics rownames rownames<-
#' 
#' @examples
#' x <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                          package = "mutscan"))
#' conversionTable <- data.frame(seqname = "f", position = 0:32) 
#' conversionTable$name = paste0((conversionTable$position - 1) %/% 7 + 1, 
#'                               c("", rep(letters[1:7], 6))[1:33])
#' out <- relabelMutPositions(x, conversionTable)
#' 
relabelMutPositions <- function(se, conversionTable, mutNameDelimiter = ".") {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertVector(x = colnames(conversionTable), type = "character")
    .assertScalar(x = mutNameDelimiter, type = "character")
    stopifnot(all(c("position", "name", "seqname") %in% 
                      colnames(conversionTable)))
    
    conversionTable$position <- as.character(conversionTable$position)
    spl <- strsplit(rownames(se), "_")
    unl <- unlist(spl)
    unl <- lapply(strsplit(unl, mutNameDelimiter, fixed = TRUE), 
                  function(w) {
                      idx <- which(conversionTable$seqname == w[1] & 
                                       conversionTable$position == w[2])
                      if (length(idx) > 0) {
                          w[2] <- conversionTable$name[idx]
                      }
                      w
                  })
    unl <- unstrsplit(unl, sep = mutNameDelimiter)
    spl <- relist(unl, skeleton = spl)
    spl <- unstrsplit(spl, "_")
    rownames(se) <- spl
    se
}

test_that("relabeling mutants works", {
    se <- SummarizedExperiment::SummarizedExperiment(
        assays = list(exprs = matrix(rnorm(14), nrow = 7))
    )
    rownames(se) <- c("f.0.WT", "f.1.CGG_r.2.AGT", "f.1.AGT_f.3.GTA",
                      "f.2.G", "f.3.G_r.2.A", "f.3.G_r.0.WT",
                      "f.2.GGT_f.3.AGT_r.1.GAG")
    colnames(se) <- c("sample1", "sample2")
    
    convTable <- data.frame(seqname = c("f", "f", "f", "f", "r", "r", "r"),
                            position = c(0, 1, 2, 3, 0, 1, 2),
                            name = LETTERS[seq_len(7)])
    
    ## Fails with the wrong arguments
    expect_error(relabelMutPositions(se = 1, conversionTable = convTable,
                                     mutNameDelimiter = "."),
                 "'se' must be of class 'SummarizedExperiment'")
    expect_error(relabelMutPositions(se = se, conversionTable = 1,
                                     mutNameDelimiter = "."),
                 "'colnamesconversionTable' must not be NULL")
    expect_error(relabelMutPositions(se = se, conversionTable = convTable[, -1],
                                     mutNameDelimiter = "."),
                 "%in% colnames(conversionTable)) is not TRUE", fixed = TRUE)
    expect_error(relabelMutPositions(se = se, conversionTable = convTable[, -2],
                                     mutNameDelimiter = "."),
                 "%in% colnames(conversionTable)) is not TRUE", fixed = TRUE)
    expect_error(relabelMutPositions(se = se, conversionTable = convTable[, -3],
                                     mutNameDelimiter = "."),
                 "%in% colnames(conversionTable)) is not TRUE", fixed = TRUE)
    expect_error(relabelMutPositions(se = se, conversionTable = convTable,
                                     mutNameDelimiter = 1),
                 "'mutNameDelimiter' must be of class 'character'")
    expect_error(relabelMutPositions(se = se, conversionTable = convTable,
                                     mutNameDelimiter = c(".", "_")),
                 "'mutNameDelimiter' must have length 1")
    
    ## Works with the right arguments
    out <- relabelMutPositions(se, convTable, ".")
    expect_equal(rownames(out),
                 c("f.A.WT", "f.B.CGG_r.G.AGT", "f.B.AGT_f.D.GTA",
                   "f.C.G", "f.D.G_r.G.A", "f.D.G_r.E.WT", 
                   "f.C.GGT_f.D.AGT_r.F.GAG"))
    expect_equal(colnames(out), colnames(se))
    
    ## Missing entries in conversion table
    convTable <- convTable[c(2, 4, 6, 7), ]
    
    out <- relabelMutPositions(se, convTable, ".")
    expect_equal(rownames(out),
                 c("f.0.WT", "f.B.CGG_r.G.AGT", "f.B.AGT_f.D.GTA",
                   "f.2.G", "f.D.G_r.G.A", "f.D.G_r.0.WT", 
                   "f.2.GGT_f.D.AGT_r.F.GAG"))
    expect_equal(colnames(out), colnames(se))
})

#' Process an experiment with multiple variable sequences
#'
#' This function enables the processing of data sets with multiple
#' variable sequences, which should potentially be handled in different
#' ways. For example, a barcode association experiment
#' with two variable sequences (the barcode and the biological variant)
#' that need to be processed differently, e.g. in terms of matching to
#' wildtype sequences or collapsing of similar sequences.
#' In contrast, while \code{digestFastqs} allow the specification
#' of multiple variable sequences (within each of the forward and reverse
#' reads), they will be concatenated and processed as a single unit.
#'
#' linkMultipleVariants will process the input in the following way:
#' \itemize{
#' \item First, run \code{digestFastqs} with the parameters provided
#'     in \code{combinedDigestParams}. Typically, this will be a
#'     "naive" counting run, where the frequencies of all observed
#'     variants are tabulated. The variable sequences
#'     within the forward and reverse reads, respectively, will be
#'     processed as a single sequence. 
#' \item Next, run \code{digestFastqs} with each of the additional
#'     parameter sets provided (\code{...}). Each of these should
#'     correspond to a single variable sequence from the combined
#'     run (i.e., if there are two Vs in the element specifications
#'     in the combined run, there should be two additional
#'     parameter sets provided, each corresponding to the
#'     processing of one variable sequence part). It is assumed
#'     that the order of the additional arguments correspond to the
#'     order of the variable sequences in the combined run, in such a way
#'     that if the variable sequences extracted in each of the separate
#'     runs are concatenated in the order that the parameter sets are
#'     provided to \code{linkMultipleVariants}, they will form the variable
#'     sequence extracted in the combined run.
#' \item The result of each of the separate runs is a 'conversion table',
#'     containing the final set of identified sequence variants as well
#'     as all individual sequences corresponding to each of them. This
#'     is then combined with the count table from the combined, "naive"
#'     run in order to create an aggregated count table. More precisely,
#'     each sequence in the combined run is split into the constituent
#'     variable sequences, and
#'     each variable sequence is then matched to the output from the right
#'     separate run, from which the final feature ID (mutant name, or
#'     collapsed sequence) will be extracted and used to replace the original
#'     sequence in the combined count table. Once all the matches are done,
#'     rows with NAs (where no match could be found in the separate run)
#'     are removed and the counts are aggregated across all identical
#'     combinations of variable sequences.
#' }
#'
#' In order to define the \code{elementsForward} and \code{elementsReverse}
#' arguments for the separate runs, a strategy that often works is to simply
#' copy the arguments from the combined run, and successively replace all
#' but one of the 'V's by 'S'. This will effectively process one variable
#' sequence at the time, while keeping all other elements of the reads
#' consistent (since this can affect e.g. filtering criteria). Note that
#' to process individual variable sequences in the reverse read, you also
#' need to swap the 'forward' and 'reverse' specifications (since
#' \code{digestFastqs} requires a forward read).
#'
#' @param combinedDigestParams A named list of arguments to
#'     \code{digestFastqs} for the combined ("naive") run.
#' @param ... Additional arguments providing arguments to \code{digestFastqs}
#'     for the separate runs (processing each variable sequence in turn).
#'     Each argument must be a named list of arguments to \code{digestFastqs}.
#'     In addition, arguments \code{collapseMaxDist}, \code{collapseMinScore}
#'     and \code{collapseMinRatio} can be specified, and will be passed on 
#'     to \code{collapseMutantsBySimilarity}.
#'
#' @author Charlotte Soneson, Michael Stadler
#'
#' @export
#'
#' @return A list with the following elements: 
#' \itemize{
#' \item countAggregated - a \code{tibble} with columns corresponding to 
#'     each of the variable sequences, and a column with the total observed 
#'     read count for the combination.
#' \item convSeparate - a list of conversion tables from the respective 
#'     separate runs.
#' \item outCombined - the \code{digestFastqs} output for the combined run.
#' }
#' 
#' @examples 
#' fqFile <- system.file("extdata", "cisInput_1.fastq.gz", 
#'                       package = "mutscan")
#' out <- linkMultipleVariants(
#'     combinedDigestParams = list(fastqForward = fqFile, 
#'                                 elementsForward = "SVCV", 
#'                                 elementLengthsForward = c(1, 10, 18, 96)),
#'     # the first variable sequence is the UMI
#'     umi = list(fastqForward = fqFile, elementsForward = "SVCS",
#'                elementLengthsForward = c(1, 10, 18, 96)),
#'     # the second variable sequence is the amplicon variant
#'     var = list(fastqForward = fqFile, elementsForward = "SSCV",
#'                elementLengthsForward = c(1, 10, 18, 96), 
#'                collapseMaxDist = 3, collapseMinScore = 1)
#' )
#' # conversion tables
#' lapply(out$convSeparate, head)
#' # aggregated count table
#' head(out$countAggregated)
#' 
#' @importFrom dplyr select rename group_by summarize across matches mutate
#' @importFrom tidyr separate separate_rows
#' @importFrom rlang .data
#'
linkMultipleVariants <- function(combinedDigestParams = list(), ...) {

    ## Process additional arguments
    paramsSeparate <- list(...)

    collapseParams <- c("collapseMaxDist",
                        "collapseMinScore", "collapseMinRatio")
    
    ## --------------------------------------------------------------------- ##
    ## Initial checks
    ## --------------------------------------------------------------------- ##
    ## combinedDigestParams and each element of paramsSeparate must be named
    ## lists
    .assertVector(x = combinedDigestParams, type = "list")
    defaults <- formals(digestFastqs)
    .assertVector(x = names(combinedDigestParams), type = "character",
                  allowNULL = FALSE, validValues = names(defaults))
    for (parms in paramsSeparate) {
        .assertVector(x = parms, type = "list")
        .assertVector(x = names(parms), type = "character", allowNULL = FALSE,
                      validValues = c(names(defaults), collapseParams))
    }

    ## Fill the parameter lists with default values for all arguments
    ## that are not provided, to avoid having to check for possible
    ## NULL values downstream
    paramsSeparate <- lapply(paramsSeparate, function(parm) {
        c(parm, as.list(defaults[!(names(defaults) %in% names(parm))]))
    })
    combinedDigestParams <- 
        c(combinedDigestParams,
          as.list(defaults[!(names(defaults) %in%
                                 names(combinedDigestParams))]))

    ## --------------------------------------------------------------------- ##
    ## Checks
    ## --------------------------------------------------------------------- ##
    ## We can't have a fwd and a rev V in the same run unless they are merged
    ## since this would disagree with the order in the concatenated V
    ## Currently, each separate run can only have one V in the fwd and/or one 
    ## in the rev (both allowed if they are matched); otherwise the lengths
    ## extracted from the combined run don't match with those from the 
    ## separate runs
    for (parm in paramsSeparate) {
        nbrFwdV <- sum(strsplit(parm$elementsForward, "")[[1]] == "V")
        nbrRevV <- sum(strsplit(parm$elementsReverse, "")[[1]] == "V")
        if (nbrFwdV == 0 && nbrRevV == 0) {
            stop("Each separate run must contain at least one variable ",
                 "segment.")
        }
        if (nbrFwdV > 0 && nbrRevV > 0 && 
            !parm$mergeForwardReverse) {
            stop("A separate run can not have variable sequences in both ",
                 "the forward and reverse reads unless 'mergeForwardReverse' ",
                 "is TRUE.")
        }
        if (nbrFwdV > 1 || nbrRevV > 1) {
            stop("A separate run can have at most one variable segment ",
                 "in each of the forward and reverse reads, respectively.")
        }
    }
    
    ## Summing UMI counts after the processing may not be accurate - will
    ## always use read counts
    if (length(grep("U", paste0(combinedDigestParams$elementsForward,
                                combinedDigestParams$elementsReverse))) > 0) {
        warning("Aggregating UMI counts may not be accurate - will use ",
                "read counts instead")
    }

    ## --------------------------------------------------------------------- ##
    ## Combined quantification - just tabulate combined variable sequences
    ## --------------------------------------------------------------------- ##
    outCombined <- do.call(digestFastqs, combinedDigestParams)

    ## Get count matrix with "raw" (uncorrected) sequences
    countCombined <- outCombined$summaryTable |>
        dplyr::select("sequence", "nbrReads", "varLengths") |>
        mutate(idx = paste0("I", seq_along(.data$sequence)))

    ## If applicable, separate into forward and reverse sequences
    if (any(grepl("_", countCombined$sequence))) {
        countCombined <- countCombined |>
            separate(.data$sequence, into = c("sequenceForward", 
                                                     "sequenceReverse"), 
                            sep = "_") |>
            separate(.data$varLengths, into = c("varLengthsForward", 
                                                       "varLengthsReverse"),
                            sep = "_") |>
            mutate(
                nCompForward = vapply(strsplit(.data$varLengthsForward, ","), 
                                      length, 0),
                nCompReverse = vapply(strsplit(.data$varLengthsReverse, ","), 
                                      length, 0))
    } else {
        countCombined <- countCombined |>
            rename(sequenceForward = "sequence",
                          varLengthsForward = "varLengths") |>
            mutate(
                nCompForward = vapply(strsplit(.data$varLengthsForward, ","), 
                                      length, 0))
    }
    
    offsetForward <- 0
    ## Split forward and reverse sequences, respectively
    if ("sequenceForward" %in% colnames(countCombined)) {
        stopifnot("varLengthsForward" %in% colnames(countCombined))
        ## Check that all sequences have the same number of components
        stopifnot(length(unique(countCombined$nCompForward)) == 1)
        offsetForward <- countCombined$nCompForward[1]
        tmp <- split(countCombined, countCombined$varLengthsForward)
        countCombined <- unsplit(lapply(tmp, function(df) {
            w <- as.numeric(strsplit(df$varLengthsForward[1], ",")[[1]])
            df <- df |>
                separate(
                    .data$sequenceForward, 
                    into = names(paramsSeparate)[seq_along(w)],
                    # into = paste0("V", seq_along(w)),
                    sep = cumsum(w[seq_len(length(w) - 1)]))
        }), countCombined$varLengthsForward)
    }
    if ("sequenceReverse" %in% colnames(countCombined)) {
        stopifnot("varLengthsReverse" %in% colnames(countCombined))
        ## Check that all sequences have the same number of components
        stopifnot(length(unique(countCombined$nCompReverse)) == 1)
        tmp <- split(countCombined, countCombined$varLengthsReverse)
        countCombined <- unsplit(lapply(tmp, function(df) {
            w <- as.numeric(strsplit(df$varLengthsReverse[1], ",")[[1]])
            df <- df |>
                separate(
                    .data$sequenceReverse, 
                    into = names(paramsSeparate)[offsetForward + seq_along(w)],
                    # into = paste0("V", offsetForward + seq_along(w)),
                    sep = cumsum(w[seq_len(length(w) - 1)]))
        }), countCombined$varLengthsReverse)
    }
    
    ## --------------------------------------------------------------------- ##
    ## Quantify each variable sequence separately
    ## --------------------------------------------------------------------- ##
    outSeparate <- lapply(paramsSeparate, function(ps) {
        tmp <- do.call(digestFastqs, ps[!names(ps) %in% collapseParams])
        if (any(collapseParams %in% names(ps))) {
            ## Collapse this variable region -> need to call 
            ## groupSimilarSequences
            tbl <- do.call(groupSimilarSequences, 
                           c(list(seqs = tmp$summaryTable$sequence,
                                  scores = tmp$summaryTable$nbrReads,
                                  verbose = FALSE),
                             ps[names(ps) %in% collapseParams]))
            colnames(tbl)[colnames(tbl) == "representative"] <- "mutantName"
            tmp$summaryTable <- tbl
        }
        tmp
    })

    ## Filter tables
    filtSeparate <- lapply(outSeparate, function(out) out$filterSummary)
    
    ## Conversion tables
    convSeparate <- lapply(outSeparate, function(out) {
        out$summaryTable |>
            dplyr::select("mutantName", "sequence") |>
            separate_rows("sequence", sep = ",")
    })

    ## --------------------------------------------------------------------- ##
    ## Replace naive sequences with corrected ones
    ## --------------------------------------------------------------------- ##
    for (i in names(paramsSeparate)) {
        countCombined[[i]] <- 
            convSeparate[[i]]$mutantName[match(countCombined[[i]],
                                               convSeparate[[i]]$sequence)]
    }

    ## Filter out rows with NAs (the sequence was not retained)
    countCombined <- countCombined[rowSums(is.na(countCombined)) == 0, ,
                                   drop = FALSE]

    ## Aggregate counts
    countAggregated <- countCombined |>
        group_by(across(names(paramsSeparate))) |>
        summarize(nbrReads = sum(.data$nbrReads), .groups = "drop")

    list(countAggregated = countAggregated, convSeparate = convSeparate,
         outCombined = outCombined, filtSeparate = filtSeparate)
}

test_that("linkMultipleVariants works", {
    ## Fails with the wrong input
    ## ------------------------------------------------------------------------
    expect_error(linkMultipleVariants(combinedDigestParams = 1),
                 "'combinedDigestParams' must be of class 'list'")
    expect_error(linkMultipleVariants(combinedDigestParams = list(1)),
                 "'namescombinedDigestParams' must not be NULL")
    expect_error(linkMultipleVariants(combinedDigestParams = list(unknown = 1)),
                 "All values in 'namescombinedDigestParams' must be one of")
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = 1),
                 "'parms' must be of class 'list'")
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = list(1)),
                 "'namesparms' must not be NULL")
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = list(unknown = 1)),
                 "All values in 'namesparms' must be one of")
    
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = list(elementsForward = "CC")),
                 "Each separate run must contain at least one variable")
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = list(elementsForward = "V",
                                                       elementsReverse = "V")),
                 "A separate run can not have variable sequences in both")
    expect_error(linkMultipleVariants(combinedDigestParams = list(fastqForward = 1),
                                      secondArg = list(elementsForward = "VV")),
                 "A separate run can have at most one variable segment")
    
    ## Works with correct input
    ## ------------------------------------------------------------------------
    ## Read truth
    truth <- readRDS(system.file("extdata", "multipleVariableRegions_truth.rds",
                                 package = "mutscan"))
    verbose <- FALSE
    
    ## ------------------------------------------------------------------------
    ## Don't filter by mutations in constant sequence
    res <- linkMultipleVariants(
        combinedDigestParams = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCVCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 20, 
            maxOverlap = 30, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            verbose = verbose),
        barcode = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            avePhredMinForward = 20, 
            collapseMaxDist = 1, collapseMinScore = 1, collapseMinRatio = 1,
            verbose = verbose),
        V2 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CSCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 10, 
            maxOverlap = 20, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV2,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE),
        V3 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCSCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 5, 
            maxOverlap = 15, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV3,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE)
    )
    
    ## Filter out only reads with N/deletions in the variable sequence
    keep <- !grepl("N|del", truth$truth$status)
    
    ## Truth
    correct <- truth$truth[keep, ] |>
        dplyr::group_by(trueBarcode, trueV2, trueV3) |> dplyr::tally() |>
        dplyr::arrange(dplyr::desc(n), trueBarcode, trueV2, trueV3)
    
    ## Obs
    obs <- res$countAggregated |> dplyr::arrange(desc(nbrReads), barcode, V2, V3)
    
    expect_equal(sum(obs$nbrReads), sum(keep))
    expect_equal(correct$trueBarcode, obs$barcode)
    expect_equal(correct$trueV2, obs$V2)
    expect_equal(correct$trueV3, obs$V3)
    expect_equal(correct$n, obs$nbrReads)
    expect_equal(res$outCombined$filterSummary[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))  ## del
    expect_equal(res$outCombined$filterSummary[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("N", truth$truth$status)))   ## N
    expect_equal(res$outCombined$filterSummary[, "nbrRetained"], 
                 sum(!grepl("N|del", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "nbrRetained"], 
                 sum(!grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "nbrRetained"], 
                 sum(!grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "nbrRetained"], 
                 sum(!grepl("del", truth$truth$status)))
    
    ## ------------------------------------------------------------------------
    ## Filter out mutations in constant sequences
    res <- linkMultipleVariants(
        combinedDigestParams = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCVCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 20, 
            maxOverlap = 30, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            verbose = verbose),
        barcode = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            constantForward = truth$constFwd,
            constantMaxDistForward = 0,
            avePhredMinForward = 20,
            collapseMaxDist = 1, collapseMinScore = 1, collapseMinRatio = 1,
            verbose = verbose),
        V2 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CSCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 10, 
            maxOverlap = 20, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV2,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE),
        V3 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCSCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 5, 
            maxOverlap = 15, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV3,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE)
    )
    
    ## Filter out reads with N/deletions in the variable sequence, or mutations in constant
    keep <- !grepl("C.*mut|N|del", truth$truth$status)
    
    ## Truth
    correct <- truth$truth[keep, ] |>
        dplyr::group_by(trueBarcode, trueV2, trueV3) |> dplyr::tally() |>
        dplyr::arrange(dplyr::desc(n), trueBarcode, trueV2, trueV3)
    
    ## Obs
    obs <- res$countAggregated |> dplyr::arrange(desc(nbrReads), barcode, V2, V3)
    
    expect_equal(sum(obs$nbrReads), sum(keep))
    expect_equal(correct$trueBarcode, obs$barcode)
    expect_equal(correct$trueV2, obs$V2)
    expect_equal(correct$trueV3, obs$V3)
    expect_equal(correct$n, obs$nbrReads)

    ## first filter out reads with deletion -> no valid overlap
    ## then reads with N in barcode -> too many N
    ## then reads with mutation in constant
    expect_equal(res$outCombined$filterSummary[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("N", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("N", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "nbrRetained"], 
                 sum(!grepl("C.*mut|N|del", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C[1-3]mut|del", truth$truth$status) & 
                         !grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "nbrRetained"], 
                 sum(!grepl("V1N|C[1-3]mut|del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "nbrRetained"], 
                 sum(!grepl("C.*mut|del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "nbrRetained"], 
                 sum(!grepl("C.*mut|del", truth$truth$status)))

    ## ------------------------------------------------------------------------
    ## Collapse to WT
    res <- linkMultipleVariants(
        combinedDigestParams = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCVCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 20, 
            maxOverlap = 30, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            verbose = verbose),
        barcode = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            avePhredMinForward = 20, 
            collapseMaxDist = 1, collapseMinScore = 1, collapseMinRatio = 1,
            verbose = verbose),
        V2 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CSCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 10, 
            maxOverlap = 20, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV2,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = TRUE,
            verbose = verbose, useTreeWTmatch = TRUE),
        V3 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCSCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 5, 
            maxOverlap = 15, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV3,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = TRUE,
            verbose = verbose, useTreeWTmatch = TRUE)
    )
    
    ## Filter out only reads with N/deletions in the variable sequence
    keep <- !grepl("N|del", truth$truth$status)
    
    ## Truth
    correct <- truth$truth[keep, ] |>
        dplyr::mutate(trueV2 = gsub("\\..*", "", trueV2),
                      trueV3 = gsub("\\..*", "", trueV3)) |>
        dplyr::group_by(trueBarcode, trueV2, trueV3) |> dplyr::tally() |>
        dplyr::arrange(dplyr::desc(n), trueBarcode, trueV2, trueV3)
    
    ## Obs
    obs <- res$countAggregated |> dplyr::arrange(desc(nbrReads), barcode, V2, V3)
    
    expect_equal(sum(obs$nbrReads), sum(keep))
    expect_equal(correct$trueBarcode, obs$barcode)
    expect_equal(correct$trueV2, obs$V2)
    expect_equal(correct$trueV3, obs$V3)
    expect_equal(correct$n, obs$nbrReads)
    expect_equal(res$outCombined$filterSummary[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))  ## del
    expect_equal(res$outCombined$filterSummary[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("N", truth$truth$status)))   ## N
    expect_equal(res$outCombined$filterSummary[, "nbrRetained"], 
                 sum(!grepl("N|del", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "nbrRetained"], 
                 sum(!grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "nbrRetained"], 
                 sum(!grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "nbrRetained"], 
                 sum(!grepl("del", truth$truth$status)))
    
    ## ------------------------------------------------------------------------
    ## Don't merge fwd and rev
    res <- linkMultipleVariants(
        combinedDigestParams = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = FALSE,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            verbose = verbose),
        barcode = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            avePhredMinForward = 20, 
            collapseMaxDist = 1, collapseMinScore = 1, collapseMinRatio = 1,
            verbose = verbose),
        V2 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            mergeForwardReverse = FALSE,
            revComplForward = FALSE, revComplReverse = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV2,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE),
        V3 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCS", elementLengthsForward = c(6, 40, 10, -1),
            mergeForwardReverse = FALSE,
            revComplForward = TRUE,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV3,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE)
    )
    
    ## Filter out only reads with N/deletions in the variable sequence
    keep <- !grepl("N|del", truth$truth$status)
    
    ## Truth
    correct <- truth$truth[keep, ] |>
        dplyr::group_by(trueBarcode, trueV2, trueV3) |> dplyr::tally() |>
        dplyr::arrange(dplyr::desc(n), trueBarcode, trueV2, trueV3)
    
    ## Obs
    obs <- res$countAggregated |> dplyr::arrange(desc(nbrReads), barcode, V2, V3)
    
    expect_equal(sum(obs$nbrReads), sum(keep))
    expect_equal(correct$trueBarcode, obs$barcode)
    expect_equal(correct$trueV2, obs$V2)
    expect_equal(correct$trueV3, obs$V3)
    expect_equal(correct$n, obs$nbrReads)
    expect_equal(res$outCombined$filterSummary[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("N", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "nbrRetained"], 
                 sum(!grepl("N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "nbrRetained"], 
                 sum(!grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f10b_nbrTooManyMutBases"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "nbrRetained"], 
                 sum(!grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "nbrRetained"], 100)
    
    ## ------------------------------------------------------------------------
    ## Don't allow any mutations
    res <- linkMultipleVariants(
        combinedDigestParams = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCVCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 20, 
            maxOverlap = 30, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            verbose = verbose),
        barcode = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            constantForward = truth$constFwd,
            constantMaxDistForward = 0,
            avePhredMinForward = 20, 
            collapseMaxDist = 0, collapseMinScore = 1, collapseMinRatio = 1,
            verbose = verbose),
        V2 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CSCV", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 10, 
            maxOverlap = 20, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV2,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 0, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE),
        V3 = list(
            fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                       package = "mutscan"),
            fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                       package = "mutscan"),
            elementsForward = "CSCSCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
            elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
            mergeForwardReverse = TRUE, minOverlap = 5, 
            maxOverlap = 15, maxFracMismatchOverlap = 0,
            revComplForward = FALSE, revComplReverse = TRUE,
            constantForward = truth$constFwd,
            constantReverse = truth$constRev,
            constantMaxDistForward = 0,
            constantMaxDistReverse = 0,
            avePhredMinForward = 20, avePhredMinReverse = 20,
            wildTypeForward = truth$WTV3,
            nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 0, 
            forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
            verbose = verbose, useTreeWTmatch = TRUE)
    )
    
    ## Filter out reads with N/deletions in the variable sequence, or mutations in constant
    keep <- !grepl("C.*mut|N|del|V2|V3", truth$truth$status)
    
    ## Truth
    correct <- truth$truth[keep, ] |>
        dplyr::group_by(obsBarcode, trueV2, trueV3) |> dplyr::tally() |>
        dplyr::arrange(dplyr::desc(n), obsBarcode, trueV2, trueV3)
    
    ## Obs
    obs <- res$countAggregated |> dplyr::arrange(desc(nbrReads), barcode, V2, V3)
    
    expect_equal(sum(obs$nbrReads), sum(keep))
    expect_equal(correct$obsBarcode, obs$barcode)
    expect_equal(correct$trueV2, obs$V2)
    expect_equal(correct$trueV3, obs$V3)
    expect_equal(correct$n, obs$nbrReads)
    
    expect_equal(res$outCombined$filterSummary[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("N", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("N", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$outCombined$filterSummary[, "nbrRetained"], 
                 sum(!grepl("C.*mut|N|del", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f6_nbrTooManyNinVar"], 
                 sum(grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C[1-3]mut|del", truth$truth$status) & 
                         !grepl("V1N", truth$truth$status)))
    expect_equal(res$filtSeparate$barcode[, "nbrRetained"], 
                 sum(!grepl("V1N|C[1-3]mut|del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f10b_nbrTooManyMutBases"],
                 sum(grepl("V2v[0-9]\\.", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("del", truth$truth$status) & 
                         !grepl("V2v[0-9]\\.", truth$truth$status)))
    expect_equal(res$filtSeparate$V2[, "nbrRetained"], 
                 sum(!grepl("C.*mut|del|V2v[0-9]\\.", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f4_nbrNoValidOverlap"], 
                 sum(grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f10b_nbrTooManyMutBases"],
                 sum(grepl("V3\\.", truth$truth$status) & 
                         !grepl("del", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "f12_nbrTooManyMutConstant"], 
                 sum(grepl("C.*mut", truth$truth$status) & 
                         !grepl("del", truth$truth$status) & 
                         !grepl("V3\\.", truth$truth$status)))
    expect_equal(res$filtSeparate$V3[, "nbrRetained"], 
                 sum(!grepl("C.*mut|del|V3\\.", truth$truth$status)))
    
    # --------------------------------------------------------------------------
    # Get warning when summing UMIs
    expect_warning({
        res <- linkMultipleVariants(
            combinedDigestParams = list(
                fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                           package = "mutscan"),
                fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                           package = "mutscan"),
                elementsForward = "UVCVCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
                elementsReverse = "CVCV", elementLengthsReverse = c(6, 40, 10, -1),
                mergeForwardReverse = TRUE, minOverlap = 20, 
                maxOverlap = 30, maxFracMismatchOverlap = 0,
                revComplForward = FALSE, revComplReverse = TRUE,
                avePhredMinForward = 20, avePhredMinReverse = 20,
                verbose = verbose),
            barcode = list(
                fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                           package = "mutscan"),
                elementsForward = "CVCSCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
                avePhredMinForward = 20, 
                collapseMaxDist = 1, collapseMinScore = 1, collapseMinRatio = 1,
                verbose = verbose),
            V2 = list(
                fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                           package = "mutscan"),
                fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                           package = "mutscan"),
                elementsForward = "CSCVCS", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
                elementsReverse = "CSCV", elementLengthsReverse = c(6, 40, 10, -1),
                mergeForwardReverse = TRUE, minOverlap = 10, 
                maxOverlap = 20, maxFracMismatchOverlap = 0,
                revComplForward = FALSE, revComplReverse = TRUE,
                avePhredMinForward = 20, avePhredMinReverse = 20,
                wildTypeForward = truth$WTV2,
                nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
                forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
                verbose = verbose, useTreeWTmatch = TRUE),
            V3 = list(
                fastqForward = system.file("extdata", "multipleVariableRegions_R1.fastq.gz",
                                           package = "mutscan"),
                fastqReverse = system.file("extdata", "multipleVariableRegions_R2.fastq.gz",
                                           package = "mutscan"),
                elementsForward = "CSCSCV", elementLengthsForward = c(6, 24, 10, 30, 10, -1),
                elementsReverse = "CVCS", elementLengthsReverse = c(6, 40, 10, -1),
                mergeForwardReverse = TRUE, minOverlap = 5, 
                maxOverlap = 15, maxFracMismatchOverlap = 0,
                revComplForward = FALSE, revComplReverse = TRUE,
                avePhredMinForward = 20, avePhredMinReverse = 20,
                wildTypeForward = truth$WTV3,
                nbrMutatedCodonsMaxForward = -1, nbrMutatedBasesMaxForward = 1, 
                forbiddenMutatedCodonsForward = "", collapseToWTForward = FALSE,
                verbose = verbose, useTreeWTmatch = TRUE)
        )}, "Aggregating UMI counts")
})

#' Collapse mutants by similarity
#' 
#' These functions can be used to collapse variants, either by similarity or 
#' according to a pre-defined grouping. The functions \code{collapseMutants} 
#' and \code{collapseMutantsByAA} assume that a grouping variable is available 
#' as a column in \code{rowData(se)} (\code{collapseMutantsByAA} is a 
#' convenience function for the case when this column is "mutantNameAA", and 
#' is provided for backwards compatibility). The 
#' \code{collapseMutantsBySimilarity} will generate the grouping variable 
#' based on user-provided thresholds on the sequence similarity (defined by 
#' the Hamming distance), and subsequently collapse based on the derived 
#' grouping. 
#' 
#' @param se A \code{\link[SummarizedExperiment]{SummarizedExperiment}}
#'     generated by \code{\link{summarizeExperiment}}
#' @param assayName The name of the assay that will be used to calculate a 
#'     "score" (typically derived from the read counts) for each variant.
#' @param scoreMethod Character scalar giving the approach used to calculate 
#'     ranking scores from the assay defined by \code{assayName}. Currently, 
#'     this can be one of \code{"rowSum"} or \code{"rowMean"}. All filtering
#'     criteria will be applied to these scores.
#' @param sequenceCol Character scalar giving the name of the column in 
#'     \code{rowData(se)} that contains the nucleotide sequence of the 
#'     variants. 
#' @param collapseMaxDist Numeric scalar defining the tolerance for collapsing 
#'     similar sequences. If the value is in [0, 1), it defines the maximal 
#'     Hamming distance in terms of a fraction of sequence length:
#'     (\code{round(collapseMaxDist * nchar(sequence))}).
#'     A value greater or equal to 1 is rounded and directly used as the 
#'     maximum allowed Hamming distance. Note that sequences can only be
#'     collapsed if they are all of the same length.
#' @param collapseMinScore Numeric scalar, indicating the minimum score for the 
#'     sequence to be considered for collapsing with similar sequences.
#' @param collapseMinRatio Numeric scalar. During collapsing of
#'     similar sequences, a low-frequency sequence will be collapsed 
#'     with a higher-frequency sequence only if the ratio between the 
#'     high-frequency and the low-frequency scores is at least this 
#'     high. The default value of 0 indicates that no such check is performed.
#' @param verbose Logical, whether to print progress messages.
#' 
#' @author Charlotte Soneson, Michael Stadler
#' 
#' @export
#' 
#' @importFrom SummarizedExperiment assayNames assay rowData
#' @importFrom Matrix rowSums rowMeans
#' 
#' @examples
#' library(SummarizedExperiment)
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                           package = "mutscan"))[1:200, ]
#' ## The rows of this object correspond to individual codon variants
#' dim(se)
#' head(rownames(se))
#' 
#' ## Collapse by amino acid
#' sec <- collapseMutantsByAA(se)
#' ## The rows of the collapsed object correspond to amino acid variants
#' dim(sec)
#' head(rownames(sec))
#' ## The mutantName column contains the individual codon variants that were 
#' ## collapsed
#' head(rowData(sec))
#' 
#' ## Collapse similar sequences
#' sec2 <- collapseMutantsBySimilarity(
#'     se = se, assayName = "counts", scoreMethod = "rowSum",
#'     sequenceCol = "sequence", collapseMaxDist = 2,
#'     collapseMinScore = 0, collapseMinRatio = 0)
#' dim(sec2)
#' head(rownames(sec2))
#' head(rowData(sec2))
#' ## collapsed count matrix
#' assay(sec2, "counts")
#'
collapseMutantsBySimilarity <- function(se, assayName, scoreMethod = "rowSum",
                                        sequenceCol = "sequence", 
                                        collapseMaxDist = 0,
                                        collapseMinScore = 0, 
                                        collapseMinRatio = 0,
                                        verbose = TRUE) {
    ## Check input arguments
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = assayName, type = "character", 
                  validValues = assayNames(se))
    .assertScalar(x = scoreMethod, type = "character",
                  validValues = c("rowSum", "rowMean"))
    .assertScalar(x = sequenceCol, type = "character",
                  validValues = colnames(rowData(se)))
    .assertScalar(x = collapseMaxDist, type = "numeric")
    .assertScalar(x = collapseMinScore, type = "numeric")
    .assertScalar(x = collapseMinRatio, type = "numeric")
    .assertScalar(x = verbose, type = "logical")
    
    ## All sequences must be the same length, and can only consist of ACGT_
    seqs <- rowData(se)[, sequenceCol]
    if (length(unique(nchar(seqs))) != 1) {
        stop("All entries in the sequence column must have the same length")
    }
    if (!all(vapply(seqs, 
                    function(x) grepl("^[ACGT_]*$", x), FALSE))) {
        stop("All entries in the sequence column must consist of [ACGT_]")
    }
    
    ## Get matching table
    if (scoreMethod == "rowSum") {
        scores <- rowSums(assay(se, assayName))
    } else if (scoreMethod == "rowMean") {
        scores <- rowMeans(assay(se, assayName))
    }
    matchTable <- groupSimilarSequences(
        seqs = seqs,
        scores = scores,
        collapseMaxDist = collapseMaxDist,
        collapseMinScore = collapseMinScore,
        collapseMinRatio = collapseMinRatio,
        verbose = verbose
    )
    matchTable$representative <- rownames(se)[
        match(matchTable$representative, 
              rowData(se)[, sequenceCol])]
    
    ## Add grouping variable
    rowData(se)$collapseCol <- 
        matchTable$representative[
            match(rowData(se)[, sequenceCol], 
                  matchTable$sequence)]
    
    ## Collapse
    se <- collapseMutants(se, nameCol = "collapseCol")
    
    se
}

#' @rdname collapseMutantsBySimilarity
#' @export
#' 
collapseMutantsByAA <- function(se) {
    collapseMutants(se = se, nameCol = "mutantNameAA")
}

#' Collapse counts by mutated amino acid
#'
#' @param se A \code{\link[SummarizedExperiment]{SummarizedExperiment}}
#'     generated by \code{\link{summarizeExperiment}}
#' @param nameCol A character scalar providing the column of 
#'     \code{rowData(se)} that contains the amino acid mutant names (that will 
#'     be the new row names).
#'
#' @return A \code{\link[SummarizedExperiment]{SummarizedExperiment}} where
#'     counts have been aggregated by the mutated amino acid(s).
#'
#' @author Charlotte Soneson, Michael Stadler
#'
#' @export
#'
#' @importFrom DelayedArray rowsum
#' @importFrom S4Vectors metadata DataFrame
#' @importFrom SummarizedExperiment assays rowData SummarizedExperiment colData
#'     rowData<-
#' @importFrom dplyr across group_by summarize full_join
#' @importFrom stats setNames
#'
#' @rdname collapseMutantsBySimilarity
#' 
collapseMutants <- function(se, nameCol) {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = nameCol, type = "character",
                  validValues = colnames(rowData(se)))
    
    ## Collapse assays
    aList <- lapply(assays(se), function(a) {
        rowsum(
            x = a,
            group = factor(rowData(se)[[nameCol]])
        )
    })
    
    ## Collapse rowData - simple columns
    rd <- mergeValues(rowData(se)[[nameCol]],
                      rowData(se)$sequence) |>
        setNames(c(nameCol, "sequence"))
    for (v in setdiff(
        intersect(c("mutantName", "mutantNameBase", "mutantNameBaseHGVS",
                    "mutantNameCodon", "mutantNameAA", "mutantNameAAHGVS",
                    "sequenceAA", "mutationTypes",
                    "nbrMutBases", "nbrMutCodons", "nbrMutAAs", "varLengths"),
                  colnames(rowData(se))),
        nameCol)) {
        tmp <- mergeValues(rowData(se)[[nameCol]],
                           rowData(se)[[v]])
        rd[[v]] <- tmp$valueColl[match(rd[[nameCol]], tmp$mutantNameColl)]
    }
    rd0 <- as.data.frame(rowData(se)) |>
        group_by(.data[[nameCol]]) |>
        summarize(
            across(c("minNbrMutBases", "minNbrMutCodons",
                     "minNbrMutAAs"),
                   function(x) min(x)),
            across(c("maxNbrMutBases", "maxNbrMutCodons",
                     "maxNbrMutAAs"),
                   function(x) max(x)))
    rd <- rd |>
        full_join(rd0, by = nameCol)
    
    rd <- DataFrame(rd)
    rownames(rd) <- rd[[nameCol]]
    
    ## Create SummarizedExperiment
    rn1 <- rownames(aList[[1]])
    cn1 <- colnames(aList[[1]])
    for (i in seq_along(aList)) {
        stopifnot(all(rownames(aList[[i]]) == rn1))
        stopifnot(all(colnames(aList[[i]]) == cn1))
    }
    stopifnot(cn1 == rownames(colData(se)))
    SummarizedExperiment(
        assays = aList,
        colData = colData(se),
        metadata = metadata(se),
        rowData = rd[rn1, ]
    )
}

## groupSimilarSequences
seqs <- c("AACGTAGCA", "ACCGTAGCA", "AACGGAGCA", "ATCGGAGCA", "TGAGGCATA")
scores <- c(5, 1, 3, 1, 8)
gr <- groupSimilarSequences(seqs = seqs, scores = scores, collapseMaxDist = 1, 
                            collapseMinScore = 0, collapseMinRatio = 0, verbose = FALSE)
expect_equal(gr$sequence, seqs)
expect_equal(gr$representative, seqs[c(1, 1, 1, 4, 5)])

gr <- groupSimilarSequences(seqs = seqs, scores = scores, collapseMaxDist = 0, 
                            collapseMinScore = 0, collapseMinRatio = 0, verbose = FALSE)
expect_equal(gr$sequence, seqs)
expect_equal(gr$representative, seqs[c(1, 2, 3, 4, 5)])

gr <- groupSimilarSequences(seqs = seqs, scores = scores, collapseMaxDist = 1, 
                            collapseMinScore = 0, collapseMinRatio = 2, verbose = FALSE)
expect_equal(gr$sequence, seqs)
expect_equal(gr$representative, seqs[c(1, 1, 3, 3, 5)])

gr <- groupSimilarSequences(seqs = seqs, scores = scores, collapseMaxDist = 1, 
                            collapseMinScore = 4, collapseMinRatio = 2, verbose = FALSE)
expect_equal(gr$sequence, seqs)
expect_equal(gr$representative, seqs[c(1, 1, 3, 4, 5)])

gr <- groupSimilarSequences(seqs = seqs, scores = scores, collapseMaxDist = 9, 
                            collapseMinScore = 4, collapseMinRatio = 2, verbose = FALSE)
expect_equal(gr$sequence, seqs)
expect_equal(gr$representative, seqs[c(1, 5, 5, 5, 5)])


## collapseMutants
Ldef <- list(
  mergeForwardReverse = TRUE, 
  minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
  revComplForward = FALSE, revComplReverse = TRUE,
  elementsForward = "SUCV", elementsReverse = "SUCVS",
  elementLengthsForward = c(1, 10, 18, 96),
  elementLengthsReverse = c(1, 7, 17, 96, -1),
  adapterForward = "GGAAGAGCACACGTC", 
  adapterReverse = "GGAAGAGCGTCGTGT",
  primerForward = "",
  primerReverse = "",
  wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
  wildTypeReverse = "", 
  constantForward = "AACCGGAGGAGGGAGCTG", 
  constantReverse = "GAGTTCATCCTGGCAGC", 
  avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
  variableNMaxForward = 0, variableNMaxReverse = 0, 
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  nbrMutatedBasesMaxForward = -1,
  nbrMutatedBasesMaxReverse = -1,
  forbiddenMutatedCodonsForward = "NNW",
  forbiddenMutatedCodonsReverse = "NNW",
  useTreeWTmatch = FALSE,
  mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
  mutNameDelimiter = ".",
  constantMaxDistForward = -1,
  constantMaxDistReverse = -1,
  umiCollapseMaxDist = 0,
  filteredReadsFastqForward = "",
  filteredReadsFastqReverse = "",
  maxNReads = -1, 
  verbose = FALSE,
  nThreads = 1, chunkSize = 1000, 
  maxReadLength = 1024
)
Ldef1 <- c(
  list(fastqForward = system.file("extdata/cisInput_1.fastq.gz", package = "mutscan"),
       fastqReverse = system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
  ), Ldef)
Ldef2 <- c(
  list(fastqForward = system.file("extdata/cisOutput_1.fastq.gz", package = "mutscan"),
       fastqReverse = system.file("extdata/cisOutput_2.fastq.gz", package = "mutscan")
  ), Ldef)

out1 <- do.call(digestFastqs, Ldef1)
out2 <- do.call(digestFastqs, Ldef2)

coldata <- data.frame(Name = c("sample1", "sample2"), 
                      Condition = c("input", "output"),
                      Replicate = c(1, 1), OD = c(0.05, 1.5),
                      stringsAsFactors = FALSE)

se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                          coldata = coldata, countType = "umis")

test_that("collapseMutantsByAA fails with incorrect arguments", {
    expect_error(collapseMutantsByAA(se = 1))
    expect_error(collapseMutantsByAA(se = coldata))
    expect_error(collapseMutantsByAA(se = out1))
    expect_error(collapseMutantsByAA(se = list(sample1 = out1, sample2 = out2)))
    expect_error(collapseMutantsByAA(se = se, nameCol = 1))
    expect_error(collapseMutantsByAA(se = se, nameCol = c("sequence", "mutantNameAA")))
    expect_error(collapseMutantsByAA(se = se, nameCol = "missing"))
})

test_that("collapseMutantsBySimilarity fails with incorrect arguments", {
    expect_error(collapseMutantsBySimilarity(se = 1))
    expect_error(collapseMutantsBySimilarity(se = coldata))
    expect_error(collapseMutantsBySimilarity(se = out1))
    expect_error(collapseMutantsBySimilarity(se = list(sample1 = out1, sample2 = out2)))
    expect_error(collapseMutantsBySimilarity(se = se, assayName = 1),
                 "'assayName' must be of class 'character'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "random"),
                 "All values in 'assayName'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = c("counts", "counts")),
                 "'assayName' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = 1),
                 "'scoreMethod' must be of class 'character'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = c("rowSum", "rowMean")),
                 "'scoreMethod' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "random"),
                 "All values in 'scoreMethod' must be one of")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = 1),
                 "'sequenceCol' must be of class 'character'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", 
                                             sequenceCol = c("sequence", "sequence")),
                 "'sequenceCol' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "abc"),
                 "All values in 'sequenceCol' must be one of")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = "1"),
                 "'collapseMaxDist' must be of class 'numeric'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = c(1, 2)),
                 "'collapseMaxDist' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = "1"),
                 "'collapseMinScore' must be of class 'numeric'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = c(1, 2)),
                 "'collapseMinScore' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = "1"),
                 "'collapseMinRatio' must be of class 'numeric'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = c(1, 2)),
                 "'collapseMinRatio' must have length 1")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = 0, verbose = "TRUE"),
                 "'verbose' must be of class 'logical'")
    expect_error(collapseMutantsBySimilarity(se = se, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = 0, verbose = c(TRUE, FALSE)),
                 "'verbose' must have length 1")
    
    se0 <- se
    SummarizedExperiment::rowData(se0)$sequence[1] <- "ACGT"
    expect_error(collapseMutantsBySimilarity(se = se0, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = 0, verbose = TRUE),
                 "All entries in the sequence column must have the same length")
    se0 <- se
    SummarizedExperiment::rowData(se0)$sequence[1] <- gsub("A", "B", rowData(se0)$sequence[1])
    expect_error(collapseMutantsBySimilarity(se = se0, assayName = "counts", 
                                             scoreMethod = "rowSum", sequenceCol = "sequence",
                                             collapseMaxDist = 1, collapseMinScore = 1,
                                             collapseMinRatio = 0, verbose = TRUE),
                 "All entries in the sequence column must consist of")
})


test_that("collapseMutantsByAA works as expected", {
  secoll <- collapseMutantsByAA(se)
  expect_s4_class(secoll, "SummarizedExperiment")
  expect_equal(SummarizedExperiment::colData(se), 
               SummarizedExperiment::colData(secoll))
  expect_equal(S4Vectors::metadata(se), 
               S4Vectors::metadata(secoll))
  expect_equal(colnames(se), colnames(secoll))
  
  aapos <- unique(unlist(rowData(se)$mutantNameAA))
  expect_equal(nrow(secoll), length(aapos))
  expect_equal(sort(rownames(secoll)), sort(unique(aapos)))
  
  tmp <- table(rep(unlist(SummarizedExperiment::rowData(se)$mutantNameAA), 
                   SummarizedExperiment::assay(se, "counts")[, 1]))
  tmp <- tmp[rownames(secoll)]
  tmp[is.na(tmp)] <- 0
  expect_equal(SummarizedExperiment::assay(secoll, "counts")[, 1],
               as.numeric(tmp), ignore_attr = TRUE)
  
  tmp <- table(rep(unlist(SummarizedExperiment::rowData(se)$mutantNameAA), 
                   SummarizedExperiment::assay(se, "counts")[, 2]))
  tmp <- tmp[rownames(secoll)]
  tmp[is.na(tmp)] <- 0
  expect_equal(SummarizedExperiment::assay(secoll, "counts")[, 2],
               as.numeric(tmp), ignore_attr = TRUE)
  
  for (mn in rownames(secoll)) {
    expect_equal(SummarizedExperiment::assay(secoll, "counts")[mn, ],
                 colSums(SummarizedExperiment::assay(se, "counts")[SummarizedExperiment::rowData(se)$mutantNameAA == mn, , drop = FALSE]))
  }
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutBases, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutBases)
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutCodons, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutCodons)
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutAAs, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutAAs)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutBases, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutBases)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutCodons, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutCodons)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutAAs, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutAAs)
})

test_that("collapseMutantsByAA works as expected - collapseToWT", {
  Ldef1$collapseToWTForward <- TRUE
  Ldef2$collapseToWTForward <- TRUE
  out1 <- do.call(digestFastqs, Ldef1)
  out2 <- do.call(digestFastqs, Ldef2)
  
  se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                            coldata = coldata, countType = "umis")
  
  secoll <- collapseMutantsByAA(se)
  expect_equal(SummarizedExperiment::colData(se), 
               SummarizedExperiment::colData(secoll))
  expect_equal(S4Vectors::metadata(se), 
               S4Vectors::metadata(secoll))
  expect_equal(colnames(se), colnames(secoll))
  
  aapos <- unique(unlist(rowData(se)$mutantNameAA))
  expect_equal(nrow(secoll), length(aapos))
  expect_equal(sort(rownames(secoll)), sort(unique(aapos)))
  
  tmp <- table(rep(unlist(SummarizedExperiment::rowData(se)$mutantNameAA), 
                   SummarizedExperiment::assay(se, "counts")[, 1]))
  tmp <- tmp[rownames(secoll)]
  tmp[is.na(tmp)] <- 0
  expect_equal(SummarizedExperiment::assay(secoll, "counts")[, 1],
               as.numeric(tmp), ignore_attr = TRUE)
  
  tmp <- table(rep(unlist(SummarizedExperiment::rowData(se)$mutantNameAA), 
                   SummarizedExperiment::assay(se, "counts")[, 2]))
  tmp <- tmp[rownames(secoll)]
  tmp[is.na(tmp)] <- 0
  expect_equal(SummarizedExperiment::assay(secoll, "counts")[, 2],
               as.numeric(tmp), ignore_attr = TRUE)
  
  for (mn in rownames(secoll)) {
    expect_equal(SummarizedExperiment::assay(secoll, "counts")[mn, ],
                 colSums(SummarizedExperiment::assay(se, "counts")[SummarizedExperiment::rowData(se)$mutantNameAA == mn, , drop = FALSE]))
  }
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutBases, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutBases)
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutCodons, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutCodons)
  expect_equal(min(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutAAs, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$minNbrMutAAs)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutBases, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutBases)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutCodons, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutCodons)
  expect_equal(max(methods::as(
      lapply(strsplit(SummarizedExperiment::rowData(se)$nbrMutAAs, ","), function(w) sort(as.integer(w))),
      "IntegerList")), 
      SummarizedExperiment::rowData(se)$maxNbrMutAAs)
  
  expect_equal(nrow(secoll), 1L)
  expect_equal(ncol(secoll), 2L)
  expect_equal(SummarizedExperiment::rowData(secoll)$nbrMutBases[[1]], "0,1,2")
  expect_equal(SummarizedExperiment::rowData(secoll)$nbrMutCodons[[1]], "0,1")
  expect_equal(SummarizedExperiment::rowData(secoll)$nbrMutAAs[[1]], "0,1")
  expect_equal(SummarizedExperiment::rowData(secoll)$mutationTypes, 
               "nonsynonymous,silent,stop")
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 4,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = TRUE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)
  
  ## Summarize single sample and collapse
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 6, 
                                        collapseMinScore = 0, collapseMinRatio = 0, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           scoreMethod = "rowMean",
                                           collapseMaxDist = 6, 
                                           collapseMinScore = 0, collapseMinRatio = 0, 
                                           verbose = TRUE)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 679L)
  
  expect_equal(SummarizedExperiment::colData(secoll)$nbrTotal, 1000L)
  expect_equal(SummarizedExperiment::colData(secoll)$f1_nbrAdapter, 314L)
  expect_equal(SummarizedExperiment::colData(secoll)$f2_nbrNoPrimer, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f3_nbrReadWrongLength, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f4_nbrNoValidOverlap, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(SummarizedExperiment::colData(secoll)$f6_nbrTooManyNinVar, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f7_nbrTooManyNinUMI, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f9_nbrMutQualTooLow, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f10b_nbrTooManyMutBases, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f11_nbrForbiddenCodons, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f12_nbrTooManyMutConstant, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(SummarizedExperiment::colData(secoll)$nbrRetained, 679L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(sum(SummarizedExperiment::assay(secoll, "counts")), 
               res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 677L)
  expect_equal(sum(res$summaryTable$nbrUmis), 679L)
  expect_equal(nrow(secoll), 294L)
  expect_equal(nrow(secollumi), 294L)
  expect_true(all(res$summaryTable$varLengths == "96_96"))
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed (relative UMI distance)", {
    fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
    fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
    ## default arguments
    Ldef <- list(
        fastqForward = fqt1, fastqReverse = fqt2,
        mergeForwardReverse = FALSE,
        minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
        revComplForward = FALSE, revComplReverse = FALSE,
        elementsForward = "SUCV", elementsReverse = "SUCV",
        elementLengthsForward = c(1, 10, 18, 96),
        elementLengthsReverse = c(1, 8, 20, 96),
        adapterForward = "GGAAGAGCACACGTC",
        adapterReverse = "GGAAGAGCGTCGTGT",
        primerForward = "",
        primerReverse = "",
        wildTypeForward = "",
        wildTypeReverse = "",
        constantForward = "AACCGGAGGAGGGAGCTG",
        constantReverse = "GAAAAAGGAAGCTGGAGAGA",
        avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
        variableNMaxForward = 0, variableNMaxReverse = 0,
        umiNMax = 0,
        nbrMutatedCodonsMaxForward = 1,
        nbrMutatedCodonsMaxReverse = 1,
        nbrMutatedBasesMaxForward = -1,
        nbrMutatedBasesMaxReverse = -1,
        forbiddenMutatedCodonsForward = "NNW",
        forbiddenMutatedCodonsReverse = "NNW",
        useTreeWTmatch = FALSE,
        mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
        mutNameDelimiter = ".",
        constantMaxDistForward = -1,
        constantMaxDistReverse = -1,
        umiCollapseMaxDist = 0.23,
        filteredReadsFastqForward = "",
        filteredReadsFastqReverse = "",
        maxNReads = -1, verbose = TRUE,
        nThreads = 1, chunkSize = 1000, 
        maxReadLength = 1024
    )
    
    res <- do.call(digestFastqs, Ldef)
    
    ## Summarize single sample and collapse
    se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                              countType = "reads")
    secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                          collapseMaxDist = 6, 
                                          collapseMinScore = 0, collapseMinRatio = 0, 
                                          verbose = FALSE)
    seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                                 countType = "umis")
    secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                             scoreMethod = "rowMean",
                                             collapseMaxDist = 6, 
                                             collapseMinScore = 0, collapseMinRatio = 0, 
                                             verbose = TRUE)
    
    expect_equal(res$filterSummary$nbrTotal, 1000L)
    expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
    expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
    expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
    expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
    expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
    expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
    expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
    expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
    expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
    expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
    expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
    expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
    expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
    expect_equal(res$filterSummary$nbrRetained, 679L)
    
    expect_equal(SummarizedExperiment::colData(secoll)$nbrTotal, 1000L)
    expect_equal(SummarizedExperiment::colData(secoll)$f1_nbrAdapter, 314L)
    expect_equal(SummarizedExperiment::colData(secoll)$f2_nbrNoPrimer, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f3_nbrReadWrongLength, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f4_nbrNoValidOverlap, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f5_nbrAvgVarQualTooLow, 7L)
    expect_equal(SummarizedExperiment::colData(secoll)$f6_nbrTooManyNinVar, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f7_nbrTooManyNinUMI, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f9_nbrMutQualTooLow, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f10a_nbrTooManyMutCodons, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f10b_nbrTooManyMutBases, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f11_nbrForbiddenCodons, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f12_nbrTooManyMutConstant, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$f13_nbrTooManyBestConstantHits, 0L)
    expect_equal(SummarizedExperiment::colData(secoll)$nbrRetained, 679L)
    
    for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
        expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
    }
    for (nm in c("fastqForward", "fastqReverse")) {
        expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                     ignore_attr = TRUE)
    }
    
    expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
    expect_equal(sum(SummarizedExperiment::assay(secoll, "counts")), 
                 res$filterSummary$nbrRetained)
    expect_equal(nrow(res$summaryTable), 677L)
    expect_equal(sum(res$summaryTable$nbrUmis), 679L)
    expect_equal(nrow(secoll), 294L)
    expect_equal(nrow(secollumi), 294L)
    expect_true(all(res$summaryTable$varLengths == "96_96"))
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed - extreme case", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 100,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 125
  )

  res <- do.call(digestFastqs, Ldef)
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 500, 
                                        collapseMinScore = 0, collapseMinRatio = 0, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           collapseMaxDist = 500, 
                                           collapseMinScore = 0, collapseMinRatio = 0, 
                                           verbose = FALSE)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 679L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 677L)
  expect_equal(sum(res$summaryTable$nbrUmis), 677L)
  expect_equal(nrow(secoll), 1L)
  expect_equal(nrow(secollumi), 1L)
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed - only forward read", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = NULL,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 5,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 10, 
                                        collapseMinScore = 0, collapseMinRatio = 0, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           collapseMaxDist = 10, 
                                           collapseMinScore = 0, collapseMinRatio = 0, 
                                           verbose = FALSE)
  
  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 297L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 703L)

  for (nm in setdiff(names(Ldef), c("fastqReverse", "forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 613L)
  expect_equal(sum(res$summaryTable$nbrUmis), 687L)
  expect_equal(nrow(se), 613L)
  expect_equal(nrow(secoll), 52L)
  expect_equal(nrow(seumi), 613L)
  expect_equal(nrow(secollumi), 52L)
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed (only abundant ones)", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 500, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 2, 
                                        collapseMinScore = 1.5, collapseMinRatio = 0, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           collapseMaxDist = 2, 
                                           collapseMinScore = 1.5, collapseMinRatio = 0, 
                                           verbose = FALSE)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 679L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "nThreads",
                                    "forbiddenMutatedCodonsReverse", "verbose",
                                    "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 677L)
  expect_equal(sum(res$summaryTable$nbrUmis), 679L)
  expect_equal(nrow(secoll), 673L)
  expect_equal(nrow(secollumi), 673L)
  
  ## Don't consider mutations here since we're collapsing (i.e., we have no wildtype)
  expect_equal(sum(res$summaryTable$nbrMutBases == 0), nrow(res$summaryTable))
  expect_equal(sum(res$summaryTable$nbrMutCodons == 0), nrow(res$summaryTable))
  expect_equal(sum(res$summaryTable$nbrMutAAs == 0), nrow(res$summaryTable))
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed (only high ratio)", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 3, 
                                        collapseMinScore = 1, collapseMinRatio = 1.5, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           collapseMaxDist = 3, 
                                           collapseMinScore = 1, collapseMinRatio = 1.5, 
                                           verbose = FALSE)
  
  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 679L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 677L)
  expect_equal(sum(res$summaryTable$nbrUmis), 679L)
  expect_equal(nrow(secoll), 656L)
  expect_equal(nrow(secollumi), 656L)
})

test_that("digestFastqs works as expected for trans experiments, when similar sequences are collapsed (only high ratio), specify distance threshold as ratio", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )
  
  res <- do.call(digestFastqs, Ldef)
  se <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                            countType = "reads")
  secoll <- collapseMutantsBySimilarity(se, assayName = "counts", 
                                        collapseMaxDist = 0.016, 
                                        collapseMinScore = 1, collapseMinRatio = 1.5, 
                                        verbose = FALSE)
  seumi <- summarizeExperiment(list(s1 = res), coldata = data.frame(Name = "s1"), 
                               countType = "umis")
  secollumi <- collapseMutantsBySimilarity(seumi, assayName = "counts", 
                                           collapseMaxDist = 0.016, 
                                           collapseMinScore = 1, collapseMinRatio = 1.5, 
                                           verbose = FALSE)
  
  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 679L)
  
  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(nrow(res$summaryTable), 677L)
  expect_equal(sum(res$summaryTable$nbrUmis), 679L)
  expect_equal(nrow(secoll), 656L)
  expect_equal(nrow(secollumi), 656L)
})

Ldef <- list(
    mergeForwardReverse = TRUE, 
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
    revComplForward = FALSE, revComplReverse = TRUE,
    elementsForward = "SUCV", elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "", 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAGTTCATCCTGGCAGC", 
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0, 
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE, 
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, 
    verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
)
Ldef1 <- c(
    list(fastqForward = system.file("extdata/cisInput_1.fastq.gz", package = "mutscan"),
         fastqReverse = system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
    ), Ldef)
Ldef2 <- c(
    list(fastqForward = system.file("extdata/cisOutput_1.fastq.gz", package = "mutscan"),
         fastqReverse = system.file("extdata/cisOutput_2.fastq.gz", package = "mutscan")
    ), Ldef)

out1 <- do.call(digestFastqs, Ldef1)
out2 <- do.call(digestFastqs, Ldef2)

coldata <- data.frame(Name = c("sample1", "sample2", "sample3", "sample4"), 
                      Condition = c("input", "output", "input", "output"),
                      Replicate = c(1, 1, 2, 2), OD = c(0.05, 1.5, 0.05, 1.5),
                      stringsAsFactors = FALSE)

se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2,
                                   sample3 = out2, sample4 = out1),
                          coldata = coldata, countType = "umis")
secoll <- collapseMutantsByAA(se)

test_that("calculateFitnessScore fails with incorrect arguments", {
    expect_error(calculateFitnessScore(se = 1, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = out1, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = list(sample1 = out1, sample2 = out2), 
                                       pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Unknown", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = 1, ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "OD", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "Unknown",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "Condition",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = 1,
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Unknown", "output", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "unknown", "input"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "output", "unknown"),
                                       WTrows = "f.0.NA"))
    expect_error(calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "OD",
                                       comparison = c("Condition", "output", "input"),
                                       WTrows = "f.0.NA", selAssay = "missing"))
})

test_that("calculateFitnessScore works as expected", {
    ppi <- calculateFitnessScore(se = secoll, pairingCol = "Replicate", ODCols = "OD",
                                 comparison = c("Condition", "output", "input"), 
                                 WTrows = "f.0.WT")
    
    expect_equal(nrow(ppi), nrow(secoll))
    expect_equal(rownames(ppi), rownames(secoll))
    expect_equal(ncol(ppi), 2)
    expect_equal(colnames(ppi), c("output_vs_input_repl1", "output_vs_input_repl2"))
    expect_equal(ppi["f.0.WT", ], c(1, 1), ignore_attr = TRUE)
    
    ## Test "replicate 1"
    w <- SummarizedExperiment::colData(secoll)$Condition == "output" & 
        SummarizedExperiment::colData(secoll)$Replicate == 1
    ncout <- as.matrix(SummarizedExperiment::assay(secoll, "counts"))[, w]
    ncout <- ncout * SummarizedExperiment::colData(secoll)$OD[w]/sum(ncout)
    
    w <- SummarizedExperiment::colData(secoll)$Condition == "input" & 
        SummarizedExperiment::colData(secoll)$Replicate == 1
    ncin <- as.matrix(SummarizedExperiment::assay(secoll, "counts"))[, w]
    ncin <- ncin * SummarizedExperiment::colData(secoll)$OD[w]/sum(ncin)
    
    ratios <- log2(ncout/ncin)
    ratios[!is.finite(ratios)] <- NA
    ratios <- ratios/ratios["f.0.WT"]
    
    expect_equal(ppi[, "output_vs_input_repl1"], ratios)
    
    ## Test "replicate 2"
    w <- SummarizedExperiment::colData(secoll)$Condition == "output" & 
        SummarizedExperiment::colData(secoll)$Replicate == 2
    ncout <- as.matrix(SummarizedExperiment::assay(secoll, "counts"))[, w]
    ncout <- ncout * SummarizedExperiment::colData(secoll)$OD[w]/sum(ncout)
    
    w <- SummarizedExperiment::colData(secoll)$Condition == "input" & 
        SummarizedExperiment::colData(secoll)$Replicate == 2
    ncin <- as.matrix(SummarizedExperiment::assay(secoll, "counts"))[, w]
    ncin <- ncin * SummarizedExperiment::colData(secoll)$OD[w]/sum(ncin)
    
    ratios <- log2(ncout/ncin)
    ratios[!is.finite(ratios)] <- NA
    
    ## Test performance if WTrows = NULL
    ppinull <- calculateFitnessScore(se = secoll, pairingCol = "Replicate",
                                     ODCols = "OD",
                                     comparison = c("Condition", "output", "input"),
                                     WTrows = NULL)
    expect_equal(ppinull[, "output_vs_input_repl2"] , ratios)
    
    ## Use the WTrows
    ratios <- ratios/ratios["f.0.WT"]
    expect_equal(ppi[, "output_vs_input_repl2"], ratios)
    
    ## Test that PPI scores of input/output gives the same as output/input
    ppirev <- calculateFitnessScore(se = secoll, pairingCol = "Replicate", 
                                    ODCols = "OD",
                                    comparison = c("Condition", "input", "output"), 
                                    WTrows = "f.0.WT")
    expect_equal(ppirev[, "input_vs_output_repl1"], ppi[, "output_vs_input_repl1"])
    
})

#' Generate QC report
#'
#' @param se A \code{SummarizedExperiment} object, typically generated with
#'     \code{summarizeExperiment()}.
#' @param outFile Character string providing the name of the output file.
#'     Should have the extension \code{.html}.
#' @param reportTitle Character string specifying the title of the QC report.
#' @param forceOverwrite Logical scalar, indicating whether an existing file
#'     with the same name as \code{outFile} should be overwritten.
#' @param ... Additional parameters to be forwarded to
#'     \code{\link[rmarkdown]{render}}, for example \code{quiet = TRUE}.
#'
#' @export
#' @author Charlotte Soneson
#'
#' @returns Invisibly, the path to the generated html file.
#'
#' @seealso \code{\link[rmarkdown]{render}} used to render the html output file.
#'
#' @importFrom rmarkdown render
#' @importFrom xfun Rscript_call
#' @importFrom DT datatable
#' @importFrom SummarizedExperiment colData
#' @importFrom dplyr bind_rows full_join
#' @importFrom S4Vectors metadata
#' @importFrom tools file_ext
#'
#' @examples
#' ## Load SummarizedExperiment object
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                           package = "mutscan"))
#' ## Define output file
#' outfile <- tempfile(fileext = ".html")
#'
#' ## Generate QC report
#' generateQCReport(se, outfile)
#'
generateQCReport <- function(se, outFile, reportTitle = "mutscan QC report",
                             forceOverwrite = FALSE, ...) {

    ## --------------------------------------------------------------------- ##
    ## Check that input arguments are valid
    ## --------------------------------------------------------------------- ##
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = outFile, type = "character")
    .assertScalar(x = forceOverwrite, type = "logical")
    if (file_ext(outFile) != "html") {
        stop("'outFile' must have the file extension '.html'.")
    }
    outDir <- dirname(outFile)
    if (!dir.exists(outDir)) {
        dir.create(outDir, recursive = TRUE)
    }
    if (file.exists(outFile)) {
        if (!forceOverwrite) {
            stop(outFile, " already exists and forceOverwrite = FALSE, ", 
                 "stopping.")
        } else {
            message(outFile, " already exists but forceOverwrite = TRUE, ", 
                    "overwriting.")
        }
    }

    ## --------------------------------------------------------------------- ##
    ## Render the Rmd file
    ## --------------------------------------------------------------------- ##
    args <- list()
    args$input <- system.file("templates", "qc_report_template.Rmd",
                              package = "mutscan")
    args$output_format <- "html_document"
    args$output_file <- basename(outFile)
    args$output_dir <- outDir
    args$intermediates_dir <- outDir
    args$run_pandoc <- TRUE
    args$params <- list(se = se, title = reportTitle)
    if (any(names(args) %in% ...names())) {
        warning("Will ignore arguments provided in '...' that are set ",
                "internally by generateQCReport: ",
                paste(intersect(names(args), ...names()), collapse = ", "))
    }
    args <- c(args, list(...)[setdiff(...names(), names(args))])

    outputReport <- Rscript_call(
        render, args
    )

    ## --------------------------------------------------------------------- ##
    ## Return (invisibly) the path to the rendered html file
    ## --------------------------------------------------------------------- ##
    invisible(outFile)
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:100, ]
outFile <- tempfile(fileext = ".html")
outFile2 <- file.path(tempdir(), "newfolder", "newreport.html")

test_that("generateQCReport fails with incorrect input", {
    expect_error(generateQCReport(se = NULL, outFile = outFile))
    expect_error(generateQCReport(se = "file", outFile = outFile))

    expect_error(generateQCReport(se = se, outFile = 1))
    expect_error(generateQCReport(se = se, outFile = "out.pdf"))
    expect_error(generateQCReport(se = se, outFile = c("out1.html", "out2.html")))

    expect_error(generateQCReport(se = se, outFile = outFile, forceOverwrite = 1))
    expect_error(generateQCReport(se = se, outFile = outFile, forceOverwrite = "wrong"))
    expect_error(generateQCReport(se = se, outFile = outFile, forceOverwrite = c(TRUE, FALSE)))
})

test_that("generateQCReport works as expected", {
    expect_warning(expect_equal(generateQCReport(se = se, outFile = outFile,
                                                 forceOverwrite = FALSE,
                                                 quiet = TRUE,
                                                 run_pandoc = FALSE),
                                outFile),
                   "Will ignore arguments")
    expect_true(file.exists(outFile))
    expect_error(generateQCReport(se = se, outFile = outFile, forceOverwrite = FALSE))
    expect_message(expect_equal(generateQCReport(se = se, outFile = outFile,
                                                 forceOverwrite = TRUE,
                                                 quiet = TRUE),
                                outFile),
                   "overwriting")
    expect_equal(generateQCReport(se = se, outFile = outFile2,
                                  forceOverwrite = FALSE, quiet = TRUE),
                 outFile2)
})

#' Construct an (interactive) scatterplot
#'
#' This function provides a convenience wrapper for constructing a scatter
#' plot of two columns in a data.frame, and color the points by whether
#' the values in a third column are below a provided value.
#' Users will interact with the function via the higher-level wrappers
#' \code{plotVolcano()} and \code{plotMeanDiff()}, which both call this
#' function internally.
#'
#' @param res A \code{data.frame} with values to plot.
#' @param xCol,yCol Character scalars defining the columns to plot on the
#'     x- and y-axis, respectively.
#' @param xLabel,yLabel Character scalars providing the x- and y-axis
#'     titles, respectively.
#' @param colorCol Character scalar defining the column to use to color
#'     the points (typically the adjusted p-value).
#' @param colorThreshold Numeric scalar. If the value in \code{colorCol}
#'     is lower than colorThreshold, the point will be indicated in red.
#' @param labelCol Character scalar defining the column to use to label
#'     individual points in the plot.
#' @param labelValues Character vector defining the points to highlight with
#'     labels. Must correspond to values in the \code{labelCol} column.
#' @param centerAxis Character vector, a subset of c("x", "y") or \code{NULL}.
#'     Indicates the axis that should be centered (i.e., where the axis
#'     limits should be extended to put 0 in the center).
#' @param pointSize Character scalar, either \code{"small"} or \code{"large"},
#'     defining the style to use to plot the points.
#' @param interactivePlot Logical scalar, whether to return an interactive
#'     plot.
#'
#' @return If \code{interactivePlot} is \code{TRUE}, a \code{plotly}
#'     object. If \code{interactivePlot} is \code{FALSE}, a \code{ggplot2}
#'     object.
#'
#' @author Charlotte Soneson
#' @keywords internal
#' @noRd
#'
#' @importFrom ggplot2 ggplot theme_minimal coord_cartesian theme labs
#'     element_text geom_point aes
#' @importFrom rlang .data
#' @importFrom ggrepel geom_text_repel
.plotScatter <- function(res, xCol, yCol, xLabel = xCol, yLabel = yCol,
                         colorCol, colorThreshold,
                         labelCol = NULL, labelValues = c(), centerAxis = "x",
                         pointSize = "small", interactivePlot = FALSE) {
    .assertVector(x = res, type = "data.frame")
    .assertScalar(x = xCol, type = "character", validValues = colnames(res))
    .assertScalar(x = yCol, type = "character", validValues = colnames(res))
    .assertScalar(x = xLabel, type = "character")
    .assertScalar(x = yLabel, type = "character")
    .assertScalar(x = colorCol, type = "character",
                  validValues = colnames(res))
    .assertScalar(x = colorThreshold, type = "numeric")
    .assertScalar(x = labelCol, type = "character",
                  validValues = colnames(res), allowNULL = TRUE)
    .assertVector(x = labelValues, type = "character", allowNULL = TRUE)
    .assertVector(x = centerAxis, type = "character",
                  validValues = c("x", "y"), allowNULL = TRUE)
    .assertScalar(x = pointSize, type = "character",
                  validValues = c("small", "large"))
    .assertScalar(x = interactivePlot, type = "logical")
    
    if (interactivePlot && !requireNamespace("plotly", quietly = TRUE)) {
        stop("The 'plotly' package is required to make interactive plots.") # nocov
    }
    xr <- range(res[[xCol]], na.rm = TRUE)
    yr <- range(res[[yCol]], na.rm = TRUE)
    if ("x" %in% centerAxis) {
        xr <- c(-max(abs(xr), na.rm = TRUE), max(abs(xr), na.rm = TRUE))
    }
    if ("y" %in% centerAxis) {
        yr <- c(-max(abs(yr), na.rm = TRUE), max(abs(yr), na.rm = TRUE))
    }
    gg <- ggplot(res, aes(x = .data[[xCol]], 
                          y = .data[[yCol]])) +
        theme_minimal() + 
        coord_cartesian(xlim = xr, ylim = yr) +
        theme(axis.text = element_text(size = 12),
              axis.title = element_text(size = 14),
              title = element_text(size = 14)) +
        labs(x = xLabel, y = yLabel)
    if (!is.null(labelCol)) {
        gg <- gg +
            aes(label = .data[[labelCol]])
    }
    if (pointSize == "large") {
        gg <- gg +
            geom_point(fill = "lightgrey", color = "grey", 
                       pch = 21, size = 1.5)
    } else {
        gg <- gg +
            geom_point(color = "black", size = 0.25, alpha = 0.5)
    }
    if (any(res[[colorCol]] < colorThreshold)) {
        gg <- gg +
            labs(subtitle = paste0("Points are indicated in red if ",
                                   colorCol, "<", colorThreshold))
        if (pointSize == "large") {
            gg <- gg +
                geom_point(
                    data = res[res[[colorCol]] < colorThreshold, , 
                               drop = FALSE],
                    fill = "red", color = "grey", pch = 21, size = 1.5
                )
        } else {
            gg <- gg +
                geom_point(
                    data = res[res[[colorCol]] < colorThreshold, , 
                               drop = FALSE],
                    color = "red", size = 0.25
                )
        }
    }
    if (interactivePlot) {
        plotly::ggplotly(gg) # nocov
    } else {
        if (!is.null(labelCol) && length(labelValues) > 0) {
            .assertVector(x = labelValues, type = "character", 
                          validValues = res[[labelCol]])
            gg <- gg +
                geom_text_repel(
                    data = res[res[[labelCol]] %in% labelValues, , 
                               drop = FALSE],
                    max.overlaps = Inf, size = 4, min.segment.length = 0.1)
        }
        gg
    }
}

.getColName <- function(res, validValues, aspect) {
    .assertVector(x = res, type = "data.frame")
    colName <- grep(paste(paste0("^", validValues, "$"), collapse = "|"),
                    colnames(res), value = TRUE)
    if (length(colName) == 0) {
        stop("No suitable column found for ", aspect, ", one of the ",
             "following expected: ", paste(validValues, collapse = ", "))
    }
    if (length(colName) > 1) {
        warning("Multiple suitable columns found for ", aspect,
                "; choosing the first: ", colName[1])
        colName <- colName[1]
    }
    colName
}

#' Construct an MA (mean-difference) plot
#'
#' @param res \code{data.frame} (typically output from
#'     \code{calculateRelativeFC()}) with columns corresponding to the
#'     average abundance (\code{logCPM} or \code{AveExpr}), log-fold
#'     change (\code{logFC}) and significance (\code{FDR} or
#'     \code{adj.P.Val}).
#' @param meanCol,logFCCol,pvalCol,padjCol Character scalars indicating the
#'     columns from \code{res} that will be used to represent the
#'     mean value (x-axis), logFC (y-axis), nominal p-value (used to find
#'     the top features to label) and adjusted p-value (used
#'     for coloring). If \code{NULL} (default), pre-specified values
#'     will be used depending on the available columns
#'     (\code{"logCPM"} or \code{"AveExpr"}, \code{"logFC"},
#'     \code{"PValue"} or \code{"P.Value"}, and
#'     \code{"FDR"} or \code{"adj.P.Val"}, respectively).
#' @param padjThreshold Numeric scalar indicating the adjusted p-value
#'     threshold to use for coloring the points. All features with
#'     adjusted p-value below the treshold will be shown in red.
#' @param pointSize Either \code{"small"} or \code{"large"}, indicating
#'     which of the two available plot styles that will be used.
#' @param interactivePlot Logical scalar, indicating whether an
#'     interactive plot should be returned, in which one can hover
#'     over the individual points and obtain further information.
#' @param nTopToLabel Numeric scalar, indicating the number of points that
#'     should be labeled in the plot. The points will be ranked by the
#'     \code{pvalCol} column, and the top \code{nTopToLabel} values will
#'     be labeled by the corresponding row names. Only used if
#'     \code{interactivePlot} is \code{FALSE}.
#'
#' @return If \code{interactivePlot} is \code{TRUE}, a \code{plotly}
#'     object. If \code{interactivePlot} is \code{FALSE}, a \code{ggplot2}
#'     object.
#'
#' @author Charlotte Soneson
#' @export
#'
#' @examples
#' library(SummarizedExperiment)
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                           package = "mutscan"))[1:200, ]
#' design <- model.matrix(~ Replicate + Condition,
#'                        data = colData(se))
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output")
#' plotMeanDiff(res, pointSize = "large", nTopToLabel = 3)
#'
plotMeanDiff <- function(res, meanCol = NULL, logFCCol = NULL, pvalCol = NULL,
                         padjCol = NULL, padjThreshold = 0.05, 
                         pointSize = "small",
                         interactivePlot = FALSE, nTopToLabel = 0) {
    .assertScalar(x = nTopToLabel, type = "numeric", rngIncl = c(0, Inf))
    .assertVector(x = res, type = "data.frame")
    ## Get the columns to use
    if (is.null(meanCol)) {
        meanCol <- .getColName(res, validValues = c("logCPM", "AveExpr"),
                               aspect = "x-axis")
    }
    if (is.null(logFCCol)) {
        logFCCol <- .getColName(res, validValues = c("logFC"),
                                aspect = "y-axis")
    }
    if (is.null(pvalCol)) {
        pvalCol <- .getColName(res, validValues = c("PValue", "P.Value"),
                               aspect = "labeling")
    }
    if (is.null(padjCol)) {
        padjCol <- .getColName(res, validValues = c("FDR", "adj.P.Val"),
                               aspect = "highlighting")
    }
    .assertScalar(pvalCol, type = "character", validValues = colnames(res))
    res$feature <- rownames(res)
    if (nTopToLabel > 0) {
        labelValues <- res[order(res[[pvalCol]]), 
                           "feature"][seq_len(nTopToLabel)]
    } else {
        labelValues <- c()
    }
    
    .plotScatter(res = res, xCol = meanCol, yCol = logFCCol,
                 xLabel = meanCol, yLabel = logFCCol,
                 colorCol = padjCol,
                 colorThreshold = padjThreshold,
                 labelCol = "feature", labelValues = labelValues,
                 centerAxis = "y", pointSize = pointSize,
                 interactivePlot = interactivePlot)
}

#' Construct a volcano plot
#'
#' @param res \code{data.frame} (typically output from
#'     \code{calculateRelativeFC()}) with columns corresponding to the
#'     log-fold change (\code{logFC}), p-value (\code{PValue} or
#'     \code{P.Value}) and significance (\code{FDR} or \code{adj.P.Val}).
#' @param logFCCol,pvalCol,padjCol Character scalars indicating the
#'     columns from \code{res} that will be used to represent the
#'     logFC (x-axis), p-value (y-axis) and adjusted p-value (used
#'     for coloring). If \code{NULL} (default), pre-specified values
#'     will be used depending on the available columns
#'     (\code{"logFC"}, \code{"PValue"} or \code{"P.Value"}, and
#'     \code{"FDR"} or \code{"adj.P.Val"}, respectively).
#' @param padjThreshold Numeric scalar indicating the adjusted p-value
#'     threshold to use for coloring the points. All features with
#'     adjusted p-value below the treshold will be shown in red.
#' @param pointSize Either \code{"small"} or \code{"large"}, indicating
#'     which of the two available plot styles that will be used.
#' @param interactivePlot Logical scalar, indicating whether an
#'     interactive plot should be returned, in which one can hover
#'     over the individual points and obtain further information.
#' @param nTopToLabel Numeric scalar, indicating the number of points that
#'     should be labeled in the plot. The points will be ranked by the
#'     \code{pvalCol} column, and the top \code{nTopToLabel} values will
#'     be labeled by the corresponding row names. Only used if
#'     \code{interactivePlot} is \code{FALSE}.
#'
#' @return If \code{interactivePlot} is \code{TRUE}, a \code{plotly}
#'     object. If \code{interactivePlot} is \code{FALSE}, a \code{ggplot2}
#'     object.
#'
#' @author Charlotte Soneson
#' @export
#'
#' @examples
#' library(SummarizedExperiment)
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                           package = "mutscan"))[1:200, ]
#' design <- model.matrix(~ Replicate + Condition,
#'                        data = colData(se))
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output")
#' plotVolcano(res, pointSize = "large", nTopToLabel = 3)
#'
plotVolcano <- function(res, logFCCol = NULL, pvalCol = NULL, padjCol = NULL,
                        padjThreshold = 0.05, pointSize = "small",
                        interactivePlot = FALSE, nTopToLabel = 0) {
    .assertScalar(x = nTopToLabel, type = "numeric", rngIncl = c(0, Inf))
    .assertVector(x = res, type = "data.frame")
    ## Get the columns to use
    if (is.null(logFCCol)) {
        logFCCol <- .getColName(res, validValues = c("logFC"),
                                aspect = "x-axis")
    }
    if (is.null(pvalCol)) {
        pvalCol <- .getColName(res, validValues = c("PValue", "P.Value"),
                               aspect = "y-axis")
    }
    if (is.null(padjCol)) {
        padjCol <- .getColName(res, validValues = c("FDR", "adj.P.Val"),
                               aspect = "highlighting")
    }
    .assertScalar(pvalCol, type = "character", validValues = colnames(res))
    res$negLog10P <- -log10(res[[pvalCol]])
    res$feature <- rownames(res)
    if (nTopToLabel > 0) {
        labelValues <- res[order(res[[pvalCol]]),
                           "feature"][seq_len(nTopToLabel)]
    } else {
        labelValues <- c()
    }
    
    .plotScatter(res = res, xCol = logFCCol, yCol = "negLog10P",
                 xLabel = logFCCol, yLabel = paste0("-log10(", pvalCol, ")"),
                 colorCol = padjCol,
                 colorThreshold = padjThreshold,
                 labelCol = "feature", labelValues = labelValues,
                 centerAxis = "x", pointSize = pointSize,
                 interactivePlot = interactivePlot)
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:200, ]
design <- stats::model.matrix(~ Replicate + Condition, 
                              data = SummarizedExperiment::colData(se))
res0 <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                            contrast = NULL, WTrows = "f.0.WT", selAssay = "counts",
                            pseudocount = 1, method = "edgeR",
                            normMethod = "sum")

test_that("plotResults functions fail with incorrect input", {

    expect_error(plotMeanDiff(as.matrix(res0)))
    expect_error(plotVolcano(as.matrix(res0)))

    expect_error(plotMeanDiff(res0, meanCol = 1))
    expect_error(plotMeanDiff(res0, meanCol = "missing"))
    expect_error(plotMeanDiff(res0, meanCol = c("AveExpr", "logCPM")))
    expect_error(plotMeanDiff(res0, logFCCol = 1))
    expect_error(plotMeanDiff(res0, logFCCol = "missing"))
    expect_error(plotMeanDiff(res0, logFCCol = c("AveExpr", "logFC")))
    expect_error(plotMeanDiff(res0, pvalCol = 1))
    expect_error(plotMeanDiff(res0, pvalCol = "missing"))
    expect_error(plotMeanDiff(res0, pvalCol = c("PValue", "P.Value")))
    expect_error(plotMeanDiff(res0, padjCol = 1))
    expect_error(plotMeanDiff(res0, padjCol = "missing"))
    expect_error(plotMeanDiff(res0, padjCol = c("FDR", "adj.P.Val")))

    expect_error(plotVolcano(res0, logFCCol = 1))
    expect_error(plotVolcano(res0, logFCCol = "missing"))
    expect_error(plotVolcano(res0, logFCCol = c("AveExpr", "logFC")))
    expect_error(plotVolcano(res0, pvalCol = 1))
    expect_error(plotVolcano(res0, pvalCol = "missing"))
    expect_error(plotVolcano(res0, pvalCol = c("PValue", "P.Value")))
    expect_error(plotVolcano(res0, padjCol = 1))
    expect_error(plotVolcano(res0, padjCol = "missing"))
    expect_error(plotVolcano(res0, padjCol = c("FDR", "adj.P.Val")))

    expect_error(plotMeanDiff(res0, padjThreshold = "0.05"))
    expect_error(plotMeanDiff(res0, padjThreshold = c(0.01, 0.05)))
    expect_error(plotVolcano(res0, padjThreshold = "0.05"))
    expect_error(plotVolcano(res0, padjThreshold = c(0.01, 0.05)))

    expect_error(plotMeanDiff(res0, pointSize = 1))
    expect_error(plotMeanDiff(res0, pointSize = c("small", "large")))
    expect_error(plotMeanDiff(res0, pointSize = "wrong"))
    expect_error(plotVolcano(res0, pointSize = 1))
    expect_error(plotVolcano(res0, pointSize = c("small", "large")))
    expect_error(plotVolcano(res0, pointSize = "wrong"))

    expect_error(plotMeanDiff(res0, interactivePlot = 1))
    expect_error(plotMeanDiff(res0, interactivePlot = c(TRUE, FALSE)))
    expect_error(plotMeanDiff(res0, interactivePlot = "TRUE"))
    expect_error(plotVolcano(res0, interactivePlot = 1))
    expect_error(plotVolcano(res0, interactivePlot = c(TRUE, FALSE)))
    expect_error(plotVolcano(res0, interactivePlot = "TRUE"))

    expect_error(plotMeanDiff(res0, nTopToLabel = "1"))
    expect_error(plotMeanDiff(res0, nTopToLabel = c(1, 2)))
    expect_error(plotMeanDiff(res0, nTopToLabel = -1))
    expect_error(plotVolcano(res0, nTopToLabel = "1"))
    expect_error(plotVolcano(res0, nTopToLabel = c(1, 2)))
    expect_error(plotVolcano(res0, nTopToLabel = -1))

    res1 <- res0
    colnames(res1)[colnames(res1) == "logCPM"] <- "wrong"
    expect_error(plotMeanDiff(res1),
                 "No suitable column found for x-axis")

    res1 <- res0
    colnames(res1)[colnames(res1) == "logFC"] <- "wrong"
    expect_error(plotMeanDiff(res1),
                 "No suitable column found for y-axis")
    expect_error(plotVolcano(res1),
                 "No suitable column found for x-axis")

    res1 <- res0
    colnames(res1)[colnames(res1) == "FDR"] <- "wrong"
    expect_error(plotMeanDiff(res1),
                 "No suitable column found for highlighting")
    expect_error(plotVolcano(res1),
                 "No suitable column found for highlighting")

    res1 <- res0
    colnames(res1)[colnames(res1) == "PValue"] <- "wrong"
    expect_error(plotVolcano(res1),
                 "No suitable column found for y-axis")

    res1 <- res0
    res1$AveExpr <- res1$logCPM
    expect_warning(plotMeanDiff(res1),
                   "Multiple suitable columns found for x-axis")

    res1 <- res0
    res1$P.Value <- res1$PValue
    expect_warning(plotVolcano(res1),
                   "Multiple suitable columns found for y-axis")

    res1 <- res0
    res1$adj.P.Val <- res1$FDR
    expect_warning(plotMeanDiff(res1),
                   "Multiple suitable columns found for highlighting")
    expect_warning(plotVolcano(res1),
                   "Multiple suitable columns found for highlighting")
})

test_that("plotResults functions work as expected", {
    res1 <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                contrast = NULL, WTrows = "f.0.WT", selAssay = "counts",
                                pseudocount = 1, method = "edgeR",
                                normMethod = "sum")
    res2 <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                contrast = NULL, WTrows = "f.0.WT", selAssay = "counts",
                                pseudocount = 1, method = "limma",
                                normMethod = "sum")

    expect_true(ggplot2::is_ggplot(plotMeanDiff(res1)))
    expect_true(ggplot2::is_ggplot(plotMeanDiff(res2)))
    expect_true(ggplot2::is_ggplot(plotVolcano(res1)))
    expect_true(ggplot2::is_ggplot(plotVolcano(res2)))

    expect_true(ggplot2::is_ggplot(
        plotMeanDiff(res1, meanCol = "logCPM", logFCCol = "logFC",
                     padjCol = "FDR", padjThreshold = 0.1)))
    expect_true(ggplot2::is_ggplot(
        plotMeanDiff(res2, meanCol = "AveExpr", logFCCol = "logFC",
                     padjCol = "adj.P.Val", padjThreshold = 0.1)))
    expect_true(ggplot2::is_ggplot(
        plotVolcano(res1, logFCCol = "logFC", pvalCol = "PValue",
                    padjCol = "FDR", padjThreshold = 0.1)))
    expect_true(ggplot2::is_ggplot(
        plotVolcano(res2, logFCCol = "logFC", pvalCol = "P.Value",
                    padjCol = "adj.P.Val", padjThreshold = 0.1)))

    expect_true(ggplot2::is_ggplot(plotMeanDiff(res1, nTopToLabel = 5)))
    expect_true(ggplot2::is_ggplot(plotMeanDiff(res2, nTopToLabel = 5)))
    expect_true(ggplot2::is_ggplot(plotVolcano(res1, nTopToLabel = 5)))
    expect_true(ggplot2::is_ggplot(plotVolcano(res2, nTopToLabel = 5)))

    expect_true(ggplot2::is_ggplot(plotMeanDiff(res1, pointSize = "large")))
    expect_true(ggplot2::is_ggplot(plotMeanDiff(res2, pointSize = "large")))
    expect_true(ggplot2::is_ggplot(plotVolcano(res1, pointSize = "large")))
    expect_true(ggplot2::is_ggplot(plotVolcano(res2, pointSize = "large")))

    ## These tests won't run if the X11 display connection can not be opened
    # skip_if_not_installed("plotly")
    # expect_s3_class(plotMeanDiff(res1, interactivePlot = TRUE), "plotly")
    # expect_s3_class(plotMeanDiff(res2, interactivePlot = TRUE), "plotly")
    # expect_s3_class(plotVolcano(res1, interactivePlot = TRUE), "plotly")
    # expect_s3_class(plotVolcano(res2, interactivePlot = TRUE), "plotly")
})

#' Make pairs plot of selected assay from a SummarizedExperiment object
#'
#' Construct a pairs plot of all columns of a given assay. The lower-triangular
#' panels display the scatter plots, the upper-triangular ones print out
#' the (Pearson or Spearman) correlations, and the diagonal panels show
#' histograms of the respective columns.
#'
#' @author Charlotte Soneson
#'
#' @export
#'
#' @return A \code{ggplot} object.
#'
#' @param se A SummarizedExperiment object, e.g. the output of
#'     \code{summarizeExperiment}
#' @param selAssay Character scalar, the assay to use as the basis for the
#'     pairs plot.
#' @param doLog Logical scalar, whether or not to log-transform the values
#'     before plotting.
#' @param pseudocount Numeric scalar, the pseudocount to add to the values
#'     before log-transforming (if \code{doLog} is \code{TRUE}).
#' @param corMethod Either "pearson" or "spearman", the type of correlation
#'     to calculate.
#' @param histBreaks Numeric scalar, the number of breaks in the histograms
#'     to put in the diagonal panels.
#' @param pointsType Either "points", "smoothscatter", "scattermore" or 
#'     "scattermost" (the latter two require the "scattermore" package to be 
#'     installed), determining the type of plots that will be made.
#' @param corSizeMult,corSizeAdd Numeric scalars determining how the
#'     absolute correlation value is transformed into a font size. The
#'     transformation is corSizeMult * abs(corr) + corSizeAdd.
#' @param pointSize,pointAlpha Numeric scalars determining the size and
#'     opacity of points in the plot.
#' @param colorByCorrelation Logical scalar, indicating whether the correlation
#'     panels should be colored according to the correlation value.
#' @param corrColorRange Numeric vector of length 2, providing the lower and 
#'     upper limits of the color scale when coloring by correlation. Both 
#'     values should be positive; the same range is used for negative 
#'     correlations. If \code{NULL} (the default), the range is inferred from 
#'     the data.
#' @param addIdentityLine Logical scalar, indicating whether the identity line 
#'     should be added (only used if \code{pointsType = "points"}).
#'
#' @importFrom GGally eval_data_col ggpairs
#' @importFrom ggplot2 ggplot annotate theme_void ylim stat_density2d
#'     scale_fill_continuous geom_point theme_bw theme element_blank aes
#'     geom_histogram scale_x_continuous scale_y_continuous geom_abline
#'     after_stat element_rect
#' @importFrom stats cor
#' @importFrom SummarizedExperiment assayNames assay
#' @importFrom grDevices hcl.colors rgb colorRamp
#' @importFrom rlang .data
#'
#' @examples
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
#'                           package = "mutscan"))[1:200, ]
#' plotPairs(se)
#' 
plotPairs <- function(se, selAssay = "counts", doLog = TRUE, pseudocount = 1,
                      corMethod = "pearson", histBreaks = 40,
                      pointsType = "points", corSizeMult = 5,
                      corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                      colorByCorrelation = TRUE, corrColorRange = NULL,
                      addIdentityLine = FALSE) {
    
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = selAssay, type = "character",
                  validValues = assayNames(se))
    .assertScalar(x = doLog, type = "logical")
    .assertScalar(x = pseudocount, type = "numeric", rngIncl = c(0, Inf))
    .assertScalar(x = corMethod, type = "character",
                  validValues = c("pearson", "spearman"))
    .assertScalar(x = histBreaks, type = "numeric", rngExcl = c(0, Inf))
    .assertScalar(x = pointsType, type = "character",
                  validValues = c("smoothscatter", "points", "scattermore",
                                  "scattermost"))
    .assertScalar(x = corSizeMult, type = "numeric", rngExcl = c(0, Inf))
    .assertScalar(x = corSizeAdd, type = "numeric", rngIncl = c(0, Inf))
    .assertScalar(x = pointSize, type = "numeric", rngExcl = c(0, Inf))
    .assertScalar(x = pointAlpha, type = "numeric", rngExcl = c(0, Inf))
    .assertScalar(x = colorByCorrelation, type = "logical")
    .assertVector(x = corrColorRange, type = "numeric", rngIncl = c(0, 1), 
                  len = 2, allowNULL = TRUE)
    if (!is.null(corrColorRange)) {
        stopifnot(corrColorRange[2] >= corrColorRange[1])
    }
    .assertScalar(x = addIdentityLine, type = "logical")
    if (pointsType %in% c("scattermore", "scattermost")) {
        .assertPackagesAvailable("scattermore")
    }
    
    ## --------------------------------------------------------------------- ##
    ## Define shared theme elements
    ## --------------------------------------------------------------------- ##
    ggtheme <- list(
        theme_bw(),
        theme(panel.grid.major = element_blank(),
              panel.grid.minor = element_blank())
    )
    
    ## --------------------------------------------------------------------- ##
    ## Correlations
    ## --------------------------------------------------------------------- ##
    ## Define function to calculate and display correlations 
    ## (for use with ggpairs)
    cor_fcn <- function(data, mapping, ...) {
        ## Get data
        xData <- eval_data_col(data, mapping$x)
        yData <- eval_data_col(data, mapping$y)
        
        ## Calculate correlation
        mainCor <- cor(xData, yData, method = corMethod, 
                       use = "pairwise.complete.obs")
        transfCor <- (abs(mainCor) - min(corRange)) / 
            (max(corRange) - min(corRange))
        ## Protect against numerical imprecision leading to values outside
        ## the allowed range
        transfCor <- min(1.0, transfCor)
        transfCor <- max(0.0, transfCor)
        
        ## Determine the color
        if (colorByCorrelation) {
            if (mainCor >= 0) {
                cols <- hcl.colors(n = 11, palette = "RdBu")[6:2]
                col <- rgb(colorRamp(
                    cols)(transfCor),
                    maxColorValue = 255)
            } else {
                # nocov start
                cols <- hcl.colors(n = 11, palette = "RdBu")[6:10]
                col <- rgb(colorRamp(
                    cols)(transfCor),
                    maxColorValue = 255)
                # nocov end
            }
        } else {
            col <- "white" #nocov
        }
        
        ## Construct plot
        ggplot(data = data, mapping = mapping) +
            annotate(x = 0.5, y = 0.5, 
                     label = round(mainCor, digits = 3),
                     geom = "text",
                     size = abs(mainCor) * corSizeMult + corSizeAdd) +
            ggtheme + ylim(c(0, 1)) +
            theme(
                panel.background = element_rect(fill = col)
            )
    }
    
    ## --------------------------------------------------------------------- ##
    ## Scatter plots
    ## --------------------------------------------------------------------- ##
    # All the point plotting functions are tested, but for some reason covr 
    # doesn't pick it up - ignore these definitions in the coverage tests
    # nocov start
    if (pointsType == "smoothscatter") {
        ## Define function to create smoothscatter-like plot 
        ## (for use with ggpairs)
        smoothscat <- function(data, mapping, ...) {
            ggplot(data = data, mapping = mapping) +
                stat_density2d(
                    aes(fill = after_stat(
                        .data$density)^0.25), geom = "tile",
                    contour = FALSE, n = 200) +
                scale_fill_continuous(low = "white", 
                                      high = "darkgreen") +
                ggtheme +
                scale_x_continuous(expand = c(0, 0)) +
                scale_y_continuous(expand = c(0, 0))
        }
    } else if (pointsType == "points") {
        ## Define function to create scatter plot (for use with ggpairs)
        if (addIdentityLine) {
            plotpoints <- function(data, mapping, ...) {
                ggplot(data = data, mapping = mapping) +
                    geom_abline(slope = 1, intercept = 0, 
                                linetype = "dashed", color = "grey") + 
                    geom_point(alpha = pointAlpha, size = pointSize) +
                    ggtheme
            }
        } else {
            plotpoints <- function(data, mapping, ...) {
                ggplot(data = data, mapping = mapping) +
                    geom_point(alpha = pointAlpha, size = pointSize) +
                    ggtheme
            }
        }
    } else if (pointsType == "scattermore") {
        ## Define function to create scattermore plot (for use with ggpairs)
        if (addIdentityLine) {
            plotscattermore <- function(data, mapping, ...) {
                ggplot(data = data, mapping = mapping) +
                    geom_abline(slope = 1, intercept = 0, 
                                linetype = "dashed", color = "grey") + 
                    scattermore::geom_scattermore(alpha = pointAlpha, 
                                                  pointsize = pointSize) +
                    ggtheme
            }
        } else {
            plotscattermore <- function(data, mapping, ...) {
                ggplot(data = data, mapping = mapping) +
                    scattermore::geom_scattermore(alpha = pointAlpha, 
                                                  pointsize = pointSize) +
                    ggtheme
            }
        }
    } else if (pointsType == "scattermost") {
        ## Define function to create scattermost plot (for use with ggpairs)
        if (addIdentityLine) {
            plotscattermost <- function(data, mapping, ...) {
                ## Get data
                xData <- eval_data_col(data, mapping$x)
                yData <- eval_data_col(data, mapping$y)
                
                ggplot(data = data, mapping = mapping) +
                    geom_abline(slope = 1, intercept = 0, 
                                linetype = "dashed", color = "grey") + 
                    scattermore::geom_scattermost(xy = cbind(xData, yData), 
                                                  pointsize = pointSize) +
                    ggtheme
            }
        } else {
            plotscattermost <- function(data, mapping, ...) {
                ## Get data
                xData <- eval_data_col(data, mapping$x)
                yData <- eval_data_col(data, mapping$y)
                
                ggplot(data = data, mapping = mapping) +
                    scattermore::geom_scattermost(xy = cbind(xData, yData), 
                                                  pointsize = pointSize) +
                    ggtheme
            }
        }
    }
    # nocov end
    
    ## Define the function to use for the plots
    if (pointsType == "smoothscatter") {
        lower <- list(continuous = smoothscat)
    } else if (pointsType == "points") {
        lower <- list(continuous = plotpoints)
    } else if (pointsType == "scattermore") {
        lower <- list(continuous = plotscattermore)
    } else if (pointsType == "scattermost") {
        lower <- list(continuous = plotscattermost)
    }
    
    ## --------------------------------------------------------------------- ##
    ## Histogram
    ## --------------------------------------------------------------------- ##
    # nocov start
    diaghist <- function(data, mapping, ...) {
        gg <- ggplot(data = data, mapping = mapping) +
            geom_histogram(fill = "#F5C710", color = "grey50", 
                           bins = histBreaks) +
            ggtheme
        if (pointsType == "smoothscatter") {
            gg <- gg + scale_x_continuous(expand = c(0, 0))
        }
        gg
    }
    # nocov end
    
    ## --------------------------------------------------------------------- ##
    ## Combined plot
    ## --------------------------------------------------------------------- ##
    ## Prepare the data and plot title, depending on whether to log-transform 
    ## or not
    if (doLog) {
        mat <- log10(as.matrix(assay(se, selAssay)) + 
                         pseudocount)
        title <- paste0("log10(", selAssay,
                        ifelse(pseudocount > 0, 
                               paste0(" + ", pseudocount), ""),
                        ")")
    } else {
        mat <- as.matrix(assay(se, selAssay))
        title <- selAssay
    }
    mat <- as.data.frame(mat)
    title <- paste0(title, ", ", toupper(substr(corMethod, 1, 1)),
                    substr(corMethod, 2, nchar(corMethod)), " correlation")
    
    ## Calculate correlations and get range
    allCors <- cor(mat, method = corMethod, 
                   use = "pairwise.complete.obs")
    if (!is.null(corrColorRange)) {
        corRange <- corrColorRange
    } else {
        corRange <- range(abs(allCors[upper.tri(allCors)]))
    }
    if (length(unique(corRange)) == 1) {
        ## Having a range with width 0 leads to problems in the rescaling of 
        ## the correlations
        corRange <- c(0, 1)
    }
    
    ## Construct the pairs plot
    ggpairs(
        data = mat,
        mapping = NULL,
        columns = seq_len(ncol(mat)),
        title = title, xlab = NULL, ylab = NULL,
        upper = list(continuous = cor_fcn),
        lower = lower,
        diag = list(continuous = diaghist),
        progress = FALSE,
        axisLabels = "show"
    )
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:1000, 1:3]

test_that("plotPairs fails with incorrect arguments", {
    expect_error(plotPairs(se = 1))
    expect_error(plotPairs(se = "x"))
    expect_error(plotPairs(se = matrix(1:6), 2, 3))

    expect_error(plotPairs(se = se, selAssay = 1))
    expect_error(plotPairs(se = se, selAssay = c("counts", "logcounts")))
    expect_error(plotPairs(se = se, selAssay = "logcounts"))

    expect_error(plotPairs(se = se, selAssay = "counts", doLog = 1))
    expect_error(plotPairs(se = se, selAssay = "counts", doLog = "TRUE"))
    expect_error(plotPairs(se = se, selAssay = "counts", doLog = c(TRUE, FALSE)))

    expect_error(plotPairs(se = se, selAssay = "counts", pseudocount = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", pseudocount = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", pseudocount = -1))

    expect_error(plotPairs(se = se, selAssay = "counts", corMethod = 1))
    expect_error(plotPairs(se = se, selAssay = "counts",
                           corMethod = c("spearman", "pearson")))
    expect_error(plotPairs(se = se, selAssay = "counts", corMethod = "pearman"))

    expect_error(plotPairs(se = se, selAssay = "counts", histBreaks = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", histBreaks = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", histBreaks = 0))

    expect_error(plotPairs(se = se, selAssay = "counts", pointsType = 1))
    expect_error(plotPairs(se = se, selAssay = "counts",
                           pointsType = c("smoothscatter", "points")))
    expect_error(plotPairs(se = se, selAssay = "counts", pointsType = "oints"))

    expect_error(plotPairs(se = se, selAssay = "counts", corSizeMult = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", corSizeMult = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", corSizeMult = 0))

    expect_error(plotPairs(se = se, selAssay = "counts", corSizeAdd = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", corSizeAdd = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", corSizeAdd = -1))

    expect_error(plotPairs(se = se, selAssay = "counts", pointSize = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", pointSize = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", pointSize = 0))

    expect_error(plotPairs(se = se, selAssay = "counts", pointAlpha = "1"))
    expect_error(plotPairs(se = se, selAssay = "counts", pointAlpha = c(1, 2)))
    expect_error(plotPairs(se = se, selAssay = "counts", pointAlpha = 0))

    expect_error(plotPairs(se = se, selAssay = "counts", colorByCorrelation = 1))
    expect_error(plotPairs(se = se, selAssay = "counts", colorByCorrelation = "TRUE"))
    expect_error(plotPairs(se = se, selAssay = "counts", colorByCorrelation = c(TRUE, FALSE)))
    
    expect_error(plotPairs(se = se, selAssay = "counts", corrColorRange = 1))
    expect_error(plotPairs(se = se, selAssay = "counts", corrColorRange = c(-1, 0.5)))
    expect_error(plotPairs(se = se, selAssay = "counts", corrColorRange = c(1, 0.5)))
    expect_error(plotPairs(se = se, selAssay = "counts", corrColorRange = c("0.5", "1")))
    expect_error(plotPairs(se = se, selAssay = "counts", corrColorRange = c(0, 2)))
    
    expect_error(plotPairs(se = se, selAssay = "counts", addIdentityLine = 1))
    expect_error(plotPairs(se = se, selAssay = "counts", addIdentityLine = "TRUE"))
    expect_error(plotPairs(se = se, selAssay = "counts", addIdentityLine = c(TRUE, FALSE)))
})

test_that("plotPairs works with correct arguments", {
    ## Defaults
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL,
                              addIdentityLine = FALSE), "ggmatrix")

    ## Not log-transformed
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = FALSE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## No pseudocount
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 0, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## Spearman correlation
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "spearman",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## Fewer breaks
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 5, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## smoothScatter
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "smoothscatter", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")
    
    ## scattermore
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "scattermore", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 3.5, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")
    
    ## scattermost
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "scattermost", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 3.5, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## Change font size to correlation relation
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 2,
                              corSizeAdd = 0, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## Point size, alpha
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 1, pointAlpha = 3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")

    ## Don't color by correlation
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = FALSE, corrColorRange = NULL, 
                              addIdentityLine = FALSE), "ggmatrix")
    
    ## Change color range
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = c(0.7, 1), 
                              addIdentityLine = FALSE), "ggmatrix")
    
    ## Add identity line
    expect_s3_class(plotPairs(se = se, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL, 
                              addIdentityLine = TRUE), "ggmatrix")

    ## Make one correlation negative
    seneg <- se
    SummarizedExperiment::assay(seneg, 1)[, 2] <-
        1/(SummarizedExperiment::assay(seneg, 1)[, 2] + 1)
    expect_s3_class(plotPairs(se = seneg, selAssay = "counts", doLog = TRUE,
                              pseudocount = 1, corMethod = "pearson",
                              histBreaks = 40, pointsType = "points", corSizeMult = 5,
                              corSizeAdd = 2, pointSize = 0.1, pointAlpha = 0.3,
                              colorByCorrelation = TRUE, corrColorRange = NULL,
                              addIdentityLine = FALSE), "ggmatrix")
})

#' Visualize the filtering procedure 
#' 
#' Display the number (or fraction) of reads remaining after each step 
#' of the internal \code{mutscan} filtering. 
#' 
#' The function assumes that the number of reads filtered out in each step 
#' are provided as columns of \code{colData(se)}, with column names 
#' of the form \code{f[0-9]_filteringreason}, and that all filtering columns 
#' occur between the columns named \code{nbrTotal} and \code{nbrRetained}. 
#' 
#' @author Charlotte Soneson
#' 
#' @export
#' 
#' @return A \code{ggplot} object. 
#' 
#' @param se A \code{SummarizedExperiment} object, e.g. from 
#'     \code{summarizeExperiment}.
#' @param valueType Either "reads" or "fractions", indicating whether to 
#'     plot the number of reads, or the fraction of the total number of reads,
#'     that are retained after/filtered out in each filtering step.
#' @param onlyActiveFilters Logical scalar, whether to only include the 
#'     active filters (i.e., where any read was filtered out in any of the 
#'     samples). Defaults to \code{TRUE}. 
#' @param displayNumbers Logical scalar, indicating whether to display the 
#'     number (or fraction) of reads retained at every filtering step. 
#' @param numberSize Numeric scalar, indicating the size of the displayed 
#'     numbers (if \code{displayNumbers} is \code{TRUE}). 
#' @param plotType Character scalar, indicating what to show in the plot. 
#'     Either \code{"remaining"} or \code{"filtered"}.
#' @param facetBy Character scalar, indicating the variable by which the plots 
#'     should be facetted. Either \code{"sample"} or \code{"step"}.
#' 
#' @importFrom tidyselect matches
#' @importFrom dplyr select mutate group_by summarize pull ungroup filter
#' @importFrom tibble rownames_to_column 
#' @importFrom tidyr gather
#' @importFrom ggplot2 ggplot aes geom_bar facet_wrap theme theme_bw labs
#'     geom_text element_text
#' @importFrom rlang .data
#' 
#' @examples 
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
#'                           package = "mutscan"))[1:200, ]
#' plotFiltering(se)
#' 
plotFiltering <- function(se, valueType = "reads", onlyActiveFilters = TRUE,
                          displayNumbers = TRUE, numberSize = 4,
                          plotType = "remaining", facetBy = "sample") {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = valueType, type = "character", 
                  validValues = c("reads", "fractions"))
    .assertScalar(x = onlyActiveFilters, type = "logical")
    .assertScalar(x = displayNumbers, type = "logical")
    .assertScalar(x = numberSize, type = "numeric", rngExcl = c(0, Inf))
    .assertScalar(x = plotType, type = "character",
                  validValues = c("remaining", "filtered"))
    .assertScalar(x = facetBy, type = "character", 
                  validValues = c("sample", "step"))
    
    ## --------------------------------------------------------------------- ##
    ## Extract relevant columns from colData(se)
    ## --------------------------------------------------------------------- ##
    cd <- as.data.frame(colData(se)) |>
        dplyr::select("nbrTotal":"nbrRetained") |>
        dplyr::select(c("nbrTotal", matches("^f.*_"), 
                        "nbrRetained"))
    
    ## Remove inactive filters if desired
    if (onlyActiveFilters) {
        cd <- cd[, colSums(cd) > 0, drop = FALSE]
    }
    
    ## Check that filtering columns are in the right order
    nbrs <- gsub("^f", "", vapply(strsplit(colnames(cd), "_"), .subset, 
                                  FUN.VALUE = "", 1))
    nbrs <- nbrs[-c(1, length(nbrs))]
    nbrs <- gsub("a$|b$", "", nbrs)
    stopifnot(all(diff(as.numeric(nbrs)) >= 0))
    
    ## Check that all columns are numeric
    stopifnot(all(apply(cd, 2, is.numeric)))
    
    ## --------------------------------------------------------------------- ##
    ## Reshape into long format
    ## --------------------------------------------------------------------- ##
    cd <- cd |> 
        rownames_to_column("sample") |>
        gather(key = "step", value = "nbrReads", -"sample") |>
        mutate(step = factor(.data$step, levels = colnames(cd)))
    
    ## Check that numbers add up (total - all filters = retained)
    stopifnot(all(
        cd |> 
            group_by(sample) |>
            summarize(
                remdiff = .data$nbrReads[.data$step == "nbrTotal"] - 
                    sum(.data$nbrReads[grepl("^f", .data$step)]),
                remlist = .data$nbrReads[.data$step == "nbrRetained"]) |>
            mutate(obsdiff = .data$remdiff - .data$remlist) |>
            pull(.data$obsdiff) == 0)) 
    
    ## --------------------------------------------------------------------- ##
    ## Calculate number of remaining reads at each step
    ## --------------------------------------------------------------------- ##
    cd <- cd |>
        group_by(sample) |>
        mutate(fracReads = signif(
            .data$nbrReads / 
                .data$nbrReads[.data$step == "nbrTotal"],
            digits = 3)) |>
        mutate(cumulsum = vapply(seq_along(.data$step), function(i) {
            sum(.data$nbrReads[as.numeric(.data$step) <= 
                                   as.numeric(.data$step[i]) & 
                                   .data$step != "nbrTotal"])
        }, NA_real_)) |>
        mutate(nbrRemaining = .data$nbrReads[.data$step == "nbrTotal"] - 
                   .data$cumulsum) |>
        mutate(fracRemaining = signif(
            .data$nbrRemaining /
                .data$nbrReads[.data$step == "nbrTotal"],
            digits = 3)) |>
        filter(.data$step != "nbrRetained") |>
        ungroup()
    
    ## --------------------------------------------------------------------- ##
    ## Create plot
    ## --------------------------------------------------------------------- ##
    
    yvar <- switch(
        paste0(valueType, "_", plotType),
        reads_remaining = "nbrRemaining",
        fractions_remaining = "fracRemaining",
        reads_filtered = "nbrReads",
        fractions_filtered = "fracReads"
    )
    ylab1 <- switch(
        valueType,
        reads = "Number of reads",
        fractions = "Fraction of reads"
    )
    ylab2 <- switch(
        plotType,
        remaining = " remaining after",
        filtered = " filtered out in"
    )
    if (plotType == "filtered") {
        cd <- cd |> 
            filter(.data$step != "nbrTotal")
    }
    gg <- ggplot(cd, aes(
        x = .data[[setdiff(c("step", "sample"), facetBy)]], 
        y = .data[[yvar]], 
        label = .data[[yvar]])) + 
        geom_bar(stat = "identity") + 
        facet_wrap(~ .data[[facetBy]], ncol = 1, scales = "free_y") + 
        theme_bw() + 
        theme(
            axis.text.x = element_text(angle = 90, hjust = 1, 
                                       vjust = 0.5, size = 12),
            axis.text.y = element_text(size = 12),
            axis.title = element_text(size = 14)) + 
        labs(x = "", y = ylab1, 
             title = paste0(ylab1, ylab2, " each filtering step"))
    
    if (displayNumbers) {
        gg <- gg + geom_text(vjust = 1.5, color = "white", 
                             size = numberSize)
    }
    gg
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:1000, 1:3]

test_that("plotFiltering fails with incorrect arguments", {
    expect_error(plotFiltering(se = 1))
    expect_error(plotFiltering(se = "x"))
    expect_error(plotFiltering(se = matrix(1:6), 2, 3))
    
    expect_error(plotFiltering(se = se, valueType = 1))
    expect_error(plotFiltering(se = se, valueType = c("reads", "fractions")))
    expect_error(plotFiltering(se = se, valueType = "wrong"))
    
    expect_error(plotFiltering(se = se, onlyActiveFilters = 1))
    expect_error(plotFiltering(se = se, onlyActiveFilters = "TRUE"))
    expect_error(plotFiltering(se = se, onlyActiveFilters = c(TRUE, FALSE)))
    
    expect_error(plotFiltering(se = se, displayNumbers = 1))
    expect_error(plotFiltering(se = se, displayNumbers = "TRUE"))
    expect_error(plotFiltering(se = se, displayNumbers = c(TRUE, FALSE)))
    
    expect_error(plotFiltering(se = se, numberSize = "1"))
    expect_error(plotFiltering(se = se, numberSize = TRUE))
    expect_error(plotFiltering(se = se, numberSize = c(1, 2)))
    expect_error(plotFiltering(se = se, numberSize = 0))
    
    expect_error(plotFiltering(se = se, plotType = 1))
    expect_error(plotFiltering(se = se, plotType = TRUE))
    expect_error(plotFiltering(se = se, plotType = c("remaining", "filtered")))
    expect_error(plotFiltering(se = se, plotType = "wrong"))
    
    expect_error(plotFiltering(se = se, facetBy = 1))
    expect_error(plotFiltering(se = se, facetBy = TRUE))
    expect_error(plotFiltering(se = se, facetBy = c("sample", "step")))
    expect_error(plotFiltering(se = se, facetBy = "wrong"))
})

test_that("plotFiltering works with correct arguments", {
    ## Defaults
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "remaining", facetBy = "sample")))
    
    ## Fractions
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "fractions", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "remaining", facetBy = "sample")))
    
    ## Only active filters
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = TRUE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "remaining", facetBy = "sample")))
    
    ## Don't display numbers
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = FALSE, numberSize = 4,
                      plotType = "remaining", facetBy = "sample")))
    
    ## Change size of displayed numbers
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = FALSE, numberSize = 2,
                      plotType = "remaining", facetBy = "sample")))
    
    ## Reads + Filtered + Sample
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "filtered", facetBy = "sample")))

    ## Reads + Filtered + Step
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "filtered", facetBy = "step")))

    ## Reads + Remaining + Step
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "reads", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "remaining", facetBy = "step")))

    ## Fractions + Filtered + Sample
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "fractions", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "filtered", facetBy = "sample")))
    
    ## Fractions + Filtered + Step
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "fractions", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "filtered", facetBy = "step")))
    
    ## Fractions + Remaining + Step
    expect_true(ggplot2::is_ggplot(
        plotFiltering(se = se, valueType = "fractions", 
                      onlyActiveFilters = FALSE, 
                      displayNumbers = TRUE, numberSize = 4,
                      plotType = "remaining", facetBy = "step")))
    
})

# This script is provided as a utility via the swissknife package 
# (https://github.com/fmicompbio/swissknife). This script is provided under 
# the MIT license, and package authors are permitted to 
# include the code as-is in other packages, as long as this note and the 
# information provided below crediting the authors of the respective 
# functions is retained. 

#' Utility function to check validity of scalar variable values.
#' 
#' This function provides a convenient way e.g. to check that provided
#' arguments to functions satisfy required criteria. 
#'  
#' @param x The variable to be checked.
#' @param type The desired type of \code{x}.
#' @param rngIncl The allowed range of the (numeric) variable \code{x}, 
#'     including the endpoints.
#' @param rngExcl The allowed range of the (numeric) variable \code{x}, 
#'     excluding the endpoints.
#' @param validValues A vector with the allowed values of \code{x}.
#' @param allowNULL Logical, whether or not \code{NULL} is an acceptable 
#'     value for \code{x}.
#' 
#' @author Michael Stadler, Charlotte Soneson
#' @noRd
#' @keywords internal
#' @importFrom methods is
.assertScalar <- function(x,
                          type = NULL,
                          rngIncl = NULL,
                          rngExcl = NULL,
                          validValues = NULL,
                          allowNULL = FALSE) {
    
    .assertVector(x = x, type = type, rngIncl = rngIncl,
                  rngExcl = rngExcl, validValues = validValues,
                  len = 1, rngLen = NULL, allowNULL = allowNULL)
    
}

#' Utility function to check validity of vector variable values.
#' 
#' This function provides a convenient way e.g. to check that provided
#' arguments to functions satisfy required criteria. 
#'  
#' @param x The variable to be checked
#' @param type The desired type of \code{x}.
#' @param rngIncl The allowed range of the (numeric) variable \code{x}, 
#'     including the endpoints.
#' @param rngExcl The allowed range of the (numeric) variable \code{x}, 
#'     excluding the endpoints.
#' @param validValues A vector with the allowed values of \code{x}.
#' @param len The required length of \code{x}.
#' @param rngLen The allowed range for the length of \code{x}.
#' @param allowNULL Logical, whether or not \code{NULL} is an acceptable 
#'     value for \code{x}.
#' 
#' @author Michael Stadler, Charlotte Soneson
#' @noRd
#' @keywords internal
#' @importFrom methods is
.assertVector <- function(x,
                          type = NULL,
                          rngIncl = NULL,
                          rngExcl = NULL,
                          validValues = NULL,
                          len = NULL, 
                          rngLen = NULL,
                          allowNULL = FALSE) {
    sc <- sys.calls()
    mycall <- sc[[length(sc)]]
    if (length(sc) >= 2 &&
        identical(as.character(sc[[length(sc) - 1]])[1], ".assertScalar")) {
        mycall <- sc[[length(sc) - 1]]
    }
    args <- lapply(mycall, as.character)[-1]
    xname <- if ("x" %in% names(args)) args$x else "argument"
    
    ## Check arguments
    stopifnot(is.null(type) || (length(type) == 1L && is.character(type)))
    stopifnot(is.null(rngIncl) || (length(rngIncl) == 2L && 
                                       is.numeric(rngIncl)))
    stopifnot(is.null(rngExcl) || (length(rngExcl) == 2L && 
                                       is.numeric(rngExcl)))
    stopifnot(is.null(len) || (length(len) == 1L && is.numeric(len)))
    stopifnot(is.null(rngLen) || (length(rngLen) == 2L && is.numeric(rngLen)))
    stopifnot(is.logical(allowNULL) && length(allowNULL) == 1L)
    if (!is.null(rngIncl) && !is.null(rngExcl)) {
        stop("'rngIncl' and 'rngExcl' can not both be specified")
    }
    
    ## If there are too many valid values, print only the first 15
    if (length(validValues) > 15) {
        vvPrint <- paste(c(validValues[seq_len(15)], 
                           "...(truncated)"),
                         collapse = ", ")
    } else {
        vvPrint <- paste(validValues, collapse = ", ")
    }
    
    if (is.null(x)) {
        if (allowNULL) {
            return(invisible(TRUE))
        } else {
            stop("'", xname, "' must not be NULL", call. = FALSE)
        }
    }
    
    if (is.null(type) && (!is.null(rngIncl) || !is.null(rngExcl))) {
        type <- "numeric"
    }
    
    if (!is.null(type) && !is(x, type)) {
        stop("'", xname, "' must be of class '", type, "'", call. = FALSE)
    }
    
    if (!is.null(rngIncl)) {
        if (!is.null(validValues)) {
            if (any((x < rngIncl[1] | x > rngIncl[2]) & 
                    !(x %in% validValues))) {
                stop("'", xname, "' must be within [", rngIncl[1], ",", 
                     rngIncl[2], "] (inclusive), or one of: ", vvPrint,
                     call. = FALSE)
            }
        } else {
            if (any(x < rngIncl[1] | x > rngIncl[2])) {
                stop("'", xname, "' must be within [", rngIncl[1], ",", 
                     rngIncl[2], "] (inclusive)", call. = FALSE)
            }
        }
    } else if (!is.null(rngExcl)) {
        if (!is.null(validValues)) {
            if (any((x <= rngExcl[1] | x >= rngExcl[2]) & 
                    !(x %in% validValues))) {
                stop("'", xname, "' must be within (", rngExcl[1], ",", 
                     rngExcl[2], ") (exclusive), or one of: ", vvPrint,
                     call. = FALSE)
            }
        } else {
            if (any(x <= rngExcl[1] | x >= rngExcl[2])) {
                stop("'", xname, "' must be within (", rngExcl[1], ",", 
                     rngExcl[2], ") (exclusive)", call. = FALSE)
            }
        }
    } else {
        if (!is.null(validValues) && !all(x %in% validValues)) {
            stop("All values in '", xname, "' must be one of: ", vvPrint,
                 call. = FALSE)
        }
    }
    
    
    if (!is.null(len) && length(x) != len) {
        stop("'", xname, "' must have length ", len, call. = FALSE)
    }
    
    if (!is.null(rngLen) && (length(x) < rngLen[1] || length(x) > rngLen[2])) {
        stop("length of '", xname, "' must be within [", rngLen[1], ",", 
             rngLen[2], "] (inclusive)", call. = FALSE)
    }
    
    return(invisible(TRUE))
}

#' Utility function that makes sure that packages are available
#' 
#' The function tries loading the namespaces of the packages given in
#' \code{pkgs}, and throws an exception with an informative error message if
#' that is not the case.
#' 
#' @param pkgs Character vector with package names. Can be either just a
#'   package name or a string of the form \code{"githubuser/packagename"} for
#'   packages hosted on GitHub.
#' @param suggestInstallation Logical scalar. If \code{TRUE}, include an
#'   expression to install the missing package(s) as part of the generated
#'   error message.
#' 
#' @author Michael Stadler, Charlotte Soneson
#' 
#' @noRd
#' @keywords internal
.assertPackagesAvailable <- function(pkgs, suggestInstallation = TRUE) {
    stopifnot(exprs = {
        is.character(pkgs)
        is.logical(suggestInstallation)
        length(suggestInstallation) == 1L
    })
    
    avail <- unlist(lapply(sub("^[^/]+/", "", pkgs),
                           function(pkg) {
                               requireNamespace(pkg, quietly = TRUE)
                           }))
    
    if (any(!avail)) {
        caller <- deparse(sys.calls()[[sys.nframe() - 1]])
        callerfunc <- sub("\\(.+$", "", caller)
        haveBioc <- requireNamespace("BiocManager", quietly = TRUE)
        msg <- paste0(
            "The package", ifelse(sum(!avail) > 1, "s '", " '"),
            paste(sub("^[^/]+/", "", pkgs[!avail]), collapse = "', '"),
            "' ",
            ifelse(sum(!avail) > 1, "are", "is"), " required for ",
            callerfunc, "(), but not installed.\n")
        if (suggestInstallation) {
            msg <- paste0(
                msg,
                "Install ", ifelse(sum(!avail) > 1, "them", "it"), " using:\n",
                ifelse(haveBioc, "", "install.packages(\"BiocManager\")\n"),
                "BiocManager::install(c(\"",
                paste(pkgs[!avail], collapse = "\", \""), "\"))")
        }
        stop(msg, call. = FALSE)
    }
    
    invisible(TRUE)
}

#' Calculate logFCs relative to WT using edgeR
#'
#' Calculate logFCs and associated p-values for a given comparison, using
#' either limma or the Negative Binomial quasi-likelihood framework of edgeR.
#' The observed counts for the WT variants can be used as offsets in the model.
#'
#' @param se SummarizedExperiment object.
#' @param design Design matrix. The rows of the design matrix must be in the
#'     same order as the columns in \code{se}.
#' @param coef Coefficient(s) to test with edgeR or limma.
#' @param contrast Numeric contrast to test with edgeR or limma.
#' @param WTrows Vector of row names that will be used as the reference when
#'     calculating logFCs and statistics. If more than one value is provided, 
#'     the sum of the corresponding counts is used to generate offsets. If 
#'     \code{NULL}, offsets will be defined as the effective library sizes 
#'     (using TMM normalization factors).
#' @param selAssay Assay to select from \code{se} for the analysis.
#' @param pseudocount Pseudocount to add when calculating log-fold changes.
#' @param method Either 'edgeR' or 'limma'. If set to 'limma', voom is used to
#'     transform the counts and estimate observation weights before applying
#'     limma. In this case, the results also contain the standard errors of the
#'     logFCs.
#' @param normMethod Character scalar indicating which normalization method
#'     should be used to calculate size factors. Should be either \code{"TMM"}
#'     or \code{"csaw"} when \code{WTrows} is \code{NULL}, and \code{"geomean"}
#'     or \code{"sum"} when \code{WTrows} is provided.
#'
#' @author Charlotte Soneson, Michael Stadler
#'
#' @export
#'
#' @return A \code{data.frame} with output from the statistical testing 
#' framework (edgeR or limma). 
#' 
#' @importFrom edgeR DGEList scaleOffset estimateDisp glmQLFit glmQLFTest
#'     topTags predFC topTags calcNormFactors getNormLibSizes
#' @importFrom SummarizedExperiment colData assay assayNames assays
#' @importFrom limma voom eBayes topTable lmFit contrasts.fit
#' @importFrom csaw normOffsets
#'
#' @examples
#' library(SummarizedExperiment)
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
#'                           package = "mutscan"))[1:200, ]
#' design <- model.matrix(~ Replicate + Condition,
#'                        data = colData(se))
#'                               
#' ## Calculate "absolute" log-fold changes with edgeR
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
#'                            method = "edgeR")
#' head(res)
#' ## Calculate log-fold changes relative to the WT sequence with edgeR
#' stopifnot("f.0.WT" %in% rownames(se))
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
#'                            method = "edgeR", WTrows = "f.0.WT")
#' head(res)
#' 
#' ## Calculate "absolute" log-fold changes with limma
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
#'                            method = "limma")
#' head(res)
#' ## Calculate log-fold changes relative to the WT sequence with limma
#' stopifnot("f.0.WT" %in% rownames(se))
#' res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
#'                            method = "limma", WTrows = "f.0.WT")
#' head(res)
#'
calculateRelativeFC <- function(se, design, coef = NULL, contrast = NULL,
                                WTrows = NULL, selAssay = "counts",
                                pseudocount = 1, method = "edgeR",
                                normMethod = ifelse(is.null(WTrows),
                                                    "TMM", "sum")) {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = selAssay, type = "character")
    if (!(selAssay %in% assayNames(se))) {
        if (is.null(assayNames(se)) &&
            length(assays(se)) == 1) {
            warning("No assayNames provided in 'se', but only one ",
                    "assay present - using that.")
            selAssay <- 1L
        } else {
            stop("The provided 'selAssay' not present in 'se'.")
        }
    }
    
    if (!is.null(WTrows)) {
        .assertVector(x = WTrows, type = "character", 
                      validValues = rownames(se))
    }
    
    if (nrow(design) != ncol(se)) {
        stop("The number of rows in 'design' (", nrow(design),
             ") is not equal to the number",
             " of columns in 'se' (", ncol(se), ").")
    }
    
    .assertScalar(x = pseudocount, type = "numeric", rngIncl = c(0, Inf))
    
    .assertScalar(x = method, type = "character",
                  validValues = c("edgeR", "limma"))
    
    if (normMethod %in% c("csaw", "TMM") && !is.null(WTrows)) {
        stop("normMethod = '", normMethod,
             "' can only be used when WTrows is NULL.")
    }
    
    if (normMethod %in% c("sum", "geomean") && is.null(WTrows)) {
        stop("normMethod = '", normMethod,
             "' can only be used when WTrows is not NULL.")
    }
    
    if (normMethod == "csaw" && method == "limma") {
        stop("normMethod = 'csaw' can only be used with method = 'edgeR'.")
    }
    
    .assertScalar(x = normMethod, type = "character",
                  validValues = c("csaw", "TMM", "geomean", "sum"))
    
    if (is.null(coef) && is.null(contrast)) {
        stop("'coef' and 'contrast' can not both be NULL.")
    }
    
    if (!is.null(contrast) && !is.null(dim(contrast))) {
        stop("'contrast' must be a vector.")
    }
    
    ## Create DGEList from SummarizedExperiment
    dge <- DGEList(
        counts = as.matrix(assay(se, selAssay)),
        samples = colData(se))
    if (normMethod == "csaw") {
        ## csaw normalization - also calculate normalization factors since
        ## aveLogCPM does not use provided offsets
        ## In this case, we know that WTrows is NULL, so all features
        ## will be used for the normalization
        dge <- calcNormFactors(dge)
        dge <- normOffsets(dge)
    } else if (normMethod == "TMM") {
        ## TMM normalization, with all features
        dge <- calcNormFactors(dge)
    } else if (normMethod == "geomean") {
        ## Use size factors (offsets) derived from the geometric mean
        ## of the WT rows
        tmp0 <- dge$counts[WTrows, , drop = FALSE]
        tmp0 <- tmp0[apply(tmp0, 1, min) > 0, , drop = FALSE]
        logoffsets <- apply(tmp0, 2, function(s) mean(log(s)))
        dge <- scaleOffset(dge, logoffsets)
    } else if (normMethod == "sum") {
        ## Use size factors (offsets) derived from the sum of the
        ## WT rows
        tmp0 <- dge$counts[WTrows, , drop = FALSE]
        dge <- scaleOffset(dge, log(colSums(tmp0)))
    }
    
    ## Fit model and perform test
    if (method == "edgeR") {
        dge <- estimateDisp(dge, design = design)
        fit <- glmQLFit(dge, design = design)
        qlf <- glmQLFTest(fit, coef = coef, contrast = contrast)
        tt <- topTags(qlf, n = Inf, sort.by = "none")$table
        ## Calculate shrunken fold changes. Only when testing
        ## a single coefficient or contrast
        predfc <- predFC(dge, design = design, 
                         prior.count = pseudocount)
        if (length(coef) == 1 && is.null(contrast)) {
            tt$logFC_shrunk <- predfc[, coef]
        } else if (!is.null(contrast)) {
            tt$logFC_shrunk <- c(predfc %*% cbind(contrast))
        }
        tt$df.total <- qlf$df.total
        tt$df.prior <- qlf$df.prior
        tt$df.test <- qlf$df.test
    } else if (method == "limma") {
        if (!is.null(dge$offset)) {
            vm <- voom(dge, design = design, lib.size = exp(dge$offset))
        } else {
            vm <- voom(dge, design = design,
                       lib.size = getNormLibSizes(dge))
        }
        fit <- lmFit(vm, design = design)
        if (!is.null(contrast)) {
            fit <- contrasts.fit(fit, contrasts = contrast)
            coef <- 1
        }
        fit <- eBayes(fit)
        tt <- topTable(fit, coef = coef,
                       confint = TRUE, number = Inf, sort.by = "none")
        if (length(coef) == 1) {
            tt$se.logFC <- sqrt(fit$s2.post) * fit$stdev.unscaled[, coef]
        }
        tt$df.total <- fit$df.total
        tt$df.prior <- fit$df.prior
    }
    tt
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:200, ]
design <- stats::model.matrix(~ Replicate + Condition, 
                              data = SummarizedExperiment::colData(se))

test_that("calculateRelativeFC fails with incorrect arguments", {
    expect_error(calculateRelativeFC(se = NULL, design = design,
                                     coef = "Conditioncis_output"),
                 "'se' must not be NULL")
    expect_error(calculateRelativeFC(se = as.matrix(assay(se)), design = design,
                                     coef = "Conditioncis_output"),
                 "'se' must be of class 'SummarizedExperiment'")
    expect_error(calculateRelativeFC(se = se, design = NULL,
                                     coef = "Conditioncis_output"),
                 "argument is of length zero")
    expect_error(calculateRelativeFC(se = se, design = design[1:3, ],
                                     coef = "Conditioncis_output"),
                 "The number of rows in")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "missing"),
                 "One or more named coef arguments")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = NULL, contrast = NULL),
                 "can not both be NULL")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = NULL,
                                     contrast = matrix(sample(c(0, 1),
                                                              ncol(se) * 2,
                                                              replace = TRUE),
                                                       ncol = 2)),
                 "'contrast' must be a vector")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = "missing"),
                 "All values in 'WTrows' must be")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     selAssay = "missing"),
                 "The provided 'selAssay' not present")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     selAssay = 1),
                 "'selAssay' must be of class 'character'")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     pseudocount = -1),
                 "must be within")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     method = "missing"),
                 "'method' must be one of: edgeR, limma")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = NULL, normMethod = "sum"),
                 "normMethod = 'sum' can only be used when WTrows is not NULL")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = NULL, normMethod = "geomean"),
                 "normMethod = 'geomean' can only be used when WTrows is not NULL")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = rownames(se)[1], normMethod = "TMM"),
                 "normMethod = 'TMM' can only be used when WTrows is NULL")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = rownames(se)[1], normMethod = "csaw"),
                 "normMethod = 'csaw' can only be used when WTrows is NULL")
    expect_error(calculateRelativeFC(se = se, design = design,
                                     coef = "Conditioncis_output",
                                     WTrows = NULL, normMethod = "csaw",
                                     method = "limma"),
                 "normMethod = 'csaw' can only be used with method = 'edgeR'")
})

test_that("calculateRelativeFC works", {
    ## With a single WT row, normMethod = "sum" or "geomean" should be identical

    res1a <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = "f.0.WT",
                                 selAssay = "counts", pseudocount = 1,
                                 method = "edgeR", normMethod = "sum")
    res1b <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = "f.0.WT",
                                 selAssay = "counts", pseudocount = 1,
                                 method = "edgeR", normMethod = "geomean")
    ## If there's only one assay, it doesn't need to be named
    senoname <- SummarizedExperiment(
        assays = list(assay(se, "counts")),
        rowData = rowData(se),
        colData = colData(se),
        metadata = metadata(se)
    )
    expect_warning(
        res1c <- calculateRelativeFC(senoname, design, coef = "Conditioncis_output",
                             contrast = NULL, WTrows = "f.0.WT", selAssay = "counts",
                             pseudocount = 1, method = "edgeR",
                             normMethod = "sum"),
        "No assayNames provided in 'se'")
    expect_equal(res1a$logFC, res1b$logFC)
    expect_equal(res1a$logFC, res1c$logFC)
    expect_equal(res1a$logFC_shrunk, res1b$logFC_shrunk)
    expect_equal(res1a$logFC_shrunk, res1c$logFC_shrunk)
    expect_equal(res1a$logCPM, res1b$logCPM)
    expect_equal(res1a$logCPM, res1c$logCPM)
    expect_equal(res1a$F, res1b$F)
    expect_equal(res1a$F, res1c$F)
    expect_equal(res1a$PValue, res1b$PValue)
    expect_equal(res1a$PValue, res1c$PValue)
    expect_equal(res1a$FDR, res1b$FDR)
    expect_equal(res1a$FDR, res1c$FDR)

    ## Specifying the coef or contrast should be equivalent
    res2b <- calculateRelativeFC(se, design, coef = NULL,
                                 contrast = as.numeric(colnames(design) == "Conditioncis_output"),
                                 WTrows = "f.0.WT", selAssay = "counts",
                                 pseudocount = 1, method = "edgeR",
                                 normMethod = "sum")
    expect_equal(res1a$logFC, res2b$logFC)
    expect_equal(res1a$logFC_shrunk, res2b$logFC_shrunk)
    expect_equal(res1a$logCPM, res2b$logCPM)
    expect_equal(res1a$F, res2b$F)
    expect_equal(res1a$PValue, res2b$PValue)
    expect_equal(res1a$FDR, res2b$FDR)

    ## With a single WT row, that one should not show any change
    expect_equal(res1a["f.0.WT", "logFC"], 0, tolerance = 1e-4)
    expect_equal(res1a["f.0.WT", "logFC_shrunk"], 0, tolerance = 1e-4)
    expect_equal(res1a["f.0.WT", "PValue"], 1, tolerance = 1e-4)
    expect_equal(res1a["f.0.WT", "FDR"], 1, tolerance = 1e-4)
    expect_named(res1a, c("logFC", "logCPM", "F", "PValue", "FDR",
                          "logFC_shrunk", "df.total", "df.prior", "df.test"))

    ## Should also work with limma
    res3a <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = "f.0.WT", selAssay = "counts",
                                 pseudocount = 1, method = "limma",
                                 normMethod = "sum")
    expect_equal(res3a["f.0.WT", "logFC"], 0, tolerance = 1e-4)
    expect_equal(res3a["f.0.WT", "P.Value"], 1, tolerance = 1e-4)
    expect_equal(res3a["f.0.WT", "adj.P.Val"], 1, tolerance = 1e-4)
    expect_named(res3a, c("logFC", "CI.L", "CI.R", "AveExpr", "t",
                          "P.Value", "adj.P.Val", "B", "se.logFC", 
                          "df.total", "df.prior"))
    
    ## Also with contrast
    res3b <- calculateRelativeFC(se, design, coef = NULL,
                                 contrast = as.numeric(colnames(design) == "Conditioncis_output"),
                                 WTrows = "f.0.WT", selAssay = "counts",
                                 pseudocount = 1, method = "limma",
                                 normMethod = "sum")
    expect_equal(res3b["f.0.WT", "logFC"], 0, tolerance = 1e-4)
    expect_equal(res3b["f.0.WT", "P.Value"], 1, tolerance = 1e-4)
    expect_equal(res3b["f.0.WT", "adj.P.Val"], 1, tolerance = 1e-4)
    expect_named(res3b, c("logFC", "CI.L", "CI.R", "AveExpr", "t",
                          "P.Value", "adj.P.Val", "B", "se.logFC",
                          "df.total", "df.prior"))
    
    ## Coef and contrast should be equivalent also for limma
    expect_equal(res3a$logFC, res3b$logFC)
    expect_equal(res3a$CI.L, res3b$CI.L)
    expect_equal(res3a$CI.R, res3b$CI.R)
    expect_equal(res3a$AveExpr, res3b$AveExpr)
    expect_equal(res3a$t, res3b$t)
    expect_equal(res3a$P.Value, res3b$P.Value)
    expect_equal(res3a$adj.P.Val, res3b$adj.P.Val)
    expect_equal(res3a$B, res3b$B)
    expect_true(all(res3a$se.logFC >= 0))
    expect_equal(res3a$se.logFC, res3b$se.logFC)
    
    ## Correlation between edgeR and limma logFCs should be high
    expect_gt(cor(res1a$logFC, res3a$logFC), 0.98)

    ## Run without WTrows
    res4a <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = NULL, selAssay = "counts",
                                 pseudocount = 1, method = "edgeR",
                                 normMethod = "TMM")
    res4b <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = NULL, selAssay = "counts",
                                 pseudocount = 1, method = "limma",
                                 normMethod = "TMM")
    res4c <- calculateRelativeFC(se, design, coef = "Conditioncis_output",
                                 contrast = NULL, WTrows = NULL, selAssay = "counts",
                                 pseudocount = 1, method = "edgeR",
                                 normMethod = "csaw")
    expect_gt(cor(res4a$logFC, res4b$logFC), 0.98)
})

.hasReadComponent <- function(composition, component) {
    tmp <- gregexpr(pattern = component, text = composition, fixed = TRUE)[[1]]
    if (length(tmp) == 1 && tmp == -1) {
        FALSE
    } else {
        TRUE
    }
}

#' Summarize and collapse multiple mutational scanning experiments
#'
#' Combine multiple sequence lists (as returned by \code{\link{digestFastqs}}
#' into a \code{\link[SummarizedExperiment]{SummarizedExperiment}}, with
#' observed variable sequences (sequence pairs) in rows and samples in columns.
#'
#' @param x A named list of objects returned by \code{\link{digestFastqs}}.
#'     Names are used to link the objects to the metadata provided in
#'     \code{coldata}.
#' @param coldata A \code{data.frame} with at least one column "Name", which
#'     will be used to link to objects in \code{x}. A potentially subset and
#'     reordered version of \code{coldata} is stored in the \code{colData} of 
#'     the returned \code{\link[SummarizedExperiment]{SummarizedExperiment}}.
#' @param countType Either "reads" or "umis". If "reads", the "count" assay of
#'     the returned object will contain the observed number of reads for each
#'     sequence (pair). If "umis", the "count" assay will contain the number of
#'     unique UMIs observed for each sequence (pair).
#'
#' @return A \code{\link[SummarizedExperiment]{SummarizedExperiment}} \code{x}
#'     with
#'     \describe{
#'         \item{assays(x)$counts}{containing the observed number of sequences 
#'         or sequence pairs (if \code{countType} = "reads"), or the observed 
#'         number of unique UMIs for each sequence or sequence pair (if 
#'         \code{countType} = "umis").}
#'         \item{rowData(x)}{containing the unique sequences or sequence 
#'         pairs.}
#'         \item{colData(x)}{containing the metadata provided by 
#'         \code{coldata}.}
#'     }
#'
#' @author Michael Stadler, Charlotte Soneson
#'
#' @export
#'
#' @importFrom SummarizedExperiment SummarizedExperiment
#' @importFrom S4Vectors DataFrame
#' @importFrom IRanges IntegerList
#' @importFrom methods is new as
#' @importFrom dplyr bind_rows distinct left_join mutate filter group_by
#'     summarize
#' @importFrom rlang .data
#' @importFrom stats setNames
#' 
#' @examples 
#' ## Input sample
#' inp <- digestFastqs(
#'     fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
#'                                package = "mutscan"), 
#'     elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
#'     constantForward = "AACCGGAGGAGGGAGCTG", 
#'     wildTypeForward = c(FOS = paste0(
#'         "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
#'         "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
#'     nbrMutatedCodonsMaxForward = 1
#' )
#' ## Output sample
#' outp <- digestFastqs(
#'     fastqForward = system.file("extdata", "cisOutput_1.fastq.gz", 
#'                                package = "mutscan"), 
#'     elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
#'     constantForward = "AACCGGAGGAGGGAGCTG", 
#'     wildTypeForward = c(FOS = paste0(
#'         "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
#'         "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
#'     nbrMutatedCodonsMaxForward = 1
#' )
#' ## Combine
#' se <- summarizeExperiment(
#'     x = list(r1inp = inp, r1outp = outp), 
#'     coldata = data.frame(Name = c("r1inp", "r1outp"), 
#'                          Condition = c("input", "output"), 
#'                          Replicate = c("rep1", "rep1")),
#'     countType = "umis"
#' )
#' se
#'
summarizeExperiment <- function(x, coldata, countType = "umis") {
    ## ------------------------------------------------------------------------
    ## Pre-flight checks
    ## ------------------------------------------------------------------------
    if (!is(x, "list") || is.null(names(x))) {
        stop("'x' must be a named list")
    }
    if (any(duplicated(names(x)))) {
        stop("duplicated names in 'x' (e.g. technical replicated to be ", 
             "merged) is not supported yet")
    }
    if (!is(coldata, "data.frame") || !("Name" %in% colnames(coldata))) {
        stop("'coldata' must be a data.frame with at least one column, named ",
             "'Name'.")
    }
    .assertScalar(x = countType, type = "character",
                  validValues = c("umis", "reads"))
    ## If no UMI sequences were given, then countType = "umis" should not be 
    ## allowed
    if (countType == "umis" &&
        !(all(vapply(x, function(w) {
            .hasReadComponent(w$parameters$elementsForward, "U") ||
                .hasReadComponent(w$parameters$elementsReverse, "U")
        }, FALSE)))) {
        stop("'countType' is set to 'umis', but no UMI sequences ",
             "were provided when quantifying. ",
             "Set 'countType' to 'reads' instead.")
    }
    ## Get the mutNameDelimiters, and make sure that they are the same
    mutnamedel <- unique(vapply(x, 
                                function(w) w$parameters$mutNameDelimiter, ""))
    if (length(mutnamedel) > 1) {
        stop("All samples must have the same 'mutNameDelimiter'")
    }
    
    coldata$Name <- as.character(coldata$Name)
    
    ## ------------------------------------------------------------------------
    ## Link elements in x with coldata
    ## ------------------------------------------------------------------------
    nms <- intersect(names(x), coldata$Name)
    if (length(nms) == 0) {
        stop("names in 'x' do not match 'coldata$Name'")
    } else if (length(nms) < length(x)) {
        nmsNotUsed <- setdiff(names(x), nms)
        warning("skipping ", length(nmsNotUsed), " elements in 'x' (",
                paste(nmsNotUsed, collapse = ", "), ") because they did not ",
                "match any 'coldata$Name'.")
    }
    x <- x[nms]
    coldata <- coldata[match(nms, coldata$Name), , drop = FALSE]
    
    ## All sample names
    allSamples <- names(x)
    names(allSamples) <- allSamples
    
    ## ------------------------------------------------------------------------
    ## Get all observed sequences for each mutant name
    ## For a given sample, the same mutant name can correspond to multiple 
    ## sequences, separated by ,
    ## ------------------------------------------------------------------------
    tmpdf <- do.call(bind_rows, lapply(x, function(w) w$summaryTable))
    
    allSequences <- mergeValues(tmpdf$mutantName, tmpdf$sequence) |>
        setNames(c("mutantName", "sequence"))
    allSequences <- DataFrame(allSequences)
    
    ## ------------------------------------------------------------------------
    ## Add info about nbr mutated bases/codons/AAs,
    ## sequenceAA, mutantNameAA, mutationTypes, varLengths
    ## Also here, each column can contain multiple values separated with ,
    ## (e.g. if variable sequences were collapsed to WT in digestFastqs)
    ## ------------------------------------------------------------------------
    for (v in intersect(c("nbrMutBases", "nbrMutCodons", "nbrMutAAs", 
                          "mutantNameBase", "mutantNameBaseHGVS",
                          "mutantNameCodon", "mutantNameAA", 
                          "mutantNameAAHGVS", "sequenceAA", "mutationTypes",
                          "varLengths"),
                        colnames(tmpdf))) {
        tmp <- mergeValues(tmpdf$mutantName, tmpdf[[v]]) |>
            setNames(c("mutantName", v))
        allSequences[[v]] <- tmp[[v]][match(allSequences$mutantName,
                                            tmp$mutantName)]
        
        if (v %in% c("nbrMutBases", "nbrMutCodons", "nbrMutAAs")) {
            tmpList <- as(
                lapply(strsplit(allSequences[[v]], ","), 
                       function(w) sort(as.integer(w))),
                "IntegerList")
            allSequences[[paste0("min", sub("^n", "N", v))]] <- min(tmpList)
            allSequences[[paste0("max", sub("^n", "N", v))]] <- max(tmpList)
        }
    }
    
    ## ------------------------------------------------------------------------
    ## Create a sparse matrix
    ## ------------------------------------------------------------------------
    countCol <- ifelse(countType == "umis", "nbrUmis", "nbrReads")
    tmp <- do.call(bind_rows, lapply(allSamples, function(s) {
        st <- x[[s]]$summaryTable
        data.frame(i = match(st$mutantName, allSequences$mutantName),
                   j = match(s, allSamples),
                   x = as.numeric(st[, countCol]))
    }))
    countMat <- new("dgTMatrix", i = tmp$i - 1L, j = tmp$j - 1L,
                    x = tmp$x, Dim = c(nrow(allSequences), 
                                       length(allSamples)))
    ## Convert to dgCMatrix for easier processing downstream
    countMat <- as(countMat, "CsparseMatrix")
    
    ## ------------------------------------------------------------------------
    ## Create the colData
    ## ------------------------------------------------------------------------
    addMeta <- do.call(bind_rows, lapply(allSamples, function(s) {
        x[[s]]$filterSummary |> mutate(Name = s)
    }))
    coldata <- left_join(coldata, addMeta, by = "Name")
    
    ## ------------------------------------------------------------------------
    ## Create SummarizedExperiment object
    ## ------------------------------------------------------------------------
    se <- SummarizedExperiment(
        assays = list(counts = countMat),
        colData = coldata[match(allSamples, coldata$Name), , drop = FALSE],
        rowData = allSequences,
        metadata = list(
            parameters = lapply(allSamples, 
                                function(w) x[[w]]$parameters),
            errorStatistics = lapply(allSamples, 
                                     function(w) x[[w]]$errorStatistics),
            countType = countType,
            mutNameDelimiter = mutnamedel)
    )
    
    if (!any(allSequences$mutantName == "")) {
        rownames(se) <- allSequences$mutantName
    }
    colnames(se) <- allSamples
    
    return(se)
}

#' Plot distribution of observed values
#' 
#' @param se A \code{SummarizedExperiment} object, typically generated by 
#'     \code{summarizeExperiment()}.
#' @param selAssay Character scalar specifying the assay in \code{se} to 
#'     use for the plotting.
#' @param groupBy Character scalar specifying a column from 
#'     \code{colData(se)} to use for coloring or stratifying the plots. 
#' @param plotType Character scalar specifying the type of plot to construct. 
#'     Either \code{'density'}, \code{'histogram'} or \code{'knee'}.
#' @param facet Logical scalar, indicating whether or not to facet the plot
#'     by the values specified in the \code{groupBy} column.
#' @param pseudocount Numeric scalar, representing the number to add to the 
#'     observed values in the \code{selAssay} assay before plotting. 
#' 
#' @export
#' @author Charlotte Soneson
#' 
#' @return A ggplot object.
#' 
#' @importFrom tibble rownames_to_column
#' @importFrom tidyr gather
#' @importFrom dplyr group_by arrange mutate desc ungroup left_join
#' @importFrom SummarizedExperiment colData assay assayNames
#' @importFrom ggplot2 ggplot scale_x_log10 scale_y_log10 labs geom_line 
#'     facet_wrap geom_density geom_histogram theme_minimal theme 
#'     element_text aes
#' @importFrom rlang .data
#' 
#' @examples 
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
#'                           package = "mutscan"))[1:200, ]
#' plotDistributions(se)
#' 
plotDistributions <- function(se, selAssay = "counts", 
                              groupBy = NULL, plotType = "density", 
                              facet = FALSE, pseudocount = 0) {
    .assertVector(x = se, type = "SummarizedExperiment")
    .assertScalar(x = selAssay, type = "character", 
                  validValues = assayNames(se))
    if (!is.null(groupBy)) {
        .assertScalar(
            x = groupBy, type = "character",
            validValues = colnames(colData(se)))
    }
    .assertScalar(x = plotType, type = "character", 
                  validValues = c("density", "knee", "histogram"))
    .assertScalar(x = facet, type = "logical")
    .assertScalar(x = pseudocount, type = "numeric", rngIncl = c(0, Inf))
    
    ## Define a common theme to use for the plots
    commonTheme <- list(
        theme_minimal(),
        theme(axis.text = element_text(size = 12),
              axis.title = element_text(size = 14))
    )
    
    df <- as.data.frame(as.matrix(
        assay(se, selAssay, withDimnames = TRUE)
    )) |>
        rownames_to_column("feature") |>
        gather(key = "Name", value = "value", -"feature") |>
        group_by(.data$Name) |>
        arrange(desc(.data$value)) |>
        mutate(idx = seq_along(.data$value), 
               value = .data$value + pseudocount) |>
        ungroup() |>
        left_join(as.data.frame(colData(se)),
                  by = "Name")
    
    ## If the user doesn't explicitly group by any variable, impose grouping
    ## by the sample ID. In that case, don't color by sample ID if facetting
    ## is used (only one curve per facet). If a variable to group by is 
    ## specified, color by sample ID even if facetting is used. 
    if (is.null(groupBy)) {
        groupBy <- "Name"
        colorFacetByName <- FALSE
    } else {
        colorFacetByName <- TRUE
    }
    
    ## Specify plot depending on desired type
    if (plotType == "knee") {
        gg <- ggplot(df, aes(x = .data$idx, 
                             y = .data$value)) + 
            scale_x_log10() + scale_y_log10() + 
            labs(x = "Feature (sorted)", 
                 y = paste0(selAssay, 
                            ifelse(pseudocount == 0, 
                                   "", paste0(" + ", pseudocount))))
        if (facet) {
            if (colorFacetByName) {
                gg <- gg + geom_line(aes(color = .data$Name))
            } else {
                gg <- gg + geom_line(aes(group = .data$Name))
            }
            gg <- gg + 
                facet_wrap(~ .data[[groupBy]])
        } else {
            gg <- gg + 
                geom_line(aes(group = .data$Name, 
                              color = .data[[groupBy]]))
        }
    } else if (plotType == "density") {
        gg <- ggplot(df, aes(x = .data$value)) + 
            scale_x_log10() + 
            labs(x = paste0(selAssay, 
                            ifelse(pseudocount == 0, 
                                   "", paste0(" + ", pseudocount))),
                 y = "Density")
        if (facet) {
            if (colorFacetByName) {
                gg <- gg + 
                    geom_density(aes(color = .data$Name))
            } else {
                gg <- gg + 
                    geom_density(aes(group = .data$Name))
            }
            gg <- gg + 
                facet_wrap(~ .data[[groupBy]])
        } else {
            gg <- gg + 
                geom_density(aes(group = .data$Name, 
                                 color = .data[[groupBy]]))
        }
    } else if (plotType == "histogram") {
        gg <- ggplot(df, aes(x = .data$value)) + 
            scale_x_log10() + 
            labs(x = paste0(selAssay, 
                            ifelse(pseudocount == 0, 
                                   "", paste0(" + ", pseudocount))),
                 y = "Count")
        if (facet) {
            if (colorFacetByName) {
                gg <- gg + 
                    geom_histogram(aes(fill = .data$Name), 
                                   bins = 50)
            } else {
                gg <- gg + 
                    geom_histogram(aes(group = .data$Name), 
                                   bins = 50)
            }
            gg <- gg +
                facet_wrap(~ .data[[groupBy]])
        } else {
            gg <- gg + 
                geom_histogram(aes(group = .data$Name,
                                   fill = .data[[groupBy]]))
        }
    }
    
    gg + commonTheme
}

se <- readRDS(system.file("extdata/GSE102901_cis_se.rds", package = "mutscan"))
se <- se[1:1000, 1:3]

test_that("plotDistributions fails with incorrect arguments", {
    expect_error(plotDistributions(se = 1))
    expect_error(plotDistributions(se = "x"))
    expect_error(plotDistributions(se = matrix(1:6), 2, 3))
    
    expect_error(plotDistributions(se = se, selAssay = 1))
    expect_error(plotDistributions(se = se, selAssay = c("counts", "counts")))
    expect_error(plotDistributions(se = se, selAssay = TRUE))
    expect_error(plotDistributions(se = se, selAssay = "wrong"))
    
    expect_error(plotDistributions(se = se, groupBy = 1))
    expect_error(plotDistributions(se = se, groupBy = "wrong"))
    expect_error(plotDistributions(se = se, groupBy = c("Name", "Condition")))

    expect_error(plotDistributions(se = se, plotType = 1))
    expect_error(plotDistributions(se = se, plotType = TRUE))
    expect_error(plotDistributions(se = se, plotType = c("density", "histogram")))
    expect_error(plotDistributions(se = se, plotType = "wrong"))
    
    expect_error(plotDistributions(se = se, facet = 1))
    expect_error(plotDistributions(se = se, facet = c(TRUE, FALSE)))
    expect_error(plotDistributions(se = se, facet = "wrong"))
    
    expect_error(plotDistributions(se = se, pseudocount = TRUE))
    expect_error(plotDistributions(se = se, pseudocount = "1"))
    expect_error(plotDistributions(se = se, pseudocount = c(0, 1)))
    expect_error(plotDistributions(se = se, pseudocount = -1))
})

test_that("plotDistributions works as expected", {
    ## Defaults
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "density",
                          facet = FALSE, pseudocount = 0)))

    ## Change plot type
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "histogram",
                          facet = FALSE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "knee",
                          facet = FALSE, pseudocount = 0)))
    
    ## groupBy Name
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "density",
                          facet = FALSE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "histogram",
                          facet = FALSE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "knee",
                          facet = FALSE, pseudocount = 0)))
    
    ## groupBy Condition
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "density",
                          facet = FALSE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "histogram",
                          facet = FALSE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "knee",
                          facet = FALSE, pseudocount = 0)))
    
    ## Facet
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "density",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "histogram",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = NULL, plotType = "knee",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "density",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "histogram",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Name", plotType = "knee",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "density",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "histogram",
                          facet = TRUE, pseudocount = 0)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "knee",
                          facet = TRUE, pseudocount = 0)))
    
    ## Increase pseudocount
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "density",
                          facet = FALSE, pseudocount = 2)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "histogram",
                          facet = FALSE, pseudocount = 3)))
    expect_true(ggplot2::is_ggplot(
        plotDistributions(se = se, selAssay = "counts",
                          groupBy = "Condition", plotType = "knee",
                          facet = FALSE, pseudocount = 4)))
    
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Calculate distances to the nearest string
#' 
#' Given a character vector, calculate the distance for each element
#' to the nearest neighbor amongst all the other elements.
#' 
#' @param x A character vector.
#' @param metric A character scalar defining the string distance metric. One
#'   of \code{"hamming"} (default), \code{"hamming_shift"} or
#'   \code{"levenshtein"}.
#' @param nThreads numeric(1), number of threads to use for parallel processing.
#' 
#' @return An integer vector of the same length as \code{x}.
#' 
#' @examples
#' calcNearestStringDist(c("lazy", "hazy", "crazy"))
#' calcNearestStringDist(c("lazy", "hazy", "crazy"), metric = "hamming_shift")
#' calcNearestStringDist(c("lazy", "hazy", "crazy"), metric = "levenshtein")
#' 
#' @export
calcNearestStringDist <- function(x, metric = "hamming", nThreads = 1L) {
    .Call(`_mutscan_calcNearestStringDist`, x, metric, nThreads)
}

complement <- function(n) {
    .Call(`_mutscan_complement`, n)
}

compareCodonPositions <- function(a, b, mutNameDelimiter) {
    .Call(`_mutscan_compareCodonPositions`, a, b, mutNameDelimiter)
}

translateString <- function(s) {
    .Call(`_mutscan_translateString`, s)
}

makeBaseHGVS <- function(mutationsSorted, mutNameDelimiter, wtSeq, varSeq) {
    .Call(`_mutscan_makeBaseHGVS`, mutationsSorted, mutNameDelimiter, wtSeq, varSeq)
}

test_makeAAHGVS <- function(mutationsSorted, mutNameDelimiter, wtSeq) {
    .Call(`_mutscan_test_makeAAHGVS`, mutationsSorted, mutNameDelimiter, wtSeq)
}

test_compareToWildtype <- function(varSeq, wtSeq, varIntQual, forbiddenCodons_vect, mutatedPhredMin = 0.0, nbrMutatedCodonsMax = -1L, codonPrefix = "c", nbrMutatedBasesMax = -1L, mutNameDelimiter = ".", collapseToWT = FALSE) {
    .Call(`_mutscan_test_compareToWildtype`, varSeq, wtSeq, varIntQual, forbiddenCodons_vect, mutatedPhredMin, nbrMutatedCodonsMax, codonPrefix, nbrMutatedBasesMax, mutNameDelimiter, collapseToWT)
}

test_decomposeRead <- function(sseq, squal, elements, elementLengths, primerSeqs, umiSeq, varSeq, varQual, varLengths, constSeq, constQual, nNoPrimer, nReadWrongLength) {
    .Call(`_mutscan_test_decomposeRead`, sseq, squal, elements, elementLengths, primerSeqs, umiSeq, varSeq, varQual, varLengths, constSeq, constQual, nNoPrimer, nReadWrongLength)
}

test_mergeReadPairPartial <- function(seqF, qualF, seqR, qualR, lenF, lenR, minOverlap = 0L, maxOverlap = 0L, minMergedLength = 0L, maxMergedLength = 0L, maxFracMismatchOverlap = 0, greedy = TRUE) {
    .Call(`_mutscan_test_mergeReadPairPartial`, seqF, qualF, seqR, qualR, lenF, lenR, minOverlap, maxOverlap, minMergedLength, maxMergedLength, maxFracMismatchOverlap, greedy)
}

findClosestRefSeq <- function(varSeq, wtSeq, upperBoundMismatch, sim) {
    .Call(`_mutscan_findClosestRefSeq`, varSeq, wtSeq, upperBoundMismatch, sim)
}

findClosestRefSeqEarlyStop <- function(varSeq, wtSeq, upperBoundMismatch, sim) {
    .Call(`_mutscan_findClosestRefSeqEarlyStop`, varSeq, wtSeq, upperBoundMismatch, sim)
}

#' Create a conversion table for collapsing similar sequences
#' @param seqs Character vector with nucleotide sequences (or pairs of
#' sequences concatenated with "_") to be collapsed. The sequences must
#' all be of the same length.
#' @param scores Numeric vector of "scores" for the sequences. Typically
#' the total read/UMI count. A higher score will be preferred when
#' deciding which sequence to use as the representative for a group of
#' collapsed sequences.
#' @param collapseMaxDist Numeric scalar defining the tolerance for collapsing
#' similar sequences. If the value is in [0, 1), it defines the maximal
#' Hamming distance in terms of a fraction of sequence length:
#' (\code{round(collapseMaxDist * nchar(sequence))}).
#' A value greater or equal to 1 is rounded and directly used as the maximum
#' allowed Hamming distance. Note that sequences can only be
#' collapsed if they are all of the same length. The default value is 0.
#' @param collapseMinScore Numeric scalar, indicating the minimum score
#' required for a sequence to be considered as a representative for a
#' group of similar sequences (i.e., to allow other sequences to be
#' collapsed into it). The default value is 0.
#' @param collapseMinRatio Numeric scalar. During collapsing of
#' similar sequences, a low-frequency sequence will be collapsed
#' with a higher-frequency sequence only if the ratio between the
#' high-frequency and the low-frequency scores is at least this
#' high. A value of 0 indicates that no such check is performed.
#' @param verbose Logical scalar, whether to print progress messages.
#'
#' @return A data.frame with two columns, containing the input sequences
#' and the representatives for the groups resulting from grouping similar
#' sequences, respectively.
#'
#' @examples
#' seqs <- c("AACGTAGCA", "ACCGTAGCA", "AACGGAGCA", "ATCGGAGCA", "TGAGGCATA")
#' scores <- c(5, 1, 3, 1, 8)
#' groupSimilarSequences(seqs = seqs, scores = scores,
#'                       collapseMaxDist = 1, collapseMinScore = 0,
#'                       collapseMinRatio = 0, verbose = FALSE)
#'
#' @export
#' @author Michael Stadler, Charlotte Soneson
groupSimilarSequences <- function(seqs, scores, collapseMaxDist = 0.0, collapseMinScore = 0.0, collapseMinRatio = 0.0, verbose = FALSE) {
    .Call(`_mutscan_groupSimilarSequences`, seqs, scores, collapseMaxDist, collapseMinScore, collapseMinRatio, verbose)
}

digestFastqsCpp <- function(fastqForwardVect, fastqReverseVect, mergeForwardReverse, minOverlap, maxOverlap, minMergedLength, maxMergedLength, maxFracMismatchOverlap, greedyOverlap, revComplForward, revComplReverse, elementsForward, elementLengthsForward, elementsReverse, elementLengthsReverse, adapterForward, adapterReverse, primerForward, primerReverse, wildTypeForward, wildTypeForwardNames, wildTypeReverse, wildTypeReverseNames, constantForward, constantReverse, avePhredMinForward = 20.0, avePhredMinReverse = 20.0, variableNMaxForward = 0L, variableNMaxReverse = 0L, umiNMax = 0L, nbrMutatedCodonsMaxForward = 1L, nbrMutatedCodonsMaxReverse = 1L, nbrMutatedBasesMaxForward = -1L, nbrMutatedBasesMaxReverse = -1L, forbiddenMutatedCodonsForward = "NNW", forbiddenMutatedCodonsReverse = "NNW", useTreeWTmatch = FALSE, collapseToWTForward = FALSE, collapseToWTReverse = FALSE, mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0, mutNameDelimiter = ".", constantMaxDistForward = -1L, constantMaxDistReverse = -1L, umiCollapseMaxDist = 0.0, filteredReadsFastqForward = "", filteredReadsFastqReverse = "", maxNReads = -1L, verbose = FALSE, nThreads = 1L, chunkSize = 100000L, maxReadLength = 1024L) {
    .Call(`_mutscan_digestFastqsCpp`, fastqForwardVect, fastqReverseVect, mergeForwardReverse, minOverlap, maxOverlap, minMergedLength, maxMergedLength, maxFracMismatchOverlap, greedyOverlap, revComplForward, revComplReverse, elementsForward, elementLengthsForward, elementsReverse, elementLengthsReverse, adapterForward, adapterReverse, primerForward, primerReverse, wildTypeForward, wildTypeForwardNames, wildTypeReverse, wildTypeReverseNames, constantForward, constantReverse, avePhredMinForward, avePhredMinReverse, variableNMaxForward, variableNMaxReverse, umiNMax, nbrMutatedCodonsMaxForward, nbrMutatedCodonsMaxReverse, nbrMutatedBasesMaxForward, nbrMutatedBasesMaxReverse, forbiddenMutatedCodonsForward, forbiddenMutatedCodonsReverse, useTreeWTmatch, collapseToWTForward, collapseToWTReverse, mutatedPhredMinForward, mutatedPhredMinReverse, mutNameDelimiter, constantMaxDistForward, constantMaxDistReverse, umiCollapseMaxDist, filteredReadsFastqForward, filteredReadsFastqReverse, maxNReads, verbose, nThreads, chunkSize, maxReadLength)
}

mergeValues <- function(mutNamesIn, valuesIn, delimiter = ',') {
    .Call(`_mutscan_mergeValues`, mutNamesIn, valuesIn, delimiter)
}

levenshtein_distance <- function(str1, str2, ignored_variable = -1L) {
    .Call(`_mutscan_levenshtein_distance`, str1, str2, ignored_variable)
}

hamming_distance <- function(str1, str2, ignored_variable = -1L) {
    .Call(`_mutscan_hamming_distance`, str1, str2, ignored_variable)
}

hamming_shift_distance <- function(str1, str2, max_abs_shift = -1L) {
    .Call(`_mutscan_hamming_shift_distance`, str1, str2, max_abs_shift)
}

test_that("levenshtein_distance() works", {
  # ground truth e.g. from:
  # stringdist::stringdist(s1, s2, method = "lv")

  s1 <- "AAAAAAACCC"
  s2 <- "AAAAAACCCC"
  s3 <- "ACGTACGTACGT"
  s4 <- "CGTACGTACGTA"

  expect_error(levenshtein_distance(c(s1, s1), s2))
  expect_error(levenshtein_distance(s1, c(s2, s2)))

  expect_identical(levenshtein_distance(s1, s1), 0L)
  expect_identical(levenshtein_distance(s2, s2), 0L)
  expect_identical(levenshtein_distance(s3, s3), 0L)
  expect_identical(levenshtein_distance(s4, s4), 0L)
  expect_identical(levenshtein_distance(s1, s2), 1L)
  expect_identical(levenshtein_distance(s1, s3), 8L)
  expect_identical(levenshtein_distance(s1, s4), 9L)
  expect_identical(levenshtein_distance(s2, s3), 9L)
  expect_identical(levenshtein_distance(s2, s4), 9L)
  expect_identical(levenshtein_distance(s3, s4), 2L)
})


test_that("hamming_distance() works", {
  # ground truth e.g. from:
  # stringdist::stringdist(s1, s2, method = "hamming")

  s1 <- "AAAAAAACCC"
  s2 <- "AAAAAACCCC"
  s3 <- "ACGTACGTACGT"
  s4 <- "CGTACGTACGTA"

  expect_error(hamming_distance(c(s1, s1), s2))
  expect_error(hamming_distance(s1, c(s2, s2)))

  expect_identical(hamming_distance(s1, s1),  0L)
  expect_identical(hamming_distance(s2, s2),  0L)
  expect_identical(hamming_distance(s3, s3),  0L)
  expect_identical(hamming_distance(s4, s4),  0L)
  expect_identical(hamming_distance(s1, s2),  1L)
  expect_identical(hamming_distance(s2, s1),  1L)
  expect_identical(hamming_distance(s3, s4), 12L)
  expect_identical(hamming_distance(s4, s3), 12L)
})




test_that("hamming_shift_distance() works", {
  # ground truth e.g. from:
  # s <- 1; stringdist::stringdist(substr(s1, 1, 15 - s), substr(s2, 1 + s, 15), "hamming") + 2 * s
  # s <- 1; stringdist::stringdist(substr(s2, 1, 15 - s), substr(s1, 1 + s, 15), "hamming") + 2 * s

  s1 <- "AAACAAAACGTTGCC"
  s2 <- "ACAAAACGTTGCCCC"
  s3 <- "AGTCATGCTTAGAAA"
  s4 <- "AAAAGTCATGCTTAG"

  expect_error(hamming_shift_distance(c(s1, s1), s2))
  expect_error(hamming_shift_distance(s1, c(s2, s2)))
  expect_error(hamming_shift_distance(s1, s2, c(1L, -1L)))

  expect_identical(hamming_shift_distance(s1, s1),  0L)
  expect_identical(hamming_shift_distance(s2, s2),  0L)
  expect_identical(hamming_shift_distance(s3, s3),  0L)
  expect_identical(hamming_shift_distance(s4, s4),  0L)
  expect_identical(hamming_shift_distance(s1, s2,  0),  9L)
  expect_identical(hamming_shift_distance(s1, s2,  1),  9L)
  expect_identical(hamming_shift_distance(s1, s2,  2),  4L)
  expect_identical(hamming_shift_distance(s1, s2,  3),  4L)
  expect_identical(hamming_shift_distance(s1, s2, -1),  4L)
  expect_identical(hamming_shift_distance(s3, s4,  0), 11L)
  expect_identical(hamming_shift_distance(s3, s4,  1), 11L)
  expect_identical(hamming_shift_distance(s3, s4,  2), 11L)
  expect_identical(hamming_shift_distance(s3, s4,  3),  6L)
  expect_identical(hamming_shift_distance(s3, s4, -1),  6L)
})


# using Rcpp-module that exposes the BKtree class to R
test_that("low-level BKtree wrapper functions work as expected", {
  # create random k-mers
  set.seed(42)
  k <- 30L
  n <- 20
  alph <- c("A","C","G","T","N")
  seqs <- unlist(lapply(seq_len(n - length(alph) + 1),
                        function(i) paste(sample(alph, size = k, replace = TRUE),
                                          collapse = "")))
  seqs <- c(seqs, paste0(setdiff(alph, substr(seqs[1], 1, 1)),
                         substr(seqs[1], 2, k)))

  # load module and create tree instances
  bk <- Rcpp::Module("mod_BKtree", PACKAGE = "mutscan")
  BKtree <- bk$BKtree
  expect_error(new(BKtree, "error"))
  # expect_error(new(BKtree, seqs, "error")) # triggers an unrecoverable crash in BioC 3.17 -> deactivate
  expect_error(new(BKtree, seqs, "hamming_shift", c(1L, -1L)))
  tree <- new(BKtree)
  tree2 <- new(BKtree, "levenshtein")
  tree3 <- new(BKtree, seqs[c(seq.int(n), seq.int(n))], "hamming", -1)
  tree4 <- new(BKtree, character(0), "hamming", -1)
  tree5 <- new(BKtree, "hamming_shift", 5L)
  tree6 <- new(BKtree, "hamming_shift", 0L)

  expect_s4_class(bk, "Module")
  expect_s4_class(BKtree, "C++Class")
  expect_s4_class(tree, "Rcpp_BKtree")
  expect_s4_class(tree2, "Rcpp_BKtree")
  expect_s4_class(tree3, "Rcpp_BKtree")
  expect_s4_class(tree4, "Rcpp_BKtree")
  expect_s4_class(tree5, "Rcpp_BKtree")
  expect_s4_class(tree6, "Rcpp_BKtree")
  expect_identical(tree$distance_metric(), "hamming")
  expect_identical(tree2$distance_metric(), "levenshtein")
  expect_identical(tree3$distance_metric(), "hamming")
  expect_identical(tree4$distance_metric(), "hamming")
  expect_identical(tree5$distance_metric(), "hamming_shift")
  expect_identical(tree6$distance_metric(), "hamming_shift")
  expect_identical(tree5$maximal_absolute_shift(), 5L)
  expect_identical(tree6$maximal_absolute_shift(), 0L)

  # add sequences
  expect_error(tree$insert(seqs))
  expect_identical(tree$size, 0)
  expect_identical(tree2$size, 0)
  for (s in seqs) {
    tree$insert(s)
    tree2$insert(s)
  }
  expect_identical(tree$size, n)
  expect_identical(tree2$size, n)
  expect_identical(tree3$size, n)
  tree$insert(seqs[1])
  tree2$insert(seqs[1])
  expect_identical(tree$size, n)
  expect_identical(tree2$size, n)

  # check presence
  expect_true(tree$has(seqs[1], 0))
  expect_true(tree2$has(seqs[1], 0))
  expect_true(tree$has(seqs[1], k))
  expect_true(tree2$has(seqs[1], k))
  expect_false(tree$has("non_existing", 0))
  expect_false(tree2$has("non_existing", 0))

  # get first element
  expect_identical(tree$first(), seqs[1])
  expect_identical(tree2$first(), seqs[1])

  # memory size
  expect_type(m <- tree$capacity(), "integer")
  expect_true(m > 0 && m < n * k * 8)
  expect_identical(tree2$capacity(), m)
  expect_identical(tree3$remove(seqs[n]), NULL)
  expect_type(m3 <- tree3$capacity(), "integer")
  expect_true(m3 > 0 && m3 < (n - 1) * k * 8)
  expect_identical(tree3$insert(seqs[n]), NULL)

  # print tree
  expect_output(tree$print(), paste0("^", n, ".+0: ", seqs[1]))
  expect_output(tree2$print(), paste0("^", n, ".+0: ", seqs[1]))
  expect_identical(tree3$remove(seqs[n]), NULL)
  expect_output(tree3$print(), paste0("^", n - 1, ".+0: ", seqs[1]))
  expect_identical(tree3$insert(seqs[n]), NULL)
  expect_output(tree4$print(), "^0$")

  # search  similar sequences
  # ... hamming distances
  # ground truth from: sum(stringdist::stringdist(seqs[1], seqs, method = "hamming") <= 24)
  res <- tree$search(seqs[1], 0)
  expect_type(res, "character")
  expect_length(res, 1L)
  expect_length(tree$search(seqs[1],  1),  5L)
  expect_length(tree$search(seqs[1], 24), 14L)
  expect_length(tree$search(seqs[2],  2),  1L)
  expect_length(tree$search(seqs[2], 17),  1L)
  expect_length(tree$search(seqs[2], 23),  9L)
  expect_length(tree$search(seqs[2], 26), 16L)
  expect_length(tree$search(seqs[2], 27), 20L)
  expect_length(tree$search(seqs[2], 30), 20L)
  expect_true(all(tree$search(seqs[2], 23) %in% seqs))
  # ... hamming+shift distances
  expect_identical(tree4$size, 0)
  expect_identical(tree5$size, 0)
  expect_identical(tree6$size, 0)
  for (s in c("CGATCGATGCAA", "AACGATCGATGC")) {
    tree4$insert(s)
    tree5$insert(s)
    tree6$insert(s)
  }
  expect_identical(tree4$size, 2)
  expect_identical(tree5$size, 2)
  expect_identical(tree6$size, 2)
  expect_length(tree4$search("CGATCGATGCAA",  0), 1L)
  expect_length(tree4$search("CGATCGATGCAA", 11), 1L)
  expect_length(tree4$search("CGATCGATGCAA", 12), 2L)
  expect_length(tree5$search("CGATCGATGCAA",  0), 1L)
  expect_length(tree5$search("CGATCGATGCAA",  3), 1L)
  expect_length(tree5$search("CGATCGATGCAA",  4), 2L)
  expect_length(tree6$search("CGATCGATGCAA",  0), 1L)
  expect_length(tree6$search("CGATCGATGCAA", 11), 1L)
  expect_length(tree6$search("CGATCGATGCAA", 12), 2L)
  # ... levenshtein distances
  # ground truth from: sum(stringdist::stringdist(seqs[1], seqs, method = "lv") <= 24)
  res <- tree2$search(seqs[1], 0)
  expect_type(res, "character")
  expect_length(res, 1L)
  expect_length(tree2$search(seqs[1],  1),  5L)
  expect_length(tree2$search(seqs[1], 18),  8L)
  expect_length(tree2$search(seqs[2],  2),  1L)
  expect_length(tree2$search(seqs[2], 17),  3L)
  expect_length(tree2$search(seqs[2], 18),  6L)
  expect_length(tree2$search(seqs[2], 19), 13L)
  expect_length(tree2$search(seqs[2], 20), 18L)
  expect_length(tree2$search(seqs[2], 30), 20L)
  expect_true(all(tree2$search(seqs[2], 19) %in% seqs))

  # remove sequences
  expect_identical(tree$remove("non_existing"), NULL)
  expect_identical(tree2$remove("non_existing"), NULL)
  expect_identical(tree$size, n)
  expect_identical(tree2$size, n)
  r <- ceiling(n * 0.7)
  for (i in seq_len(r)) {
    tree$remove(seqs[i])
    tree2$remove(seqs[i])
  }
  expect_identical(tree$size, n - r)
  expect_identical(tree2$size, n - r)
  expect_identical(tree$remove_all(), NULL)
  expect_identical(tree2$remove_all(), NULL)
  expect_identical(tree$size, 0)
  expect_identical(tree2$size, 0)
  expect_identical(tree3$size, n)
  expect_type(tree3$remove(seqs[1]), "NULL")
  expect_identical(tree3$size, n - 1)
  expect_type(tree3$insert(seqs[1]), "NULL")
  expect_identical(tree3$size, n)

  # check existing elements
  expect_identical(tree$remove_all(), NULL)
  expect_identical(tree2$remove_all(), NULL)
  expect_identical(tree$search(seqs[1], 0), character(0))
  expect_identical(tree2$has(seqs[1], 0), FALSE)
  for (s in seqs) {
    tree$insert(s)
    tree2$insert(s)
  }
  expect_type(res <- tree$get_all(), "character")
  expect_type(res2 <- tree2$get_all(), "character")
  expect_identical(seqs, res)
  expect_identical(seqs, res2)
  set.seed(43)
  i <- sample(length(seqs), round(length(seqs) / 3))
  for (ii in i) {
    tree$remove(seqs[ii])
    tree2$remove(seqs[ii])
  }
  expect_type(res <- tree$get_all(), "character")
  expect_type(res2 <- tree2$get_all(), "character")
  expect_identical(seqs[-i], res)
  expect_identical(seqs[-i], res2)

})

## ----------------------------------------------------------------------------
## complement
## ----------------------------------------------------------------------------
test_that("complement works", {
    expect_identical(complement("a"), "T")
    expect_identical(complement("A"), "T")
    expect_identical(complement("c"), "G")
    expect_identical(complement("C"), "G")
    expect_identical(complement("g"), "C")
    expect_identical(complement("G"), "C")
    expect_identical(complement("t"), "A")
    expect_identical(complement("T"), "A")
    expect_identical(complement("n"), "N")
    expect_identical(complement("N"), "N")
    expect_error(complement("B"), "Invalid DNA base")
    expect_error(complement("Q"), "Invalid DNA base")
})

## ----------------------------------------------------------------------------
## compareCodonPositions
## ----------------------------------------------------------------------------
test_that("compareCodonPositions works", {
  expect_true(compareCodonPositions("f.1.AAA_", "f.2.ACT_", "."))
  expect_true(compareCodonPositions("f.2.ACT_", "f.10.TCA_", "."))
})

## ----------------------------------------------------------------------------
## findClosestRefSeq
## ----------------------------------------------------------------------------
test_that("findClosestRefSeq works", {
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("ATTT", "ACTT"),
                                 upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("ACGT", "ACTT"),
                                 upperBoundMismatch = 4L, sim = 0L), 0L)
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("ACG", "ACGT"),
                                 upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("AACGT", "ACCTA"),
                                 upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("ATTT", "ACTT"),
                                 upperBoundMismatch = 0L, sim = 0L), -1L)
  expect_equal(findClosestRefSeq(varSeq = "ACGT", wtSeq = c("ATGT", "ACTT"),
                                 upperBoundMismatch = 1L, sim = 0L), -2L)

  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("ATTT", "ACTT"),
                                          upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("ACGT", "ACTT"),
                                          upperBoundMismatch = 4L, sim = 0L), 0L)
  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("ACG", "ACGT"),
                                          upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("AACGT", "ACCTA"),
                                          upperBoundMismatch = 4L, sim = 0L), 1L)
  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("ATTT", "ACTT"),
                                          upperBoundMismatch = 0L, sim = 0L), -1L)
  expect_equal(findClosestRefSeqEarlyStop(varSeq = "ACGT", wtSeq = c("ATGT", "ACTT"),
                                          upperBoundMismatch = 1L, sim = 0L), -2L)
})

## ----------------------------------------------------------------------------
## decomposeRead
## ----------------------------------------------------------------------------
test_that("decomposeRead works", {
  ## All element lengths specified, no primer
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVCUSV",
    elementLengths = c(2, 1, 3, 3, 1, 2, 1, 1), primerSeqs = "",
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "", nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "ACCT")
  expect_equal(res$varSeq, "AAGG")
  expect_equal(res$varQual, "gABG")
  expect_equal(res$varLengths, c(3L, 1L))
  expect_equal(res$constSeq, "TTAC")
  expect_equal(res$constQual, "defC")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## All element lengths specified, no primer, 0 lengths included
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTG", squal = "abcdefgABCDEF", elements = "SUSCVUCUS",
    elementLengths = c(0, 2, 1, 3, 3, 0, 1, 2, 1), primerSeqs = "",
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "ACCT")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTAC")
  expect_equal(res$constQual, "defC")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Derive variable length from other lengths
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTG", squal = "abcdefgABCDEF", elements = "USCVCUS",
    elementLengths = c(2, 1, 3, -1, 1, 2, 1), primerSeqs = "",
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "ACCT")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTAC")
  expect_equal(res$constQual, "defC")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Primer
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVPS",
    elementLengths = c(2, 1, 3, 3, 3, -1), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "AC")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTA")
  expect_equal(res$constQual, "def")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Only variable sequence
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "V",
    elementLengths = -1, primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "ACGTTAAAGCCTGG")
  expect_equal(res$varQual, "abcdefgABCDEFG")
  expect_equal(res$varLengths, 14L)
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Primer + variable sequence
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "PV",
    elementLengths = c(-1, -1), primerSeqs = c("ACG"),
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "TTAAAGCCTGG")
  expect_equal(res$varQual, "defgABCDEFG")
  expect_equal(res$varLengths, 11L)
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Variable sequence + primer
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "VP",
    elementLengths = c(-1, -1), primerSeqs = c("ACG"),
    varSeq = "A", varQual = "E", varLengths = 1L, umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "A")
  expect_equal(res$varQual, "E")
  expect_equal(res$varLengths, c(1L, 0L))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Primer, not the right length of sequence
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "VPU",
    elementLengths = c(10, 3, 2), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0L), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "")
  expect_equal(res$varQual, "")
  expect_equal(res$varLengths, integer(0))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 1)

  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "VPU",
    elementLengths = c(8, 3, 2), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "")
  expect_equal(res$varQual, "")
  expect_equal(res$varLengths, integer(0))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 1)

  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "CVPU",
    elementLengths = c(10, -1, 3, 2), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "")
  expect_equal(res$varQual, "")
  expect_equal(res$varLengths, integer(0))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 1)

  ## Primer, but don't explicitly specify that the part after the primer should be skipped
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVP",
    elementLengths = c(2, 1, 3, 3, 3), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "AC")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTA")
  expect_equal(res$constQual, "def")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Multiple primers, only one present
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVP",
    elementLengths = c(2, 1, 3, 3, 3), primerSeqs = c("TTT", "CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "AC")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTA")
  expect_equal(res$constQual, "def")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Multiple primers, none present
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVP",
    elementLengths = c(2, 1, 3, 3, 3), primerSeqs = c("TTT", "GGG"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "")
  expect_equal(res$varQual, "")
  expect_equal(res$varLengths, integer(0))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 1)
  expect_equal(res$nReadWrongLength, 0)

  ## Multiple primers, all present (use the first one)
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "USCVP",
    elementLengths = c(2, 1, 3, 3, 3), primerSeqs = c("CCT", "TTA"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "AC")
  expect_equal(res$varSeq, "AAG")
  expect_equal(res$varQual, "gAB")
  expect_equal(res$varLengths, 3L)
  expect_equal(res$constSeq, "TTA")
  expect_equal(res$constQual, "def")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Primer, infer length of sequences
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "UPV",
    elementLengths = c(-1, 3, -1), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "ACGTTAAAG")
  expect_equal(res$varSeq, "GG")
  expect_equal(res$varQual, "FG")
  expect_equal(res$varLengths, 2L)
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Primer, don't specify primer length
  res <- mutscan:::test_decomposeRead(
    sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "UPV",
    elementLengths = c(-1, -1, -1), primerSeqs = c("CCT"),
    varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
    constSeq = "", constQual = "",
    nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "ACGTTAAAG")
  expect_equal(res$varSeq, "GG")
  expect_equal(res$varQual, "FG")
  expect_equal(res$varLengths, 2L)
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)

  ## Multiple variable segments with different lengths
  res <- mutscan:::test_decomposeRead(
      sseq = "ACGTTAAAGCCTGG", squal = "abcdefgABCDEFG", elements = "VPV",
      elementLengths = c(-1, -1, -1), primerSeqs = c("CCT"),
      varSeq = "", varQual = "", varLengths = integer(0), umiSeq = "",
      constSeq = "", constQual = "",
      nNoPrimer = 0, nReadWrongLength = 0)
  expect_equal(res$umiSeq, "")
  expect_equal(res$varSeq, "ACGTTAAAGGG")
  expect_equal(res$varQual, "abcdefgABFG")
  expect_equal(res$varLengths, c(9L, 2L))
  expect_equal(res$constSeq, "")
  expect_equal(res$constQual, "")
  expect_equal(res$nNoPrimer, 0)
  expect_equal(res$nReadWrongLength, 0)
})

## ----------------------------------------------------------------------------
## mergeReadPairsPartial
## ----------------------------------------------------------------------------
test_that("mergeReadPairsPartial works", {
  ## don't count N as mismatch, pick base with higher quality
  sF1 <- "AAAANA"; qF1 <- rep(40L, nchar(sF1)); lF1 <- 6L
  sR1 <- "AACCCC"; qR1 <- rep(42L, nchar(sR1)); lR1 <- 6L
  res1a <- mutscan:::test_mergeReadPairPartial(sF1, qF1, sR1, qR1, lF1, lR1, 1, 6, 0, 0, 1/4, TRUE)
  res1b <- mutscan:::test_mergeReadPairPartial(sF1, qF1, sR1, qR1, lF1, lR1, 1, 6, 0, 0, 1/4, FALSE)
  res1c <- mutscan:::test_mergeReadPairPartial(sF1, qR1, sR1, qF1, lF1, lR1, 2, 2, 0, 0, 1/4, TRUE)
  res1d <- mutscan:::test_mergeReadPairPartial(sF1, qF1, sR1, qR1, lF1, lR1, 2, 2, 0, 0, 0, TRUE)
  expect_type(res1a, "list")
  expect_type(res1b, "list")
  expect_type(res1c, "list")
  expect_type(res1d, "list")
  expect_identical(res1a$mergedSeq, "AAAACCCC")
  expect_identical(res1a$mergedQual, rep(c(40L, 42L), c(2, 6)))
  expect_identical(res1a$mergedLengths, 8L)
  expect_identical(res1a, res1b)
  expect_identical(res1c$mergedSeq, "AAAANACCCC")
  expect_identical(res1d$mergedSeq, "AAAAAACCCC")
  expect_identical(res1c$mergedLengths, 10L)
  expect_identical(res1d$mergedLengths, 10L)

  ## no valid overlap, multiple possible overlaps with single mismatch
  sF2 <- "TTACACG"; qF2 <- rep(10L, nchar(sF2)); lF2 <- 7L
  sR2 <- "ACACACA"; qR2 <- rep(40L, nchar(sR2)); lR2 <- 7L
  res2a <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 0, 0, 0, 0, maxFracMismatchOverlap = 3/7, TRUE)
  res2b <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 1, 7, 0, 0, 1/5, TRUE)
  expect_type(res2a, "list")
  expect_type(res2b, "list")
  expect_identical(res2a$mergedSeq, sR2)
  expect_identical(res2a$mergedQual, qR2)
  expect_identical(res2a$mergedLengths, 7L)
  expect_identical(res2b$mergedSeq, "TTACACACA")
  expect_identical(res2b$mergedQual, rep(c(10L,40L), c(2,7)))
  expect_identical(res2b$mergedLengths, 9L)

  ## specify min/max merged length instead
  res2c <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 1, 7, 9, 9, 1, FALSE)
  res2d <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 1, 7, 11, 11, 1, FALSE)
  res2e <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 1, 7, 12, 12, 1, FALSE)
  expect_type(res2c, "list")
  expect_type(res2d, "list")
  expect_type(res2e, "list")
  expect_identical(res2c$mergedSeq, "TTACACACA")
  expect_identical(res2c$mergedQual, rep(c(10L, 40L), c(2, 7)))
  expect_identical(res2c$mergedLengths, 9L)
  expect_identical(res2d$mergedSeq, "TTACACACACA")
  expect_identical(res2d$mergedQual, rep(c(10L, 40L), c(4, 7)))
  expect_identical(res2d$mergedLengths, 11L)
  expect_identical(res2e$mergedSeq, "TTACAACACACA")
  expect_identical(res2e$mergedQual, rep(c(10L, 40L), c(5, 7)))
  expect_identical(res2e$mergedLengths, 12L)

  ## invalid overlaps specified/overlaps modified internally
  ## minOverlap > lenF -> no valid overlap
  res0a <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 8, 7, 9, 9, 1, FALSE)
  expect_type(res0a, "list")
  expect_identical(res0a$mergedSeq, sF2)
  expect_identical(res0a$mergedQual, qF2)
  expect_identical(res0a$mergedLengths, 7L)
  expect_true(res0a$return)
  ## minOverlap = 0, lenF > lenR -> minOverlap = lenR
  sF0 <- "TTACACG"; qF0 <- rep(10L, nchar(sF0)); lF0 <- nchar(sF0)
  sR0 <- "ACACG"; qR0 <- rep(40L, nchar(sR0)); lR0 <- nchar(sR0)
  res0b <- mutscan:::test_mergeReadPairPartial(sF0, qF0, sR0, qR0, lF0, lR0, 0, 7, 5, 9, 1, FALSE)
  expect_type(res0b, "list")
  expect_identical(res0b$mergedSeq, sF0)
  expect_identical(res0b$mergedQual, rep(c(10L, 40L), c(2, 5)))
  expect_identical(res0b$mergedLengths, 7L)
  expect_false(res0b$return)
  ## maxOverlap = 0, lenF > lenR -> maxOverlap = lenR
  res0c <- mutscan:::test_mergeReadPairPartial(sF0, qF0, sR0, qR0, lF0, lR0, 1, 0, 5, 9, 1, FALSE)
  expect_type(res0c, "list")
  expect_identical(res0c$mergedSeq, sF0)
  expect_identical(res0c$mergedQual, rep(c(10L, 40L), c(2, 5)))
  expect_identical(res0c$mergedLengths, 7L)
  expect_false(res0c$return)
  ## minOverlap > maxOverlap -> no valid overlap
  res0d <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 4, 3, 9, 9, 1, FALSE)
  expect_type(res0d, "list")
  expect_identical(res0d$mergedSeq, sF2)
  expect_identical(res0d$mergedQual, qF2)
  expect_identical(res0d$mergedLengths, 7L)
  expect_true(res0d$return)

  ## minMergedLength > lenF + lenR -> no valid overlap
  res0e <- mutscan:::test_mergeReadPairPartial(sF2, qF2, sR2, qR2, lF2, lR2, 1, 7, 15, 9, 1, FALSE)
  expect_type(res0e, "list")
  expect_identical(res0e$mergedSeq, sF2)
  expect_identical(res0e$mergedQual, qF2)
  expect_identical(res0e$mergedLengths, 7L)
  expect_true(res0e$return)
  ## maxMergedLength > lenF + lenR -> maxMergedLength = lenF + lenR
  res0f <- mutscan:::test_mergeReadPairPartial(sF0, qF0, sR0, qR0, lF0, lR0, 1, 0, 5, 19, 1, FALSE)
  expect_type(res0f, "list")
  expect_identical(res0f$mergedSeq, sF0)
  expect_identical(res0f$mergedQual, rep(c(10L, 40L), c(2, 5)))
  expect_identical(res0f$mergedLengths, 7L)
  expect_false(res0f$return)

  ## padded reads
  for (i in 1:10) {
    sF <- paste(rep(c("C","A"), c(i, 6)), collapse = "")
    qF <- rep(30L, nchar(sF))
    sR <- paste(rep(c("A","C"), c(6, i)), collapse = "")
    qR <- rep(30L, nchar(sR))
    lF <- i + 6
    lR <- 6 + i
    res <- mutscan:::test_mergeReadPairPartial(sF, qF, sR, qR, lF, lR, 6, 6, 0, 0, 0)
    expect_identical(res$mergedSeq, paste(rep(c("C","A","C"), c(i, 6, i)), collapse = ""))
    expect_equal(res$mergedLengths, i + 6 + i)
  }

  ## reads of unequal length
  sR <- paste(rep("A", 6), collapse = "")
  qR <- rep(30L, nchar(sR))
  lR <- 6L
  for (i in 1:10) {
    sF <- paste(rep(c("C","A"), c(i, 6)), collapse = "")
    qF <- rep(30L, nchar(sF))
    lF <- i + 6
    res <- mutscan:::test_mergeReadPairPartial(sF, qF, sR, qR, lF, lR, 6, 6, 0, 0, 0)
    expect_identical(res$mergedSeq, sF)
    expect_equal(res$mergedLengths, lF)
    res <- mutscan:::test_mergeReadPairPartial(sR, qR, sF, qF, lR, lF, 6, 6, 0, 0, 0)
    expect_identical(res$mergedSeq, sR)
    expect_equal(res$mergedLengths, lR)
  }

  ## test variable segment lengths merging
  ## ... that must work
  sF3 <- "AAAAAAAAACGTCCCAAC";   qF3 <- rep(40L, nchar(sF3)); lF3 <- c(13L,  5L)
  sR3 <- "CGTCCCAACCCGGGGGGGGG"; qR3 <- rep(42L, nchar(sR3)); lR3 <- c( 4L, 16L)
  res3a <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, lF3, lR3, 1, 0, 1, 0, 0.1, FALSE)
  res3b <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(5L, 13L), 20L, 1, 0, 1, 0, 0.1, FALSE)
  res3c <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(5L, 13L), c(14L, 6L), 1, 0, 1, 0, 0.1, FALSE)
  res3d <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(5L, 8L, 5L), c(4L, 13L, 3L), 1, 0, 1, 0, 0.1, FALSE)
  res3e <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(3L, 2L, 8L, 2L, 3L), c(4L, 2L, 11L, 3L), 1, 0, 1, 0, 0.1, FALSE)
  res3f <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(5L, 13L), c(9L, 5L, 6L), 1, 0, 1, 0, 0.1, FALSE)
  expect_false(res3a$return)
  expect_identical(res3a$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3a$mergedLengths, c(13L, 16L))
  expect_identical(res3b$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3b$mergedLengths, c(5L, 24L))
  expect_identical(res3c$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3c$mergedLengths, c(5L, 18L, 6L))
  expect_identical(res3d$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3d$mergedLengths, c(5L, 8L, 13L, 3L))
  expect_identical(res3e$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3e$mergedLengths, c(3L, 2L, 8L, 2L, 11L, 3L))
  expect_identical(res3f$mergedSeq, "AAAAAAAAACGTCCCAACCCGGGGGGGGG")
  expect_equal(res3f$mergedLengths, c(5L, 13L, 5L, 6L))
  ## ... that must fail
  res3g <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(12L, 6L), c( 4L, 16L), 1, 0, 1, 0, 0.1, FALSE)
  res3h <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(5L, 13L), c(2L, 18L), 1, 0, 1, 0, 0.1, FALSE)
  res3i <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sR3, qR3, c(12L, 6L), c(3L, 4L, 2L, 11L), 1, 0, 1, 0, 0.1, FALSE)
  expect_identical(res3g$mergedSeq, sF3)
  expect_true(res3g$return)
  expect_identical(res3h$mergedSeq, sF3)
  expect_true(res3h$return)
  expect_identical(res3i$mergedSeq, sF3)
  expect_true(res3i$return)
  ## ... that completely overlaps and must work
  res3j <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sF3, qF3, 18L, 18L, 1, 0, 1, 0, 0.1, FALSE)
  res3k <- mutscan:::test_mergeReadPairPartial(sF3, qF3, sF3, qF3, c(5L, 13L), c(5L, 13L), 1, 0, 1, 0, 0.1, FALSE)
  expect_identical(res3j$mergedSeq, sF3)
  expect_false(res3j$return)
  expect_equal(res3j$mergedLengths, 18L)
  expect_identical(res3k$mergedSeq, sF3)
  expect_false(res3k$return)
  expect_equal(res3k$mergedLengths, c(5L, 13L))
})

## ----------------------------------------------------------------------------
## makeBaseHGVS/makeAAHGVS
## ----------------------------------------------------------------------------
test_that("makeBaseHGVS works", {
    expect_equal(makeBaseHGVS(c("FOS.1.A", "FOS.2.T"), ".", "TAG", "ATG"), "1_2delinsAT_")
    expect_equal(makeBaseHGVS(c("f.1.A", "r.4.A"), ".", "TAGT", "AAGA"), "[1T>A;4T>A]_")
    expect_equal(makeBaseHGVS(c("f.1.A", "r.4.A", "f.6.C"), ".", "TAGTGTAGTCCGT", "AAGAGCAGTCCGT"),
                 "[1T>A;4_6delinsAGC]_")
    expect_equal(makeBaseHGVS(c("r.4.A"), ".", "TAGTGTAGTCCGT", "TAGAGTAGTCCGT"), "4T>A_")
    expect_equal(makeBaseHGVS(character(0), ".", "AAA", "AAA"), "_")
    expect_equal(makeBaseHGVS(c("f.1.A", "r.4.A", "r.8.A"), ".",
                              "TACTTTAT", "AAGAAAAA"),
                 "[1T>A;4T>A;8T>A]_")
})

test_that("makeAAHGVS works", {
    expect_equal(test_makeAAHGVS("FOS.1.H", ".", "TDTLQAETDQLEDEKSALQTEIANLLKEKEKL"),
                 "(Thr1His)_")
    expect_equal(test_makeAAHGVS(c("f.1.L", "r.4.M"), ".", "TDTLQAETDQLEDEKSALQTEIANLLKEKEKL"),
                 "[(Thr1Leu);(Leu4Met)]_")
    expect_equal(test_makeAAHGVS(c("f.1.L", "f.2.M"), ".", "TDTLQAETDQLEDEKSALQTEIANLLKEKEKL"),
                 "[(Thr1Leu);(Asp2Met)]_")
    expect_equal(test_makeAAHGVS(c("f.1.L", "f.2.M", "f.7.M"), ".", "TDTLQAETDQLEDEKSALQTEIANLLKEKEKL"),
                 "[(Thr1Leu);(Asp2Met);(Glu7Met)]_")
    expect_equal(test_makeAAHGVS(character(0), ".", "TDTL"), "_")
})

## ----------------------------------------------------------------------------
## compareToWildtype
## ----------------------------------------------------------------------------
test_that("compareToWildtype works", {
    # read is kept
    expect_identical(
        test_compareToWildtype(varSeq = "AAAGGACGA", wtSeq = "AATGGACGT",
                               varIntQual = rep(32L, 9L),
                               forbiddenCodons_vect = character(0),
                               mutatedPhredMin = 0.0,
                               nbrMutatedCodonsMax = 3L,
                               codonPrefix = "xyz",
                               nbrMutatedBasesMax = 3L,
                               mutNameDelimiter = ".",
                               collapseToWT = TRUE),
        list(nMutQualTooLow = 0L, nTooManyMutCodons = 0L, nForbiddenCodons = 0L,
             nTooManyMutBases = 0L, nMutBases = 2L, nMutCodons = 2L,
             nMutAAs = 1L, mutantName = "xyz_", mutantNameBase = "xyz_",
             mutantNameCodon = "xyz_", mutantNameBaseHGVS = "xyz:c_",
             mutantNameAA = "xyz_", mutantNameAAHGVS = "xyz:p_",
             mutationTypes = c("nonsynonymous", "silent")))
    expect_identical(
        test_compareToWildtype(varSeq = "AAAGGACGA", wtSeq = "AATGGACGT",
                               varIntQual = rep(32L, 9L),
                               forbiddenCodons_vect = character(0),
                               mutatedPhredMin = 0.0,
                               nbrMutatedCodonsMax = 3L,
                               codonPrefix = "xyz:c",
                               nbrMutatedBasesMax = 3L,
                               mutNameDelimiter = ".",
                               collapseToWT = TRUE),
        list(nMutQualTooLow = 0L, nTooManyMutCodons = 0L, nForbiddenCodons = 0L,
             nTooManyMutBases = 0L, nMutBases = 2L, nMutCodons = 2L,
             nMutAAs = 1L, mutantName = "xyz:c_", mutantNameBase = "xyz:c_",
             mutantNameCodon = "xyz:c_", mutantNameBaseHGVS = "xyz:c_",
             mutantNameAA = "xyz:c_", mutantNameAAHGVS = "xyz:p_",
             mutationTypes = c("nonsynonymous", "silent")))
    # read is filtered out
    expect_identical(
        test_compareToWildtype(varSeq = "CGTCGTCGA", wtSeq = "CGTCGTCGT",
                               varIntQual = rep(32L, 9L),
                               forbiddenCodons_vect = character(0),
                               mutatedPhredMin = 33.0, # <---
                               nbrMutatedCodonsMax = 3L,
                               codonPrefix = "xyz",
                               nbrMutatedBasesMax = 3L,
                               mutNameDelimiter = ".",
                               collapseToWT = TRUE),
        list())
    expect_identical(
        test_compareToWildtype(varSeq = "CGACGACGA", wtSeq = "CGTCGTCGT",
                               varIntQual = rep(32L, 9L),
                               forbiddenCodons_vect = character(0),
                               mutatedPhredMin = 0.0,
                               nbrMutatedCodonsMax = -1L,
                               codonPrefix = "xyz",
                               nbrMutatedBasesMax = 2L, # <---
                               mutNameDelimiter = ".",
                               collapseToWT = TRUE),
        list())
})

## ----------------------------------------------------------------------------
## groupSimilarSequences
## ----------------------------------------------------------------------------
test_that("groupSimilarSequences works", {
    expect_identical(
        groupSimilarSequences(c("AA", "AT", "AC", "TT"), 1:4, collapseMaxDist = 0),
        data.frame(sequence = c("AA", "AT", "AC", "TT"),
                   representative = c("AA", "AT", "AC", "TT"))
    )
    expect_identical(
        groupSimilarSequences(c("AA", "AT", "AC", "TT"), 1:4, collapseMaxDist = 1),
        data.frame(sequence = c("AA", "AT", "AC", "TT"),
                   representative = c("AC", "TT", "AC", "TT"))
    )
    expect_warning(groupSimilarSequences(c("AA", "AT", "AC", "TTT"), 1:4,
                                       collapseMaxDist = 1),
                 "not all of the same length")
})

test_that("translateString works", {
  code <- Biostrings::GENETIC_CODE
  for (i in seq_along(code)) {
    expect_identical(translateString(names(code)[i]), unname(code[i]))
  }
  expect_identical(translateString(paste(names(code), collapse = "_")),
                   paste(code, collapse = "_"))
  expect_identical(translateString(paste(names(code), collapse = "")),
                   paste(code, collapse = ""))
  expect_identical(translateString(tolower(paste(names(code), collapse = ""))),
                   paste(code, collapse = ""))
  expect_identical(translateString(paste(gsub("T", "U", names(code)), collapse = "")),
                   paste(code, collapse = ""))
  expect_identical(translateString("AAN,AAA.A"), "XXX")
})

test_that("calcNearestStringDist fails with incorrect arguments", {
    expect_error(calcNearestStringDist(), "argument \"x\" is missing")
    expect_error(calcNearestStringDist(x = 1:3), "Expecting a string vector")
    expect_error(calcNearestStringDist(x = letters, metric = "error"), "unknown distance metric 'error'")
    expect_error(calcNearestStringDist(x = letters, nThreads = c(1L, 2L)), "Expecting a single value")
})

test_that("calcNearestStringDist works as expected", {
    # strs1 <- c("lazy", "hazy", "craz")  # same lengths
    # strs2 <- c("lazy", "hazy", "crazy") # unequal lengths

    set.seed(42)
    strs1 <- sample(x = Biostrings::mkAllStrings(alphabet = Biostrings::DNA_BASES,
                                                 width = 7),
                    size = 200)
    strs2 <- c(sample(x = Biostrings::mkAllStrings(alphabet = Biostrings::DNA_BASES,
                                                   width = 7),
                      size = 200),
               sample(x = Biostrings::mkAllStrings(alphabet = Biostrings::DNA_BASES,
                                                   width = 8),
                      size = 200))

    d0 <- pwalign::stringDist(x = strs1, method = "hamming")
    e0 <- pwalign::stringDist(x = strs2, method = "levenshtein")

    dm <- as.matrix(d0)
    em <- as.matrix(e0)

    diag(dm) <- diag(em) <- NA

    d <- as.integer(unname(apply(dm, 1, min, na.rm = TRUE)))
    e <- as.integer(unname(apply(em, 1, min, na.rm = TRUE)))

    expect_type(res1 <- calcNearestStringDist(x = strs1, metric = "hamming", nThreads = 1L), "integer")
    expect_type(res2 <- calcNearestStringDist(x = strs1, metric = "hamming", nThreads = 4L), "integer")
    expect_type(res3 <- calcNearestStringDist(x = strs2, metric = "levenshtein", nThreads = 1L), "integer")
    expect_type(res4 <- calcNearestStringDist(x = strs2, metric = "levenshtein", nThreads = 4L), "integer")
    expect_type(res5 <- calcNearestStringDist(x = strs1, metric = "hamming_shift", nThreads = 1L), "integer")
    expect_type(res6 <- calcNearestStringDist(x = strs1, metric = "hamming_shift", nThreads = 4L), "integer")

    expect_identical(d, res1)
    expect_identical(e, res3)
    expect_true(all(res5 <= res1))
    expect_identical(res1, res2)
    expect_identical(res3, res4)
    expect_identical(res5, res6)
})

Ldef <- list(
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    collapseToWTForward = FALSE,
    collapseToWTReverse = FALSE, 
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    maxNReads = -1, verbose = FALSE
)
Ldef1 <- c(
    list(fastqForward = system.file("extdata/transInput_1.fastq.gz", package = "mutscan"),
         fastqReverse = system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
    ), Ldef)
Ldef2 <- c(
    list(fastqForward = system.file("extdata/transOutput_1.fastq.gz", package = "mutscan"),
         fastqReverse = system.file("extdata/transOutput_2.fastq.gz", package = "mutscan")
    ), Ldef)

out1 <- do.call(digestFastqs, Ldef1)
out2 <- do.call(digestFastqs, Ldef2)

coldata <- data.frame(Name = c("sample1", "sample2"),
                      Condition = c("input", "output"),
                      Replicate = c(1, 1), OD = c(0.05, 1.5),
                      stringsAsFactors = FALSE)

test_that("summarizeExperiment fails with incorrect arguments", {
    ## x must be a named list, with names matching coldata$Name
    expect_error(summarizeExperiment(x = 1, coldata = coldata))
    expect_error(summarizeExperiment(x = list(out1), coldata = coldata))
    expect_error(summarizeExperiment(x = out1, coldata = coldata))
    expect_error(summarizeExperiment(x = list(s1 = out1, s1 = out2), coldata = coldata))

    ## coldata must be a data.frame with a column Name
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = 1))
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = coldata[, c("Condition", "OD")]))
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = as.list(coldata)))

    ## countType must be reads or umis
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = coldata, countType = 1))
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = coldata, countType = "umi"))
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                                     coldata = coldata, countType = c("reads", "umis")))

    ## names must match
    expect_error(summarizeExperiment(x = list(s1 = out1, s2 = out2),
                                     coldata = coldata, countType = "umis"))
    expect_warning(summarizeExperiment(x = list(sample1 = out1, sample2 = out2,
                                                sample3 = out1),
                                       coldata = coldata, countType = "umis"))

    ## samples must have the same mutNameDelimiter
    tmpout2 <- out2
    tmpout2$parameters$mutNameDelimiter <- ":"
    expect_error(summarizeExperiment(x = list(sample1 = out1, sample2 = tmpout2),
                                     coldata = coldata, countType = "reads"))
})

test_that("summarizeExperiment works as expected with reads output", {
    se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                              coldata = coldata, countType = "reads")
    expect_equal(nrow(se), length(union(out1$summaryTable$mutantName,
                                        out2$summaryTable$mutantName)))
    expect_equal(ncol(se), 2)
    expect_equal(sort(rownames(se)),
                 sort(union(out1$summaryTable$mutantName, out2$summaryTable$mutantName)))
    expect_equal(Matrix::colSums(SummarizedExperiment::assay(se, "counts")),
                 c(out1$filterSummary$nbrRetained, out2$filterSummary$nbrRetained),
                 ignore_attr = TRUE)
    expect_equal(Matrix::colSums(SummarizedExperiment::assay(se, "counts")),
                 c(sum(out1$summaryTable$nbrReads), sum(out2$summaryTable$nbrReads)),
                 ignore_attr = TRUE)

    for (cn in colnames(out1$filterSummary)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     out1$filterSummary[, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     out2$filterSummary[, cn])
    }

    for (cn in colnames(coldata)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     coldata[1, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     coldata[2, cn])
    }

    expect_equal(S4Vectors::metadata(se)$parameters[["sample1"]], out1$parameters)
    expect_equal(S4Vectors::metadata(se)$parameters[["sample2"]], out2$parameters)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample1"]], out1$errorStatistics)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample2"]], out2$errorStatistics)

    expect_true(all(c("mutantName", "sequence", "nbrMutBases",
                      "minNbrMutBases", "maxNbrMutBases",
                      "nbrMutCodons", "minNbrMutCodons", "maxNbrMutCodons",
                      "nbrMutAAs", "minNbrMutAAs", "maxNbrMutAAs",
                      "sequenceAA", "mutantNameAA", "mutationTypes",
                      "varLengths") %in% 
                        colnames(SummarizedExperiment::rowData(se))))
    
    ## Check that the number of mutated codons are correct (=equal to the number of
    ## entries in the mutant name that don't contain WT)
    expect_equal(sapply(strsplit(SummarizedExperiment::rowData(se)$mutantName, "_"),
                        function(w) length(w[!grepl("WT", w)])),
                 SummarizedExperiment::rowData(se)$minNbrMutCodons, ignore_attr = TRUE)
    expect_equal(sapply(strsplit(SummarizedExperiment::rowData(se)$mutantName, "_"),
                        function(w) length(w[!grepl("WT", w)])),
                 SummarizedExperiment::rowData(se)$maxNbrMutCodons, ignore_attr = TRUE)
    
    ## Check that the number of mutated bases is not larger than 3x the number of mutated codons
    expect_true(all(SummarizedExperiment::rowData(se)$maxNbrMutBases <=  
                        3 * SummarizedExperiment::rowData(se)$maxNbrMutCodons))
    expect_true(all(SummarizedExperiment::rowData(se)$maxNbrMutAAs <= 
                        SummarizedExperiment::rowData(se)$maxNbrMutCodons))
    
    ## variable lengths
    expect_equal(SummarizedExperiment::rowData(se)$varLengths, 
                 rep("96_96", nrow(se)), ignore_attr = TRUE)
    
    ## All variants with no mutated AAs must have a WT in the name
    expect_true(all(grepl("WT", SummarizedExperiment::rowData(se)$mutantNameAA[SummarizedExperiment::rowData(se)$maxNbrMutAAs == 0])))
    ## No mutation for the complete WT
    expect_equal(SummarizedExperiment::rowData(se)["f.0.WT_r.0.WT", ]$mutationTypes, "")
    ## Mutation types for variants with no mutated AAs, but mutated bases, should be silent
    expect_true(all(SummarizedExperiment::rowData(se)$mutationTypes[SummarizedExperiment::rowData(se)$maxNbrMutAAs == 0 & SummarizedExperiment::rowData(se)$maxNbrMutBases > 0] == "silent"))
    ## check translation
    expect_equal(SummarizedExperiment::rowData(se)$sequenceAA[3], 
                 mutscan:::translateString(SummarizedExperiment::rowData(se)$sequence[3]))
    
    ## Spot checks
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutBases[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.CTC"], 3)  ## WT: GCT
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutCodons[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.CTC"], 1)  ## WT: GCT
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutAAs[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.CTC"], 1)  ## WT: A
    expect_equal(SummarizedExperiment::rowData(se)$mutantNameAA[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.CTC"], "f.0.WT_r.13.L")  ## WT: A
    expect_equal(SummarizedExperiment::rowData(se)$mutationTypes[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.CTC"], "nonsynonymous")  ## WT: A
    
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutBases[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.GCG"], 1)  ## WT: GCT
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutCodons[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.GCG"], 1)  ## WT: GCT
    expect_equal(SummarizedExperiment::rowData(se)$minNbrMutAAs[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.GCG"], 0)  ## WT: A
    expect_equal(SummarizedExperiment::rowData(se)$mutantNameAA[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.GCG"], "f.0.WT_r.0.WT")  ## WT: A
    expect_equal(SummarizedExperiment::rowData(se)$mutationTypes[SummarizedExperiment::rowData(se)$mutantName == "f.0.WT_r.13.GCG"], "silent")  ## WT: A
    
})

test_that("summarizeExperiment works as expected with umis output", {
    se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                              coldata = coldata, countType = "umis")
    expect_equal(nrow(se), length(union(out1$summaryTable$mutantName,
                                        out2$summaryTable$mutantName)))
    expect_equal(ncol(se), 2)
    expect_equal(sort(rownames(se)),
                 sort(union(out1$summaryTable$mutantName, out2$summaryTable$mutantName)))
    expect_equal(Matrix::colSums(SummarizedExperiment::assay(se, "counts")),
                 c(sum(out1$summaryTable$nbrUmis), sum(out2$summaryTable$nbrUmis)),
                 ignore_attr = TRUE)

    for (cn in colnames(out1$filterSummary)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     out1$filterSummary[, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     out2$filterSummary[, cn])
    }

    for (cn in colnames(coldata)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     coldata[1, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     coldata[2, cn])
    }

    expect_equal(S4Vectors::metadata(se)$parameters[["sample1"]], out1$parameters)
    expect_equal(S4Vectors::metadata(se)$parameters[["sample2"]], out2$parameters)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample1"]], out1$errorStatistics)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample2"]], out2$errorStatistics)
})

test_that("summarizeExperiment orders samples equally in count matrix/colData", {
    se1 <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                               coldata = coldata, countType = "umis")
    se2 <- summarizeExperiment(x = list(sample2 = out2, sample1 = out1),
                               coldata = coldata, countType = "umis")
    m1 <- se1[, match(coldata$Name, colnames(se1))]
    m2 <- se2[rownames(m1), match(coldata$Name, colnames(se2))]
    expect_equal(SummarizedExperiment::colData(m1), SummarizedExperiment::colData(m2))
    expect_equal(SummarizedExperiment::assay(m1, "counts"),
                 SummarizedExperiment::assay(m2, "counts"))
})

test_that("summarizeExperiment recognizes the presence of UMI counts correctly", {
    L1 <- Ldef1; L1$elementsForward <- "SSCV"; L1$elementsReverse <- "SSCV"
    L2 <- Ldef2; L2$elementsForward <- "SSCV"; L2$elementsReverse <- "SSCV"
    outl1 <- do.call(digestFastqs, L1)
    outl2 <- do.call(digestFastqs, L2)
    expect_error(summarizeExperiment(x = list(sample1 = outl1, sample2 = outl2),
                                     coldata = coldata, countType = "umis"))

})

test_that("summarizeExperiment works as expected when collapsing to WT", {
    ## Also test that errorStatistics is empty if there is no constant sequence
    Ldef1$collapseToWTForward <- TRUE
    Ldef2$collapseToWTForward <- TRUE
    Ldef1$elementsForward = "SUSV"
    Ldef1$elementsReverse = "SUSV"
    Ldef2$elementsForward = "SUSV"
    Ldef2$elementsReverse = "SUSV"
    
    out1 <- do.call(digestFastqs, Ldef1)
    out2 <- do.call(digestFastqs, Ldef2)
    
    se <- summarizeExperiment(x = list(sample1 = out1, sample2 = out2),
                              coldata = coldata, countType = "reads")
    expect_equal(nrow(se), length(union(out1$summaryTable$mutantName,
                                        out2$summaryTable$mutantName)))
    expect_equal(ncol(se), 2)
    expect_equal(sort(rownames(se)),
                 sort(union(out1$summaryTable$mutantName, out2$summaryTable$mutantName)))
    expect_equal(Matrix::colSums(SummarizedExperiment::assay(se, "counts")),
                 c(out1$filterSummary$nbrRetained, out2$filterSummary$nbrRetained),
                 ignore_attr = TRUE)
    expect_equal(Matrix::colSums(SummarizedExperiment::assay(se, "counts")),
                 c(sum(out1$summaryTable$nbrReads), sum(out2$summaryTable$nbrReads)),
                 ignore_attr = TRUE)
    
    for (cn in colnames(out1$filterSummary)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     out1$filterSummary[, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     out2$filterSummary[, cn])
    }
    
    for (cn in colnames(coldata)) {
        expect_equal(SummarizedExperiment::colData(se)["sample1", cn],
                     coldata[1, cn])
        expect_equal(SummarizedExperiment::colData(se)["sample2", cn],
                     coldata[2, cn])
    }
    
    expect_equal(S4Vectors::metadata(se)$parameters[["sample1"]], out1$parameters)
    expect_equal(S4Vectors::metadata(se)$parameters[["sample2"]], out2$parameters)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample1"]], out1$errorStatistics)
    expect_equal(S4Vectors::metadata(se)$errorStatistics[["sample2"]], out2$errorStatistics)
    expect_equal(sum(S4Vectors::metadata(se)$errorStatistics[["sample1"]][, c("nbrMatchForward", "nbrMismatchForward", "nbrMatchReverse", "nbrMismatchReverse")]), 0)
    expect_equal(sum(S4Vectors::metadata(se)$errorStatistics[["sample2"]][, c("nbrMatchForward", "nbrMismatchForward", "nbrMatchReverse", "nbrMismatchReverse")]), 0)
    
    expect_true(all(c("mutantName", "sequence", "nbrMutBases",
                      "minNbrMutBases", "maxNbrMutBases",
                      "nbrMutCodons", "minNbrMutCodons", "maxNbrMutCodons",
                      "nbrMutAAs", "minNbrMutAAs", "maxNbrMutAAs",
                      "sequenceAA", "mutantNameAA", "mutationTypes",
                      "varLengths") %in% 
                        colnames(SummarizedExperiment::rowData(se))))
    
    ## variable lengths
    expect_equal(SummarizedExperiment::rowData(se)$varLengths, 
                 rep("96_96", nrow(se)), ignore_attr = TRUE)
    
    expect_false(any(grepl("^,", SummarizedExperiment::rowData(se)$mutationTypes)))
    expect_false(any(grepl(",$", SummarizedExperiment::rowData(se)$mutationTypes)))
    expect_equal(SummarizedExperiment::rowData(se)$sequenceAA[3], 
                 mutscan:::translateString(SummarizedExperiment::rowData(se)$sequence[3]))
    expect_type(SummarizedExperiment::rowData(se)$nbrMutBases, "character")
    expect_type(SummarizedExperiment::rowData(se)$nbrMutCodons, "character")
    expect_type(SummarizedExperiment::rowData(se)$nbrMutAAs, "character")
    expect_equal(SummarizedExperiment::rowData(se)$nbrMutBases[[which(SummarizedExperiment::rowData(se)$mutantName == "f_r.0.WT")]], "0,1,2,3")
    expect_equal(SummarizedExperiment::rowData(se)$nbrMutCodons[[which(SummarizedExperiment::rowData(se)$mutantName == "f_r.0.WT")]], "0,1")
    expect_equal(SummarizedExperiment::rowData(se)$nbrMutAAs[[which(SummarizedExperiment::rowData(se)$mutantName == "f_r.0.WT")]], "0,1")
    
    expect_true(all(grepl("stop", SummarizedExperiment::rowData(se)$mutationTypes[grep("\\*", SummarizedExperiment::rowData(se)$mutantNameAA)])))
})

test_that("mergeValues works", {
    res <- mergeValues(c("A", "B", "C", "A", "D", "B"),
                       c("a,b", "b,c", "c", "b,c", "b,a", "d"))
    expect_s3_class(res, "data.frame")
    expect_named(res, c("mutantNameColl", "valueColl"))
    expect_equal(res$mutantNameColl, c("A", "B", "C", "D"))
    expect_equal(res$value, c("a,b,c", "b,c,d", "c", "a,b"))
    
    res <- mergeValues(c("B", "A", "C", "D"),
                       c("a,b", "c", "c", "d"))
    expect_s3_class(res, "data.frame")
    expect_named(res, c("mutantNameColl", "valueColl"))
    expect_equal(res$mutantNameColl, c("A", "B", "C", "D"))
    expect_equal(res$value, c("c", "a,b", "c", "d"))
    
    res <- mergeValues(c("A", "B", "C", "A", "D", "B"),
                       c("a:b", "b:c", "c", "b:c", "b:a", "d"),
                       delimiter = ":")
    expect_s3_class(res, "data.frame")
    expect_named(res, c("mutantNameColl", "valueColl"))
    expect_equal(res$mutantNameColl, c("A", "B", "C", "D"))
    expect_equal(res$value, c("a:b:c", "b:c:d", "c", "a:b"))
    
    res <- mergeValues(c("A", "B", "C", "A", "D", "B"),
                       c("a,b", "b:c", "c", "b:c", "b:a", "d"),
                       delimiter = ":")
    expect_s3_class(res, "data.frame")
    expect_named(res, c("mutantNameColl", "valueColl"))
    expect_equal(res$mutantNameColl, c("A", "B", "C", "D"))
    expect_equal(res$value, c("a,b:b:c", "b:c:d", "c", "a:b"))
})

#' Calculate fitness scores.
#' 
#' Using sequence counts before and after selection, calculate fitness scores 
#' as described by Diss and Lehner (2018).
#' 
#' @param se SummarizedExperiment object as returned by 
#'     \code{\link{summarizeExperiment}}.
#' @param pairingCol Name of column in \code{colData(se)} with 
#'     replicate/pairing information. Samples with the same value in this 
#'     column will be paired.
#' @param ODCols Name(s) of column(s) in \code{colData(se)} with OD values 
#'     (numeric), used to normalize for different numbers of cells.
#' @param comparison 3-element character vector of the form 
#'     \code{(column, numerator, denominator)}. \code{column} is the name of 
#'     the column in \code{colData(se)} with experimental conditions. 
#'     \code{numerator} and \code{denominator} define the comparison, 
#'     e.g. \code{c("cond", "output", "input")} will look in the 
#'     \code{"cond"} column and calculate fitness for the ratio of 
#'     \code{"output"} over \code{"input"} counts.
#' @param WTrows Vector of row names that will be used as the reference when
#'     calculating fitness scores. If more than one value is provided, the 
#'     average of the corresponding fitness scores is used as a reference. 
#'     If NULL, no division by WT scores will be done.
#' @param selAssay Assay to select from \code{se} for the analysis.
#'   
#' @return A numeric vector with fitness scores.
#' 
#' @author Michael Stadler and Charlotte Soneson
#' 
#' @references "The genetic landscape of a physical interaction."
#'     Diss G and Lehner B. Elife. 2018;7:e32472. doi: 10.7554/eLife.32472.
#' 
#' @importFrom methods is
#' @importFrom SummarizedExperiment colData assay
#' @importFrom Matrix colSums
#' 
#' @export
#' 
#' @examples 
#' se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
#'                           package = "mutscan"))
#' ## Check that the wildtype sequence is present in the data
#' stopifnot("f.0.WT" %in% rownames(se))
#' ## Calculate PPI scores as defined in Diss & Lehner (2018)
#' ppis <- calculateFitnessScore(
#'     se = se, pairingCol = "Replicate", 
#'     ODCols = c("OD1", "OD2"),
#'     comparison = c("Condition", "cis_output", "cis_input"),
#'     WTrows = "f.0.WT")
#' ## Matrix with PPI scores for each replicate
#' head(ppis)
#' 
calculateFitnessScore <- function(se, pairingCol, ODCols, comparison, WTrows,
                                  selAssay = "counts") {
    ## ------------------------------------------------------------------------
    ## pre-flight checks
    ## ------------------------------------------------------------------------
    .assertVector(x = se, type = "SummarizedExperiment")
    
    ## pairingCol is in colData(se)
    .assertScalar(x = pairingCol, type = "character", 
                  validValues = colnames(colData(se)))
    
    ## ODCols are all in colData(se) and contain numeric values
    .assertVector(x = ODCols, type = "character", rngLen = c(1, Inf),
                  validValues = colnames(colData(se)))
    for (odc in ODCols) {
        .assertVector(x = colData(se)[[odc]], 
                      type = "numeric")
    }
    
    ## comparison is length(3)-character with column and values in colData(se)
    .assertVector(x = comparison, type = "character", len = 3)
    .assertScalar(x = comparison[1], type = "character", 
                  validValues = colnames(colData(se)))
    .assertVector(
        x = comparison[2:3], type = "character", 
        validValues = colData(se)[[comparison[1]]])
    
    ## there is exactly one observation per pairing and condition
    if (any(table(colData(se)[colData(se)[, comparison[1]] %in% 
                              comparison[2:3], pairingCol],
                  colData(se)[colData(se)[, comparison[1]] %in% 
                              comparison[2:3], comparison[1]]) != 1)) {
        stop("There must be exactly one sample for every ", 
             "replicate-condition combination ",
             "defined by 'pairingCol' and 'comparison'")
    }
    
    ## WTrows exist in the SE
    .assertVector(x = WTrows, type = "character", validValues = rownames(se), 
                  allowNULL = TRUE)
    
    ## ------------------------------------------------------------------------
    ## subset se and reorder samples by shared replicates
    ## ------------------------------------------------------------------------
    se_numerator <- se[, colData(se)[, comparison[1]] == comparison[2]]
    se_denominator <- se[, colData(se)[, comparison[1]] == comparison[3]]
    
    shared_repl <- intersect(colData(se_numerator)[, pairingCol], 
                             colData(se_denominator)[, pairingCol])
    se_numerator <- se_numerator[, match(shared_repl, 
                                         colData(se_numerator)[, pairingCol])]
    se_denominator <- 
        se_denominator[, match(shared_repl, 
                               colData(se_denominator)[, pairingCol])]
    
    ## ------------------------------------------------------------------------
    ## calculate normalized counts (n_i)
    ## ------------------------------------------------------------------------
    norm_counts_numerator <- sweep(
        as.matrix(assay(se_numerator, selAssay)), MARGIN = 2, 
        STATS = apply(colData(se_numerator)[, ODCols, drop = FALSE], 
                      MARGIN = 1, prod) /
            colSums(assay(se_numerator, selAssay)), 
        FUN = "*")
    norm_counts_denominator <- sweep(
        as.matrix(assay(se_denominator, selAssay)), MARGIN = 2, 
        STATS = apply(colData(se_denominator)[, ODCols, drop = FALSE], 
                      MARGIN = 1, prod) /
            colSums(assay(se_denominator, selAssay)), 
        FUN = "*")
    n <- log2(norm_counts_numerator/norm_counts_denominator)
    n[!is.finite(n)] <- NA
    colnames(n) <- paste0(comparison[2], "_vs_", comparison[3],
                          "_repl", shared_repl)
    
    
    ## ------------------------------------------------------------------------
    ## calculate fitness = n_i / n_WT
    ## ------------------------------------------------------------------------
    if (is.null(WTrows)) {
        nWT <- rep(1, ncol(n))
    } else {
        nWT <- colMeans(n[WTrows, , drop = FALSE])
    }
    fitness <- sweep(n, MARGIN = 2, STATS = nWT, FUN = "/")
    return(fitness)
}

## usethis namespace: start
#' @useDynLib mutscan, .registration = TRUE
#' @importFrom Rcpp sourceCpp
## usethis namespace: end
NULL

# global variables definition for R CMD check
# (avoid NOTE about "no visible binding for global variable")
#' @importFrom utils globalVariables
utils::globalVariables(c("."))

test_that("parallel processing works (cis)", {
  fqc1 <- system.file("extdata/cisInput_1.fastq.gz", package = "mutscan")
  fqc2 <- system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqc1, fastqReverse = fqc2, 
    mergeForwardReverse = TRUE, 
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 1, greedyOverlap = TRUE, 
    revComplForward = FALSE, revComplReverse = TRUE,
    elementsForward = "SUCV", elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = c(wt1 = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                        wt2 = "CCTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                        wt3 = "ACTGATACACTCCAAGCGGAGACTGACCAACTAGAAGATGAGAAGTCTTCTTTGCAGACCGAGATTGCCAACGTGCTGAAGGAGAAGGAAAAACTA"),
    wildTypeReverse = "", 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAGTTCATCCTGGCAGC", 
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0, 
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = -1,
    nbrMutatedCodonsMaxReverse = -1,
    nbrMutatedBasesMaxForward = 2,
    nbrMutatedBasesMaxReverse = 2,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1, 
    maxReadLength = 1024
  )
  
  res <- do.call(digestFastqs, Ldef)
  
  ## Increase number of threads
  Ldef1 <- Ldef; Ldef1$nThreads <- 3
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))

  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
  
  ## Increase chunk size
  Ldef1 <- Ldef; Ldef1$nThreads <- 3; Ldef1$chunkSize <- 100
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))
  
  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
  
  ## With tree matching
  Ldef1 <- Ldef; Ldef1$useTreeWTmatch <- TRUE; Ldef1$nThreads <- 3; Ldef1$chunkSize <- 100
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))
  
  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
})

test_that("parallel processing works (trans)", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2, 
    mergeForwardReverse = FALSE, 
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0, 
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE, 
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = 0,
    constantMaxDistReverse = 0,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1, 
    maxReadLength = 1024
  )
  
  res <- do.call(digestFastqs, Ldef)
  
  ## Increase number of threads
  Ldef1 <- Ldef; Ldef1$nThreads <- 3
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))
  
  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
  
  ## Increase chunk size
  Ldef1 <- Ldef; Ldef1$nThreads <- 3; Ldef1$chunkSize <- 100
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))
  
  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
  
  ## With tree matching
  Ldef1 <- Ldef; Ldef1$useTreeWTmatch <- TRUE; Ldef1$nThreads <- 3; Ldef1$chunkSize <- 100
  ## Capture warnings from missing OpenMP/not enough threads
  a <- capture_warnings({
      res1 <- do.call(digestFastqs, Ldef1)
  })
  expect_equal(sum(grepl("OpenMP|available", a)), length(a))
  
  expect_equal(res$filterSummary$nbrTotal, res1$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res1$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res1$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res1$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res1$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res1$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res1$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res1$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res1$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res1$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res1$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res1$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res1$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res1$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res1$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res1$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res1$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res1$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res1$summaryTable$sequence)
  
})

test_that("digestFastqs works as expected for trans experiments", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCVV", elementsReverse = "SUCVV",
    elementLengthsForward = c(1, 10, 18, 80, 16),
    elementLengthsReverse = c(1, 8, 20, 16, 80),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)

  L <- Ldef; L$fastqForward <- c(fqt1, fqt1); L$fastqReverse <- c(fqt2, fqt2)
  res2 <- do.call(digestFastqs, L)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 192L + 95L + 68L + 37L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 279L)

  expect_equal(res$filterSummary * 2, res2$filterSummary)
  expect_equal(res$summaryTable$nbrReads * 2, res2$summaryTable$nbrReads)
  expect_equal(res$summaryTable$nbrUmis, res2$summaryTable$nbrUmis)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_equal(sum(res$summaryTable$nbrReads == 2), 2L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.13.GAG_r.0.WT", "f.26.TAG_r.8.AGG")))
  expect_equal(sort(res$summaryTable$mutantNameCodon[res$summaryTable$nbrReads == 2]),
               sort(c("f.13.GAG_r.0.WT", "f.26.TAG_r.8.AGG")))
  expect_equal(sort(res$summaryTable$mutantNameBase[res$summaryTable$nbrReads == 2]),
               sort(c("f.39.G_r.0.WT", "f.76.T_f.77.A_r.22.A_r.23.G")))
  expect_equal(sort(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("f:c.39T>G_r:c", "f:c.76_77delinsTA_r:c.22_23delinsAG")))
  expect_equal(sort(res$summaryTable$mutantNameAAHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("f:p.(Asp13Glu)_r:p", "f:p.(Leu26*)_r:p.(Val8Arg)")))
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))

  ## Check the number of reads with a given number of mismatches
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 1]), 4L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 2]), 14L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 3]), 52L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 4]), 83L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 5]), 87L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 6]), 39L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 1]), 15L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 2]), 264L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 0]), 1L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 1]), 37L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 2]), 241L)

  ## Similarly but for the number of unique sequences
  expect_equal(sum(res$summaryTable$nbrMutBases == 1), 3L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 2), 14L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 3), 52L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 4), 82L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 5), 87L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 6), 39L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 1), 14L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 2), 263L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 1), 36L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 2), 240L)

  ## check that variable segment lengths are reported correctly
  expect_true(all(res$summaryTable$varLengths == "80,16_16,80"))

  ## Check that mutant naming worked (compare to manual matching)
  example_seq <- paste0("ACTGATACAACCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTG",
                        "CAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA_ATCGCCCGGCT",
                        "GGAGGAAAAAGTGGGCACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGC",
                        "CAACATGCTCAGGGAACAGGTGGCACAGCTT")
  expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
               "f.4.ACC_r.9.GGC")
  expect_equal(res$summaryTable$mutantNameAA[res$summaryTable$sequence == example_seq],
               "f.4.T_r.9.G")
  expect_equal(res$summaryTable$mutantNameCodon[res$summaryTable$sequence == example_seq],
               "f.4.ACC_r.9.GGC")
  expect_equal(res$summaryTable$mutantNameBase[res$summaryTable$sequence == example_seq],
               "f.10.A_f.11.C_r.25.G_r.26.G_r.27.C")
  expect_equal(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$sequence == example_seq],
               "f:c.10_11delinsAC_r:c.25_27delinsGGC")
  expect_equal(res$summaryTable$mutantNameAAHGVS[res$summaryTable$sequence == example_seq],
               "f:p.(Leu4Thr)_r:p.(Lys9Gly)")
  
  expect_equal(sum(grepl("WT", res$summaryTable$mutantNameAA) & res$summaryTable$nbrMutBases > 0), 37L)
  expect_true(all(grepl("silent", res$summaryTable$mutationTypes[grepl("WT", res$summaryTable$mutantNameAA) & grepl("f\\.[1-9]", res$summaryTable$mutantName) & grepl("r\\.[1-9]", res$summaryTable$mutantName)])))
  expect_true(all(grepl("nonsynonymous|stop", res$summaryTable$mutationTypes[grepl("\\.[1-9][0-9]*\\.", res$summaryTable$mutantNameAA)])))
  expect_true(all(grepl("stop", res$summaryTable$mutationTypes[grepl("TAG|TAA|TGA", res$summaryTable$mutantName)])))

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 160L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 52L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 302L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 472L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 4020L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 11L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 4L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 1L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 204L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 14L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 474L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 462L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 4405L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 17L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 27], 3L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 1L)

})

## ----------------------------------------------------------------------------
## Constant seq mismatch filtering
## ----------------------------------------------------------------------------
test_that("digestFastqs works as expected for trans experiments, filter based on constant seq mismatches", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = 0,
    constantMaxDistReverse = 0,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 28L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 251L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  expect_equal(sum(res$summaryTable$nbrReads == 2), 1L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.26.TAG_r.8.AGG")))

  ## Check the number of reads with a given number of mismatches
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 1]), 3L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 2]), 14L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 3]), 46L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 4]), 76L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 5]), 75L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 6]), 37L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 1]), 14L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 2]), 237L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 0]), 1L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 1]), 35L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 2]), 215L)
  
  ## Similarly but for the number of unique sequences
  expect_equal(sum(res$summaryTable$nbrMutBases == 1), 3L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 2), 14L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 3), 46L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 4), 75L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 5), 75L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 6), 37L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 1), 14L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 2), 236L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 1), 35L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 2), 214L)
  
  
  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMismatchForward, rep(0L, nrow(res$errorStatistics)))
  expect_equal(res$errorStatistics$nbrMismatchReverse, rep(0L, nrow(res$errorStatistics)))

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 126L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 43L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 248L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 398L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 3703L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 137L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 11L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 380L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 407L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 4085L)

  ## --------------------------------------------------------------------------
  ## Allow 1 mismatch
  ## --------------------------------------------------------------------------
  L <- Ldef; L$constantMaxDistForward <- 1L; L$constantMaxDistReverse <- 1L
  res <- do.call(digestFastqs, L)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 4L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 275L)

  for (nm in setdiff(names(L), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], L[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(L[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  expect_equal(sum(res$summaryTable$nbrReads == 2), 2L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.13.GAG_r.0.WT", "f.26.TAG_r.8.AGG")))

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(L$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(L$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 147L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 49L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 288L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 459L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 3994L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 9L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 3L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 1L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 178L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 12L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 450L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 454L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 4394L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 10L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 27], 1L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 1L)

  ## --------------------------------------------------------------------------
  ## Provide 2 constant sequences
  ## --------------------------------------------------------------------------
  L <- Ldef; L$constantForward <- c("AACCGGAGGAGGGAGCTG", "AACCGGCGGAGGGAGCTG")
  res <- do.call(digestFastqs, L)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 26L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 253L)

  for (nm in setdiff(names(L), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], L[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(L[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  expect_equal(sum(res$summaryTable$nbrReads == 2), 1L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.26.TAG_r.8.AGG")))

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(L$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(L$constantReverse[1]) * res$filterSummary$nbrRetained)

  ## --------------------------------------------------------------------------
  ## Split constant sequence in two and let the function concatenate them
  ## --------------------------------------------------------------------------
  L <- Ldef; L$elementsForward <- "SUCCV"; L$elementsReverse <- "SUCCV"
  L$elementLengthsForward <- c(1, 10, 10, 8, 96); L$elementLengthsReverse <- c(1, 8, 9, 11, 96)

  res <- do.call(digestFastqs, L)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 28L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 251L)

  for (nm in setdiff(names(L), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], L[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(L[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  expect_equal(sum(res$summaryTable$nbrReads == 2), 1L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.26.TAG_r.8.AGG")))

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(L$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(L$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMismatchForward, rep(0L, nrow(res$errorStatistics)))
  expect_equal(res$errorStatistics$nbrMismatchReverse, rep(0L, nrow(res$errorStatistics)))

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 126L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 43L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 248L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 398L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 3703L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 137L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 11L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 380L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 407L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 4085L)

})

## ----------------------------------------------------------------------------
## Read composition specification
## ----------------------------------------------------------------------------
test_that("digestFastqs works as expected for trans experiments - no UMI specified", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SSCV", elementsReverse = "SSCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 500, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 279L)

  expect_equal(res$summaryTable$nbrUmis, rep(0, nrow(res$summaryTable)))
})

test_that("digestFastqs works as expected for trans experiments, when variable sequence length is not given", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, -1),
    elementLengthsReverse = c(1, 8, 20, -1),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 700, 
    maxReadLength = 125
  )

  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 279L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_equal(sum(res$summaryTable$nbrReads == 2), 2L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.13.GAG_r.0.WT", "f.26.TAG_r.8.AGG")))
  expect_equal(sort(res$summaryTable$mutantNameCodon[res$summaryTable$nbrReads == 2]),
               sort(c("f.13.GAG_r.0.WT", "f.26.TAG_r.8.AGG")))
  expect_equal(sort(res$summaryTable$mutantNameBase[res$summaryTable$nbrReads == 2]),
               sort(c("f.39.G_r.0.WT", "f.76.T_f.77.A_r.22.A_r.23.G")))
  expect_equal(sort(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("f:c.39T>G_r:c", "f:c.76_77delinsTA_r:c.22_23delinsAG")))
  expect_equal(sort(res$summaryTable$mutantNameAAHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("f:p.(Asp13Glu)_r:p", "f:p.(Leu26*)_r:p.(Val8Arg)")))
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))

  ## Check that mutant naming worked (compare to manual matching)
  example_seq <- paste0("ACTGATACAACCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTG",
                        "CAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA_ATCGCCCGGCT",
                        "GGAGGAAAAAGTGGGCACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGC",
                        "CAACATGCTCAGGGAACAGGTGGCACAGCTT")
  expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
               "f.4.ACC_r.9.GGC")
  expect_equal(res$summaryTable$mutantNameAA[res$summaryTable$sequence == example_seq],
               "f.4.T_r.9.G")
  expect_equal(res$summaryTable$mutantNameCodon[res$summaryTable$sequence == example_seq],
               "f.4.ACC_r.9.GGC")
  expect_equal(res$summaryTable$mutantNameBase[res$summaryTable$sequence == example_seq],
               "f.10.A_f.11.C_r.25.G_r.26.G_r.27.C")
  expect_equal(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$sequence == example_seq],
               "f:c.10_11delinsAC_r:c.25_27delinsGGC")
  expect_equal(res$summaryTable$mutantNameAAHGVS[res$summaryTable$sequence == example_seq],
               "f:p.(Leu4Thr)_r:p.(Lys9Gly)")

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 160L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 52L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 302L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 472L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 4020L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 11L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 4L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 1L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 204L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 14L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 474L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 462L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 4405L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 17L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 27], 3L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 1L)

})

## ----------------------------------------------------------------------------
## Multiple wildtype sequences
## ----------------------------------------------------------------------------
test_that("digestFastqs works as expected for trans experiments when multiple reference sequences are provided", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = c(forward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",  ## this is the right one
                        wrong = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"),
    wildTypeReverse = c(wrong = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                        reverse1 = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), ## this is the right one
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 30.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNA",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 25.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = "=",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = TRUE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  expect_output(res <- do.call(digestFastqs, Ldef))

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 88L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 189L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 149L + 5L + 34L + 25L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 1L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 193L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "nThreads",
                                    "forbiddenMutatedCodonsReverse", "verbose",
                                    "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_equal(sum(res$summaryTable$nbrReads == 2), 1L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("forward=26=TAG_reverse1=8=AGG")))
  expect_equal(sort(res$summaryTable$mutantNameCodon[res$summaryTable$nbrReads == 2]),
               sort(c("forward=26=TAG_reverse1=8=AGG")))
  expect_equal(sort(res$summaryTable$mutantNameBase[res$summaryTable$nbrReads == 2]),
               sort(c("forward=76=T_forward=77=A_reverse1=22=A_reverse1=23=G")))
  expect_equal(sort(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("forward:c.76_77delinsTA_reverse1:c.22_23delinsAG")))
  expect_equal(sort(res$summaryTable$mutantNameAAHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("forward:p.(Leu26*)_reverse1:p.(Val8Arg)")))
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))

  ## Check that mutant naming worked (compare to manual matching)
  example_seq <- paste0("ACTGATACAACCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTG",
                        "CAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA_ATCGCCCGGCT",
                        "GGAGGAAAAAGTGGGCACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGC",
                        "CAACATGCTCAGGGAACAGGTGGCACAGCTT")
  expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
               "forward=4=ACC_reverse1=9=GGC")
  expect_equal(res$summaryTable$mutantNameAA[res$summaryTable$sequence == example_seq],
               "forward=4=T_reverse1=9=G")
  expect_equal(res$summaryTable$mutantNameCodon[res$summaryTable$sequence == example_seq],
               "forward=4=ACC_reverse1=9=GGC")
  expect_equal(res$summaryTable$mutantNameBase[res$summaryTable$sequence == example_seq],
               "forward=10=A_forward=11=C_reverse1=25=G_reverse1=26=G_reverse1=27=C")
  expect_equal(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$sequence == example_seq],
               "forward:c.10_11delinsAC_reverse1:c.25_27delinsGGC")
  expect_equal(res$summaryTable$mutantNameAAHGVS[res$summaryTable$sequence == example_seq],
               "forward:p.(Leu4Thr)_reverse1:p.(Lys9Gly)")

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 96L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 26L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 179L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 302L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 2865L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 2L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 3L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 1L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 116L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 22], 5L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 308L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 322L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 3101L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 6L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 27], 1L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 1L)

  ## Test that we get the same results with useTreeWTmatch = TRUE
  Ldef$useTreeWTmatch <- TRUE
  expect_output(res2 <- do.call(digestFastqs, Ldef))

  expect_equal(res$filterSummary$nbrTotal, res2$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res2$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res2$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res2$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res2$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res2$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res2$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res2$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res2$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res2$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res2$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res2$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res2$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res2$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res2$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res2$filterSummary$nbrRetained)

  expect_equal(res$summaryTable$nbrReads, res2$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res2$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res2$summaryTable$sequence)

})

## ----------------------------------------------------------------------------
## Only forward read
## ----------------------------------------------------------------------------
test_that("digestFastqs works as expected for experiments with only forward read", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = NULL,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = numeric(0),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "",
    avePhredMinForward = 20.0, avePhredMinReverse = 30.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNA",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 25.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 300, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)

  L <- Ldef; L$fastqForward <- c(fqt1, fqt1); L$fastqReverse <- NULL
  res2 <- do.call(digestFastqs, L)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 297L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 211L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 212L + 7L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 1L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
    expect_equal(res$filterSummary$nbrRetained, 272L)
    
    expect_equal(res$filterSummary * 2, res2$filterSummary)
    expect_equal(res$summaryTable$nbrReads * 2, res2$summaryTable$nbrReads)
    expect_equal(res$summaryTable$nbrUmis, res2$summaryTable$nbrUmis)
    
    for (nm in setdiff(names(Ldef), c("fastqReverse", "forbiddenMutatedCodonsForward",
                                      "forbiddenMutatedCodonsReverse", "verbose",
                                      "elementLengthsReverse", "fastqForward"))) {
        expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
    }
    expect_equal(res$parameters[["fastqReverse"]], "", ignore_attr = TRUE)
    expect_equal(res$parameters[["fastqForward"]], 
                 normalizePath(Ldef[["fastqForward"]], mustWork = FALSE), 
                 ignore_attr = TRUE)

    
    expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
    expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
    expect_equal(sum(res$summaryTable$nbrReads == 2), 29L)
    
    ## Check the number of reads with a given number of mismatches
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 0]), 10L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 1]), 47L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 2]), 123L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 3]), 92L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 0]), 10L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 1]), 262L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 0]), 21L)
    expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 1]), 251L)
    
    ## Similarly but for the number of unique sequences
    expect_equal(sum(res$summaryTable$nbrMutBases == 0), 1L)
    expect_equal(sum(res$summaryTable$nbrMutBases == 1), 43L)
    expect_equal(sum(res$summaryTable$nbrMutBases == 2), 102L)
    expect_equal(sum(res$summaryTable$nbrMutBases == 3), 78L)
    expect_equal(sum(res$summaryTable$nbrMutCodons == 0), 1L)
    expect_equal(sum(res$summaryTable$nbrMutCodons == 1), 223L)
    expect_equal(sum(res$summaryTable$nbrMutAAs == 0), 10L)
    expect_equal(sum(res$summaryTable$nbrMutAAs == 1), 214L)
    
    
    ## Check that mutant naming worked (compare to manual matching)
    example_seq <- paste0("ACTGATACAACCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTG",
                          "CAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")
    expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
                 "f.4.ACC")
    
    expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
                 nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
    expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
                 nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)
    
    expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 187L)
    expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 47L)
    expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 314L)
    expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 469L)
    expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 3868L)
    
    expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 6L)
    expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 4L)
    expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 1L)
    
})

## ----------------------------------------------------------------------------
## Only forward read, provide a too short WT sequence
## ----------------------------------------------------------------------------
test_that("digestFastqs works as expected when the WT sequence is too short", {
    fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
    fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
    ## default arguments
    Ldef <- list(
        fastqForward = fqt1, fastqReverse = NULL,
        mergeForwardReverse = FALSE,
        minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
        revComplForward = FALSE, revComplReverse = FALSE,
        elementsForward = "SUCV", elementsReverse = "",
        elementLengthsForward = c(1, 10, 18, 96),
        elementLengthsReverse = numeric(0),
        adapterForward = "GGAAGAGCACACGTC",
        adapterReverse = "",
        primerForward = "",
        primerReverse = "",
        wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACT",
        wildTypeReverse = "",
        constantForward = "AACCGGAGGAGGGAGCTG",
        constantReverse = "",
        avePhredMinForward = 20.0, avePhredMinReverse = 30.0,
        variableNMaxForward = 0, variableNMaxReverse = 0,
        umiNMax = 0,
        nbrMutatedCodonsMaxForward = 1,
        nbrMutatedCodonsMaxReverse = 1,
        nbrMutatedBasesMaxForward = -1,
        nbrMutatedBasesMaxReverse = -1,
        forbiddenMutatedCodonsForward = "NNA",
        forbiddenMutatedCodonsReverse = "NNW",
        useTreeWTmatch = FALSE,
        mutatedPhredMinForward = 25.0, mutatedPhredMinReverse = 0.0,
        mutNameDelimiter = ".",
        constantMaxDistForward = -1,
        constantMaxDistReverse = -1,
        umiCollapseMaxDist = 0,
        filteredReadsFastqForward = "",
        filteredReadsFastqReverse = "",
        maxNReads = -1, verbose = FALSE,
        nThreads = 1, chunkSize = 300, 
        maxReadLength = 1024
    )
    
    res <- do.call(digestFastqs, Ldef)
    
    expect_equal(res$filterSummary$nbrTotal, 1000L)
    expect_equal(res$filterSummary$f1_nbrAdapter, 297L)
    expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
    expect_equal(res$filterSummary$f3_nbrReadWrongLength, 284L) ## generated as 1000-297-419 (all var seqs are too long)
    expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
    expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
    expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
    expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
    expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
    expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 419L) ## this value was generated and checked manually
    expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
    expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
    expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
    expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
    expect_equal(res$filterSummary$nbrRetained, 0L)
    
    expect_equal(nrow(res$summaryTable), 0L)
    
    for (nm in setdiff(names(Ldef), c("fastqReverse", "forbiddenMutatedCodonsForward",
                                      "forbiddenMutatedCodonsReverse", "verbose",
                                      "elementLengthsReverse", "fastqForward"))) {
        expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
    }
    expect_equal(res$parameters[["fastqReverse"]], "", ignore_attr = TRUE)
    expect_equal(res$parameters[["fastqForward"]], 
                 normalizePath(Ldef[["fastqForward"]], mustWork = FALSE), 
                 ignore_attr = TRUE)
    
})

## ----------------------------------------------------------------------------
## Save filtered reads to fastq files
## ----------------------------------------------------------------------------
test_that("writing filtered reads to file works", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")

  outfile1 <- file.path(tempdir(), "mutscan_test_1.fastq.gz")
  outfile2 <- gsub("_1.fastq", "_2.fastq", outfile1)

  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = outfile1,
    filteredReadsFastqReverse = outfile2,
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )

  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 287L + 105L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 279L)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse",
                                    "filteredReadsFastqForward", 
                                    "filteredReadsFastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse", "filteredReadsFastqForward",
               "filteredReadsFastqReverse")) {
    expect_equal(normalizePath(res$parameters[[nm]], mustWork = FALSE), 
                 normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  

  # Biostrings 2.63.1 on arm64 produces the following warning:
  # Warning message:
  # In XStringSet("DNA", x, start = start, end = end, width = width,  :
  #   metadata columns on input DNAStringSet object were dropped
  suppressWarnings({
      out1 <- Biostrings::readQualityScaledDNAStringSet(outfile1)
      out2 <- Biostrings::readQualityScaledDNAStringSet(outfile2)
  })

  expect_equal(length(out1), length(out2))
  expect_equal(names(out1), gsub(" 2", " 1", names(out2)))
  expect_equal(length(out1), res$filterSummary$nbrTotal - res$filterSummary$nbrRetained)
  expect_equal(length(grep("adapter", names(out1))), res$filterSummary$f1_nbrAdapter)
  expect_equal(length(grep("noPrimer", names(out1))), res$filterSummary$f2_nbrNoPrimer)
  expect_equal(length(grep("readWrongLength", names(out1))), res$filterSummary$f3_nbrReadWrongLength)
  expect_equal(length(grep("noValidOverlap", names(out1))), res$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(length(grep("avgVarQualTooLow", names(out1))), res$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(length(grep("tooManyNinVar", names(out1))), res$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(length(grep("tooManyNinUMI", names(out1))), res$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(length(grep("tooManyBestWTHits", names(out1))), res$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(length(grep("mutQualTooLow", names(out1))),
               res$filterSummary$f9_nbrMutQualTooLow + res$filterSummary$f10a_nbrTooManyMutCodons +
                 res$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(length(grep("tooManyMutConstant", names(out1))), res$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(length(grep("tooManyBestConstantHits", names(out1))), res$filterSummary$f13_nbrTooManyBestConstantHits)

  ## Test that we get the same results with useTreeWTmatch = TRUE
  Ldef$useTreeWTmatch <- TRUE
  res2 <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, res2$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res2$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res2$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res2$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res2$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res2$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res2$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res2$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res2$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res2$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res2$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res2$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res2$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res2$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res2$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res2$filterSummary$nbrRetained)

  expect_equal(res$summaryTable$nbrReads, res2$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res2$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res2$summaryTable$sequence)

})

test_that("digestFastqs works as expected for trans experiments, if collapsing to WT", {
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2,
    mergeForwardReverse = FALSE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCVV", elementsReverse = "SUCVV",
    elementLengthsForward = c(1, 10, 18, 80, 16),
    elementLengthsReverse = c(1, 8, 20, 16, 80),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    collapseToWTForward = TRUE,
    collapseToWTReverse = TRUE, 
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )
  
  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 192L + 95L + 68L + 37L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 279L)
  
  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(res$summaryTable$nbrReads, res$summaryTable$maxNbrReads)
  expect_equal(nrow(res$summaryTable), 1L)
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))
  
  ## Check the number of reads with a given number of mismatches
  expect_equal(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == "1,2,3,4,5,6"], 279L)
  expect_equal(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == "1,2"], 279L)
  expect_equal(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == "0,1,2"], 279L)
  expect_equal(res$summaryTable$mutationTypes, "nonsynonymous,silent,stop")
  expect_equal(res$summaryTable$mutantNameAA, "f_r")
  expect_equal(res$summaryTable$mutantNameBase, "f_r")
  expect_equal(res$summaryTable$mutantNameCodon, "f_r")
  expect_equal(res$summaryTable$mutantNameBaseHGVS, "f:c_r:c")
  expect_equal(res$summaryTable$mutantNameAAHGVS, "f:p_r:p")
  
  ## check that variable segment lengths are reported correctly
  expect_equal(res$summaryTable$varLengths, "80,16_16,80")
})

test_that("reads are filtered out if there are too many constant hits", {
    fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
    fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
    ## default arguments
    Ldef <- list(
        fastqForward = fqt1, fastqReverse = fqt2,
        mergeForwardReverse = FALSE,
        minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
        revComplForward = FALSE, revComplReverse = FALSE,
        elementsForward = "SUCVV", elementsReverse = "SUCVV",
        elementLengthsForward = c(1, 10, 18, 80, 16),
        elementLengthsReverse = c(1, 8, 20, 16, 80),
        adapterForward = "GGAAGAGCACACGTC",
        adapterReverse = "GGAAGAGCGTCGTGT",
        primerForward = "",
        primerReverse = "",
        wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
        wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
        constantForward = "AACCGGAGGAGGGAGCTG",
        constantReverse = "GAAAAAGGAAGCTGGAGAGA",
        avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
        variableNMaxForward = 0, variableNMaxReverse = 0,
        umiNMax = 0,
        nbrMutatedCodonsMaxForward = 1,
        nbrMutatedCodonsMaxReverse = 1,
        nbrMutatedBasesMaxForward = -1,
        nbrMutatedBasesMaxReverse = -1,
        forbiddenMutatedCodonsForward = "NNW",
        forbiddenMutatedCodonsReverse = "NNW",
        useTreeWTmatch = FALSE,
        collapseToWTForward = TRUE,
        collapseToWTReverse = TRUE, 
        mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
        mutNameDelimiter = ".",
        constantMaxDistForward = -1,
        constantMaxDistReverse = -1,
        umiCollapseMaxDist = 0,
        filteredReadsFastqForward = "",
        filteredReadsFastqReverse = "",
        maxNReads = -1, verbose = FALSE,
        nThreads = 1, chunkSize = 1000, 
        maxReadLength = 1024
    )
    
    ## Forward
    Ldef1 <- Ldef; Ldef1$constantForward <- c(c1 = "AACCGGAGGAGGGAGCTA", c2 = "AACCGGAGGAGGGAGCTC")
    res1 <- do.call(digestFastqs, Ldef1)
    
    expect_equal(res1$filterSummary$nbrTotal, 1000L)
    expect_equal(res1$filterSummary$f1_nbrAdapter, 314L)
    expect_equal(res1$filterSummary$f2_nbrNoPrimer, 0L)
    expect_equal(res1$filterSummary$f3_nbrReadWrongLength, 0L)
    expect_equal(res1$filterSummary$f4_nbrNoValidOverlap, 0L)
    expect_equal(res1$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
    expect_equal(res1$filterSummary$f6_nbrTooManyNinVar, 0L)
    expect_equal(res1$filterSummary$f7_nbrTooManyNinUMI, 0L)
    expect_equal(res1$filterSummary$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(res1$filterSummary$f9_nbrMutQualTooLow, 0L)
    expect_equal(res1$filterSummary$f10a_nbrTooManyMutCodons, 192L + 95L + 68L + 37L)
    expect_equal(res1$filterSummary$f10b_nbrTooManyMutBases, 0L)
    expect_equal(res1$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
    expect_equal(res1$filterSummary$f12_nbrTooManyMutConstant, 0L)
    expect_equal(res1$filterSummary$f13_nbrTooManyBestConstantHits, 279L)
    expect_equal(res1$filterSummary$nbrRetained, 0L)
    
    ## Reverse
    Ldef1 <- Ldef; Ldef1$constantReverse <- c(c1 = "GAAAAAGGAAGCTGGAGAGG", c2 = "GAAAAAGGAAGCTGGAGAGT")
    res1 <- do.call(digestFastqs, Ldef1)
    
    expect_equal(res1$filterSummary$nbrTotal, 1000L)
    expect_equal(res1$filterSummary$f1_nbrAdapter, 314L)
    expect_equal(res1$filterSummary$f2_nbrNoPrimer, 0L)
    expect_equal(res1$filterSummary$f3_nbrReadWrongLength, 0L)
    expect_equal(res1$filterSummary$f4_nbrNoValidOverlap, 0L)
    expect_equal(res1$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
    expect_equal(res1$filterSummary$f6_nbrTooManyNinVar, 0L)
    expect_equal(res1$filterSummary$f7_nbrTooManyNinUMI, 0L)
    expect_equal(res1$filterSummary$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(res1$filterSummary$f9_nbrMutQualTooLow, 0L)
    expect_equal(res1$filterSummary$f10a_nbrTooManyMutCodons, 192L + 95L + 68L + 37L)
    expect_equal(res1$filterSummary$f10b_nbrTooManyMutBases, 0L)
    expect_equal(res1$filterSummary$f11_nbrForbiddenCodons, 6L + 2L)
    expect_equal(res1$filterSummary$f12_nbrTooManyMutConstant, 0L)
    expect_equal(res1$filterSummary$f13_nbrTooManyBestConstantHits, 279L)
    expect_equal(res1$filterSummary$nbrRetained, 0L)
})

test_that("digestFastqs runs with revComplForward = TRUE", {
    fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
    fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
    ## default arguments
    Ldef <- list(
        fastqForward = fqt1, fastqReverse = fqt2,
        mergeForwardReverse = FALSE,
        minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE,
        revComplForward = TRUE, revComplReverse = FALSE,
        elementsForward = "SUCVV", elementsReverse = "SUCVV",
        elementLengthsForward = c(1, 10, 18, 80, 16),
        elementLengthsReverse = c(1, 8, 20, 16, 80),
        adapterForward = "GGAAGAGCACACGTC",
        adapterReverse = "GGAAGAGCGTCGTGT",
        primerForward = "",
        primerReverse = "",
        wildTypeForward = "TAGTTTTTCCTTCTCCTTCAGCAGGTTGGCAATCTCGGTCTGCAAAGCAGACTTCTCATCTTCTAGTTGGTCTGTCTCCGCTTGGAGTGTATCAGT",
        wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT",
        constantForward = "CAGCTCCCTCCTCCGGTT",
        constantReverse = "GAAAAAGGAAGCTGGAGAGA",
        avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
        variableNMaxForward = 0, variableNMaxReverse = 0,
        umiNMax = 0,
        nbrMutatedCodonsMaxForward = 1,
        nbrMutatedCodonsMaxReverse = 1,
        nbrMutatedBasesMaxForward = -1,
        nbrMutatedBasesMaxReverse = -1,
        forbiddenMutatedCodonsForward = "",
        forbiddenMutatedCodonsReverse = "",
        useTreeWTmatch = FALSE,
        collapseToWTForward = TRUE,
        collapseToWTReverse = TRUE, 
        mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
        mutNameDelimiter = ".",
        constantMaxDistForward = -1,
        constantMaxDistReverse = -1,
        umiCollapseMaxDist = 0,
        filteredReadsFastqForward = "",
        filteredReadsFastqReverse = "",
        maxNReads = -1, verbose = FALSE,
        nThreads = 1, chunkSize = 1000, 
        maxReadLength = 1024
    )
    
    res <- do.call(digestFastqs, Ldef)

    expect_equal(res$filterSummary$nbrTotal, 1000L)
    expect_equal(res$filterSummary$f1_nbrAdapter, 314L)
    expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
    expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
    expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
    expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 7L)
    expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 0L)
    expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
    expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
    expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
    expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 394L)
    expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
    expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 0L)
    expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
    expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
    expect_equal(res$filterSummary$nbrRetained, 285L)
})

test_that("digestFastqs fails with incorrect arguments", {
  ## example "trans" fastq files 
  fqt1 <- system.file("extdata/transInput_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/transInput_2.fastq.gz", package = "mutscan")
  
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2, 
    mergeForwardReverse = FALSE, 
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SUCV", elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = c(""),
    primerReverse = c(""),
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0, 
    variableNMaxForward = 0, variableNMaxReverse = 0, 
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = TRUE,
    collapseToWTForward = FALSE,
    collapseToWTReverse = FALSE,
    mutatedPhredMinForward = 0.0, 
    mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )
  
  ## Incorrect specification of merging/rev complementing arguments
  for (var in c("mergeForwardReverse", "revComplForward", "revComplReverse")) {
    L <- Ldef; L[[var]] <- "str"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- ""
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- c(TRUE, TRUE)
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- c(TRUE, FALSE)
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- 1
    expect_error(do.call(digestFastqs, L))
  }
  
  ## Nonexistent fastqForward/fastqReverse file
  L <- Ldef; L$fastqForward <- "nonexistent.fastq.gz"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqForward <- ""
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqForward <- NULL
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqReverse <- "nonexistent.fastq.gz"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqReverse <- ""
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqForward <- c(fqt1, fqt2)
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L$fastqReverse <- c(fqt1, fqt2)
  expect_error(do.call(digestFastqs, L))
  
  ## No reverse reads if mergeForwardReverse is TRUE
  L <- Ldef; L[["fastqReverse"]] <- NULL; L[["mergeForwardReverse"]] <- TRUE
  expect_error(do.call(digestFastqs, L))
  
  ## Wrong type of numeric argument
  for (var in c("avePhredMinForward", "avePhredMinReverse", "variableNMaxForward",
                "variableNMaxReverse", "umiNMax", 
                "nbrMutatedCodonsMaxForward", "nbrMutatedCodonsMaxReverse", 
                "nbrMutatedBasesMaxForward", "nbrMutatedBasesMaxReverse", 
                "mutatedPhredMinForward", "mutatedPhredMinReverse",
                "umiCollapseMaxDist",
                "constantMaxDistForward", "constantMaxDistReverse", "maxNReads",
                "nThreads", "chunkSize", "maxReadLength")) {
    L <- Ldef; L[[var]] <- "str"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- c(1, 2)
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- -2
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- TRUE
    expect_error(do.call(digestFastqs, L))
    if (!(var %in% c("nbrMutatedCodonsMaxForward", "nbrMutatedCodonsMaxReverse", 
                     "nbrMutatedBasesMaxForward", "nbrMutatedBasesMaxReverse",
                     "constantMaxDistForward", "constantMaxDistReverse",
                     "maxNReads", "maxReadLength"))) {
      L <- Ldef; L[[var]] <- -1
      expect_error(do.call(digestFastqs, L))
    }
    if (var %in% c("nThreads", "chunkSize", "maxReadLength")) {
      L <- Ldef; L[[var]] <- 0
      expect_error(do.call(digestFastqs, L))
    }
  }
  
  ## Both or none of max number of codons and bases specified
  ## Note that this should only give an error if a wildtype sequence is specified (which it is here)
  L <- Ldef; L[["nbrMutatedCodonsMaxForward"]] <- L[["nbrMutatedBasesMaxForward"]] <- 1
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["nbrMutatedCodonsMaxReverse"]] <- L[["nbrMutatedBasesMaxReverse"]] <- 1
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["nbrMutatedCodonsMaxForward"]] <- L[["nbrMutatedBasesMaxForward"]] <- -1
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["nbrMutatedCodonsMaxReverse"]] <- L[["nbrMutatedBasesMaxReverse"]] <- -1
  expect_error(do.call(digestFastqs, L))
  
  
  for (var in c("elementLengthsForward", "elementLengthsReverse")) {
    L <- Ldef; L[[var]] <- as.character(L[[var]])
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- as.logical(L[[var]])
    expect_error(do.call(digestFastqs, L))
  }
  
  for (var in c("elementsForward", "elementsReverse")) {
    L <- Ldef; L[[var]] <- "ABC"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- "UPPV"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- c("UPV", "CUS")
    expect_error(do.call(digestFastqs, L))
  }
  
  for (var in c("Forward", "Reverse")) {
    L <- Ldef; L[[paste0("elements", var)]] <- "CUV"
    L[[paste0("elementLengths", var)]] <- c(-1, -1, 4)
    expect_error(do.call(digestFastqs, L))
    L[[paste0("elementLengths", var)]] <- c(1, 4)
    expect_error(do.call(digestFastqs, L))
    
    L <- Ldef; L[[paste0("elementLengths", var)]][2] <- -2
    expect_error(do.call(digestFastqs, L))
    
    L <- Ldef; L[[paste0("elements", var)]] <- ""
    L[[paste0("elementLengths", var)]] <- c()
    expect_error(do.call(digestFastqs, L))
    
    L <- Ldef; L[[paste0("elements", var)]] <- "CUVPUV"
    L[[paste0("elementLengths", var)]] <- c(-1, -1, 4, 4, -1, 2)
    expect_error(do.call(digestFastqs, L))
    L[[paste0("elementLengths", var)]] <- c(-1, 1, 4, 4, -1, -1)
    expect_error(do.call(digestFastqs, L))
    L[[paste0("elementLengths", var)]] <- c(1, 4)
    expect_error(do.call(digestFastqs, L))
  }
  
  ## Invalid sequences
  for (var in c("adapterForward", "adapterReverse", "wildTypeForward",
                "wildTypeReverse", "constantForward", "constantReverse")) {
    if (!(var %in% c("constantForward", "constantReverse"))) {
      L <- Ldef; L[[var]] <- c("ACGT", "ACGT")
      expect_error(do.call(digestFastqs, L))
    }
    L <- Ldef; L[[var]] <- 1
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- "EF"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- paste0(L[[var]], " ")
    expect_error(do.call(digestFastqs, L))
  }
  
  for (var in c("primerForward", "primerReverse")) {
    L <- Ldef; L[[var]] <- 1
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- "EF"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- c("ACGT", "EF")
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[var]] <- "ACGT "
    expect_error(do.call(digestFastqs, L))
  }
  
  ## Wild type sequence not in named vector (or unnamed string)
  L <- Ldef
  L$wildTypeForward <- c("ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA", 
                         "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT")
  expect_error(do.call(digestFastqs, L))
  
  ## Duplicated wild type sequences
  L <- Ldef
  L$wildTypeForward <- c(wt1 = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                         wt2 = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")
  expect_error(do.call(digestFastqs, L))
  
  ## Constant sequence of wrong length
  for (var in c("constantForward", "constantReverse")) {
    L <- Ldef; L[[var]] <- substr(L[[var]], 1, 10)
    expect_error(do.call(digestFastqs, L))
  }
  
  ## Invalid forbidden codons
  L <- Ldef; L[["forbiddenMutatedCodonsForward"]] <- c("EFI")
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["forbiddenMutatedCodonsReverse"]] <- c("EFI")
  expect_error(do.call(digestFastqs, L))
  
  ## Invalid mutNameDelimiter
  L <- Ldef; L[["mutNameDelimiter"]] <- ""
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["mutNameDelimiter"]] <- "__"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["mutNameDelimiter"]] <- "f"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["mutNameDelimiter"]] <- "r"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["mutNameDelimiter"]] <- 1
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["mutNameDelimiter"]] <- c("a", "B")
  expect_error(do.call(digestFastqs, L))
  
  ## Invalid value of useTreeWTmatch
  L <- Ldef; L[["useTreeWTmatch"]] <- 2
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["useTreeWTmatch"]] <- "TRUE"
  expect_error(do.call(digestFastqs, L))
  
  ## Invalid value of output FASTQ files
  L <- Ldef; L[["filteredReadsFastqForward"]] <- 1; L[["filteredReadsFastqReverse"]] <- 1
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["filteredReadsFastqForward"]] <- c("file1.gz", "file2.gz")
  L[["filteredReadsFastqReverse"]] <- "file3.gz"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["filteredReadsFastqForward"]] <- TRUE; L[["filteredReadsFastqReverse"]] <- FALSE
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["filteredReadsFastqForward"]] <- "file1.fastq"
  L[["filteredReadsFastqReverse"]] <- "file3.fastq"
  expect_error(do.call(digestFastqs, L))
  
  ## Invalid combination of input and output FASTQ
  L <- Ldef; L[["fastqReverse"]] <- NULL; L[["filteredReadsFastqForward"]] <- "file1.fastq.gz";
  L[["filteredReadsFastqReverse"]] <- "file2.fastq.gz"
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["filteredReadsFastqForward"]] <- "file1.fastq.gz";
  L[["filteredReadsFastqReverse"]] <- ""
  expect_error(do.call(digestFastqs, L))
  L <- Ldef; L[["filteredReadsFastqForward"]] <- "";
  L[["filteredReadsFastqReverse"]] <- "file2.fastq.gz"
  expect_error(do.call(digestFastqs, L))
  
  ## Invalid value of verbose
  for (v in c("verbose", "collapseToWTForward", "collapseToWTReverse")) {
    L <- Ldef; L[[v]] <- 2
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[v]] <- "TRUE"
    expect_error(do.call(digestFastqs, L))
    L <- Ldef; L[[v]] <- c(TRUE, FALSE)
    expect_error(do.call(digestFastqs, L))
  }
  
  ## Too long read
  expect_error(digestFastqs(
      fastqForward = system.file("extdata", "cisInput_1.fastq.gz",
                                 package = "mutscan"),
      elementsForward = "V", elementLengthsForward = -1, 
      maxReadLength = 100),
      "Encountered a read exceeding the maximal allowed length")
  ## Check border cases
  expect_error(digestFastqs(
      fastqForward = system.file("extdata", "cisInput_1.fastq.gz",
                                 package = "mutscan"),
      elementsForward = "V", elementLengthsForward = -1, 
      maxReadLength = 124),
      "Encountered a read exceeding the maximal allowed length")
  ## Works if the maxReadLength is specified accordingly
  out <- digestFastqs(
      fastqForward = system.file("extdata", "cisInput_1.fastq.gz",
                                 package = "mutscan"),
      elementsForward = "V", elementLengthsForward = -1, 
      maxReadLength = 125)
  expect_type(out, "list")
})

test_that("digestFastqs works as expected for cis experiments", {
  fqc1 <- system.file("extdata/cisInput_1.fastq.gz", package = "mutscan")
  fqc2 <- system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqc1, fastqReverse = fqc2,
    mergeForwardReverse = TRUE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 1, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = TRUE,
    elementsForward = "SUCV", elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAGTTCATCCTGGCAGC",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000,
    maxReadLength = 1024
  )

  ## hit maxNReads
  Ldef0 <- Ldef
  Ldef0$maxNReads <- 10
  res0 <- do.call(digestFastqs, Ldef0)
  expect_identical(res0$filterSummary$nbrTotal, 10L)

  ## hit nTooManyNinUMI
  Ldef0 <- Ldef
  Ldef0$elementsForward <- "UV"
  Ldef0$elementsReverse <- "UVS"
  Ldef0$elementLengthsForward <- c(115, 10)
  Ldef0$elementLengthsReverse = c(111, 10, -1)
  Ldef0$adapterForward <- Ldef0$adapterReverse <- ""
  Ldef0$primerForward <- Ldef0$primerReverse <- ""
  Ldef0$wildTypeForward <- "GGAAAAACTA"
  Ldef0$constantForward <- Ldef0$constantReverse <- ""
  res0 <- do.call(digestFastqs, Ldef0)
  expect_identical(res0$filterSummary$f7_nbrTooManyNinUMI, 497L)

  ## hit nTooManyMutBases
  Ldef0 <- Ldef
  Ldef0$nbrMutatedBasesMaxForward <- 0
  Ldef0$nbrMutatedBasesMaxReverse <- 0
  Ldef0$nbrMutatedCodonsMaxForward <- -1
  Ldef0$nbrMutatedCodonsMaxReverse <- -1
  res0 <- do.call(digestFastqs, Ldef0)
  expect_identical(res0$filterSummary$f10b_nbrTooManyMutBases, 753L)

  ## hit not all WT sequences are of the same length
  Ldef0 <- Ldef
  Ldef0$maxNReads <- 100
  Ldef0$useTreeWTmatch <- TRUE
  Ldef0$mergeForwardReverse <- FALSE
  Ldef0$wildTypeForward <- ""
  Ldef0$wildTypeReverse <- c(w1 = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                             w2 = "GACAGACCAACTAGAAGTTACATGAGAAGTCTGCTTTGCAGACCGAGATTGCCA")
  res0 <- do.call(digestFastqs, Ldef0)
  expect_true(any(grepl("_w1", res0$summaryTable$mutantNameBase)))

  ## now run with the unmodified arguments
  res <- do.call(digestFastqs, Ldef)

  ## Specify reverse WT - check that it's ignored
  LdefrevWT <- Ldef
  LdefrevWT$wildTypeReverse <- "ACGT"
  expect_warning(resrevWT <- do.call(digestFastqs, LdefrevWT),
                 "Ignoring 'wildTypeReverse'")

  expect_equal(res$filterSummary, resrevWT$filterSummary)
  expect_equal(res$summaryTable, resrevWT$summaryTable)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 126L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 44L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 581L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 82L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 167)

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE),
                 ignore_attr = TRUE)
  }

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(sum(res$summaryTable$nbrReads == 2), 11L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("f.14.AAG", "f.15.ATG", "f.19.CAC", "f.1.ACC", "f.20.AAC", "f.21.GTG",
                      "f.24.AGC", "f.4.CGC", "f.7.GTG", "f.9.GGC", "f.9.GTC")))
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))

  ## Check the number of reads with a given number of mismatches
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 0]), 77L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 1]), 88L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 2]), 2L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 0]), 77L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 1]), 90L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 0]), 97L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 1]), 70L)

  ## Similarly but for the number of unique sequences
  expect_equal(sum(res$summaryTable$nbrMutBases == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 1), 64L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 2), 2L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 1), 66L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 0), 14L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 1), 53L)

  ## Check that mutant naming worked (compare to manual matching)
  example_seq <- paste0("ACTGATACACGCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCT",
                        "GCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")
  expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
               "f.4.CGC")
  expect_equal(res$summaryTable$mutantNameAA[res$summaryTable$sequence == example_seq],
               "f.4.R")
  expect_equal(res$summaryTable$mutantNameBase[res$summaryTable$sequence == example_seq],
               "f.11.G")
  expect_equal(res$summaryTable$mutantNameCodon[res$summaryTable$sequence == example_seq],
               "f.4.CGC")
  expect_equal(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$sequence == example_seq],
               "f:c.11T>G")
  expect_equal(res$summaryTable$mutantNameAAHGVS[res$summaryTable$sequence == example_seq],
               "f:p.(Leu4Arg)")

  expect_equal(res$summaryTable$sequenceAA,
               as.character(Biostrings::translate(
                 Biostrings::DNAStringSet(res$summaryTable$sequence))))
  expect_equal(as.numeric(res$summaryTable$varLengths),
               nchar(res$summaryTable$sequence))
  expect_equal(sum(grepl("WT", res$summaryTable$mutantNameAA) & res$summaryTable$nbrMutBases > 0), 13)
  expect_true(all(grepl("silent", res$summaryTable$mutationTypes[grepl("WT", res$summaryTable$mutantNameAA) & res$summaryTable$nbrMutBases > 0])))
  expect_true(all(grepl("nonsynonymous|stop", res$summaryTable$mutationTypes[grepl("\\.[1-9][0-9]*\\.", res$summaryTable$mutantNameAA)])))
  expect_true(all(grepl("stop", res$summaryTable$mutationTypes[grepl("TAG|TAA|TGA", res$summaryTable$mutantName)])))

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 51L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 19L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 79L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 163L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 2690L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 1L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 1L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 2L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 41L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 56L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 96L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 2620L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 2], 4L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 7L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 1L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 37], 14L)

  ## Add an additional constant sequence of the wrong length - should fail
  Ldefcs <- Ldef
  Ldefcs$constantForward <- c("AACCGGAGGAGGGAGCTG", "AACCGGAGGAGGGAGCT")
  expect_error(do.call(digestFastqs, Ldefcs),
               "All constant sequences must be of the same length")

})

test_that("digestFastqs works as expected when specifying max nbr of mutated bases rather than codons", {
  fqc1 <- system.file("extdata/cisInput_1.fastq.gz", package = "mutscan")
  fqc2 <- system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqc1, fastqReverse = fqc2,
    mergeForwardReverse = TRUE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 1, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = TRUE,
    elementsForward = "SUCV", elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAGTTCATCCTGGCAGC",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = -1,
    nbrMutatedCodonsMaxReverse = -1,
    nbrMutatedBasesMaxForward = 2,
    nbrMutatedBasesMaxReverse = 2,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000,
    maxReadLength = 125
  )

  res <- do.call(digestFastqs, Ldef)

  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 126L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 0L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 44L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 0L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 416L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 212L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 202L)

  expect_named(res$summaryTable, c("mutantName", "sequence","nbrReads",
                                   "maxNbrReads", "nbrUmis", "nbrMutBases",
                                   "nbrMutCodons", "nbrMutAAs", "varLengths",
                                   "mutantNameBase", "mutantNameCodon",
                                   "mutantNameBaseHGVS", "mutantNameAA",
                                   "mutantNameAAHGVS", "mutationTypes",
                                   "sequenceAA"))

  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE),
                 ignore_attr = TRUE)
  }

  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(sum(res$summaryTable$nbrReads == 3), 5L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 3]),
               sort(c("f.12.G", "f.17.T", "f.60.G", "f.85.T", "f.96.G")))
  expect_true(all(res$summaryTable$nbrReads == res$summaryTable$nbrUmis))

  ## Check the number of reads with a given number of mismatches
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 0]), 77L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 1]), 88L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutBases == 2]), 37L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 0]), 77L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 1]), 90L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutCodons == 2]), 35L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 0]), 98L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 1]), 86L)
  expect_equal(sum(res$summaryTable$nbrReads[res$summaryTable$nbrMutAAs == 2]), 18L)

  ## Similarly but for the number of unique sequences
  expect_equal(sum(res$summaryTable$nbrMutBases == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 1), 64L)
  expect_equal(sum(res$summaryTable$nbrMutBases == 2), 36L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 0), 1L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 1), 66L)
  expect_equal(sum(res$summaryTable$nbrMutCodons == 2), 34L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 0), 15L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 1), 69L)
  expect_equal(sum(res$summaryTable$nbrMutAAs == 2), 17L)

  ## Check that mutant naming worked (compare to manual matching)
  example_seq <- paste0("ACTGATACACGCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCT",
                        "GCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")
  expect_equal(res$summaryTable$mutantName[res$summaryTable$sequence == example_seq],
               "f.11.G")
  expect_equal(res$summaryTable$mutantNameAA[res$summaryTable$sequence == example_seq],
               "f.4.R")
  expect_equal(res$summaryTable$mutantNameBase[res$summaryTable$sequence == example_seq],
               "f.11.G")
  expect_equal(res$summaryTable$mutantNameCodon[res$summaryTable$sequence == example_seq],
               "f.4.CGC")
  expect_equal(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$sequence == example_seq],
               "f:c.11T>G")
  expect_equal(res$summaryTable$mutantNameAAHGVS[res$summaryTable$sequence == example_seq],
               "f:p.(Leu4Arg)")

  expect_equal(sum(res$errorStatistics$nbrMatchForward + res$errorStatistics$nbrMismatchForward),
               nchar(Ldef$constantForward[1]) * res$filterSummary$nbrRetained)
  expect_equal(sum(res$errorStatistics$nbrMatchReverse + res$errorStatistics$nbrMismatchReverse),
               nchar(Ldef$constantReverse[1]) * res$filterSummary$nbrRetained)

  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 14], 61L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 22], 22L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 27], 97L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 33], 203L)
  expect_equal(res$errorStatistics$nbrMatchForward[res$errorStatistics$PhredQuality == 37], 3249L)

  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 14], 1L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 27], 1L)
  expect_equal(res$errorStatistics$nbrMismatchForward[res$errorStatistics$PhredQuality == 37], 2L)

  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 14], 42L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 27], 63L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 33], 117L)
  expect_equal(res$errorStatistics$nbrMatchReverse[res$errorStatistics$PhredQuality == 37], 3184L)

  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 2], 5L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 14], 7L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 33], 2L)
  expect_equal(res$errorStatistics$nbrMismatchReverse[res$errorStatistics$PhredQuality == 37], 14L)

})

test_that("digestFastqs gives the same results regardless of how WT matching is done", {
  fqc1 <- system.file("extdata/cisInput_1.fastq.gz", package = "mutscan")
  fqc2 <- system.file("extdata/cisInput_2.fastq.gz", package = "mutscan")
  ## default arguments
  Ldef <- list(
    fastqForward = fqc1, fastqReverse = fqc2,
    mergeForwardReverse = TRUE,
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 1, greedyOverlap = TRUE,
    revComplForward = FALSE, revComplReverse = TRUE,
    elementsForward = "SUCV", elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    adapterForward = "GGAAGAGCACACGTC",
    adapterReverse = "GGAAGAGCGTCGTGT",
    primerForward = "",
    primerReverse = "",
    wildTypeForward = c(wt1 = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                        wt2 = "CCTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA",
                        wt3 = "ACTGATACACTCCAAGCGGAGACTGACCAACTAGAAGATGAGAAGTCTTCTTTGCAGACCGAGATTGCCAACGTGCTGAAGGAGAAGGAAAAACTA"),
    wildTypeReverse = "",
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAGTTCATCCTGGCAGC",
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0,
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = -1,
    nbrMutatedCodonsMaxReverse = -1,
    nbrMutatedBasesMaxForward = 2,
    nbrMutatedBasesMaxReverse = 2,
    forbiddenMutatedCodonsForward = "NNW",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".",
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000,
    maxReadLength = 1024
  )

  res1 <- do.call(digestFastqs, Ldef)
  Ldef$useTreeWTmatch <- TRUE
  res2 <- do.call(digestFastqs, Ldef)

  expect_equal(res1$filterSummary$nbrTotal, res2$filterSummary$nbrTotal)
  expect_equal(res1$filterSummary$f1_nbrAdapter, res2$filterSummary$f1_nbrAdapter)
  expect_equal(res1$filterSummary$f2_nbrNoPrimer, res2$filterSummary$f2_nbrNoPrimer)
  expect_equal(res1$filterSummary$f3_nbrReadWrongLength, res2$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res1$filterSummary$f4_nbrNoValidOverlap, res2$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res1$filterSummary$f5_nbrAvgVarQualTooLow, res2$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res1$filterSummary$f6_nbrTooManyNinVar, res2$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res1$filterSummary$f7_nbrTooManyNinUMI, res2$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res1$filterSummary$f8_nbrTooManyBestWTHits, res2$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res1$filterSummary$f9_nbrMutQualTooLow, res2$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res1$filterSummary$f10a_nbrTooManyMutCodons, res2$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res1$filterSummary$f10b_nbrTooManyMutBases, res2$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res1$filterSummary$f11_nbrForbiddenCodons, res2$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res1$filterSummary$f12_nbrTooManyMutConstant, res2$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res1$filterSummary$f13_nbrTooManyBestConstantHits, res2$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res1$filterSummary$nbrRetained, res2$filterSummary$nbrRetained)

  expect_equal(res1$summaryTable$nbrReads, res2$summaryTable$nbrReads)
})

test_that("digestFastqs works as expected for primer experiments", {
  fqt1 <- system.file("extdata/leujunt0_1.fastq.gz", package = "mutscan")
  fqt2 <- system.file("extdata/leujunt0_2.fastq.gz", package = "mutscan")
  leu <- c(ATF2 = "GATCCTGATGAAAAAAGGAGAAAGTTTTTAGAGCGAAATAGAGCAGCAGCTTCAAGATGCCGACAAAAAAGGAAAGTCTGGGTTCAGTCTTTAGAGAAGAAAGCTGAAGACTTGAGTTCATTAAATGGTCAGCTGCAGAGTGAAGTCACCCTGCTGAGAAATGAAGTGGCACAGCTGAAACAGCTTCTTCTGGCT",
           ATF7 = "GATCCAGATGAGCGACGGCAGCGCTTTCTGGAGCGCAACCGGGCTGCAGCCTCCCGCTGCCGCCAAAAGCGAAAGCTGTGGGTGTCCTCCCTAGAGAAGAAGGCCGAAGAACTCACTTCTCAGAACATTCAGCTGAGTAATGAAGTCACATTACTACGCAATGAGGTGGCCCAGTTGAAACAGCTACTGTTAGCT",
           CREB5 = "GATCCGGACGAGAGGCGGCGGAAATTTCTGGAACGGAACCGGGCAGCTGCCACCCGCTGCAGACAGAAGAGGAAGGTCTGGGTGATGTCATTGGAAAAGAAAGCAGAAGAACTCACCCAGACAAACATGCAGCTTCAGAATGAAGTGTCTATGTTGAAAAATGAGGTGGCCCAGCTGAAACAGTTGTTGTTAACA",
           ATF3 = "GAAGAAGATGAAAGGAAAAAGAGGCGACGAGAAAGAAATAAGATTGCAGCTGCAAAGTGCCGAAACAAGAAGAAGGAGAAGACGGAGTGCCTGCAGAAAGAGTCGGAGAAGCTGGAAAGTGTGAATGCTGAACTGAAGGCTCAGATTGAGGAGCTCAAGAACGAGAAGCAGCATTTGATATACATGCTCAACCTT",
           JDP2 = "GAGGAAGAGGAGCGAAGGAAAAGGCGCCGGGAGAAGAACAAAGTCGCAGCAGCCCGATGCCGGAACAAGAAGAAGGAGCGCACGGAGTTTCTGCAGCGGGAATCCGAGCGGCTGGAACTCATGAACGCAGAGCTGAAGACCCAGATTGAGGAGCTGAAGCAGGAGCGGCAGCAGCTCATCCTGATGCTGAACCGA",
           ATF4 = "GAGAAACTGGATAAGAAGCTGAAAAAAATGGAGCAAAACAAGACAGCAGCCACTAGGTACCGCCAGAAGAAGAGGGCGGAGCAGGAGGCTCTTACTGGTGAGTGCAAAGAGCTGGAAAAGAAGAACGAGGCTCTAAAAGAGAGGGCGGATTCCCTGGCCAAGGAGATCCAGTACCTGAAAGATTTGATAGAAGAG",
           ATF5 = "ACCCGAGGGGACCGCAAGCAAAAGAAGAGAGACCAGAACAAGTCGGCGGCTCTGAGGTACCGCCAGCGGAAGCGGGCAGAGGGTGAGGCCCTGGAGGGCGAGTGCCAGGGGCTGGAGGCACGGAATCGCGAGCTGAAGGAACGGGCAGAGTCCGTGGAGCGCGAGATCCAGTACGTCAAGGACCTGCTCATCGAG",
           CREBZF = "AGTCCCCGGAAGGCGGCGGCGGCCGCTGCCCGCCTTAATCGACTGAAGAAGAAGGAGTACGTGATGGGGCTGGAGAGTCGAGTCCGGGGTCTGGCAGCCGAGAACCAGGAGCTGCGGGCCGAGAATCGGGAGCTGGGCAAACGCGTACAGGCACTGCAGGAGGAGAGTCGCTACCTACGGGCAGTCTTAGCCAAC",
           BATF2 = "CCCAAGGAGCAACAAAGGCAGCTGAAGAAGCAGAAGAACCGGGCAGCCGCCCAGCGAAGCCGGCAGAAGCACACAGACAAGGCAGACGCCCTGCACCAGCAGCACGAGTCTCTGGAAAAAGACAACCTCGCCCTGCGGAAGGAGATCCAGTCCCTGCAGGCCGAGCTGGCGTGGTGGAGCCGGACCCTGCACGTG",
           BATF3 = "GAGGATGATGACAGGAAGGTCCGAAGGAGAGAAAAAAACCGAGTTGCTGCTCAGAGAAGTCGGAAGAAGCAGACCCAGAAGGCTGACAAGCTCCATGAGGAATATGAGAGCCTGGAGCAAGAAAACACCATGCTGCGGAGAGAGATCGGGAAGCTGACAGAGGAGCTGAAGCACCTGACAGAGGCACTGAAGGAG",
           CEBPE = "AAAGATAGCCTTGAGTACCGGCTGAGGCGGGAGCGCAACAACATCGCCGTGCGCAAGAGCCGAGACAAGGCCAAGAGGCGCATTCTGGAGACGCAGCAGAAGGTGCTGGAGTACATGGCAGAGAACGAGCGCCTCCGCAGCCGCGTGGAGCAGCTCACCCAGGAGCTAGACACCCTCCGCAACCTCTTCCGCCAG",
           BACH1 = "CTGGATTGTATCCATGATATTCGAAGAAGAAGTAAAAACAGAATTGCTGCACAGCGCTGTCGCAAGAGAAAACTTGACTGTATACAGAATCTTGAATCAGAAATTGAGAAGCTGCAAAGTGAAAAGGAGAGCTTGTTGAAGGAAAGAGATCACATTTTGTCAACTCTGGGTGAGACAAAGCAGAACCTAACTGGA",
           BACH2 = "TTAGAGTTTATTCATGATGTCCGACGGCGCAGCAAGAACCGCATCGCGGCCCAGCGCTGCCGCAAAAGGAAACTGGACTGTATTCAGAATTTAGAATGTGAAATCCGCAAATTGGTGTGTGAGAAAGAGAAACTGTTGTCAGAGAGGAATCAACTGAAAGCATGCATGGGGGAACTGTTGGACAACTTCTCCTGC",
           NFE2L1 = "CTGAGCCTCATCCGAGACATCCGGCGCCGGGGCAAGAACAAGATGGCGGCGCAGAACTGCCGCAAGCGCAAGCTGGACACCATCCTGAATCTGGAGCGTGATGTGGAGGACCTGCAGCGTGACAAAGCCCGGCTGCTGCGGGAGAAAGTGGAGTTCCTGCGCTCCCTGCGACAGATGAAGCAGAAGGTCCAGAGC",
           NFE2 = "CTAGCGCTAGTCCGGGACATCCGACGACGGGGCAAAAACAAGGTGGCAGCCCAGAACTGCCGCAAGAGGAAGCTGGAAACCATTGTGCAGCTGGAGCGGGAGCTGGAGCGGCTGACCAATGAACGGGAGCGGCTTCTCAGGGCCCGCGGGGAGGCAGACCGGACCCTGGAGGTCATGCGCCAACAGCTGACAGAG",
           NFIL3 = "AAGAAAGATGCTATGTATTGGGAAAAAAGGCGGAAAAATAATGAAGCTGCCAAAAGATCTCGTGAGAAGCGTCGACTGAATGACCTGGTTTTAGAGAACAAACTAATTGCACTGGGAGAAGAAAACGCCACTTTAAAAGCTGAGCTGCTTTCACTAAAATTAAAGTTTGGTTTAATTAGCTCCACAGCATATGCT",
           FOS = "GAAGAAGAAGAGAAAAGGAGAATCCGAAGGGAAAGGAATAAGATGGCTGCAGCCAAATGCCGCAACCGGAGGAGGGAGCTGACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTAGAGTTCATCCTGGCAGCT",
           FOSB = "GAGGAAGAGGAGAAGCGAAGGGTGCGCCGGGAACGAAATAAACTAGCAGCAGCTAAATGCAGGAACCGGCGGAGGGAGCTGACCGACCGACTCCAGGCGGAGACAGATCAGTTGGAGGAAGAAAAAGCAGAGCTGGAGTCGGAGATCGCCGAGCTCCAAAAGGAGAAGGAACGTCTGGAGTTTGTGCTGGTGGCC",
           FOSL1 = "GAGGAAGAGGAGCGCCGCCGAGTAAGGCGCGAGCGGAACAAGCTGGCTGCGGCCAAGTGCAGGAACCGGAGGAAGGAACTGACCGACTTCCTGCAGGCGGAGACTGACAAACTGGAAGATGAGAAATCTGGGCTGCAGCGAGAGATTGAGGAGCTGCAGAAGCAGAAGGAGCGCCTAGAGCTGGTGCTGGAAGCC",
           FOSL2 = "GAAGAGGAGGAGAAGCGTCGCATCCGGCGGGAGAGGAACAAGCTGGCTGCAGCCAAGTGCCGGAACCGACGCCGGGAGCTGACAGAGAAGCTGCAGGCGGAGACAGAGGAGCTGGAGGAGGAGAAGTCAGGCCTGCAGAAGGAGATTGCTGAGCTGCAGAAGGAGAAGGAGAAGCTGGAGTTCATGTTGGTGGCT",
           MAFB = "GTGATCCGCCTGAAGCAGAAGCGGCGGACCCTGAAGAACCGGGGCTACGCCCAGTCTTGCAGGTATAAACGCGTCCAGCAGAAGCACCACCTGGAGAATGAGAAGACGCAGCTCATTCAGCAGGTGGAGCAGCTTAAGCAGGAGGTGTCCCGGCTGGCCCGCGAGAGAGACGCCTACAAGGTCAAGTGCGAGAAA",
           JUN = "CAGGAGCGGATCAAGGCGGAGAGGAAGCGCATGAGGAACCGCATCGCTGCCTCCAAGTGCCGAAAAAGGAAGCTGGAGAGAATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTTAAACAGAAAGTCATGAAC",
           JUNB = "CAAGAGCGCATCAAAGTGGAGCGCAAGCGGCTGCGGAACCGGCTGGCGGCCACCAAGTGCCGGAAGCGGAAGCTGGAGCGCATCGCGCGCCTGGAGGACAAGGTGAAGACGCTCAAGGCCGAGAACGCGGGGCTGTCGAGTACCGCCGGCCTCCTCCGGGAGCAGGTGGCCCAGCTCAAACAGAAGGTCATGACC",
           JUND = "CAGGAGCGCATCAAGGCGGAGCGCAAGCGGCTGCGCAACCGCATCGCCGCCTCCAAGTGCCGCAAGCGCAAGCTGGAGCGCATCTCGCGCCTGGAAGAGAAAGTGAAGACCCTCAAGAGTCAGAACACGGAGCTGGCGTCCACGGCGAGCCTGCTGCGCGAGCAGGTGGCGCAGCTCAAGCAGAAAGTCCTCAGC",
           CREB3 = "GAACAAATTCTGAAACGTGTGCGGAGGAAGATTCGAAATAAAAGATCTGCTCAAGAGAGCCGCAGGAAAAAGAAGGTGTATGTTGGGGGTTTAGAGAGCAGGGTCTTGAAATACACAGCCCAGAATATGGAGCTTCAGAACAAAGTACAGCTTCTGGAGGAACAGAATTTGTCCCTTCTAGATCAACTGAGGAAA",
           HLF = "CTGAAGGATGACAAGTACTGGGCAAGGCGCAGAAAGAACAACATGGCAGCCAAGCGCTCCCGCGACGCCCGGAGGCTGAAAGAGAACCAGATCGCCATCCGGGCCTCGTTCCTGGAGAAGGAGAACTCGGCCCTCCGCCAGGAGGTGGCTGACTTGAGGAAGGAGCTGGGCAAATGCAAGAACATACTTGCCAAG",
           MAFG = "ATCGTCCAGCTGAAGCAGCGCCGGCGCACGCTCAAGAACCGCGGCTACGCTGCCAGCTGCCGCGTGAAGCGGGTGACGCAGAAGGAGGAGCTGGAGAAGCAGAAGGCGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCCTCAGAGAACGCCAGCATGAAGCTGGAGCTCGACGCGCTGCGCTCCAAGTACGAGGCG",
           MAFK = "GTGACCCGCCTGAAGCAGCGTCGGCGCACACTCAAGAACCGCGGCTACGCGGCCAGCTGCCGCATCAAGCGGGTGACGCAGAAGGAGGAGCTGGAGCGGCAGCGCGTGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCGCGTGAGAACAGCAGCATGCGGCTGGAGCTGGACGCCCTGCGCTCCAAGTACGAGGCG",
           XBP1 = "AGCCCCGAGGAGAAGGCGCTGAGGAGGAAACTGAAAAACAGAGTAGCAGCTCAGACTGCCAGAGATCGAAAGAAGGCTCGAATGAGTGAGCTGGAACAGCAAGTGGTAGATTTAGAAGAAGAGAACCAAAAACTTTTGCTAGAAAATCAGCTTTTACGAGAGAAAACTCATGGCCTTGTAGTTGAGAACCAGGAG",
           ATF6 = "ATTGCTGTGCTAAGGAGACAGCAACGTATGATAAAAAATCGAGAATCCGCTTGTCAGTCTCGCAAGAAGAAGAAAGAATATATGCTAGGGTTAGAGGCGAGATTAAAGGCTGCCCTCTCAGAAAACGAGCAACTGAAGAAAGAAAATGGAACACTGAAGCGGCAGCTGGATGAAGTTGTGTCAGAGAACCAGAGG",
           ATF6B = "GCAAAGCTGCTGAAGCGGCAGCAGCGAATGATCAAGAACCGGGAGTCAGCCTGCCAGTCCCGGAGAAAGAAGAAAGAGTATCTGCAGGGACTGGAGGCTCGGCTGCAAGCAGTACTGGCTGACAACCAGCAGCTCCGCCGAGAGAATGCTGCCCTCCGGCGGCGGCTGGAGGCCCTGCTGGCTGAAAACAGCGAG",
           CEBPA = "AAGAACAGCAACGAGTACCGGGTGCGGCGCGAGCGCAACAACATCGCGGTGCGCAAGAGCCGCGACAAGGCCAAGCAGCGCAACGTGGAGACGCAGCAGAAGGTGCTGGAGCTGACCAGTGACAATGACCGCCTGCGCAAGCGGGTGGAACAGCTGAGCCGCGAACTGGACACGCTGCGGGGCATCTTCCGCCAG",
           CEBPB = "AAGCACAGCGACGAGTACAAGATCCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGATGCGCAACCTGGAGACGCAGCACAAGGTCCTGGAGCTCACGGCCGAGAACGAGCGGCTGCAGAAGAAGGTGGAGCAGCTGTCGCGCGAGCTCAGCACCCTGCGGAACTTGTTCAAGCAG",
           CEBPD = "CGCGGCAGCCCCGAGTACCGGCAGCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGCGGCGCAACCAGGAGATGCAGCAGAAGTTGGTGGAGCTGTCGGCTGAGAACGAGAAGCTGCACCAGCGCGTGGAGCAGCTCACGCGGGACCTGGCCGGCCTCCGGCAGTTCTTCAAGCAG",
           CEBPG = "CGAAACAGTGACGAGTATCGGCAACGCCGAGAGAGGAACAACATGGCTGTGAAAAAGAGCCGGTTGAAAAGCAAGCAGAAAGCACAAGACACACTGCAGAGAGTCAATCAGCTCAAAGAAGAGAATGAACGGTTGGAAGCAAAAATCAAATTGCTGACCAAGGAATTAAGTGTACTCAAAGATTTGTTTCTTGAG",
           CREB1 = "GAAGCAGCACGAAAGAGAGAGGTCCGTCTAATGAAGAACAGGGAAGCAGCTCGAGAGTGTCGTAGAAAGAAGAAAGAATATGTGAAATGTTTAGAAAACAGAGTGGCAGTGCTTGAAAATCAAAACAAGACATTGATTGAGGAGCTAAAAGCACTTAAGGACCTTTACTGCCACAAATCAGAT",
           CREB3L1 = "GAGAAGGCCTTGAAGAGAGTCCGGAGGAAAATCAAGAACAAGATCTCAGCCCAGGAGAGCCGTCGTAAGAAGAAGGAGTATGTGGAGTGTCTAGAAAAGAAGGTGGAGACATTTACATCTGAGAACAATGAACTGTGGAAGAAGGTGGAGACCCTGGAGAATGCCAACAGGACCCTGCTCCAGCAGCTGCAGAAA",
           CREB3L2 = "GAGAAGGCCCTGAAGAAAATTCGGAGGAAGATCAAGAATAAGATTTCTGCTCAGGAAAGTAGGAGAAAGAAGAAAGAATACATGGACAGCCTGGAGAAAAAAGTGGAGTCTTGTTCAACTGAGAACTTGGAGCTTCGGAAGAAGGTAGAGGTTCTAGAGAACACTAATAGGACTCTCCTTCAGCAACTCCAGAAG",
           CREB3L3 = "GAGCGAGTGCTGAAAAAAATCCGCCGGAAAATCCGGAACAAGCAGTCGGCGCAAGAAAGCAGGAAGAAGAAGAAGGAATATATCGATGGCCTGGAGACTCGGATGTCAGCTTGCACTGCTCAGAATCAGGAGTTACAGAGGAAAGTCTTGCATCTCGAGAAGCAAAACCTGTCCCTCTTGGAGCAACTGAAGAAA",
           CREB3L4 = "GAGAGGGTCCTCAAGAAGGTCAGGAGGAAAATCCGTAACAAGCAGTCAGCTCAGGACAGTCGGCGGCGGAAGAAGGAGTACATTGATGGGCTGGAGAGCAGGGTGGCAGCCTGTTCTGCACAGAACCAAGAATTACAGAAAAAAGTCCAGGAGCTGGAGAGGCACAACATCTCCTTGGTAGCTCAGCTCCGCCAG",
           CREBL2 = "CCAGCCAAAATTGACTTGAAAGCAAAACTTGAGAGGAGCCGGCAGAGTGCAAGAGAATGCCGAGCCCGAAAAAAGCTGAGATATCAGTATTTGGAAGAGTTGGTATCCAGTCGAGAAAGAGCTATATGTGCCCTCAGAGAGGAACTGGAAATGTACAAGCAGTGGTGCATGGCAATGGACCAAGGAAAAATCCCT",
           CREBRF = "CCCTTAACAGCCCGACCAAGGTCAAGGAAGGAAAAAAATAAGCTGGCTTCCAGAGCTTGTCGGTTAAAGAAGAAAGCCCAGTATGAAGCTAATAAAGTGAAATTATGGGGCCTCAACACAGAATATGATAATTTATTGTTTGTAATCAACTCCATCAAGCAAGAGATTGTAAACCGGGTACAGAATCCAAGAGAT",
           DBP = "CAGAAGGATGAGAAATACTGGAGCCGGCGGTACAAGAACAACGAGGCAGCCAAGCGGTCCCGTGACGCCCGGCGGCTCAAGGAGAACCAGATATCGGTGCGGGCGGCCTTCCTGGAGAAGGAGAACGCCCTGCTGCGGCAGGAAGTTGTGGCCGTGCGCCAGGAGCTGTCCCACTACCGCGCCGTGCTGTCCCGA",
           NFE2L2 = "CTTGCATTAATTCGGGATATACGTAGGAGGGGTAAGAATAAAGTGGCTGCTCAGAATTGCAGAAAAAGAAAACTGGAAAATATAGTAGAACTAGAGCAAGATTTAGATCATTTGAAAGATGAAAAAGAAAAATTGCTCAAAGAAAAAGGAGAAAATGACAAAAGCCTTCACCTACTGAAAAAACAACTCAGCACC",
           NFE2L3 = "GTCTCACTTATCCGTGACATCAGACGAAGAGGGAAAAATAAAGTTGCTGCGCAGAACTGTCGTAAACGCAAATTGGACATAATTTTGAATTTAGAAGATGATGTATGTAACTTGCAAGCAAAGAAGGAAACTCTTAAGAGAGAGCAAGCACAATGTAACAAAGCTATTAACATAATGAAACAGAAACTGCATGAC",
           TEF = "CAGAAGGATGAAAAGTACTGGACAAGACGCAAGAAGAACAACGTGGCAGCTAAACGGTCACGGGATGCCCGGCGCCTGAAAGAGAATCAGATCACCATCCGGGCAGCCTTCCTGGAGAAGGAGAACACAGCCCTGCGGACGGAGGTGGCCGAGCTACGCAAGGAGGTGGGCAAGTGCAAGACCATCGTGTCCAAG"
  )
  ## default arguments
  Ldef <- list(
    fastqForward = fqt1, fastqReverse = fqt2, 
    mergeForwardReverse = FALSE, 
    minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
    revComplForward = FALSE, revComplReverse = FALSE,
    elementsForward = "SPV", elementsReverse = "SPVS",
    elementLengthsForward = c(-1, -1, -1),
    elementLengthsReverse = c(-1, -1, 96, -1),
    adapterForward = "", 
    adapterReverse = "",
    primerForward = "GTCAGGTGGAGGCGGATCC",
    primerReverse = "GAAAAAGGAAGCTGGAGAGA",
    wildTypeForward = leu,
    wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
    constantForward = "", 
    constantReverse = "", 
    avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
    variableNMaxForward = 0, variableNMaxReverse = 0, 
    umiNMax = 0,
    nbrMutatedCodonsMaxForward = 0,
    nbrMutatedCodonsMaxReverse = 1,
    nbrMutatedBasesMaxForward = -1,
    nbrMutatedBasesMaxReverse = -1,
    forbiddenMutatedCodonsForward = "",
    forbiddenMutatedCodonsReverse = "NNW",
    useTreeWTmatch = FALSE,
    mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    umiCollapseMaxDist = 0,
    filteredReadsFastqForward = "",
    filteredReadsFastqReverse = "",
    maxNReads = -1, verbose = FALSE,
    nThreads = 1, chunkSize = 1000, 
    maxReadLength = 1024
  )
  
  res <- do.call(digestFastqs, Ldef)
  
  expect_equal(res$filterSummary$nbrTotal, 1000L)
  expect_equal(res$filterSummary$f1_nbrAdapter, 0L)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, 126L)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, 0L)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, 0L)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, 0L)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, 76L)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, 0L)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, 0L)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, 0L)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, 58L + 137L)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, 0L)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, 3L)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, 0L)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, 0L)
  expect_equal(res$filterSummary$nbrRetained, 600L)
  
  for (nm in setdiff(names(Ldef), c("forbiddenMutatedCodonsForward", "forbiddenMutatedCodonsReverse", "verbose", "fastqForward", "fastqReverse"))) {
    expect_equal(res$parameters[[nm]], Ldef[[nm]], ignore_attr = TRUE)
  }
  for (nm in c("fastqForward", "fastqReverse")) {
    expect_equal(res$parameters[[nm]], normalizePath(Ldef[[nm]], mustWork = FALSE), 
                 ignore_attr = TRUE)
  }
  
  expect_equal(sum(res$summaryTable$nbrReads), res$filterSummary$nbrRetained)
  expect_equal(sum(res$summaryTable$nbrReads == 2), 11L)
  expect_equal(sort(res$summaryTable$mutantName[res$summaryTable$nbrReads == 2]),
               sort(c("BACH2.0.WT_r.6.CCC", "BATF2.0.WT_r.21.TGG", "BATF2.0.WT_r.7.GGG", "CEBPB.0.WT_r.12.CTC",
                      "CEBPD.0.WT_r.13.CCC", "CEBPD.0.WT_r.23.GAG", "CREB3L3.0.WT_r.22.GGG",
                      "FOSL1.0.WT_r.22.CCC", "FOSL2.0.WT_r.13.GGG", "JUNB.0.WT_r.0.WT", "XBP1.0.WT_r.0.WT")))
  expect_equal(sort(res$summaryTable$mutantNameCodon[res$summaryTable$nbrReads == 2]),
               sort(c("BACH2.0.WT_r.6.CCC", "BATF2.0.WT_r.21.TGG", "BATF2.0.WT_r.7.GGG", "CEBPB.0.WT_r.12.CTC",
                      "CEBPD.0.WT_r.13.CCC", "CEBPD.0.WT_r.23.GAG", "CREB3L3.0.WT_r.22.GGG",
                      "FOSL1.0.WT_r.22.CCC", "FOSL2.0.WT_r.13.GGG", "JUNB.0.WT_r.0.WT", "XBP1.0.WT_r.0.WT")))
  expect_equal(sort(res$summaryTable$mutantNameBase[res$summaryTable$nbrReads == 2]),
               sort(c("BACH2.0.WT_r.16.C_r.17.C_r.18.C", "BATF2.0.WT_r.61.T_r.62.G", 
                      "BATF2.0.WT_r.19.G_r.20.G_r.21.G", "CEBPB.0.WT_r.34.C_r.35.T_r.36.C",
                      "CEBPD.0.WT_r.37.C_r.39.C", "CEBPD.0.WT_r.67.G_r.69.G", 
                      "CREB3L3.0.WT_r.65.G_r.66.G", "FOSL1.0.WT_r.64.C", 
                      "FOSL2.0.WT_r.38.G_r.39.G", "JUNB.0.WT_r.0.WT", "XBP1.0.WT_r.0.WT")))
  expect_equal(sort(res$summaryTable$mutantNameBaseHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("BACH2:c_r:c.16_18delinsCCC", "BATF2:c_r:c.61_62delinsTG", 
                      "BATF2:c_r:c.19_21delinsGGG", "CEBPB:c_r:c.34_36delinsCTC",
                      "CEBPD:c_r:c.37_39delinsCCC", "CEBPD:c_r:c.67_69delinsGAG", 
                      "CREB3L3:c_r:c.65_66delinsGG", "FOSL1:c_r:c.64G>C", 
                      "FOSL2:c_r:c.38_39delinsGG", "JUNB:c_r:c", "XBP1:c_r:c")))
  expect_equal(sort(res$summaryTable$mutantNameAAHGVS[res$summaryTable$nbrReads == 2]),
               sort(c("BACH2:p_r:p.(Glu6Pro)", "BATF2:p_r:p.(Thr21Trp)", 
                      "BATF2:p_r:p.(Lys7Gly)", "CEBPB:p_r:p.(Lys12Leu)",
                      "CEBPD:p_r:p.(Ala13Pro)", "CEBPD:p_r:p.(Asn23Glu)",
                      "CREB3L3:p_r:p.(Ala22Gly)", "FOSL1:p_r:p.(Ala22Pro)", 
                      "FOSL2:p_r:p.(Ala13Gly)", "JUNB:p_r:p", "XBP1:p_r:p")))
  expect_equal(res$summaryTable$nbrUmis, rep(0L, nrow(res$summaryTable))) ## no UMIs in this experiment
  
  ## ---------------------------------------------------------------------------
  ## Test that we get the same results with useTreeWTmatch = TRUE
  ## ---------------------------------------------------------------------------
  Ldef$useTreeWTmatch <- TRUE
  res2 <- do.call(digestFastqs, Ldef)
  
  expect_equal(res$filterSummary$nbrTotal, res2$filterSummary$nbrTotal)
  expect_equal(res$filterSummary$f1_nbrAdapter, res2$filterSummary$f1_nbrAdapter)
  expect_equal(res$filterSummary$f2_nbrNoPrimer, res2$filterSummary$f2_nbrNoPrimer)
  expect_equal(res$filterSummary$f3_nbrReadWrongLength, res2$filterSummary$f3_nbrReadWrongLength)
  expect_equal(res$filterSummary$f4_nbrNoValidOverlap, res2$filterSummary$f4_nbrNoValidOverlap)
  expect_equal(res$filterSummary$f5_nbrAvgVarQualTooLow, res2$filterSummary$f5_nbrAvgVarQualTooLow)
  expect_equal(res$filterSummary$f6_nbrTooManyNinVar, res2$filterSummary$f6_nbrTooManyNinVar)
  expect_equal(res$filterSummary$f7_nbrTooManyNinUMI, res2$filterSummary$f7_nbrTooManyNinUMI)
  expect_equal(res$filterSummary$f8_nbrTooManyBestWTHits, res2$filterSummary$f8_nbrTooManyBestWTHits)
  expect_equal(res$filterSummary$f9_nbrMutQualTooLow, res2$filterSummary$f9_nbrMutQualTooLow)
  expect_equal(res$filterSummary$f10a_nbrTooManyMutCodons, res2$filterSummary$f10a_nbrTooManyMutCodons)
  expect_equal(res$filterSummary$f10b_nbrTooManyMutBases, res2$filterSummary$f10b_nbrTooManyMutBases)
  expect_equal(res$filterSummary$f11_nbrForbiddenCodons, res2$filterSummary$f11_nbrForbiddenCodons)
  expect_equal(res$filterSummary$f12_nbrTooManyMutConstant, res2$filterSummary$f12_nbrTooManyMutConstant)
  expect_equal(res$filterSummary$f13_nbrTooManyBestConstantHits, res2$filterSummary$f13_nbrTooManyBestConstantHits)
  expect_equal(res$filterSummary$nbrRetained, res2$filterSummary$nbrRetained)
  
  expect_equal(res$summaryTable$nbrReads, res2$summaryTable$nbrReads)
  expect_equal(res$summaryTable$mutantName, res2$summaryTable$mutantName)
  expect_equal(res$summaryTable$sequence, res2$summaryTable$sequence)
  
  ## ---------------------------------------------------------------------------
  ## Test aggregating across groups of wildtype sequences
  ## ---------------------------------------------------------------------------
  Ldef <- list(
      fastqForward = fqt1, fastqReverse = fqt2, 
      mergeForwardReverse = FALSE, 
      minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
      revComplForward = FALSE, revComplReverse = FALSE,
      elementsForward = "SPV", elementsReverse = "SPSS",
      elementLengthsForward = c(-1, -1, -1),
      elementLengthsReverse = c(-1, -1, 96, -1),
      adapterForward = "", 
      adapterReverse = "",
      primerForward = "GTCAGGTGGAGGCGGATCC",
      primerReverse = "GAAAAAGGAAGCTGGAGAGA",
      wildTypeForward = leu,
      wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
      constantForward = "", 
      constantReverse = "", 
      avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
      variableNMaxForward = 0, variableNMaxReverse = 0, 
      umiNMax = 0,
      nbrMutatedCodonsMaxForward = 0,
      nbrMutatedCodonsMaxReverse = 0,
      nbrMutatedBasesMaxForward = -1,
      nbrMutatedBasesMaxReverse = -1,
      forbiddenMutatedCodonsForward = "",
      forbiddenMutatedCodonsReverse = "NNW",
      useTreeWTmatch = FALSE,
      mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
      mutNameDelimiter = ".", 
      constantMaxDistForward = -1,
      constantMaxDistReverse = -1,
      umiCollapseMaxDist = 0,
      filteredReadsFastqForward = "",
      filteredReadsFastqReverse = "",
      maxNReads = -1, verbose = FALSE,
      nThreads = 1, chunkSize = 1000, 
      maxReadLength = 1024
  )
  res1 <- do.call(digestFastqs, Ldef)
  leu2 <- leu
  names(leu2)[1:5] <- "ATF2"
  Ldef$wildTypeForward <- leu2
  res2 <- do.call(digestFastqs, Ldef)
  
  expect_identical(res1$filterSummary, res2$filterSummary)
  expect_identical(res1$errorStatistics, res2$errorStatistics)
  expect_equal(nrow(res1$summaryTable), nrow(res2$summaryTable) + 4)
  idx <- which(!res1$summaryTable$mutantName %in% paste0(names(leu)[1:5], ".0.WT_r.0.WT"))
  expect_identical(res1$summaryTable[idx, ], 
                   res2$summaryTable[match(res1$summaryTable$mutantName[idx], 
                                           res2$summaryTable$mutantName), ],
                   ignore_attr = TRUE)
  expect_equal(sum(res1$summaryTable[-idx, "nbrReads"]), 
               res2$summaryTable[res2$summaryTable$mutantName == 
                                     paste0(names(leu2)[1], ".0.WT_r.0.WT"), "nbrReads"])
  
  ## ---------------------------------------------------------------------------
  ## Swap forward and reverse reads, specify max nbr of mutated bases
  ## ---------------------------------------------------------------------------
  Ldef <- list(
      fastqForward = fqt2, fastqReverse = fqt1, 
      mergeForwardReverse = FALSE, 
      minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
      revComplForward = FALSE, revComplReverse = FALSE,
      elementsForward = "SPSS", elementsReverse = "SPV", 
      elementLengthsForward = c(-1, -1, 96, -1),
      elementLengthsReverse = c(-1, -1, -1),
      adapterForward = "", 
      adapterReverse = "",
      primerForward = "GAAAAAGGAAGCTGGAGAGA",
      primerReverse = "GTCAGGTGGAGGCGGATCC",
      wildTypeForward = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", 
      wildTypeReverse = leu,
      constantForward = "", 
      constantReverse = "", 
      avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
      variableNMaxForward = 0, variableNMaxReverse = 0, 
      umiNMax = 0,
      nbrMutatedCodonsMaxForward = -1,
      nbrMutatedCodonsMaxReverse = -1,
      nbrMutatedBasesMaxForward = 0,
      nbrMutatedBasesMaxReverse = 0,
      forbiddenMutatedCodonsForward = "NNW",
      forbiddenMutatedCodonsReverse = "",
      useTreeWTmatch = FALSE,
      mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
      mutNameDelimiter = ".", 
      constantMaxDistForward = -1,
      constantMaxDistReverse = -1,
      umiCollapseMaxDist = 0,
      filteredReadsFastqForward = "",
      filteredReadsFastqReverse = "",
      maxNReads = -1, verbose = FALSE,
      nThreads = 1, chunkSize = 1000, 
      maxReadLength = 1024
  )
  res1 <- do.call(digestFastqs, Ldef)
  leu2 <- leu
  names(leu2)[1:5] <- "ATF2"
  Ldef$wildTypeReverse <- leu2
  res2 <- do.call(digestFastqs, Ldef)
  
  expect_identical(res1$filterSummary, res2$filterSummary)
  expect_identical(res1$errorStatistics, res2$errorStatistics)
  expect_equal(nrow(res1$summaryTable), nrow(res2$summaryTable) + 4)
  idx <- which(!res1$summaryTable$mutantName %in% paste0("f.0.WT_", names(leu)[1:5], ".0.WT"))
  expect_identical(res1$summaryTable[idx, ], 
                   res2$summaryTable[match(res1$summaryTable$mutantName[idx], 
                                           res2$summaryTable$mutantName), ],
                   ignore_attr = TRUE)
  expect_equal(sum(res1$summaryTable[-idx, "nbrReads"]), 
               res2$summaryTable[res2$summaryTable$mutantName == 
                                     paste0("f.0.WT_", names(leu2)[1], ".0.WT"), "nbrReads"])
  
  ## ---------------------------------------------------------------------------
  ## Too many WT hits, reverse reads
  ## ---------------------------------------------------------------------------
  Ldef <- list(
      fastqForward = fqt1, fastqReverse = fqt2, 
      mergeForwardReverse = FALSE, 
      minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
      revComplForward = FALSE, revComplReverse = FALSE,
      elementsForward = "SPV", elementsReverse = "SPSS",
      elementLengthsForward = c(-1, -1, -1),
      elementLengthsReverse = c(-1, -1, 96, -1),
      adapterForward = "", 
      adapterReverse = "",
      primerForward = "GTCAGGTGGAGGCGGATCC",
      primerReverse = "GAAAAAGGAAGCTGGAGAGA",
      wildTypeForward = leu,
      wildTypeReverse = c(A = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT", B = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTTA"), 
      constantForward = "", 
      constantReverse = "", 
      avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
      variableNMaxForward = 0, variableNMaxReverse = 0, 
      umiNMax = 0,
      nbrMutatedCodonsMaxForward = 0,
      nbrMutatedCodonsMaxReverse = 0,
      nbrMutatedBasesMaxForward = -1,
      nbrMutatedBasesMaxReverse = -1,
      forbiddenMutatedCodonsForward = "",
      forbiddenMutatedCodonsReverse = "NNW",
      useTreeWTmatch = FALSE,
      mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
      mutNameDelimiter = ".", 
      constantMaxDistForward = -1,
      constantMaxDistReverse = -1,
      umiCollapseMaxDist = 0,
      filteredReadsFastqForward = "",
      filteredReadsFastqReverse = "",
      maxNReads = -1, verbose = FALSE,
      nThreads = 1, chunkSize = 1000, 
      maxReadLength = 1024
  )
  res1 <- do.call(digestFastqs, Ldef)
  
  expect_identical(res1$filterSummary$f8_nbrTooManyBestWTHits, 809L)
  
  ## ---------------------------------------------------------------------------
  ## Too short WT hit, reverse reads
  ## ---------------------------------------------------------------------------
  Ldef <- list(
      fastqForward = fqt1, fastqReverse = fqt2, 
      mergeForwardReverse = FALSE, 
      minOverlap = 0, maxOverlap = 0, maxFracMismatchOverlap = 0, greedyOverlap = TRUE, 
      revComplForward = FALSE, revComplReverse = FALSE,
      elementsForward = "SPV", elementsReverse = "SPVS",
      elementLengthsForward = c(-1, -1, -1),
      elementLengthsReverse = c(-1, -1, 96, -1),
      adapterForward = "", 
      adapterReverse = "",
      primerForward = "GTCAGGTGGAGGCGGATCC",
      primerReverse = "GAAAAAGGAAGCTGGAGAGA",
      wildTypeForward = leu,
      wildTypeReverse = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCT", 
      constantForward = "", 
      constantReverse = "", 
      avePhredMinForward = 20.0, avePhredMinReverse = 20.0,
      variableNMaxForward = 0, variableNMaxReverse = 0, 
      umiNMax = 0,
      nbrMutatedCodonsMaxForward = 0,
      nbrMutatedCodonsMaxReverse = 1,
      nbrMutatedBasesMaxForward = -1,
      nbrMutatedBasesMaxReverse = -1,
      forbiddenMutatedCodonsForward = "",
      forbiddenMutatedCodonsReverse = "NNW",
      useTreeWTmatch = FALSE,
      mutatedPhredMinForward = 0.0, mutatedPhredMinReverse = 0.0,
      mutNameDelimiter = ".", 
      constantMaxDistForward = -1,
      constantMaxDistReverse = -1,
      umiCollapseMaxDist = 0,
      filteredReadsFastqForward = "",
      filteredReadsFastqReverse = "",
      maxNReads = -1, verbose = FALSE,
      nThreads = 1, chunkSize = 1000, 
      maxReadLength = 1024
  )
  res1 <- do.call(digestFastqs, Ldef)
  
  expect_identical(res1$filterSummary$f3_nbrReadWrongLength, 439L)
  expect_identical(res1$filterSummary$nbrRetained, 0L)
})

---
title: "Multiplexed assays of variant effect analysis with `mutscan`"
author: "Charlotte Soneson and Michael Stadler"
date: "`r BiocStyle::doc_date()`"
package: "`r BiocStyle::pkg_ver('mutscan')`"
output: 
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{mutscan}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r chunk-opts, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

```{r setup, message = FALSE}
library(SummarizedExperiment)
library(mutscan)
```

# Introduction

`mutscan` provides functionality for processing and visualizing Multiplexed 
Assays of Variant Effect (MAVE) and other similar types of data, starting from 
single-end or paired-end FASTQ files. 
A broad range of library designs can be processed, encompassing experiments 
considering either a single protein or combinations of multiple proteins (e.g., 
aimed at studying protein-protein interactions).

The figure below provides a high-level overview of the `mutscan` functionality,
which will be described in more detail in the following sections. 

```{r overview-figure, fig.align = 'center', out.width = "100%", fig.cap = "Overview of the `mutscan` functionality. The `digestFastqs()` function processes each sequencing library, possibly consisting of multiple (pairs of) FASTQ files, separately, and generates an output object containing, among other things, the count table and a summary of the filtering steps. The `summarizeExperiment()` function takes one or more of these objects and combines them into a `SummarizedExperiment` object, that can then be used for downstream analysis such as plotting and statistical testing.", echo = FALSE}
knitr::include_graphics("mutscan-overview.001.png")
```


# Example data

The `mutscan` package contains several small example FASTQ files 
representing different types of experiments:

```{r list-files}
datadir <- system.file("extdata", package = "mutscan")
dir(datadir)
```

- `transInput_{1,2}.fastq.gz`, `transOutput_{1,2}.fastq.gz` - data from TRANS
experiment in [@Diss2018]. The forward and reverse reads correspond to the
mutated _FOS_ and _JUN_ sequences, respectively. Each read consists of a UMI
sequence, followed by a constant sequence and the variable region.
- `cisInput_{1,2}.fastq.gz`, `cisOutput_{1,2}.fastq.gz` - data from CIS
experiment in [@Diss2018]. The forward and reverse reads both correspond to the
mutated _FOS_ sequence. Each read consists of a UMI sequence, followed by a
constant sequence and the variable region.
- `GSE102901_cis_se.rds` - a `SummarizedExperiment` object obtained by
processing the full CIS data from [@Diss2018].
- `leujunt0_{1,2}.fastq.gz` - unpublished data; the forward read corresponds to
an unmutated sequence of one of 46 leucine zipper sequences, and the reverse
read corresponds to mutated _JUN_ sequences. Each read contains a
(frame-shifted) primer sequence, followed by the variable region.

# Processing Multiplexed Assays of Variant Effect data

## Read composition specification

The main function for processing the Multiplexed Assays of Variant Effect 
FASTQ files is `digestFastqs()`. This function requires the specification of 
paths to compressed FASTQ file(s) and the composition of the reads in these 
files. The composition is specified by the user, and is given in the form of 
a character string indicating the parts constituting the respective read, and 
an integer vector specifying the lengths of the individual parts. The 
permitted components are: 

* `S` (skip) - nucleotide(s) to skip
* `U` (UMI) - UMI sequence 
* `C` (constant) - constant sequence, can be used e.g. to estimate the 
sequencing error rate if available
* `V` (variable) - nucleotides corresponding to the variable region
* `P` (primer) - a primer sequence, which may appear in a nonspecified 
position in the read but whose position defines the start or end 
position of other components. 

A read can have several components of the same type (e.g., skipped nucleotides
both in the beginning or the end, or two variable regions separated by 
a primer). In such cases, `mutscan` will concatenate all components of the 
same type in the processing, while recording the lengths of the individual 
components.

As an example, a read with the following composition:

[1 skipped nt] - [10 nt UMI] - [18 nt constant sequence] - [96 nt variable region] 

would be specified to `digestFastqs()` by an elements string `"SUCV"`, 
and an element length vector `c(1, 10, 18, 96)`. 

The package can also accommodate designs with primer sequences. In this 
situation, the primer acts as an 'anchor', and the read composition before and 
after the primer is specified. For example, a read with the following 
composition:

[unknown sequence] - [10 nt primer] - [variable region, constituting the remainder of the read]

would be represented by an elements string `"SPV"`, and an element length 
vector `c(-1, 10, -1)`, where the -1 indicates that the corresponding read part
consists of the remaining part of the read, not accounted for by any of the 
other specified parts. In addition, the sequence of the primer must be 
specified, and any read where the primer is not present will be discarded. 

The forward and reverse reads can have different compositions. The user can 
also specify whether the variable parts of the forward and/or
reverse reads should be reverse complemented before being processed, and 
whether the variable regions of the forward and reverse reads should be merged 
into a single consensus sequence.

## Filtering 

In addition to reading the FASTQ files, the `digestFastqs()` function will 
perform a series of filtering steps, in the following order:

* Search for perfect matches to forward/reverse adapter sequences, filter out
the read pair if a match is found in either the forward or reverse read.
* Search for perfect matches to forward/reverse primer sequences, filter out 
the read pair if not both are found.
* Filter out reads whose length is not compatible with the indicated 
composition.
* If forward and reverse variable regions should be merged, filter out read
pairs where no valid overlap could be found.
* Filter out read pair if the average quality in the variable region is below
`avePhredMin` in either the forward or reverse read (or the consensus sequence
if they are merged).
* Filter out read pair if the number of Ns in the variable region exceeds
`variableNMax`.
* Filter out read pair if the number of Ns in the combined forward and reverse
UMI sequence exceeds `umiNMax`.
* If wild type (reference) sequences are provided (via `wildTypeForward` and/or 
`wildTypeReverse`), filter out reads if there are multiple best matches among 
the provided wild type sequences.
* Filter out read pair if any mutated base has a quality below
`mutatedPhredMin`.
* Filter out read pair if the number of mutated codons exceeds
`nbrMutatedCodonsMax` or the number of mutated nucleotides exceeds 
`nbrMutatedBasesMax`.
* Filter out read pair if any of the mutated codons match any of the
codons encoded by `forbiddenMutatedCodons`.
* Filter out read pair if there are too many mutations in the constant 
sequence, or if there are multiple equally close best hits among the 
provided constant sequences. 

Reads that are filtered out in one step will not be processed further (and 
thus will not contribute to the count of reads being filtered out for any 
other reasons). 

# Processing TRANS data

Here, we illustrate the processing of the provided TRANS experiment example
data. We filter out reads with any adapter match, an average phred quality 
below 20, any Ns in the UMI or variable sequence, more than one mutated codon, 
or a mutated codon ending with A or T (represented by the `NNW` value of the
`forbiddenMutatedCodonsForward` and `forbiddenMutatedCodonsReverse` arguments). 
In order to identify the number of mutations, we have to provide the wild type 
sequences of _FOS_ and _JUN_ (forward and reverse reads, respectively). This 
will additionally name the final mutants according to the positions and 
sequences of the mutated codons. 

```{r trans-digest}
transInput <- digestFastqs(
    fastqForward = file.path(datadir, "transInput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "transInput_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV",
    elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1,
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = FALSE
)
```

The `digestFastqs()` function returns a list with four elements. The 
`parameters` list records all parameter values used during the processing, as 
well as the `mutscan` version and time of processing.

```{r trans-pars}
transInput$parameters
```

The `filterSummary` data.frame contains a summary of the number of reads
filtered out in each step. Note that for any given filtering step, only the
reads retained by the previous filters are considered. The numbers in the 
filter column names indicate the order of the filters.

```{r trans-filter}
transInput$filterSummary
```

The `summaryTable` provides the number of reads and unique UMIs observed for
each variable sequence pair. The underscore in the strings in the sequence 
column indicate the separation between the forward and reverse wild type 
sequences. In addition, the table contains a column `mutantName`, which
provides a shorthand notation for each mutant. The format of the values in this
column is a combination of `WTS.xx.NNN` (separated by `_`), where `WTS` 
provides the name of the closest matching wild type sequence (if only one, 
unnamed wild type sequence is provided, the name will be `f/r` depending on 
if it corresponds to the forward/reverse read, respectively). The `xx` part 
indicates the mutated codon or nucleotide number, and `NNN` gives the 
observed sequence for the mutated codon or nucleotide. A sequence without 
mutations is named `WTS.0.WT`, where, again, `WTS` is the name of the 
matching wild type sequence.

```{r trans-summary}
head(transInput$summaryTable)
```

Finally, the `errorStatistics` data.frame lists the number of matching and
mismatching bases in the constant sequences, stratified by the phred quality
score (from 0 to 99).

```{r trans-error}
transInput$errorStatistics[rowSums(transInput$errorStatistics[, -1]) != 0, ]
```

The `errorStatistics` output can be used to estimate the sequencing error rate:

```{r trans-error-stats}
(propErrorsConstantF <- sum(transInput$errorStatistics$nbrMismatchForward) /
   (nchar(transInput$parameters$constantForward) * 
        transInput$filterSummary$nbrRetained))
(propErrorsConstantR <- sum(transInput$errorStatistics$nbrMismatchReverse) /
   (nchar(transInput$parameters$constantReverse) * 
        transInput$filterSummary$nbrRetained))
```

# Processing CIS data

Next, we illustrate the processing of the provided CIS experiment example data.
Recall that in this case, both the forward and reverse variable sequence 
correspond to the same mutated _FOS_ sequence, and before matching to the 
wild type sequence, the two components need to be merged to generate a single 
variable sequence (this is specified via the `mergeForwardReverse` argument).
As for the TRANS data, we filter out reads with any adapter match, an average
phred quality below 20, any Ns in the UMI or variable sequence, more than one
mutated codon, or a mutated codon ending with A or T (represented by the `NNW`
value of the `forbiddenMutatedCodonsForward` argument). Since the 
forward and reverse variable sequences are merged into one variable 
sequence, only one wild type sequence is provided (the `wildTypeReverse`
argument will be ignored if specified).

```{r cis-digest}
cisInput <- digestFastqs(
    fastqForward = file.path(datadir, "cisInput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "cisInput_2.fastq.gz"),
    mergeForwardReverse = TRUE, 
    minOverlap = 0, 
    maxOverlap = 0, 
    maxFracMismatchOverlap = 1, 
    greedyOverlap = TRUE, 
    revComplForward = FALSE, 
    revComplReverse = TRUE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV", 
    elementsReverse = "SUCVS",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 7, 17, 96, -1),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAGTTCATCCTGGCAGC", 
    primerForward = "", 
    primerReverse = "",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = "", 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = TRUE
)
```

```{r cis-out}
cisInput$parameters
cisInput$filterSummary
cisInput$errorStatistics[rowSums(cisInput$errorStatistics[, -1]) != 0, ]
```

The `summaryTable` now provides the number of reads and unique UMIs observed 
for each variable sequence, and all values in the `mutNames` column will start
with `FOS`.

```{r cis-summary}
head(cisInput$summaryTable)
```

# Processing TRANS data with primers

Finally, we illustrate the processing of the provided example data, where the 
first read corresponds to one of 46 leucine zipper sequences, and the second 
read is a mutated _JUN_ sequence. We first need to define the possible 
sequences for the forward read. If multiple wild type sequences are provided 
like here, `mutscan` will match each read against all of them and find the 
most similar one for each read.

```{r def-primers}
leu <- c(
    ATF2 = "GATCCTGATGAAAAAAGGAGAAAGTTTTTAGAGCGAAATAGAGCAGCAGCTTCAAGATGCCGACAAAAAAGGAAAGTCTGGGTTCAGTCTTTAGAGAAGAAAGCTGAAGACTTGAGTTCATTAAATGGTCAGCTGCAGAGTGAAGTCACCCTGCTGAGAAATGAAGTGGCACAGCTGAAACAGCTTCTTCTGGCT",
    ATF7 = "GATCCAGATGAGCGACGGCAGCGCTTTCTGGAGCGCAACCGGGCTGCAGCCTCCCGCTGCCGCCAAAAGCGAAAGCTGTGGGTGTCCTCCCTAGAGAAGAAGGCCGAAGAACTCACTTCTCAGAACATTCAGCTGAGTAATGAAGTCACATTACTACGCAATGAGGTGGCCCAGTTGAAACAGCTACTGTTAGCT",
    CREB5 = "GATCCGGACGAGAGGCGGCGGAAATTTCTGGAACGGAACCGGGCAGCTGCCACCCGCTGCAGACAGAAGAGGAAGGTCTGGGTGATGTCATTGGAAAAGAAAGCAGAAGAACTCACCCAGACAAACATGCAGCTTCAGAATGAAGTGTCTATGTTGAAAAATGAGGTGGCCCAGCTGAAACAGTTGTTGTTAACA",
    ATF3 = "GAAGAAGATGAAAGGAAAAAGAGGCGACGAGAAAGAAATAAGATTGCAGCTGCAAAGTGCCGAAACAAGAAGAAGGAGAAGACGGAGTGCCTGCAGAAAGAGTCGGAGAAGCTGGAAAGTGTGAATGCTGAACTGAAGGCTCAGATTGAGGAGCTCAAGAACGAGAAGCAGCATTTGATATACATGCTCAACCTT",
    JDP2 = "GAGGAAGAGGAGCGAAGGAAAAGGCGCCGGGAGAAGAACAAAGTCGCAGCAGCCCGATGCCGGAACAAGAAGAAGGAGCGCACGGAGTTTCTGCAGCGGGAATCCGAGCGGCTGGAACTCATGAACGCAGAGCTGAAGACCCAGATTGAGGAGCTGAAGCAGGAGCGGCAGCAGCTCATCCTGATGCTGAACCGA",
    ATF4 = "GAGAAACTGGATAAGAAGCTGAAAAAAATGGAGCAAAACAAGACAGCAGCCACTAGGTACCGCCAGAAGAAGAGGGCGGAGCAGGAGGCTCTTACTGGTGAGTGCAAAGAGCTGGAAAAGAAGAACGAGGCTCTAAAAGAGAGGGCGGATTCCCTGGCCAAGGAGATCCAGTACCTGAAAGATTTGATAGAAGAG",
    ATF5 = "ACCCGAGGGGACCGCAAGCAAAAGAAGAGAGACCAGAACAAGTCGGCGGCTCTGAGGTACCGCCAGCGGAAGCGGGCAGAGGGTGAGGCCCTGGAGGGCGAGTGCCAGGGGCTGGAGGCACGGAATCGCGAGCTGAAGGAACGGGCAGAGTCCGTGGAGCGCGAGATCCAGTACGTCAAGGACCTGCTCATCGAG",
    CREBZF = "AGTCCCCGGAAGGCGGCGGCGGCCGCTGCCCGCCTTAATCGACTGAAGAAGAAGGAGTACGTGATGGGGCTGGAGAGTCGAGTCCGGGGTCTGGCAGCCGAGAACCAGGAGCTGCGGGCCGAGAATCGGGAGCTGGGCAAACGCGTACAGGCACTGCAGGAGGAGAGTCGCTACCTACGGGCAGTCTTAGCCAAC",
    BATF2 = "CCCAAGGAGCAACAAAGGCAGCTGAAGAAGCAGAAGAACCGGGCAGCCGCCCAGCGAAGCCGGCAGAAGCACACAGACAAGGCAGACGCCCTGCACCAGCAGCACGAGTCTCTGGAAAAAGACAACCTCGCCCTGCGGAAGGAGATCCAGTCCCTGCAGGCCGAGCTGGCGTGGTGGAGCCGGACCCTGCACGTG",
    BATF3 = "GAGGATGATGACAGGAAGGTCCGAAGGAGAGAAAAAAACCGAGTTGCTGCTCAGAGAAGTCGGAAGAAGCAGACCCAGAAGGCTGACAAGCTCCATGAGGAATATGAGAGCCTGGAGCAAGAAAACACCATGCTGCGGAGAGAGATCGGGAAGCTGACAGAGGAGCTGAAGCACCTGACAGAGGCACTGAAGGAG",
    CEBPE = "AAAGATAGCCTTGAGTACCGGCTGAGGCGGGAGCGCAACAACATCGCCGTGCGCAAGAGCCGAGACAAGGCCAAGAGGCGCATTCTGGAGACGCAGCAGAAGGTGCTGGAGTACATGGCAGAGAACGAGCGCCTCCGCAGCCGCGTGGAGCAGCTCACCCAGGAGCTAGACACCCTCCGCAACCTCTTCCGCCAG",
    BACH1 = "CTGGATTGTATCCATGATATTCGAAGAAGAAGTAAAAACAGAATTGCTGCACAGCGCTGTCGCAAGAGAAAACTTGACTGTATACAGAATCTTGAATCAGAAATTGAGAAGCTGCAAAGTGAAAAGGAGAGCTTGTTGAAGGAAAGAGATCACATTTTGTCAACTCTGGGTGAGACAAAGCAGAACCTAACTGGA",
    BACH2 = "TTAGAGTTTATTCATGATGTCCGACGGCGCAGCAAGAACCGCATCGCGGCCCAGCGCTGCCGCAAAAGGAAACTGGACTGTATTCAGAATTTAGAATGTGAAATCCGCAAATTGGTGTGTGAGAAAGAGAAACTGTTGTCAGAGAGGAATCAACTGAAAGCATGCATGGGGGAACTGTTGGACAACTTCTCCTGC",
    NFE2L1 = "CTGAGCCTCATCCGAGACATCCGGCGCCGGGGCAAGAACAAGATGGCGGCGCAGAACTGCCGCAAGCGCAAGCTGGACACCATCCTGAATCTGGAGCGTGATGTGGAGGACCTGCAGCGTGACAAAGCCCGGCTGCTGCGGGAGAAAGTGGAGTTCCTGCGCTCCCTGCGACAGATGAAGCAGAAGGTCCAGAGC",
    NFE2 = "CTAGCGCTAGTCCGGGACATCCGACGACGGGGCAAAAACAAGGTGGCAGCCCAGAACTGCCGCAAGAGGAAGCTGGAAACCATTGTGCAGCTGGAGCGGGAGCTGGAGCGGCTGACCAATGAACGGGAGCGGCTTCTCAGGGCCCGCGGGGAGGCAGACCGGACCCTGGAGGTCATGCGCCAACAGCTGACAGAG",
    NFIL3 = "AAGAAAGATGCTATGTATTGGGAAAAAAGGCGGAAAAATAATGAAGCTGCCAAAAGATCTCGTGAGAAGCGTCGACTGAATGACCTGGTTTTAGAGAACAAACTAATTGCACTGGGAGAAGAAAACGCCACTTTAAAAGCTGAGCTGCTTTCACTAAAATTAAAGTTTGGTTTAATTAGCTCCACAGCATATGCT",
    FOS = "GAAGAAGAAGAGAAAAGGAGAATCCGAAGGGAAAGGAATAAGATGGCTGCAGCCAAATGCCGCAACCGGAGGAGGGAGCTGACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTAGAGTTCATCCTGGCAGCT",
    FOSB = "GAGGAAGAGGAGAAGCGAAGGGTGCGCCGGGAACGAAATAAACTAGCAGCAGCTAAATGCAGGAACCGGCGGAGGGAGCTGACCGACCGACTCCAGGCGGAGACAGATCAGTTGGAGGAAGAAAAAGCAGAGCTGGAGTCGGAGATCGCCGAGCTCCAAAAGGAGAAGGAACGTCTGGAGTTTGTGCTGGTGGCC",
    FOSL1 = "GAGGAAGAGGAGCGCCGCCGAGTAAGGCGCGAGCGGAACAAGCTGGCTGCGGCCAAGTGCAGGAACCGGAGGAAGGAACTGACCGACTTCCTGCAGGCGGAGACTGACAAACTGGAAGATGAGAAATCTGGGCTGCAGCGAGAGATTGAGGAGCTGCAGAAGCAGAAGGAGCGCCTAGAGCTGGTGCTGGAAGCC",
    FOSL2 = "GAAGAGGAGGAGAAGCGTCGCATCCGGCGGGAGAGGAACAAGCTGGCTGCAGCCAAGTGCCGGAACCGACGCCGGGAGCTGACAGAGAAGCTGCAGGCGGAGACAGAGGAGCTGGAGGAGGAGAAGTCAGGCCTGCAGAAGGAGATTGCTGAGCTGCAGAAGGAGAAGGAGAAGCTGGAGTTCATGTTGGTGGCT",
    MAFB = "GTGATCCGCCTGAAGCAGAAGCGGCGGACCCTGAAGAACCGGGGCTACGCCCAGTCTTGCAGGTATAAACGCGTCCAGCAGAAGCACCACCTGGAGAATGAGAAGACGCAGCTCATTCAGCAGGTGGAGCAGCTTAAGCAGGAGGTGTCCCGGCTGGCCCGCGAGAGAGACGCCTACAAGGTCAAGTGCGAGAAA",
    JUN = "CAGGAGCGGATCAAGGCGGAGAGGAAGCGCATGAGGAACCGCATCGCTGCCTCCAAGTGCCGAAAAAGGAAGCTGGAGAGAATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTTAAACAGAAAGTCATGAAC",
    JUNB = "CAAGAGCGCATCAAAGTGGAGCGCAAGCGGCTGCGGAACCGGCTGGCGGCCACCAAGTGCCGGAAGCGGAAGCTGGAGCGCATCGCGCGCCTGGAGGACAAGGTGAAGACGCTCAAGGCCGAGAACGCGGGGCTGTCGAGTACCGCCGGCCTCCTCCGGGAGCAGGTGGCCCAGCTCAAACAGAAGGTCATGACC",
    JUND = "CAGGAGCGCATCAAGGCGGAGCGCAAGCGGCTGCGCAACCGCATCGCCGCCTCCAAGTGCCGCAAGCGCAAGCTGGAGCGCATCTCGCGCCTGGAAGAGAAAGTGAAGACCCTCAAGAGTCAGAACACGGAGCTGGCGTCCACGGCGAGCCTGCTGCGCGAGCAGGTGGCGCAGCTCAAGCAGAAAGTCCTCAGC",
    CREB3 = "GAACAAATTCTGAAACGTGTGCGGAGGAAGATTCGAAATAAAAGATCTGCTCAAGAGAGCCGCAGGAAAAAGAAGGTGTATGTTGGGGGTTTAGAGAGCAGGGTCTTGAAATACACAGCCCAGAATATGGAGCTTCAGAACAAAGTACAGCTTCTGGAGGAACAGAATTTGTCCCTTCTAGATCAACTGAGGAAA",
    HLF = "CTGAAGGATGACAAGTACTGGGCAAGGCGCAGAAAGAACAACATGGCAGCCAAGCGCTCCCGCGACGCCCGGAGGCTGAAAGAGAACCAGATCGCCATCCGGGCCTCGTTCCTGGAGAAGGAGAACTCGGCCCTCCGCCAGGAGGTGGCTGACTTGAGGAAGGAGCTGGGCAAATGCAAGAACATACTTGCCAAG",
    MAFG = "ATCGTCCAGCTGAAGCAGCGCCGGCGCACGCTCAAGAACCGCGGCTACGCTGCCAGCTGCCGCGTGAAGCGGGTGACGCAGAAGGAGGAGCTGGAGAAGCAGAAGGCGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCCTCAGAGAACGCCAGCATGAAGCTGGAGCTCGACGCGCTGCGCTCCAAGTACGAGGCG",
    MAFK = "GTGACCCGCCTGAAGCAGCGTCGGCGCACACTCAAGAACCGCGGCTACGCGGCCAGCTGCCGCATCAAGCGGGTGACGCAGAAGGAGGAGCTGGAGCGGCAGCGCGTGGAGCTGCAGCAGGAGGTGGAGAAGCTGGCGCGTGAGAACAGCAGCATGCGGCTGGAGCTGGACGCCCTGCGCTCCAAGTACGAGGCG",
    XBP1 = "AGCCCCGAGGAGAAGGCGCTGAGGAGGAAACTGAAAAACAGAGTAGCAGCTCAGACTGCCAGAGATCGAAAGAAGGCTCGAATGAGTGAGCTGGAACAGCAAGTGGTAGATTTAGAAGAAGAGAACCAAAAACTTTTGCTAGAAAATCAGCTTTTACGAGAGAAAACTCATGGCCTTGTAGTTGAGAACCAGGAG",
    ATF6 = "ATTGCTGTGCTAAGGAGACAGCAACGTATGATAAAAAATCGAGAATCCGCTTGTCAGTCTCGCAAGAAGAAGAAAGAATATATGCTAGGGTTAGAGGCGAGATTAAAGGCTGCCCTCTCAGAAAACGAGCAACTGAAGAAAGAAAATGGAACACTGAAGCGGCAGCTGGATGAAGTTGTGTCAGAGAACCAGAGG",
    ATF6B = "GCAAAGCTGCTGAAGCGGCAGCAGCGAATGATCAAGAACCGGGAGTCAGCCTGCCAGTCCCGGAGAAAGAAGAAAGAGTATCTGCAGGGACTGGAGGCTCGGCTGCAAGCAGTACTGGCTGACAACCAGCAGCTCCGCCGAGAGAATGCTGCCCTCCGGCGGCGGCTGGAGGCCCTGCTGGCTGAAAACAGCGAG",
    CEBPA = "AAGAACAGCAACGAGTACCGGGTGCGGCGCGAGCGCAACAACATCGCGGTGCGCAAGAGCCGCGACAAGGCCAAGCAGCGCAACGTGGAGACGCAGCAGAAGGTGCTGGAGCTGACCAGTGACAATGACCGCCTGCGCAAGCGGGTGGAACAGCTGAGCCGCGAACTGGACACGCTGCGGGGCATCTTCCGCCAG",
    CEBPB = "AAGCACAGCGACGAGTACAAGATCCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGATGCGCAACCTGGAGACGCAGCACAAGGTCCTGGAGCTCACGGCCGAGAACGAGCGGCTGCAGAAGAAGGTGGAGCAGCTGTCGCGCGAGCTCAGCACCCTGCGGAACTTGTTCAAGCAG",
    CEBPD = "CGCGGCAGCCCCGAGTACCGGCAGCGGCGCGAGCGCAACAACATCGCCGTGCGCAAGAGCCGCGACAAGGCCAAGCGGCGCAACCAGGAGATGCAGCAGAAGTTGGTGGAGCTGTCGGCTGAGAACGAGAAGCTGCACCAGCGCGTGGAGCAGCTCACGCGGGACCTGGCCGGCCTCCGGCAGTTCTTCAAGCAG",
    CEBPG = "CGAAACAGTGACGAGTATCGGCAACGCCGAGAGAGGAACAACATGGCTGTGAAAAAGAGCCGGTTGAAAAGCAAGCAGAAAGCACAAGACACACTGCAGAGAGTCAATCAGCTCAAAGAAGAGAATGAACGGTTGGAAGCAAAAATCAAATTGCTGACCAAGGAATTAAGTGTACTCAAAGATTTGTTTCTTGAG",
    CREB1 = "GAAGCAGCACGAAAGAGAGAGGTCCGTCTAATGAAGAACAGGGAAGCAGCTCGAGAGTGTCGTAGAAAGAAGAAAGAATATGTGAAATGTTTAGAAAACAGAGTGGCAGTGCTTGAAAATCAAAACAAGACATTGATTGAGGAGCTAAAAGCACTTAAGGACCTTTACTGCCACAAATCAGAT",
    CREB3L1 = "GAGAAGGCCTTGAAGAGAGTCCGGAGGAAAATCAAGAACAAGATCTCAGCCCAGGAGAGCCGTCGTAAGAAGAAGGAGTATGTGGAGTGTCTAGAAAAGAAGGTGGAGACATTTACATCTGAGAACAATGAACTGTGGAAGAAGGTGGAGACCCTGGAGAATGCCAACAGGACCCTGCTCCAGCAGCTGCAGAAA",
    CREB3L2 = "GAGAAGGCCCTGAAGAAAATTCGGAGGAAGATCAAGAATAAGATTTCTGCTCAGGAAAGTAGGAGAAAGAAGAAAGAATACATGGACAGCCTGGAGAAAAAAGTGGAGTCTTGTTCAACTGAGAACTTGGAGCTTCGGAAGAAGGTAGAGGTTCTAGAGAACACTAATAGGACTCTCCTTCAGCAACTCCAGAAG",
    CREB3L3 = "GAGCGAGTGCTGAAAAAAATCCGCCGGAAAATCCGGAACAAGCAGTCGGCGCAAGAAAGCAGGAAGAAGAAGAAGGAATATATCGATGGCCTGGAGACTCGGATGTCAGCTTGCACTGCTCAGAATCAGGAGTTACAGAGGAAAGTCTTGCATCTCGAGAAGCAAAACCTGTCCCTCTTGGAGCAACTGAAGAAA",
    CREB3L4 = "GAGAGGGTCCTCAAGAAGGTCAGGAGGAAAATCCGTAACAAGCAGTCAGCTCAGGACAGTCGGCGGCGGAAGAAGGAGTACATTGATGGGCTGGAGAGCAGGGTGGCAGCCTGTTCTGCACAGAACCAAGAATTACAGAAAAAAGTCCAGGAGCTGGAGAGGCACAACATCTCCTTGGTAGCTCAGCTCCGCCAG",
    CREBL2 = "CCAGCCAAAATTGACTTGAAAGCAAAACTTGAGAGGAGCCGGCAGAGTGCAAGAGAATGCCGAGCCCGAAAAAAGCTGAGATATCAGTATTTGGAAGAGTTGGTATCCAGTCGAGAAAGAGCTATATGTGCCCTCAGAGAGGAACTGGAAATGTACAAGCAGTGGTGCATGGCAATGGACCAAGGAAAAATCCCT",
    CREBRF = "CCCTTAACAGCCCGACCAAGGTCAAGGAAGGAAAAAAATAAGCTGGCTTCCAGAGCTTGTCGGTTAAAGAAGAAAGCCCAGTATGAAGCTAATAAAGTGAAATTATGGGGCCTCAACACAGAATATGATAATTTATTGTTTGTAATCAACTCCATCAAGCAAGAGATTGTAAACCGGGTACAGAATCCAAGAGAT",
    DBP = "CAGAAGGATGAGAAATACTGGAGCCGGCGGTACAAGAACAACGAGGCAGCCAAGCGGTCCCGTGACGCCCGGCGGCTCAAGGAGAACCAGATATCGGTGCGGGCGGCCTTCCTGGAGAAGGAGAACGCCCTGCTGCGGCAGGAAGTTGTGGCCGTGCGCCAGGAGCTGTCCCACTACCGCGCCGTGCTGTCCCGA",
    NFE2L2 = "CTTGCATTAATTCGGGATATACGTAGGAGGGGTAAGAATAAAGTGGCTGCTCAGAATTGCAGAAAAAGAAAACTGGAAAATATAGTAGAACTAGAGCAAGATTTAGATCATTTGAAAGATGAAAAAGAAAAATTGCTCAAAGAAAAAGGAGAAAATGACAAAAGCCTTCACCTACTGAAAAAACAACTCAGCACC",
    NFE2L3 = "GTCTCACTTATCCGTGACATCAGACGAAGAGGGAAAAATAAAGTTGCTGCGCAGAACTGTCGTAAACGCAAATTGGACATAATTTTGAATTTAGAAGATGATGTATGTAACTTGCAAGCAAAGAAGGAAACTCTTAAGAGAGAGCAAGCACAATGTAACAAAGCTATTAACATAATGAAACAGAAACTGCATGAC",
    TEF = "CAGAAGGATGAAAAGTACTGGACAAGACGCAAGAAGAACAACGTGGCAGCTAAACGGTCACGGGATGCCCGGCGCCTGAAAGAGAATCAGATCACCATCCGGGCAGCCTTCCTGGAGAAGGAGAACACAGCCCTGCGGACGGAGGTGGCCGAGCTACGCAAGGAGGTGGGCAAGTGCAAGACCATCGTGTCCAAG"
)
```

Next we process the data, not allowing any mismatches in the forward read, but 
1 mismatching codon in the reverse read. Now, we assume that the variable 
sequence starts immediately after the provided primers, and hence we don't 
specify any UMI/constant sequence lengths. For the forward read, the variable 
region is taken to be the remainder of the read (after the primer), whereas 
for the reverse read, we specify the variable sequence length to 96.

```{r primers-digest}
leujunt0 <- digestFastqs(
    fastqForward = file.path(datadir, "leujunt0_1.fastq.gz"),
    fastqReverse = file.path(datadir, "leujunt0_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    elementsForward = "SPV", 
    elementsReverse = "SPVS",
    elementLengthsForward = c(-1, 19, -1),
    elementLengthsReverse = c(-1, 20, 96, -1),
    constantForward = "", 
    constantReverse = "", 
    adapterForward = "",
    adapterReverse = "", 
    primerForward = "GTCAGGTGGAGGCGGATCC", 
    primerReverse = "GAAAAAGGAAGCTGGAGAGA",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20, 
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = leu, 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 0, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "",
    forbiddenMutatedCodonsReverse = "NNW",
    mutatedPhredMinForward = 0.0, 
    mutatedPhredMinReverse = 0.0,
    mutNameDelimiter = ".", 
    verbose = TRUE
)
```

```{r primers-out}
leujunt0$parameters
leujunt0$filterSummary
```

```{r primers-summary}
head(leujunt0$summaryTable)
```

# Combining samples

The `digestFastqs()` function processes each sample (pair of FASTQ files)
separately. In order to prepare the data for downstream statistical analysis
and plotting, the `summarizeExperiment()` function takes a named list 
of outputs from `digestFastqs()`, and a `data.frame` with sample annotations, 
and generates a `SummarizedExperiment` object, 
with an `assay` containing either UMI or read counts. To illustrate this, we 
process also the output sample for the TRANS experiment for which we 
processed the input sample above, and feed both outputs to 
`summarizeExperiment()`.

```{r combine-digest}
transOutput <- digestFastqs(
    fastqForward = file.path(datadir, "transOutput_1.fastq.gz"),
    fastqReverse = file.path(datadir, "transOutput_2.fastq.gz"),
    mergeForwardReverse = FALSE, 
    revComplForward = FALSE, 
    revComplReverse = FALSE,
    adapterForward = "GGAAGAGCACACGTC", 
    adapterReverse = "GGAAGAGCGTCGTGT", 
    elementsForward = "SUCV", 
    elementsReverse = "SUCV",
    elementLengthsForward = c(1, 10, 18, 96),
    elementLengthsReverse = c(1, 8, 20, 96),
    constantForward = "AACCGGAGGAGGGAGCTG",
    constantReverse = "GAAAAAGGAAGCTGGAGAGA", 
    primerForward = "", 
    primerReverse = "",
    avePhredMinForward = 20, 
    avePhredMinReverse = 20,
    variableNMaxForward = 0, 
    variableNMaxReverse = 0, 
    umiNMax = 0,
    wildTypeForward = c(FOS = "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTCTGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA"), 
    wildTypeReverse = c(JUN = "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTCGGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT"), 
    nbrMutatedCodonsMaxForward = 1, 
    nbrMutatedCodonsMaxReverse = 1, 
    forbiddenMutatedCodonsForward = "NNW", 
    forbiddenMutatedCodonsReverse = "NNW",
    mutNameDelimiter = ".", 
    constantMaxDistForward = -1,
    constantMaxDistReverse = -1,
    verbose = FALSE
)

## Generate SummarizedExperiment object
se <- summarizeExperiment(
    x = list(sample1 = transInput,
             sample2 = transOutput),
    coldata = data.frame(Name = c("sample1", "sample2"),
                         Condition = c("input", "output"),
                         Replicate = c("R1", "R1"),
                         OD = c(0.05, 1.5))
)

## The SummarizedExperiment contains a count matrix, and annotations 
## for samples and variants in the `colData` and `rowData`, 
## respectively.
head(assay(se, "counts"))
Matrix::colSums(assay(se, "counts"))
head(rowData(se))
colData(se)

## Count type (reads or UMIs)
metadata(se)$countType
```

# Collapsing count matrix to amino acids

The \link{\code[SummarizedExperiment]{SummarizedExperiment}} object generated 
by `summarizeExperiment()` contains one row for each observed variant
(combination). This can be further collapsed by replacing the mutated codon by
the corresponding amino acid, and aggregating the counts corresponding to the
same mutated amino acid (combination).

```{r collapse-aa}
se_collapsed <- collapseMutantsByAA(se)
head(assay(se_collapsed, "counts"))
Matrix::colSums(assay(se_collapsed, "counts"))
colData(se_collapsed)
```

# Diagnostic plots

`mutscan` contains functionality for 
generating a variety of diagnostic plots. Here we illustrate these using the 
full CIS data set from [@Diss2018], which has been processed using 
`digestFastqs()` as illustrated above, and summarized in a 
`SummarizedExperiment` object provided with the package. 

```{r read-se}
se <- readRDS(file.path(datadir, "GSE102901_cis_se.rds"))
```

First, we can plot a summary of the filtering 
process, indicating the number of reads that were filtered out by 
(or retained after) each step of the `mutscan` filtering. 

```{r plot-filter-reads, fig.height=10, fig.width=8}
plotFiltering(se, valueType = "reads", onlyActiveFilters = TRUE, 
              plotType = "remaining", facetBy = "sample", numberSize = 3)
```

```{r plot-filter-frac, fig.height=10, fig.width=8}
plotFiltering(se, valueType = "fractions", onlyActiveFilters = TRUE,
              plotType = "filtered", facetBy = "step", numberSize = 3)
```

We can also generate a pairs plot displaying the correlation among the samples 
in the data set. 

```{r plot-pairs, fig.height=7}
plotPairs(se, selAssay = "counts")
```

Additional plot functions can be used to visualize the total count per sample, 
across all variants, or the distribution of variant counts per sample. 

```{r plot-more}
plotTotals(se, selAssay = "counts")
plotDistributions(se, selAssay = "counts", plotType = "density", 
                  pseudocount = 1)
```

Finally, we can create a full QC report as follows:

```{r make-qc-report, eval=FALSE}
generateQCReport(se, outFile = tempfile(fileext = ".html"))
```


# Calculating fitness scores

The function `calculateFitnessScore()` can be used to calculate fitness scores 
as described in [@Diss2018]. The function requires the user to specify a 
`pairingCol`, containing the replicate ID for each sample; one or more 
`ODCols`, containing the optical density for each sample, and a 
`comparison`, which is a character vector of length 3 specifying the 
comparison to perform, of the form (`groupColumn`, `numerator`, 
`denominator`). Here, `groupColumn` is the name of the column in `colData(se)` 
that contains the grouping information, and `numerator` and `denominator` 
specify the values of this column representing the two groups to be compared.

Here, we illustrate the application of `calculateFitnessScore()` on the
SummarizedExperiment object containing all the three CIS
replicates from [@Diss2018].

```{r calc-ppi}
se_collapsed <- collapseMutantsByAA(se)
ppis <- calculateFitnessScore(
    se = se_collapsed, pairingCol = "Replicate", 
    ODCols = c("OD1", "OD2"),
    comparison = c("Condition", "cis_output", "cis_input"),
    WTrows = "f.0.WT"
)
head(ppis[order(abs(rowMeans(ppis)), decreasing = TRUE), , drop = FALSE])

## The PPI score for the WT sequence is 1, by construction
ppis["f.0.WT", , drop = FALSE]
```

# Scoring mutations with `edgeR` or `limma`

As an alternative to the fitness scoring, `mutscan` can be used to model the
observed counts using a generalized linear model (with `edgeR`) or a general 
linear model (with `limma`) and estimate a logFC and a p-value for the 
enrichment of each variant betwen two conditions (or more generally, in 
association with any predictor), compared to one or more WT sequences. Note 
that for this, at least two replicates are required.

We start by looking at the design matrix, in order to determine which of the 
coefficients to specify for the testing in `calculateRelativeFC()`. For more 
information about how to set up and interpret design matrices in `edgeR` or 
`limma`, see e.g. 
[Law et al (2020)](https://f1000research.com/articles/9-1444), 
or [Soneson et al (2020)](https://f1000research.com/articles/9-512).

```{r def-mm}
model.matrix(~ Replicate + Condition,
             data = colData(se_collapsed))
```

Next, we apply either `edgeR` or `limma` to extract the logFCs of the mutants, 
compared to the wildtype sequence.

```{r calc-rel-fc}
## edgeR
edger_scores <- calculateRelativeFC(
    se = se_collapsed,
    design = model.matrix(~ Replicate + Condition,
                          data = colData(se_collapsed)),
    coef = "Conditioncis_output", pseudocount = 1, WTrows = "f.0.WT",
    method = "edgeR")
head(edger_scores[order(edger_scores$PValue), , drop = FALSE])
## As before, the WT sequence has a logFC close to 0, by construction
edger_scores["f.0.WT", , drop = FALSE]

## limma
limma_scores <- calculateRelativeFC(
    se = se_collapsed,
    design = model.matrix(~ Replicate + Condition,
                          data = colData(se_collapsed)),
    coef = "Conditioncis_output", pseudocount = 1, WTrows = "f.0.WT",
    method = "limma")
head(limma_scores[order(limma_scores$P.Value), , drop = FALSE])
## As before, the WT sequence has a logFC close to 0, by construction
limma_scores["f.0.WT", , drop = FALSE]
```

`mutscan` also contains functions for plotting the results of the statistical 
testing - in particular, MA (mean-difference) plots and volcano plots can be 
easily generated. 

```{r plot-rel-fc}
plotMeanDiff(edger_scores, pointSize = "large")
plotVolcano(edger_scores, pointSize = "large")
```

# FAQ

### Can `digestFastqs` process a sample where the reads are spread across multiple (pairs of) FASTQ files?

Yes, you can specify a vector of FASTQ files to `fastqForward`/`fastqReverse`. 
In this case, the reads from all (pairs of) files will be analyzed as if they 
came from a single FASTQ file. In case of paired-end data, take care to specify 
the forward and reverse FASTQ files in the same order. 

# Session info

This vignette was compiled on the following system:

```{r session-info}
sessionInfo()
```

---
title: "`r params$title`"
subtitle: "`r paste0('Generated with mutscan v', utils::packageVersion('mutscan'))`"
date: "`r Sys.time()`"
output: 
    html_document:
        theme: united
        toc: true
        toc_float: true
        code_folding: hide
params: 
    se: "unspecified"
    title: "mutscan QC report"
editor_options: 
  chunk_output_type: console
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r}
suppressPackageStartupMessages({
    library(SummarizedExperiment)
    library(mutscan)
    library(DT)
    library(dplyr)
})
```

# Data object

The input data use in this report is contained in the following 
SummarizedExperiment object:

```{r, class.source = "fold-show"}
params$se
```

# Sample annotation and filtering summary

The table below summarizes the available sample information, including 
annotations provided by the data analyst as well as filtering information 
collected by `mutscan`. 

```{r, warning = FALSE}
tmpdf <- as.data.frame(colData(params$se))
extradf <- do.call(
    dplyr::bind_rows, 
    lapply(names(metadata(params$se)$parameters), function(nm) {
        data.frame(Name = nm, 
                   lapply(metadata(params$se)$parameters[[nm]], 
                          function(x) paste(x, collapse = "; ")))
    })
)
DT::datatable(
    dplyr::full_join(tmpdf, extradf, by = "Name"),
    extensions = "FixedColumns",
    rownames = FALSE,
    options = list(
        paging = TRUE, searching = TRUE, info = FALSE,
        pageLength = 20, 
        sort = TRUE, scrollX = TRUE, fixedColumns = list(leftColumns = 1)
    )) %>%
    formatStyle("Name", "vertical-align" = "center")
```

The plots below focus on the filtering information, and show the number of 
reads remaining after each filtering step performed by `mutscan`, as well as the 
fraction of reads that are filtered out by each step.

```{r, fig.height = 1.5 * ncol(params$se), fig.width = 10}
tryCatch({mutscan::plotFiltering(
    params$se, valueType = "reads", onlyActiveFilters = TRUE, 
    displayNumbers = TRUE, plotType = "remaining", facetBy = "sample")},
    error = function(e) message("Couldn't generate filtering plots."))
```

```{r, fig.height = 1.5 * ncol(params$se), fig.width = 10}
tryCatch({mutscan::plotFiltering(
    params$se, valueType = "fractions", onlyActiveFilters = TRUE, 
    displayNumbers = TRUE, plotType = "filtered", facetBy = "step")},
    error = function(e) message("Couldn't generate filtering plots."))
```

# Pairs plot

Next, we show a pairs plot, displaying the correlation between each pair of
samples. 

```{r, fig.height = 10, fig.width = 10}
tryCatch({mutscan::plotPairs(params$se)},
         error = function(e) message("Couldn't generate pairs plot."))
```

# Total count plot

The plot below shows the total count for each sample, across all features
in the SummarizedExperiment object.

```{r}
tryCatch({mutscan::plotTotals(params$se, selAssay = "counts")},
         error = function(e) message("Couldn't generate total count plot."))
```

# Distribution plots

The next group of plots displays the distribution of counts for each sample 
(across all features in the SummarizedExperiment object). The same information 
is displayed as a 'knee plot', where the counts for each sample are 
arranged in decreasing order, and as a density plot. Note that both plots 
display the values on a log scale; hence any zero values are not shown. 

```{r, warning = FALSE}
tryCatch({mutscan::plotDistributions(
    params$se, selAssay = "counts", pseudocount = 1,
    groupBy = NULL, plotType = "knee", facet = FALSE)},
    error = function(e) message("Couldn't generate distribution plot."))
```

```{r, warning = FALSE}
tryCatch({mutscan::plotDistributions(
    params$se, selAssay = "counts", pseudocount = 1, 
    groupBy = NULL, plotType = "density", facet = FALSE)},
    error = function(e) message("Couldn't generate distribution plot."))
```

# Session info

<details>
<summary><b>
Click here to see the package versions used to generate this report.
</b></summary>
```{r}
sessionInfo()
```
</details>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarizeExperiment.R
\name{summarizeExperiment}
\alias{summarizeExperiment}
\title{Summarize and collapse multiple mutational scanning experiments}
\usage{
summarizeExperiment(x, coldata, countType = "umis")
}
\arguments{
\item{x}{A named list of objects returned by \code{\link{digestFastqs}}.
Names are used to link the objects to the metadata provided in
\code{coldata}.}

\item{coldata}{A \code{data.frame} with at least one column "Name", which
will be used to link to objects in \code{x}. A potentially subset and
reordered version of \code{coldata} is stored in the \code{colData} of 
the returned \code{\link[SummarizedExperiment]{SummarizedExperiment}}.}

\item{countType}{Either "reads" or "umis". If "reads", the "count" assay of
the returned object will contain the observed number of reads for each
sequence (pair). If "umis", the "count" assay will contain the number of
unique UMIs observed for each sequence (pair).}
}
\value{
A \code{\link[SummarizedExperiment]{SummarizedExperiment}} \code{x}
    with
    \describe{
        \item{assays(x)$counts}{containing the observed number of sequences 
        or sequence pairs (if \code{countType} = "reads"), or the observed 
        number of unique UMIs for each sequence or sequence pair (if 
        \code{countType} = "umis").}
        \item{rowData(x)}{containing the unique sequences or sequence 
        pairs.}
        \item{colData(x)}{containing the metadata provided by 
        \code{coldata}.}
    }
}
\description{
Combine multiple sequence lists (as returned by \code{\link{digestFastqs}}
into a \code{\link[SummarizedExperiment]{SummarizedExperiment}}, with
observed variable sequences (sequence pairs) in rows and samples in columns.
}
\examples{
## Input sample
inp <- digestFastqs(
    fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
                               package = "mutscan"), 
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    wildTypeForward = c(FOS = paste0(
        "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
        "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
    nbrMutatedCodonsMaxForward = 1
)
## Output sample
outp <- digestFastqs(
    fastqForward = system.file("extdata", "cisOutput_1.fastq.gz", 
                               package = "mutscan"), 
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    wildTypeForward = c(FOS = paste0(
        "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
        "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
    nbrMutatedCodonsMaxForward = 1
)
## Combine
se <- summarizeExperiment(
    x = list(r1inp = inp, r1outp = outp), 
    coldata = data.frame(Name = c("r1inp", "r1outp"), 
                         Condition = c("input", "output"), 
                         Replicate = c("rep1", "rep1")),
    countType = "umis"
)
se

}
\author{
Michael Stadler, Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotDistributions.R
\name{plotDistributions}
\alias{plotDistributions}
\title{Plot distribution of observed values}
\usage{
plotDistributions(
  se,
  selAssay = "counts",
  groupBy = NULL,
  plotType = "density",
  facet = FALSE,
  pseudocount = 0
)
}
\arguments{
\item{se}{A \code{SummarizedExperiment} object, typically generated by 
\code{summarizeExperiment()}.}

\item{selAssay}{Character scalar specifying the assay in \code{se} to 
use for the plotting.}

\item{groupBy}{Character scalar specifying a column from 
\code{colData(se)} to use for coloring or stratifying the plots.}

\item{plotType}{Character scalar specifying the type of plot to construct. 
Either \code{'density'}, \code{'histogram'} or \code{'knee'}.}

\item{facet}{Logical scalar, indicating whether or not to facet the plot
by the values specified in the \code{groupBy} column.}

\item{pseudocount}{Numeric scalar, representing the number to add to the 
observed values in the \code{selAssay} assay before plotting.}
}
\value{
A ggplot object.
}
\description{
Plot distribution of observed values
}
\examples{
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
                          package = "mutscan"))[1:200, ]
plotDistributions(se)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseMutantsByAA.R
\name{collapseMutantsBySimilarity}
\alias{collapseMutantsBySimilarity}
\alias{collapseMutantsByAA}
\alias{collapseMutants}
\title{Collapse mutants by similarity}
\usage{
collapseMutantsBySimilarity(
  se,
  assayName,
  scoreMethod = "rowSum",
  sequenceCol = "sequence",
  collapseMaxDist = 0,
  collapseMinScore = 0,
  collapseMinRatio = 0,
  verbose = TRUE
)

collapseMutantsByAA(se)

collapseMutants(se, nameCol)
}
\arguments{
\item{se}{A \code{\link[SummarizedExperiment]{SummarizedExperiment}}
generated by \code{\link{summarizeExperiment}}}

\item{assayName}{The name of the assay that will be used to calculate a 
"score" (typically derived from the read counts) for each variant.}

\item{scoreMethod}{Character scalar giving the approach used to calculate 
ranking scores from the assay defined by \code{assayName}. Currently, 
this can be one of \code{"rowSum"} or \code{"rowMean"}. All filtering
criteria will be applied to these scores.}

\item{sequenceCol}{Character scalar giving the name of the column in 
\code{rowData(se)} that contains the nucleotide sequence of the 
variants.}

\item{collapseMaxDist}{Numeric scalar defining the tolerance for collapsing 
similar sequences. If the value is in [0, 1), it defines the maximal 
Hamming distance in terms of a fraction of sequence length:
(\code{round(collapseMaxDist * nchar(sequence))}).
A value greater or equal to 1 is rounded and directly used as the 
maximum allowed Hamming distance. Note that sequences can only be
collapsed if they are all of the same length.}

\item{collapseMinScore}{Numeric scalar, indicating the minimum score for the 
sequence to be considered for collapsing with similar sequences.}

\item{collapseMinRatio}{Numeric scalar. During collapsing of
similar sequences, a low-frequency sequence will be collapsed 
with a higher-frequency sequence only if the ratio between the 
high-frequency and the low-frequency scores is at least this 
high. The default value of 0 indicates that no such check is performed.}

\item{verbose}{Logical, whether to print progress messages.}

\item{nameCol}{A character scalar providing the column of 
\code{rowData(se)} that contains the amino acid mutant names (that will 
be the new row names).}
}
\value{
A \code{\link[SummarizedExperiment]{SummarizedExperiment}} where
    counts have been aggregated by the mutated amino acid(s).
}
\description{
These functions can be used to collapse variants, either by similarity or 
according to a pre-defined grouping. The functions \code{collapseMutants} 
and \code{collapseMutantsByAA} assume that a grouping variable is available 
as a column in \code{rowData(se)} (\code{collapseMutantsByAA} is a 
convenience function for the case when this column is "mutantNameAA", and 
is provided for backwards compatibility). The 
\code{collapseMutantsBySimilarity} will generate the grouping variable 
based on user-provided thresholds on the sequence similarity (defined by 
the Hamming distance), and subsequently collapse based on the derived 
grouping.
}
\examples{
library(SummarizedExperiment)
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
                          package = "mutscan"))[1:200, ]
## The rows of this object correspond to individual codon variants
dim(se)
head(rownames(se))

## Collapse by amino acid
sec <- collapseMutantsByAA(se)
## The rows of the collapsed object correspond to amino acid variants
dim(sec)
head(rownames(sec))
## The mutantName column contains the individual codon variants that were 
## collapsed
head(rowData(sec))

## Collapse similar sequences
sec2 <- collapseMutantsBySimilarity(
    se = se, assayName = "counts", scoreMethod = "rowSum",
    sequenceCol = "sequence", collapseMaxDist = 2,
    collapseMinScore = 0, collapseMinRatio = 0)
dim(sec2)
head(rownames(sec2))
head(rowData(sec2))
## collapsed count matrix
assay(sec2, "counts")

}
\author{
Charlotte Soneson, Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculateRelativeFC.R
\name{calculateRelativeFC}
\alias{calculateRelativeFC}
\title{Calculate logFCs relative to WT using edgeR}
\usage{
calculateRelativeFC(
  se,
  design,
  coef = NULL,
  contrast = NULL,
  WTrows = NULL,
  selAssay = "counts",
  pseudocount = 1,
  method = "edgeR",
  normMethod = ifelse(is.null(WTrows), "TMM", "sum")
)
}
\arguments{
\item{se}{SummarizedExperiment object.}

\item{design}{Design matrix. The rows of the design matrix must be in the
same order as the columns in \code{se}.}

\item{coef}{Coefficient(s) to test with edgeR or limma.}

\item{contrast}{Numeric contrast to test with edgeR or limma.}

\item{WTrows}{Vector of row names that will be used as the reference when
calculating logFCs and statistics. If more than one value is provided, 
the sum of the corresponding counts is used to generate offsets. If 
\code{NULL}, offsets will be defined as the effective library sizes 
(using TMM normalization factors).}

\item{selAssay}{Assay to select from \code{se} for the analysis.}

\item{pseudocount}{Pseudocount to add when calculating log-fold changes.}

\item{method}{Either 'edgeR' or 'limma'. If set to 'limma', voom is used to
transform the counts and estimate observation weights before applying
limma. In this case, the results also contain the standard errors of the
logFCs.}

\item{normMethod}{Character scalar indicating which normalization method
should be used to calculate size factors. Should be either \code{"TMM"}
or \code{"csaw"} when \code{WTrows} is \code{NULL}, and \code{"geomean"}
or \code{"sum"} when \code{WTrows} is provided.}
}
\value{
A \code{data.frame} with output from the statistical testing 
framework (edgeR or limma).
}
\description{
Calculate logFCs and associated p-values for a given comparison, using
either limma or the Negative Binomial quasi-likelihood framework of edgeR.
The observed counts for the WT variants can be used as offsets in the model.
}
\examples{
library(SummarizedExperiment)
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
                          package = "mutscan"))[1:200, ]
design <- model.matrix(~ Replicate + Condition,
                       data = colData(se))
                              
## Calculate "absolute" log-fold changes with edgeR
res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
                           method = "edgeR")
head(res)
## Calculate log-fold changes relative to the WT sequence with edgeR
stopifnot("f.0.WT" \%in\% rownames(se))
res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
                           method = "edgeR", WTrows = "f.0.WT")
head(res)

## Calculate "absolute" log-fold changes with limma
res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
                           method = "limma")
head(res)
## Calculate log-fold changes relative to the WT sequence with limma
stopifnot("f.0.WT" \%in\% rownames(se))
res <- calculateRelativeFC(se, design, coef = "Conditioncis_output", 
                           method = "limma", WTrows = "f.0.WT")
head(res)

}
\author{
Charlotte Soneson, Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{calcNearestStringDist}
\alias{calcNearestStringDist}
\title{Calculate distances to the nearest string}
\usage{
calcNearestStringDist(x, metric = "hamming", nThreads = 1L)
}
\arguments{
\item{x}{A character vector.}

\item{metric}{A character scalar defining the string distance metric. One
of \code{"hamming"} (default), \code{"hamming_shift"} or
\code{"levenshtein"}.}

\item{nThreads}{numeric(1), number of threads to use for parallel processing.}
}
\value{
An integer vector of the same length as \code{x}.
}
\description{
Given a character vector, calculate the distance for each element
to the nearest neighbor amongst all the other elements.
}
\examples{
calcNearestStringDist(c("lazy", "hazy", "crazy"))
calcNearestStringDist(c("lazy", "hazy", "crazy"), metric = "hamming_shift")
calcNearestStringDist(c("lazy", "hazy", "crazy"), metric = "levenshtein")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotPairs.R
\name{plotPairs}
\alias{plotPairs}
\title{Make pairs plot of selected assay from a SummarizedExperiment object}
\usage{
plotPairs(
  se,
  selAssay = "counts",
  doLog = TRUE,
  pseudocount = 1,
  corMethod = "pearson",
  histBreaks = 40,
  pointsType = "points",
  corSizeMult = 5,
  corSizeAdd = 2,
  pointSize = 0.1,
  pointAlpha = 0.3,
  colorByCorrelation = TRUE,
  corrColorRange = NULL,
  addIdentityLine = FALSE
)
}
\arguments{
\item{se}{A SummarizedExperiment object, e.g. the output of
\code{summarizeExperiment}}

\item{selAssay}{Character scalar, the assay to use as the basis for the
pairs plot.}

\item{doLog}{Logical scalar, whether or not to log-transform the values
before plotting.}

\item{pseudocount}{Numeric scalar, the pseudocount to add to the values
before log-transforming (if \code{doLog} is \code{TRUE}).}

\item{corMethod}{Either "pearson" or "spearman", the type of correlation
to calculate.}

\item{histBreaks}{Numeric scalar, the number of breaks in the histograms
to put in the diagonal panels.}

\item{pointsType}{Either "points", "smoothscatter", "scattermore" or 
"scattermost" (the latter two require the "scattermore" package to be 
installed), determining the type of plots that will be made.}

\item{corSizeMult, corSizeAdd}{Numeric scalars determining how the
absolute correlation value is transformed into a font size. The
transformation is corSizeMult * abs(corr) + corSizeAdd.}

\item{pointSize, pointAlpha}{Numeric scalars determining the size and
opacity of points in the plot.}

\item{colorByCorrelation}{Logical scalar, indicating whether the correlation
panels should be colored according to the correlation value.}

\item{corrColorRange}{Numeric vector of length 2, providing the lower and 
upper limits of the color scale when coloring by correlation. Both 
values should be positive; the same range is used for negative 
correlations. If \code{NULL} (the default), the range is inferred from 
the data.}

\item{addIdentityLine}{Logical scalar, indicating whether the identity line 
should be added (only used if \code{pointsType = "points"}).}
}
\value{
A \code{ggplot} object.
}
\description{
Construct a pairs plot of all columns of a given assay. The lower-triangular
panels display the scatter plots, the upper-triangular ones print out
the (Pearson or Spearman) correlations, and the diagonal panels show
histograms of the respective columns.
}
\examples{
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
                          package = "mutscan"))[1:200, ]
plotPairs(se)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculateFitnessScore.R
\name{calculateFitnessScore}
\alias{calculateFitnessScore}
\title{Calculate fitness scores.}
\usage{
calculateFitnessScore(
  se,
  pairingCol,
  ODCols,
  comparison,
  WTrows,
  selAssay = "counts"
)
}
\arguments{
\item{se}{SummarizedExperiment object as returned by 
\code{\link{summarizeExperiment}}.}

\item{pairingCol}{Name of column in \code{colData(se)} with 
replicate/pairing information. Samples with the same value in this 
column will be paired.}

\item{ODCols}{Name(s) of column(s) in \code{colData(se)} with OD values 
(numeric), used to normalize for different numbers of cells.}

\item{comparison}{3-element character vector of the form 
\code{(column, numerator, denominator)}. \code{column} is the name of 
the column in \code{colData(se)} with experimental conditions. 
\code{numerator} and \code{denominator} define the comparison, 
e.g. \code{c("cond", "output", "input")} will look in the 
\code{"cond"} column and calculate fitness for the ratio of 
\code{"output"} over \code{"input"} counts.}

\item{WTrows}{Vector of row names that will be used as the reference when
calculating fitness scores. If more than one value is provided, the 
average of the corresponding fitness scores is used as a reference. 
If NULL, no division by WT scores will be done.}

\item{selAssay}{Assay to select from \code{se} for the analysis.}
}
\value{
A numeric vector with fitness scores.
}
\description{
Using sequence counts before and after selection, calculate fitness scores 
as described by Diss and Lehner (2018).
}
\examples{
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
                          package = "mutscan"))
## Check that the wildtype sequence is present in the data
stopifnot("f.0.WT" \%in\% rownames(se))
## Calculate PPI scores as defined in Diss & Lehner (2018)
ppis <- calculateFitnessScore(
    se = se, pairingCol = "Replicate", 
    ODCols = c("OD1", "OD2"),
    comparison = c("Condition", "cis_output", "cis_input"),
    WTrows = "f.0.WT")
## Matrix with PPI scores for each replicate
head(ppis)

}
\references{
"The genetic landscape of a physical interaction."
    Diss G and Lehner B. Elife. 2018;7:e32472. doi: 10.7554/eLife.32472.
}
\author{
Michael Stadler and Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotFiltering.R
\name{plotFiltering}
\alias{plotFiltering}
\title{Visualize the filtering procedure}
\usage{
plotFiltering(
  se,
  valueType = "reads",
  onlyActiveFilters = TRUE,
  displayNumbers = TRUE,
  numberSize = 4,
  plotType = "remaining",
  facetBy = "sample"
)
}
\arguments{
\item{se}{A \code{SummarizedExperiment} object, e.g. from 
\code{summarizeExperiment}.}

\item{valueType}{Either "reads" or "fractions", indicating whether to 
plot the number of reads, or the fraction of the total number of reads,
that are retained after/filtered out in each filtering step.}

\item{onlyActiveFilters}{Logical scalar, whether to only include the 
active filters (i.e., where any read was filtered out in any of the 
samples). Defaults to \code{TRUE}.}

\item{displayNumbers}{Logical scalar, indicating whether to display the 
number (or fraction) of reads retained at every filtering step.}

\item{numberSize}{Numeric scalar, indicating the size of the displayed 
numbers (if \code{displayNumbers} is \code{TRUE}).}

\item{plotType}{Character scalar, indicating what to show in the plot. 
Either \code{"remaining"} or \code{"filtered"}.}

\item{facetBy}{Character scalar, indicating the variable by which the plots 
should be facetted. Either \code{"sample"} or \code{"step"}.}
}
\value{
A \code{ggplot} object.
}
\description{
Display the number (or fraction) of reads remaining after each step 
of the internal \code{mutscan} filtering.
}
\details{
The function assumes that the number of reads filtered out in each step 
are provided as columns of \code{colData(se)}, with column names 
of the form \code{f[0-9]_filteringreason}, and that all filtering columns 
occur between the columns named \code{nbrTotal} and \code{nbrRetained}.
}
\examples{
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
                          package = "mutscan"))[1:200, ]
plotFiltering(se)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotResults.R
\name{plotVolcano}
\alias{plotVolcano}
\title{Construct a volcano plot}
\usage{
plotVolcano(
  res,
  logFCCol = NULL,
  pvalCol = NULL,
  padjCol = NULL,
  padjThreshold = 0.05,
  pointSize = "small",
  interactivePlot = FALSE,
  nTopToLabel = 0
)
}
\arguments{
\item{res}{\code{data.frame} (typically output from
\code{calculateRelativeFC()}) with columns corresponding to the
log-fold change (\code{logFC}), p-value (\code{PValue} or
\code{P.Value}) and significance (\code{FDR} or \code{adj.P.Val}).}

\item{logFCCol, pvalCol, padjCol}{Character scalars indicating the
columns from \code{res} that will be used to represent the
logFC (x-axis), p-value (y-axis) and adjusted p-value (used
for coloring). If \code{NULL} (default), pre-specified values
will be used depending on the available columns
(\code{"logFC"}, \code{"PValue"} or \code{"P.Value"}, and
\code{"FDR"} or \code{"adj.P.Val"}, respectively).}

\item{padjThreshold}{Numeric scalar indicating the adjusted p-value
threshold to use for coloring the points. All features with
adjusted p-value below the treshold will be shown in red.}

\item{pointSize}{Either \code{"small"} or \code{"large"}, indicating
which of the two available plot styles that will be used.}

\item{interactivePlot}{Logical scalar, indicating whether an
interactive plot should be returned, in which one can hover
over the individual points and obtain further information.}

\item{nTopToLabel}{Numeric scalar, indicating the number of points that
should be labeled in the plot. The points will be ranked by the
\code{pvalCol} column, and the top \code{nTopToLabel} values will
be labeled by the corresponding row names. Only used if
\code{interactivePlot} is \code{FALSE}.}
}
\value{
If \code{interactivePlot} is \code{TRUE}, a \code{plotly}
    object. If \code{interactivePlot} is \code{FALSE}, a \code{ggplot2}
    object.
}
\description{
Construct a volcano plot
}
\examples{
library(SummarizedExperiment)
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
                          package = "mutscan"))[1:200, ]
design <- model.matrix(~ Replicate + Condition,
                       data = colData(se))
res <- calculateRelativeFC(se, design, coef = "Conditioncis_output")
plotVolcano(res, pointSize = "large", nTopToLabel = 3)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotTotals.R
\name{plotTotals}
\alias{plotTotals}
\title{Plot the column totals of a selected assay}
\usage{
plotTotals(se, selAssay = "counts", groupBy = NULL)
}
\arguments{
\item{se}{A \code{SummarizedExperiment} object, typically generated by 
\code{summarizeExperiment()}.}

\item{selAssay}{Character scalar specifying the assay in \code{se} to 
use for the plotting.}

\item{groupBy}{Character scalar indicating a column in 
\code{rowData(se)} to group the features by before calculating the 
column sums.}
}
\value{
A ggplot object.
}
\description{
Plot the column totals of a selected assay
}
\examples{
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds", 
                          package = "mutscan"))[1:200, ]
plotTotals(se)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkMultipleVariants.R
\name{linkMultipleVariants}
\alias{linkMultipleVariants}
\title{Process an experiment with multiple variable sequences}
\usage{
linkMultipleVariants(combinedDigestParams = list(), ...)
}
\arguments{
\item{combinedDigestParams}{A named list of arguments to
\code{digestFastqs} for the combined ("naive") run.}

\item{...}{Additional arguments providing arguments to \code{digestFastqs}
for the separate runs (processing each variable sequence in turn).
Each argument must be a named list of arguments to \code{digestFastqs}.
In addition, arguments \code{collapseMaxDist}, \code{collapseMinScore}
and \code{collapseMinRatio} can be specified, and will be passed on 
to \code{collapseMutantsBySimilarity}.}
}
\value{
A list with the following elements: 
\itemize{
\item countAggregated - a \code{tibble} with columns corresponding to 
    each of the variable sequences, and a column with the total observed 
    read count for the combination.
\item convSeparate - a list of conversion tables from the respective 
    separate runs.
\item outCombined - the \code{digestFastqs} output for the combined run.
}
}
\description{
This function enables the processing of data sets with multiple
variable sequences, which should potentially be handled in different
ways. For example, a barcode association experiment
with two variable sequences (the barcode and the biological variant)
that need to be processed differently, e.g. in terms of matching to
wildtype sequences or collapsing of similar sequences.
In contrast, while \code{digestFastqs} allow the specification
of multiple variable sequences (within each of the forward and reverse
reads), they will be concatenated and processed as a single unit.
}
\details{
linkMultipleVariants will process the input in the following way:
\itemize{
\item First, run \code{digestFastqs} with the parameters provided
    in \code{combinedDigestParams}. Typically, this will be a
    "naive" counting run, where the frequencies of all observed
    variants are tabulated. The variable sequences
    within the forward and reverse reads, respectively, will be
    processed as a single sequence. 
\item Next, run \code{digestFastqs} with each of the additional
    parameter sets provided (\code{...}). Each of these should
    correspond to a single variable sequence from the combined
    run (i.e., if there are two Vs in the element specifications
    in the combined run, there should be two additional
    parameter sets provided, each corresponding to the
    processing of one variable sequence part). It is assumed
    that the order of the additional arguments correspond to the
    order of the variable sequences in the combined run, in such a way
    that if the variable sequences extracted in each of the separate
    runs are concatenated in the order that the parameter sets are
    provided to \code{linkMultipleVariants}, they will form the variable
    sequence extracted in the combined run.
\item The result of each of the separate runs is a 'conversion table',
    containing the final set of identified sequence variants as well
    as all individual sequences corresponding to each of them. This
    is then combined with the count table from the combined, "naive"
    run in order to create an aggregated count table. More precisely,
    each sequence in the combined run is split into the constituent
    variable sequences, and
    each variable sequence is then matched to the output from the right
    separate run, from which the final feature ID (mutant name, or
    collapsed sequence) will be extracted and used to replace the original
    sequence in the combined count table. Once all the matches are done,
    rows with NAs (where no match could be found in the separate run)
    are removed and the counts are aggregated across all identical
    combinations of variable sequences.
}

In order to define the \code{elementsForward} and \code{elementsReverse}
arguments for the separate runs, a strategy that often works is to simply
copy the arguments from the combined run, and successively replace all
but one of the 'V's by 'S'. This will effectively process one variable
sequence at the time, while keeping all other elements of the reads
consistent (since this can affect e.g. filtering criteria). Note that
to process individual variable sequences in the reverse read, you also
need to swap the 'forward' and 'reverse' specifications (since
\code{digestFastqs} requires a forward read).
}
\examples{
fqFile <- system.file("extdata", "cisInput_1.fastq.gz", 
                      package = "mutscan")
out <- linkMultipleVariants(
    combinedDigestParams = list(fastqForward = fqFile, 
                                elementsForward = "SVCV", 
                                elementLengthsForward = c(1, 10, 18, 96)),
    # the first variable sequence is the UMI
    umi = list(fastqForward = fqFile, elementsForward = "SVCS",
               elementLengthsForward = c(1, 10, 18, 96)),
    # the second variable sequence is the amplicon variant
    var = list(fastqForward = fqFile, elementsForward = "SSCV",
               elementLengthsForward = c(1, 10, 18, 96), 
               collapseMaxDist = 3, collapseMinScore = 1)
)
# conversion tables
lapply(out$convSeparate, head)
# aggregated count table
head(out$countAggregated)

}
\author{
Charlotte Soneson, Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generateQCReport.R
\name{generateQCReport}
\alias{generateQCReport}
\title{Generate QC report}
\usage{
generateQCReport(
  se,
  outFile,
  reportTitle = "mutscan QC report",
  forceOverwrite = FALSE,
  ...
)
}
\arguments{
\item{se}{A \code{SummarizedExperiment} object, typically generated with
\code{summarizeExperiment()}.}

\item{outFile}{Character string providing the name of the output file.
Should have the extension \code{.html}.}

\item{reportTitle}{Character string specifying the title of the QC report.}

\item{forceOverwrite}{Logical scalar, indicating whether an existing file
with the same name as \code{outFile} should be overwritten.}

\item{...}{Additional parameters to be forwarded to
\code{\link[rmarkdown]{render}}, for example \code{quiet = TRUE}.}
}
\value{
Invisibly, the path to the generated html file.
}
\description{
Generate QC report
}
\examples{
## Load SummarizedExperiment object
se <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
                          package = "mutscan"))
## Define output file
outfile <- tempfile(fileext = ".html")

## Generate QC report
generateQCReport(se, outfile)

}
\seealso{
\code{\link[rmarkdown]{render}} used to render the html output file.
}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relabelMutPositions.R
\name{relabelMutPositions}
\alias{relabelMutPositions}
\title{Relabel the positions of mutations in the designated ID}
\usage{
relabelMutPositions(se, conversionTable, mutNameDelimiter = ".")
}
\arguments{
\item{se}{SummarizedExperiment object, with row names of the form
XX\{.\}AA\{.\}NNN, where XX is the name of the reference sequence, 
AA is the position of the mutated codon, and NNN is the mutated codon 
or amino acid. \{.\} is the delimiter, to be specified in the 
\code{mutNameDelimiter} argument. For rows corresponding to sequences 
with multiple mutated codons, the row names contain multiple names of 
the form above in a single string, separated by "_".}

\item{conversionTable}{\code{data.frame} with at least three columns:
\itemize{
\item seqname The reference sequence name (should match XX in the 
mutation name)
\item position The codon position (should match AA in the mutation name)
\item name The new name for the codon (will replace AA in the mutation 
name, if the reference sequence matches seqname)
}}

\item{mutNameDelimiter}{The delimiter used in the mutation name 
(\{.\} above).}
}
\value{
A SummarizedExperiment object with modified row names.
}
\description{
Relabel the positions of mutations in the designated ID
}
\examples{
x <- readRDS(system.file("extdata", "GSE102901_cis_se.rds",
                         package = "mutscan"))
conversionTable <- data.frame(seqname = "f", position = 0:32) 
conversionTable$name = paste0((conversionTable$position - 1) \%/\% 7 + 1, 
                              c("", rep(letters[1:7], 6))[1:33])
out <- relabelMutPositions(x, conversionTable)

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/digestFastqs.R
\name{digestFastqs}
\alias{digestFastqs}
\title{Read, filter and digest sequences from fastq file(s).}
\usage{
digestFastqs(
  fastqForward,
  fastqReverse = NULL,
  mergeForwardReverse = FALSE,
  minOverlap = 0,
  maxOverlap = 0,
  minMergedLength = 0,
  maxMergedLength = 0,
  maxFracMismatchOverlap = 1,
  greedyOverlap = TRUE,
  revComplForward = FALSE,
  revComplReverse = FALSE,
  adapterForward = "",
  adapterReverse = "",
  elementsForward = "",
  elementLengthsForward = numeric(0),
  elementsReverse = "",
  elementLengthsReverse = numeric(0),
  primerForward = c(""),
  primerReverse = c(""),
  wildTypeForward = "",
  wildTypeReverse = "",
  constantForward = c(""),
  constantReverse = c(""),
  avePhredMinForward = 20,
  avePhredMinReverse = 20,
  variableNMaxForward = 0,
  variableNMaxReverse = 0,
  umiNMax = 0,
  nbrMutatedCodonsMaxForward = 1,
  nbrMutatedCodonsMaxReverse = 1,
  nbrMutatedBasesMaxForward = -1,
  nbrMutatedBasesMaxReverse = -1,
  forbiddenMutatedCodonsForward = "",
  forbiddenMutatedCodonsReverse = "",
  useTreeWTmatch = FALSE,
  collapseToWTForward = FALSE,
  collapseToWTReverse = FALSE,
  mutatedPhredMinForward = 0,
  mutatedPhredMinReverse = 0,
  mutNameDelimiter = ".",
  constantMaxDistForward = -1,
  constantMaxDistReverse = -1,
  umiCollapseMaxDist = 0,
  filteredReadsFastqForward = "",
  filteredReadsFastqReverse = "",
  maxNReads = -1,
  verbose = FALSE,
  nThreads = 1,
  chunkSize = 1e+05,
  maxReadLength = 1024
)
}
\arguments{
\item{fastqForward, fastqReverse}{Character vectors, paths to gzipped FASTQ 
files corresponding to forward and reverse reads, respectively. If more 
than one forward/reverse sequence file is given, they need to be 
provided in the same order. Note that if multiple fastq files are 
provided, they are all assumed to correspond to the same sample, and 
will effectively be concatenated.}

\item{mergeForwardReverse}{Logical scalar, whether to fuse the forward and
reverse variable sequences.}

\item{minOverlap, maxOverlap}{Numeric scalar, the minimal and maximal allowed
overlap between the forward and reverse reads when merging. Only used if
\code{mergeForwardReverse} is \code{TRUE}. If set to 0, only overlaps
covering the full length of the shortest of the two reads will be
considered.}

\item{minMergedLength, maxMergedLength}{Numeric scalar, the minimal and 
maximal allowed total length of the merged product (if 
\code{mergeForwardReverse} is \code{TRUE}). If set to 0, any length is 
allowed.}

\item{maxFracMismatchOverlap}{Numeric scalar, maximal mismatch rate in the
overlap. Only used if \code{mergeForwardReverse} is \code{TRUE}.}

\item{greedyOverlap}{Logical scalar. If \code{TRUE}, the first overlap 
satisfying \code{minOverlap}, \code{maxOverlap}, \code{minMergedLength},
\code{maxMergedLength} and \code{maxFracMismatchOverlap} will
be retained. If \code{FALSE}, all valid overlaps will be scored and the 
one with the highest score (largest number of matches) will be retained.}

\item{revComplForward, revComplReverse}{Logical scalar, whether to reverse
complement the forward/reverse variable and constant sequences, 
respectively.}

\item{adapterForward, adapterReverse}{Character scalars, the adapter sequence 
for forward/reverse reads, respectively. If a forward/reverse read 
contains the corresponding adapter sequence, the sequence pair will be 
filtered out. If set to \code{NULL}, no adapter filtering is performed. 
The number of filtered read pairs are reported in the return value.}

\item{elementsForward, elementsReverse}{Character scalars representing the
composition of the forward and reverse reads, respectively. The strings 
should consist only of the letters S (skip), C (constant), U (umi), 
P (primer), V (variable), and cover the full extent of the read. Most 
combinations are allowed (and a given letter can appear multiple times), 
but there can be at most one occurrence of P. If a given letter is 
included multiple times, the corresponding sequences will be 
concatenated in the output.}

\item{elementLengthsForward, elementLengthsReverse}{Numeric vectors 
containing the lengths of each read component from 
\code{elementsForward}/\code{elementsReverse}, respectively. If the 
length of one element is set to -1, it will be inferred from the other 
lengths (as the remainder of the read). At most one number (or one 
number on each side of the primer P) can be set to -1. The indicated
length of the primer is not used (instead it's inferred from the 
provided primer sequence) and can also be set to -1.}

\item{primerForward, primerReverse}{Character vectors, representing the 
primer sequence(s) for forward/reverse reads, respectively. Only read 
pairs that contain perfect matches to both the forward and reverse 
primers (if given) will be retained. Multiple primers can be specified - 
they will be considered in order and the first match will be used.}

\item{wildTypeForward, wildTypeReverse}{Character scalars or named character
vectors, the wild type sequence for the forward and reverse variable 
region. If given as a single string, the reference sequence will be 
named 'f' (for forward) or 'r' (for reverse).}

\item{constantForward, constantReverse}{Character vectors giving,
the expected constant forward and reverse sequences. If more than one 
sequence is provided, they must all have the same length.}

\item{avePhredMinForward, avePhredMinReverse}{Numeric scalar, the minimum 
average Phred score in the variable region for a read to be retained. 
If a read pair contains both forward and reverse variable regions, the 
minimum average Phred score has to be achieved in both for a read pair 
to be retained.}

\item{variableNMaxForward, variableNMaxReverse}{Numeric scalar, the maximum 
number of Ns allowed in the variable region for a read to be retained.}

\item{umiNMax}{Numeric scalar, the maximum number of Ns allowed in the UMI 
for a read to be retained.}

\item{nbrMutatedCodonsMaxForward, nbrMutatedCodonsMaxReverse}{Numeric
scalar, the maximum number of mutated codons that are allowed. Note that 
for the forward and reverse sequence, respectively, exactly one of
\code{nbrMutatedCodonsMax} and \code{nbrMutatedBasesMax} must be -1,
and the other must be a non-negative number. The one that is not -1
will be used to filter and name the identified mutants.}

\item{nbrMutatedBasesMaxForward, nbrMutatedBasesMaxReverse}{Numeric scalar,
the maximum number of mutated bases that are allowed. Note that for the
forward and reverse sequence, respectively, exactly one of
\code{nbrMutatedCodonsMax} and \code{nbrMutatedBasesMax} must be -1,
and the other must be a non-negative number. The one that is not -1
will be used to filter and name the identified mutants.}

\item{forbiddenMutatedCodonsForward, forbiddenMutatedCodonsReverse}{Character
vector of codons (can contain ambiguous IUPAC characters, see
\code{\link[Biostrings]{IUPAC_CODE_MAP}}). If a read pair contains a
mutated codon matching this pattern, it will be filtered out.}

\item{useTreeWTmatch}{Logical scalar. Should a tree-based matching
to wild type sequences be used if possible? If the number of allowed
mismatches is small, and the number of wild type sequences is large,
this is typically faster.}

\item{collapseToWTForward, collapseToWTReverse}{Logical scalar, indicating
whether to just represent the observed variable sequence by the
closest wildtype sequence rather than retaining the information about
the mutations.}

\item{mutatedPhredMinForward, mutatedPhredMinReverse}{Numeric scalar, the 
minimum Phred score of a mutated base for the read to be retained. If 
any mutated base has a Phred score lower than \code{mutatedPhredMin}, 
the read (pair) will be discarded.}

\item{mutNameDelimiter}{Character scalar, the delimiter used in the naming 
of mutants. Generally, mutants will be named as XX\{.\}YY\{.\}NNN, where 
XX is the closest provided reference sequence, YY is the mutated base or 
codon number (depending on whether \code{nbrMutatedBases*} or 
\code{nbrMutatedCodons*} is specified), and NNN is the
mutated base or codon. Here, \{.\} is the provided 
\code{mutNameDelimiter}. The delimiter must be a single character 
(not "_"), and can not appear in any of the provided reference sequence 
names.}

\item{constantMaxDistForward, constantMaxDistReverse}{Numeric scalars, the 
maximum allowed Hamming distance between the extracted and expected 
constant sequence. If multiple constant sequences are provided, the most 
similar one is used. Reads with a larger distance to the expected 
constant sequence are discarded. If set to -1, no filtering is done.}

\item{umiCollapseMaxDist}{Numeric scalar defining 
the tolerances for collapsing similar UMI sequences. If the value is 
in [0, 1), it defines the maximal Hamming distance in terms of a 
fraction of sequence length:
(\code{round(umiCollapseMaxDist * nchar(umiSeq))}).
A value greater or equal to 1 is rounded and directly used as the 
maximum allowed Hamming distance.}

\item{filteredReadsFastqForward, filteredReadsFastqReverse}{Character 
scalars, the names of a (pair of) FASTQ file(s) where filtered-out reads 
will be written. The name(s) should end in .gz (the output will always 
be compressed). If empty, filtered reads will not be written to a file.}

\item{maxNReads}{Integer scalar, the maximum number of reads to process. 
The first \code{maxNReads} read (pairs) in the FASTQ file(s) will be 
used. If set to -1, all reads in the FASTQ file(s) will be processed.}

\item{verbose}{Logical scalar, whether to print out progress messages.}

\item{nThreads}{Numeric scalar, the number of threads to use for parallel 
processing.}

\item{chunkSize}{Numeric scalar, the number of read (pairs) to keep in 
memory for parallel processing. Reduce from the default value if you 
run out of memory.}

\item{maxReadLength}{Numeric scalar, the maximum allowed read length. Longer 
read lengths lead to higher memory allocation, and may require 
the \code{chunkSize} to be decreased.}
}
\value{
A list with four entries:
\describe{
\item{summaryTable}{A \code{data.frame} that contains, for each observed
mutation combination, the corresponding variable region sequences (or pair 
of sequences), the number of observed such sequences, and the number of 
unique UMIs observed for the sequence. It also has additional columns: 
'maxNbrReads' contains the number of reads for the most frequent observed 
sequence represented by the feature (only relevant if similar variable 
regions are collapsed). 'nbrMutBases', 'nbrMutCodons' and 'nbrMutAAs' give 
the number of mutated bases, codons or amino acids in each variant. 
Alternative variant names based on base, codon or amino acid sequence are 
provided in columns mutantNameBase', 'mutantNameCodon', 'mutantNameAA'. In 
addition, mutantNameBaseHGVS' and 'mutantNameAAHGVS' give base- and amino 
acid-based names following the HGVS nomenclature 
(https://varnomen.hgvs.org/). Please note that the provided reference 
sequence names are used for the HGVS sequence identifiers. It is up to the 
user to use appropriately named reference sequences in order to obtain 
valid HGVS variant names.}
\item{filterSummary}{A \code{data.frame} that contains the number of input
reads, the number of reads filtered out in the processing, and the number of
retained reads. The filters are named according to the convention
"fxx_filter", where "xx" indicates the order in which the filters were
applied, and "filter" indicates the type of filter. Note that filters are
applied successively, and the reads filtered out in one step are not
considered for successive filtering steps.}
\item{errorStatistics}{A \code{data.frame} that contains, for each Phred
quality score between 0 and 99, the number of bases in the extracted 
constant sequences with that quality score that match/mismatch with the 
provided reference constant sequence.}
\item{parameters}{A \code{list} with all parameter settings that were used 
in the processing. Also contains the version of the package and the time of
processing.}
}
}
\description{
Read sequences for one or a pair of fastq files and digest them (extract
umis, constant and variable parts, filter, extract mismatch information from
constant and count the observed unique variable parts). Alternatively, 
primer sequences could be specified, in which case the sequence immediately
following the primer will be considered the variable sequence.
}
\details{
The processing of a read pair goes as follows:
\enumerate{
 \item Search for perfect matches to forward/reverse adapter sequences,
 filter out the read pair if a match is found in either the forward or
 reverse read.
 \item If primer sequences are provided, search for perfect matches, and
 filter out the read pair if not all provided primer sequences can be found.
 \item Extract the UMI, constant and variable sequence from forward and
 reverse reads, based on the definition of the respective read composition.
 \item If requested, collapse forward and reverse variable regions by
 retaining, for each position, the base with the highest reported base
 quality.
 \item Filter out the read (pair) if the average quality in the variable 
 region is below \code{avePhredMinForward}/\code{avePhredMinReverse}, in 
 either the forward or reverse read (or the merged read).
 \item Filter out the read (pair) if the number of Ns in the variable region
 exceeds \code{variableNMaxForward}/\code{variableNMaxReverse}.
 \item Filter out the read (pair) if the number of Ns in the combined 
 forward and reverse UMI sequence exceeds \code{umiNMax}
 \item If one or more wild type sequences (for the variable region) are
 provided, find the mismatches between the (forward/reverse) variable region
 and the provided wild type sequence (if more than one wild type sequence is
 provided, first find the one that is closest to the read).
 \item Filter out the read (pair) if any mutated base has a quality below
 \code{mutatedPhredMinForward}/\code{mutatedPhredMinReverse}.
 \item Filter out the read (pair) if the number of mutated codons exceeds
 \code{nbrMutatedCodonsMaxForward}/\code{nbrMutatedCodonsMaxReverse}.
 \item Filter out the read (pair) if any of the mutated codons match any of 
 the codons encoded by
 \code{forbiddenMutatedCodonsForward}/\code{forbiddenMutatedCodonsReverse}.
 \item Assign a 'mutation name' to the read (pair). This name is a 
 combination of parts of the form XX\{.\}YY\{.\}NNN, where XX is the name of 
 the most similar reference sequence, YY is the mutated codon number, and 
 NNN is the mutated codon. \{.\} is a delimiter, specified via 
 \code{mutNameDelimiter}. If no wildtype sequences are provided, the
 variable sequence will be used as the mutation name'.
}

Based on the retained reads following this filtering process, count the
number of reads, and the number of unique UMIs, for each variable sequence
(or pair of variable sequences).
}
\examples{
## See the vignette for complete worked-out examples for different types of 
## data sets

## ---------------------------------------------------------------------- ## 
## Process a single-end data set, assume that the full read represents  
## the variable region
out <- digestFastqs(
    fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
                               package = "mutscan"), 
    elementsForward = "V", elementLengthsForward = -1
)
## Table with read counts and mutant information
head(out$summaryTable)
## Filter summary
out$filterSummary

## ---------------------------------------------------------------------- ## 
## Process a single-end data set, specify the read as a combination of 
## UMI, constant region and variable region (skip the first base)
out <- digestFastqs(
    fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
                               package = "mutscan"), 
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
    constantForward = "AACCGGAGGAGGGAGCTG"
)
## Table with read counts and mutant information
head(out$summaryTable)
## Filter summary
out$filterSummary
## Error statistics
out$errorStatistics

## ---------------------------------------------------------------------- ## 
## Process a single-end data set, specify the read as a combination of 
## UMI, constant region and variable region (skip the first base), provide 
## the wild type sequence to compare the variable region to and limit the 
## number of allowed mutated codons to 1
out <- digestFastqs(
    fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
                               package = "mutscan"), 
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96), 
    constantForward = "AACCGGAGGAGGGAGCTG", 
    wildTypeForward = c(FOS = paste0(
        "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
        "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
    nbrMutatedCodonsMaxForward = 1
)
## Table with read counts and mutant information
head(out$summaryTable)
## Filter summary
out$filterSummary
## Error statistics
out$errorStatistics

## ---------------------------------------------------------------------- ## 
## Process a paired-end data set where both the forward and reverse reads 
## contain the same variable region and thus should be merged to generate 
## the final variable sequence, specify the reads as a combination of 
## UMI, constant region and variable region (skip the first and/or last
## base), provide the wild type sequence to compare the variable region to 
## and limit the number of allowed mutated codons to 1
out <- digestFastqs(
    fastqForward = system.file("extdata", "cisInput_1.fastq.gz", 
                               package = "mutscan"),
    fastqReverse = system.file("extdata", "cisInput_2.fastq.gz",
                               package = "mutscan"), 
    mergeForwardReverse = TRUE, 
    revComplForward = FALSE, revComplReverse = TRUE, 
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96),
    elementsReverse = "SUCVS", elementLengthsReverse = c(1, 7, 17, 96, -1),
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAGTTCATCCTGGCAGC",
    wildTypeForward = c(FOS = paste0(
        "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
        "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
    nbrMutatedCodonsMaxForward = 1
)
## Table with read counts and mutant information
head(out$summaryTable)
## Filter summary
out$filterSummary
## Error statistics
out$errorStatistics

## ---------------------------------------------------------------------- ## 
## Process a paired-end data set where the forward and reverse reads 
## contain variable regions corresponding to different proteins, and thus 
## should not be merged, specify the reads as a combination of 
## UMI, constant region and variable region (skip the first base), provide 
## the wild type sequence to compare the variable region to and limit the 
## number of allowed mutated codons to 1
out <- digestFastqs(
    fastqForward = system.file("extdata", "transInput_1.fastq.gz", 
                               package = "mutscan"),
    fastqReverse = system.file("extdata", "transInput_2.fastq.gz",
                               package = "mutscan"), 
    mergeForwardReverse = FALSE,  
    elementsForward = "SUCV", elementLengthsForward = c(1, 10, 18, 96),
    elementsReverse = "SUCV", elementLengthsReverse = c(1, 8, 20, 96),
    constantForward = "AACCGGAGGAGGGAGCTG", 
    constantReverse = "GAAAAAGGAAGCTGGAGAGA",
    wildTypeForward = c(FOS = paste0(
        "ACTGATACACTCCAAGCGGAGACAGACCAACTAGAAGATGAGAAGTC", 
        "TGCTTTGCAGACCGAGATTGCCAACCTGCTGAAGGAGAAGGAAAAACTA")),
    wildTypeReverse = c(JUN = paste0(
        "ATCGCCCGGCTGGAGGAAAAAGTGAAAACCTTGAAAGCTCAGAACTC", 
        "GGAGCTGGCGTCCACGGCCAACATGCTCAGGGAACAGGTGGCACAGCTT")), 
    nbrMutatedCodonsMaxForward = 1,
    nbrMutatedCodonsMaxReverse = 1
)
## Table with read counts and mutant information
head(out$summaryTable)
## Filter summary
out$filterSummary
## Error statistics
out$errorStatistics

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{groupSimilarSequences}
\alias{groupSimilarSequences}
\title{Create a conversion table for collapsing similar sequences}
\usage{
groupSimilarSequences(
  seqs,
  scores,
  collapseMaxDist = 0,
  collapseMinScore = 0,
  collapseMinRatio = 0,
  verbose = FALSE
)
}
\arguments{
\item{seqs}{Character vector with nucleotide sequences (or pairs of 
sequences concatenated with "_") to be collapsed. The sequences must 
all be of the same length.}

\item{scores}{Numeric vector of "scores" for the sequences. Typically
the total read/UMI count. A higher score will be preferred when 
deciding which sequence to use as the representative for a group of 
collapsed sequences.}

\item{collapseMaxDist}{Numeric scalar defining the tolerance for collapsing 
similar sequences. If the value is in [0, 1), it defines the maximal 
Hamming distance in terms of a fraction of sequence length:
(\code{round(collapseMaxDist * nchar(sequence))}).
A value greater or equal to 1 is rounded and directly used as the maximum
allowed Hamming distance. Note that sequences can only be
collapsed if they are all of the same length. The default value is 0.}

\item{collapseMinScore}{Numeric scalar, indicating the minimum score 
required for a sequence to be considered as a representative for a 
group of similar sequences (i.e., to allow other sequences to be 
collapsed into it). The default value is 0.}

\item{collapseMinRatio}{Numeric scalar. During collapsing of
similar sequences, a low-frequency sequence will be collapsed 
with a higher-frequency sequence only if the ratio between the 
high-frequency and the low-frequency scores is at least this 
high. A value of 0 indicates that no such check is performed.}

\item{verbose}{Logical scalar, whether to print progress messages.}
}
\value{
A data.frame with two columns, containing the input sequences 
and the representatives for the groups resulting from grouping similar
sequences, respectively.
}
\description{
Create a conversion table for collapsing similar sequences
}
\examples{
seqs <- c("AACGTAGCA", "ACCGTAGCA", "AACGGAGCA", "ATCGGAGCA", "TGAGGCATA")
scores <- c(5, 1, 3, 1, 8)
groupSimilarSequences(seqs = seqs, scores = scores, 
                      collapseMaxDist = 1, collapseMinScore = 0, 
                      collapseMinRatio = 0, verbose = FALSE)
                            
}
\author{
Michael Stadler, Charlotte Soneson
}
